(* Module implementing oracles with ATI *)
module E = Errormsg
module L = Logic
module P = Proof
module B = Bitmap 
module H = Hashtbl
module U = Util

module A = Ati
open Pretty

let debug = false
let checkSubgoalsBackward = true

(** We use that ATI module to construct oracles *)
class oracleClauseData 
    (name: string) 
    (vars: string list)
    (hyps: L.pred list) 
    (conc: L.pred)
    = object
      method name = name
      method conc = conc
      method hyps = hyps
      method maxUnif = List.length vars - 1;
      method clause = ()
    end
    
type t = unit A.state

(**** The BITSTREAM *****)
module FastBits : Oracle.BITSTREAM = 
  struct
    type data = B of int32 | S of string
    type bits = 
        { mutable current: int32;
          mutable used:    int;         (* 0 <= ??? < 32; how many bits of 
                                         * current are used. They are always 
                                         * on low-order positions  *) 
          mutable rest:    data list;   (* The rest of the bits *)
          mutable count:   int          (* How many bits total *)
        } 
          
    type outbits = bits   (* The "rest" is in reverse order *)

    let start name = 
      if debug then
        ignore (E.log "ATI: creating empty oracle for %s\n" name)
      
    let empty () = 
      { current = Int32.zero;
        used    = 0;
        count   = 0;
        rest    = []}

    let log2 n =                        (* Calculate the number of bits 
                                         * necessary *)
      assert (n > 0);
      let rec countBits n = 
        if n = Int32.zero then 0 else 
        1 + (countBits (Int32.shift_right_logical n 1)) 
      in
      countBits (Int32.of_int (n - 1))

      
    let put bits ~nrchoices ~thechoice = 
      assert (thechoice >= 0 && thechoice < nrchoices);
      let nrbits = log2 nrchoices in
      let toPut = Int32.of_int thechoice in
      bits.count <- bits.count + nrbits;
      if 32 - bits.used > nrbits then begin
        bits.current <- 
           Int32.logor bits.current (Int32.shift_left toPut bits.used);
        bits.used <- bits.used + nrbits
      end else begin
        let usenow = 32 - bits.used in
        let current =            
          Int32.logor bits.current (Int32.shift_left toPut bits.used) in
        bits.rest <- B current :: bits.rest;
        bits.current <- Int32.shift_right_logical toPut usenow;
        bits.used <- nrbits - usenow
      end

    let putString bits str = bits.rest <- S str :: bits.rest;

    type inbits  = bits   (* The "rest" is in the right order *)

    let in_of_out bits = 
      let rec revonto acc = function
          [] -> acc
        | r :: rest -> revonto (r :: acc) rest  
      in
      let res = 
        { current = Int32.zero;
          used    = 0;
          count   = bits.count;
          rest    = revonto [B bits.current] bits.rest } in
      let ddata = function
          B i -> text ("0x" ^ (Int32.format "%x" i))
        | S s -> dprintf "!%s!" s
      in
      if debug then 
        ignore (E.log "ATI: created %d inbits: @[%a@]@!"
                  res.count
                  (docList ~sep:(chr ',' ++ break) ddata) res.rest);
      res

        
    let rec replenish bits = 
      match bits.rest with
        [] -> E.s (E.bug "Cannot replenish")
      | S _ :: rest -> begin
          bits.rest <- rest;
          replenish bits
      end
      | B b :: rest -> begin
          assert (bits.used = 0);
          bits.current <- b;
          bits.used <- min 32 bits.count;
          bits.rest <- rest
      end

    let get bits ~nrchoices = 
      let nrbits = log2 nrchoices in
      assert (nrbits > 0);
      if nrbits > bits.count || nrbits > 32 then
        E.s (E.bug "Trying to fetch too many bits");
      let start, left = 
        if bits.used < nrbits then begin
          let first = bits.current in
          let usenow = bits.used in
          bits.used <- 0;
          bits.count <- bits.count - usenow;
          replenish bits;
          first, nrbits - usenow
        end else 
          Int32.zero, nrbits
      in
      let res = 
        Int32.logand 
          (Int32.logor start (Int32.shift_left bits.current (nrbits - left))) 
          (Int32.sub (Int32.shift_left Int32.one nrbits) Int32.one) in
      bits.current <- Int32.shift_right_logical bits.current left;
      bits.used <- bits.used - left;
      bits.count <- bits.count - left;
      let ires = Int32.to_int res in
      if ires >= nrchoices then 
        E.s (E.bug "Got choice %d when selecting out of %d choices\b"
               ires nrchoices);
      ires
        
    exception OneComment

    let dump out bits = 
      let bits = in_of_out bits in
      let output_byte separator = 
        if bits.count <= 0 then ()
        else begin
          if bits.used = 0 then begin
            match bits.rest with
              S s :: rest -> begin
                output_string out "\nANN_ARCHCOMMENT ";
                output_string out s; 
                bits.rest <- rest;
                raise OneComment
              end 
            | _ -> replenish bits
          end;
          assert (bits.used > 0);
          output_string out separator;
          output_string out (Int32.format "0x%02x" (Int32.logand bits.current
                                                      (Int32.of_int 255)));
          bits.current <- Int32.shift_right_logical bits.current 8;
          bits.used <- bits.used - 8;
          bits.count <- bits.count - 8
        end
      in
      while bits.count > 0 do
        try begin
          output_byte "\nANN_ARCHW1DECL\t";
          output_byte ", ";
          output_byte ", ";
          output_byte ", ";
          output_byte ", ";
          output_byte ", ";
          output_byte ", ";
          output_byte ", "
        end with OneComment -> ()
      done
       
      
  end

module SlowBits : Oracle.BITSTREAM = 
  struct
    type choice = { which: int; outof: int }
    type bits = choice list ref
    type inbits = bits
    type outbits = bits

    let start name = 
      if debug then
        ignore (E.log "ATI: creating empty oracle for %s\n" name)

    let empty () = ref []
    let put bits ~nrchoices ~thechoice = 
      bits := { which = thechoice; outof = nrchoices } :: !bits

    let putString _ _ = ()

    let in_of_out bits = 
      let res = ref (List.rev !bits) in
      if debug then 
        ignore (E.log "ATI: created inbits: @[%a@]@!"
                  (docList ~sep:(chr ',' ++ break) 
                     (fun c -> num c.which ++ text "/" ++ num c.outof))
                  !res);
      res
        
    let get bits ~nrchoices = 
      match !bits with 
        c :: rest when c.outof = nrchoices -> begin
          bits := rest;
          c.which
        end
      | c :: _ -> E.s (E.bug "ATI: choice does not match: need %d, found %d/%d"
                         nrchoices c.which c.outof);
      | [] -> E.s (E.bug "ATI: no -> int choice available")

    let dump out bits = 
      let inbits = in_of_out bits in
      List.iter 
        (fun c -> 
          Printf.fprintf out "ANN_ARCHW1DECL %d / %d\n" c.which c.outof)
        !inbits
  end

(* module Bits = SlowBits  For debugging *)
module Bits = FastBits (* *)


(******* A NON-DETERMINISTIC INTERPRETER **********)
let rec interpreter (state: unit A.state) 
                             (* The state of the oracle, the possible 
                              * choices. Returns the pick (index and clause) 
                              * and the next state *)
                    (choose: 'a -> unit A.clauseData list 
                                -> int * unit A.clauseData * 'a) 
                    (initstate: 'a)
                    (atom: L.pred)
                    (subst: L.subst) : 'a = 
  (* Find out what are the possible choices *)
  let choices: unit A.clauseData list = A.findChoices state (atom, subst) in
  (* Choose one *)
  let (pick: int), 
      (clause: unit A.clauseData),
      (nextstate: 'a) = choose initstate choices in
  if debug then 
    ignore (E.log "  -picked clause %d (of %d)\n" 
              pick (List.length choices));
  (* Process the subgoals *)
  let csubst = L.newSubst (clause#maxUnif + 1) in
  if not (L.punify (fun s i es -> s.(i) <- L.Done es; true) 
            (atom, subst) (clause#conc, csubst)) then
    E.s (E.bug "Clause %a is not the right choice for %a\n"
           A.dclause clause L.d_spred (atom, subst));
  (* Do the subgoals, threading the chooser state *)
  if checkSubgoalsBackward then
    List.fold_right (fun h cstate -> interpreter state choose cstate h csubst)
      clause#hyps
      nextstate
  else
    List.fold_left (fun cstate h -> interpreter state choose cstate h csubst) 
      nextstate
      clause#hyps

(***** PROOF CHECKING *****)
type checkstate = unit
let check state atom inbits = 
  if debug then
    ignore (E.log "ATI:Checking proof of: %a@!" L.d_pred atom);

  let oracle () chlist = 
    let nrChoices = List.length chlist in
    if nrChoices <= 0 then 
      E.s (E.bug "Atioracle.check: no choices");

    let pick = if nrChoices = 1 then 0 else Bits.get inbits nrChoices in
    pick, List.nth chlist pick, ()
  in
  interpreter state oracle () atom L.emptySubst
  

(**** PROOF CONVERSION ****)
type convstate = P.proof list    (* A stack of proofs to look at next *)
let convert (state: unit A.state) 
            outbits (prf: P.proof) (atom: L.pred) = 
  if debug then
    ignore (E.log "ATI:Converting proof: %a@! of atom: %a@!"
              P.d_prf prf L.d_pred atom);
  let localoutbits = Bits.empty () in   (* For debugging purposes *)
  let oracle (cstate: convstate) (chlist: unit A.clauseData list) = 
    let nrChoices = List.length chlist in
    if nrChoices = 0 then
      E.s (E.bug "no choices");

    let (pname: string), (pargs: P.parg list), (reststate: convstate) = 
      match cstate with
        P.G(cname, pargs) :: reststate -> cname, pargs, reststate
      | P.Hyp v :: reststate -> U.symbolName v, [], reststate
      | _ -> E.s (E.unimp "proof conversion supports only P.G proofs")
    in
    (* Search for the named clause in the list *)
    let (pickidx: int), (cpick: unit A.clauseData) = 
      let rec findPick idx = function
          c :: _ when pname = c#name -> idx, c
        | _ :: rest -> findPick (idx + 1) rest
        | [] -> E.s (E.bug "Cannot match pick %s in choices: %a"
                       pname 
                       (docList
                          (fun c -> text c#name))
                       chlist)
      in
      findPick 0 chlist
    in
    if nrChoices > 1 then begin
      Bits.put outbits nrChoices pickidx;
      if debug then 
        Bits.put localoutbits nrChoices pickidx
    end;
    
    (* Drop syntax arguments from the pargs *)
    let nrHyps = List.length cpick#hyps in
    let nrArgs = List.length pargs in
    let rec dropNPush n = function
        l when n = 0 -> 
          push (if checkSubgoalsBackward then List.rev l else l)
      | [] -> E.s (E.bug "Dropping too many\n")
      | _ :: rest -> dropNPush (n - 1) rest
    and push = function
        [] -> reststate
      | P.P prf :: rest -> prf :: push rest
      | _ -> E.s (E.bug "Proof arguments are not P.P")
    in
    let newstate = dropNPush (nrArgs - nrHyps) pargs in
    pickidx, cpick, newstate

  in
  let final = interpreter state oracle [prf] atom L.emptySubst in
  if final <> [] then E.s (E.bug "Left over proof: %a :: ..."
                             P.d_prf (List.hd final));
  if debug then begin
    let localinbits = Bits.in_of_out localoutbits in
    check state atom localinbits
  end

    
  

let make () = 
  let dummyclause = 
    new oracleClauseData "dummy" [] [] (L.pFun "dummy" [])
  in
  A.make dummyclause


(* Add an axiom to the ATI *)
let addAxiom (state : unit A.state) 
             ((name, vars, hyps, conc) : Input.axiom) : unit =
          (* Construct a map from variables to Subs *)
  let vHash : (string, int) H.t = H.create 7 in
  let varId = ref 0 in
  List.iter (fun v -> H.add vHash v !varId; incr varId) vars;
          (* Now rewrite the predicates to use substs instead of variables *)
  let doSubst level (e: L.exp) : L.exp option = 
    match e with 
      L.NVar v -> begin
        try Some (L.UVar (H.find vHash v))
        with Not_found -> 
          E.s (E.bug "Cannot find mapping for variable %s\n" v)
      end
    | L.UVar _ -> E.s (E.bug "Unexpected sub in axioms")
    | _ -> None
  in
  let doExp (e: L.exp) : L.exp = L.substExp doSubst e in
  let doPred (p: L.pred) : L.pred = L.substPred doSubst p in
  begin try
    let hyps = List.map doPred hyps in
    let conc = doPred conc in
    A.addClause state 
      ~clause:(new oracleClauseData name vars hyps conc)
  with Not_found -> ()
  end
                                              
    
    (* Assume an atom with no free variables *)
let assume (state: unit A.state) ~name atom = 
  let newclause = 
    new oracleClauseData name [] [] atom 
  in
  A.addClause state ~clause:newclause
  

let undo state = A.deleteLastClause state



