module L = Logic
module I = Input
module P = Proof
module E = Errormsg
module H = Hashtbl

open Pretty

let debug = ref false
let debugUndo = false

let spyTable : (string, unit) H.t = H.create 17
let setSpy yes what = 
  if yes then
    H.add spyTable what ()
  else
    try H.remove spyTable what with _ -> ()

(**************** DEBUGGING *****************)
(*  Turn on debugging and you will see a series of lines:

    AX:n.l: operation goal

    where n is the invocation id. Every time prove is invoked internally a 
    new invocation id is generated. The invocation id is reset at the 
    external invocation of prove. 

    l is the level. It starts at 0 and as prove is invoked recursively it 
    grows. 

    operation is one of: Call, Success, Redo, Fail or Cut
    an invocation is created by prove and the first operation is a Call. If 
    one clause matches then the next operation on the invocation is Success 
    and the subgoals are processed. If one of the subgoals backtracks the 
    Redo operation is initiated and the next clause is tried followed by
    Success and subgoals, a.s.o. When Call or Redo fail the Fail operation is
    initiated. Cut is a special case of Redo which does not attempt any other  
    matches but Fails immediately. 


    This debugger model is inspired by Prolog debuggers. See for example the 
    Sicstus Prolog manual online:
     http://www.cs.bham.ac.uk/~pjh/prolog_course/sicstus_manual_v3_5/sicstus_9.html#SEC75
    
      
*)

     (* We will keep enough information to allow us to undo unification *)
type 'a undoElem = 
    Snapshot of int
  | DoneUnif of L.subst * int
  | AddAxiom of 'a axiom


and 'a program =                            (* A collection of axioms  *)
    { mutable undo: 'a undoElem list;
      dummyInit: 'a;                    (* Used to initialize subgoals, when 
                                         * no specific initializer is given *)
      axioms: (string, 'a axiom) Hashtbl.t;(* The axioms, indexed by their 
                                            * head  *)
    } 


(* A DSL for specifying axioms for Prolog-style proof search *)
and 'a axiom = 
    { name:    string;                  (* For debugging *)
      maxUnif: int;                      (* The maximal value of a 
                                          * unification variable index. *)
      conc:    Logic.pred;              (* The conclusion, typically 
                                         * containing unification variables.  *)
      sub:     'a subgoal;        
      init:    'a;                      (* The initial accumulator *)
                                        (* A proof constructor that is given 
                                         * the program, the current 
                                         * substitution and name of the axiom 
                                         * and the accumulator *)
      isassume: bool;
      mkprf:   'a program -> Logic.subst -> 'a -> Proof.proof
}

and 'a subgoal = 
              'a program -> (* The current program *)
              Logic.subst -> (* The substitution for this axiom *)
              'a cont -> (* Success continuation. Pass on the accumulator + 
                          * everything new *)
              'a ->      (* The accumulated results *)
              back ->    (* Backtracking function *)
              result 

    (* The success continuation of a subgoal *)
and 'a cont = 'a -> back -> result

    (* Retry means to retry the last successful subproof, independent of its 
     * level *)
    (* Commit i means to retry the last successful subproof at level lower 
     * than i. This means that a different clause for the parent might be 
     * tried *)
    (* Cut i means to skip all subgoals at level >= i and the parent (the 
     * last success at level i - 1). This means that no more clauses of the 
     * parent are retried  *)
and backinfo = Retry | Cut of int | Commit of int

    (* The failure continuation of a subgoal. Used for implementing 
     * backtraching. If the argument is Some i then it means that we must not 
     * backtrack on all successful goals at levels > i and also on the last 
     * successful subgoal at level i *)
and back = backinfo -> result

and result = Failure of backinfo | Success of Proof.proof * back

(**************** PRINTING **************)
let dax () ax = 
  dprintf "Axiom %s(@[%a@])"
    ax.name L.d_pred ax.conc


(**************** THE UNDO STACK ************)
let snapshotId   = ref 0 

let snapshot prg = incr snapshotId;
                   if !debug && debugUndo then
                     ignore (E.log "AX:snapshot %d\n" !snapshotId);
                   prg.undo <- Snapshot !snapshotId :: prg.undo

let getAtomIndex = function
    L.PFun(n, _) -> n
  | L.PNez(L.BinOp(L.SetEq, _, _)) -> "__eq"
  | L.PNez(L.BinOp(L.SetNe, _, _)) -> "__ne"
  | a -> E.s (E.unimp "getAtomIndex: %a" L.d_pred a)


  (* Undo all the unifications *)
let undo prg = 
  let rec loop = function
      [] -> E.s (E.bug "AX:empty undo stack")
    | Snapshot id :: rest -> 
        if !debug && debugUndo then
          ignore (E.log "AX:Undoing to snapshot %d\n" id);
        rest
    | DoneUnif (s, i) :: rest -> begin
        assert(i >= 0 && i < Array.length s);
        assert (s.(i) != L.Unif);
        s.(i) <- L.Unif;
        loop rest
    end
    | AddAxiom a :: rest -> begin
        if !debug then
            ignore (E.log "Retracting axiom %s\n" a.name);
        let name = getAtomIndex a.conc in
        H.remove prg.axioms name;
        loop rest
    end
  in
  prg.undo <- loop prg.undo

  (* Change a substitution and record it *)
let changeSub prg s i es = 
  prg.undo <- DoneUnif (s, i) :: prg.undo;
  assert (i >= 0 && i < Array.length s);
  s.(i) <- L.Done es;
  true

let undoBack (prg: 'a program) (back: back) : back = 
  fun cut -> undo prg; back cut

let punify prg = L.punify (changeSub prg)
let eunify prg = L.eunify (changeSub prg)

let dbinfo () = function
    Retry -> text "Retry"
  | Cut i -> dprintf "Cut %d" i
  | Commit i -> dprintf "Commit %d" i

(****************** PROGRAMS *************)
let emptyProgram (dummyInit : 'a) : 'a program = 
  { undo      = [];
    dummyInit = dummyInit;
    axioms    = H.create 17;
  } 


  (* Add a new axiom to the program *)
let extendProgram prg ax = 
  if !debug then
    ignore (E.log "AX:Adding %a@!" dax ax);
  let name = getAtomIndex ax.conc in
  prg.undo <- AddAxiom ax :: prg.undo;
  H.add prg.axioms name ax

  (* Do nothing tactical *)
let sgID = 
  fun prg s cont acc back -> cont acc back


  (* Add an assumption to the program. Model it as a trivial axiom *)
let assume prg (at: L.pred) (prf: P.proof) (name: string) = 
  extendProgram prg { name    = name;
               conc    = at;
               maxUnif = -1;             (* No free unification variables *)
               sub     = sgID;
               init    = prg.dummyInit;
               isassume = true;
               mkprf   = fun _ _ _ -> prf;
             } 

let getCandidateAxioms prg axselector (a : L.pred) = 
  let aname = getAtomIndex a in
  List.filter axselector (H.find_all prg.axioms aname)

let checkMatchAxiom prg s ax a = 
  (* Create a subtitution appropriate for axiom *)
  let axs = Array.create (ax.maxUnif + 1) L.Unif in
  if punify prg (ax.conc, axs) (a, s) then
    Some axs
  else begin
    None
  end

let goalId = ref 0
let level  = ref 0  (* The level we are at. Starts at 0 on the first 
                     * invocation and then increases with each nested 
                     * invocation *)

(************ TACTICALS **************)

  (* Always succeeds with a given result. *)
let sgTRUE (prg : 'a program) 
           (s : L.subst) (cont: 'a cont) (acc: 'a) (back: back) = 
       (* Add the res to the accumulator *)
  sgID prg s cont acc back

 (* Always fails *)
let sgFALSE prg s cont acc back = back Retry

 (* Always succeeds but on backtracking initiates a CUT *)
let sgCUT prg s cont acc back = 
  let thisId = !goalId in
  incr goalId;
  let thisLevel = !level in
  let id = dprintf "%d.%d:" thisId thisLevel in
  if !debug then
    ignore (E.log "AX:%a Install Cut@!" insert id);
  let back' binfo = 
    let newbinfo = match binfo with
      Retry -> Cut thisLevel
    | Cut i -> if i < thisLevel then binfo else Cut thisLevel
    | Commit i -> if i < thisLevel then binfo else Cut thisLevel
    in
    if !debug then
      ignore (E.log "AX:%a Processing cut. Propagating %a@!" 
                insert id dbinfo newbinfo);
    back newbinfo
  in
  cont acc back'

let sgCOMMIT prg s cont acc back = 
  let thisId = !goalId in
  incr goalId;
  let thisLevel = !level in
  let id = dprintf "%d.%d:" thisId thisLevel in
  if !debug then
    ignore (E.log "AX:%a Install Commit@!" insert id);
  let back' binfo = 
    let newbinfo = match binfo with
      Retry -> Commit thisLevel
    | Commit i -> if i < thisLevel then binfo else Commit thisLevel
    | Cut i -> if i <= thisLevel then binfo else Commit thisLevel
    in
    if !debug then
      ignore (E.log "AX:%a Processing commit. Propagating %a@!" 
                insert id dbinfo newbinfo);
    back newbinfo
  in
  cont acc back'

 (* Conditional *)
let sgIF (cond: 'a program -> L.subst -> 'a -> bool * 'a) 
         (tsg: 'a subgoal) 
         (fsg: 'a subgoal) prg s cont acc back = 
  snapshot prg;  (* Take a snapshot because the conditional can have 
                  * side-effects *)
  let which, res = cond prg s acc in
  let back' = undoBack prg back in
  if which then 
    tsg prg s cont res back'
  else
    fsg prg s cont res back'

 (* Sequencing *)
let rec sgSEQ = function
  [] -> sgID
  | sg :: sglist -> 
      fun prg s cont acc back ->
        sg prg s (fun acc1 back1 -> sgSEQ sglist prg s cont acc1 back1)
          acc
          back
    
let (&&&) sg1 sg2 = 
  fun prg s cont acc back -> 
    sg1 prg s (fun acc1 back1 -> sg2 prg s (fun acc2 back2 -> cont acc2 back2)
                                           acc1
                                           back1)
              acc
              back
                                        
 (* A selection tactical. The selection function can use unification *)
let sgCHOICE (choose: 'a program -> L.subst -> 'a -> 'a subgoal *'a)
    : 'a subgoal = 
  fun prg s cont acc back -> 
    snapshot prg;
    let sg', acc' = choose prg s acc in
    sg' prg s cont acc' (undoBack prg back)

 (* A disjunction. Consumes CUT *)
let rec sgOR : 'a subgoal list -> 'a subgoal = function
    [] -> sgFALSE
  | sg :: sglist -> 
      fun prg s cont acc back -> 
        sg prg s (fun acc1 back1 -> cont acc1 back1)
                 acc
                 (function Retry -> sgOR sglist prg s cont acc back
                   | cut -> back cut)

   (*** MAIN ENTRY POINT ***)
let rec prove (prg: 'a program) 
          (axselector: 'a axiom -> bool)
          ((a, s): L.spred) : P.proof option = 
  goalId := 0;
  level  := 0;
  (* Record the undo stack *)
  let oldUndo = prg.undo in
  let res = proveIntern prg axselector (a,s) in
  (* Now revert back to the last state *)
  while prg.undo != oldUndo do
    undo prg
  done;
  match res with
    Success (x, _) -> Some x
  | Failure _ -> None

   (*** INTERNAL PROVER ****)
and proveIntern (prg: 'a program) 
          (axselector: 'a axiom -> bool)
          ((a, s): L.spred) : result = 
  let thisId = !goalId in
  let thisLevel = !level in
  incr level;  (* Prepare the level for the subgoals *)
  let id = dprintf "%d.%d:" thisId thisLevel in
  incr goalId;
  (* Get the list of axioms that match this goal *)
  let axlist = getCandidateAxioms prg axselector a in
  let showCandidates () axlist = 
    (docList (chr ',' ++ break) (fun ax -> text ax.name)) () axlist
  in
  if !debug then
    ignore (E.log "AX:%a Call %a@!" insert id L.d_spred (a, s));

  (* Check the axioms in turn *)
  let rec checkAxioms = function
      [] -> begin
        if !debug then
          ignore (E.log "            no match. Fail@!");
        Failure Retry
      end
    | ax :: rest -> begin
        snapshot prg;
        match checkMatchAxiom prg s ax a with
          None -> (* This one does not actually match *)
            undo prg;
            checkAxioms rest
        | Some axs ->   (* This is a match. Get the substitution for the 
                         * axiom  *)
            if !debug then
              ignore (E.log "            match with %s@!" ax.name);
            let back binfo =
              undo prg; 
              let retry, newbinfo = match binfo with
                Retry -> true, Retry
              | Cut i -> false, if i > thisLevel then Retry else binfo
              | Commit i -> 
                  if i > thisLevel then true, Retry else false, binfo
              in
              if retry then begin
                if !debug then
                  ignore (E.log "AX:%a Redo %a@!"
                            insert id L.d_spred (a, s));
                checkAxioms rest 
              end else begin
                if !debug then
                  ignore (E.log "AX:%a Fail/Cut through with %a. Now=%a@!" 
                            insert id dbinfo binfo dbinfo newbinfo);
                Failure newbinfo
              end
            in
                (* Run the subgoals *)
            match ax.sub prg axs 
                  (fun acc back -> Success (ax.mkprf prg axs acc, back)) 
                  ax.init 
                  (fun cut -> Failure cut) 
            with
                    (* Report success but be ready to continue scanning the 
                     * assumptions if somebody backtracks here *)
              Success (prf, back') -> begin
                if !debug then
                  ignore (E.log "AX:%a Success %a@!" 
                            insert id L.d_spred (a, s));
                (* Chain two backtracking functions together *)
                let rec prependToBack back1 = 
                  fun cut -> 
                    match back1 cut with
                      Failure cut -> back cut
                    | Success(pf, back') -> 
                        Success (pf, prependToBack back')
                in
                Success (prf, 
                         prependToBack back')
              end
            | Failure cut -> back cut
    end
  in
  let res = checkAxioms axlist in
  decr level;
  res
    
  (* Search for the proof of an atom. Uses a selector to rule out some 
   * axioms. The selector CANNOT do unification, or must undo. Consume CUTS *)
let sgPROVE_sel (axselector: 'a axiom -> bool)
                (a: L.pred) 
                (trans: 'a -> P.proof -> 'a)  
                 : 'a subgoal = 
  fun prg s cont acc back ->
    match proveIntern prg axselector (a, s) with
      Failure binfo -> back binfo
    | Success (prf, back') ->
        let rec backtrack (getNext : back) : back = 
          fun binfo -> begin
            match getNext binfo with
              Failure binfo' -> back binfo'
            | Success (prf, back') -> 
                cont (trans acc prf) (backtrack back')
          end
        in
        cont (trans acc prf) (backtrack back')
  
let sgPROVE a trans = sgPROVE_sel (fun _ -> true) a trans

let sgASSUMPTION a trans = sgPROVE_sel (fun ax -> ax.maxUnif = -1) a trans


let sgIsGround e = 
  fun prg s cont acc back ->
    match L.unrollExp (e, s) with
      (L.UVar i, s') when s'.(i) = L.Unif -> back Retry
    | _ -> cont acc back


let sgIsInt e f = 
  fun prg s cont acc back ->
    match L.unrollExp (e, s) with
      (L.Int i, _) when f (L.to_int i) -> cont acc back
    | _ -> back Retry

let sgIsConst e = 
  fun prg s cont acc back ->
    match L.unrollExp (e, s) with
      (L.Fun(_, []), _) -> cont acc back
    | _ -> back Retry

let sgDO f = 
  fun prg s cont acc back -> 
    f s acc;
    cont acc back

let sgCHECK f = 
  sgIF f sgTRUE sgFALSE
      
let sgACCUM f =
  fun prg s cont acc back ->
    cont (f prg s acc) back

let sgUNIF t1 t2 = sgCHECK (fun prg s acc -> (eunify prg (t1, s) (t2, s), acc))
      
 
