
val debug : bool ref

type 'a program                         (* A collection of axioms  *)

type 'a subgoal                         (* Subgoals *)

(* A Domain Specific Language for specifying axioms for Prolog-style proof 
 * search  *)
type 'a axiom = 
    { name:    string;                  (* For debugging *)
      maxUnif: int;                      (* The maximal value of a 
                                          * unification variable index.  *)
      conc:    Logic.pred;              (* The conclusion, typically 
                                         * containing unification variables. *)
      sub:     'a subgoal;        
      init:    'a;                      (* The initial accumulator *)
                                        (* A proof constructor that is given 
                                         * the program, the current 
                                         * substitution and name of the axiom 
                                         * and the accumulator *)
      isassume: bool;
      mkprf:   'a program -> Logic.subst -> 'a -> Proof.proof
}



(*** TACTICALS ***)
val sgID: 'a subgoal   (* Just continue *)

                 (* Like sgID but runs a thunk whenever it is invoked *)
val sgDO: (Logic.subst -> 'a -> unit) -> 'a subgoal

                 (* Always succeeds and knows how to finish off a proof. The 
                  * transformer function CANNOT use unification *)
val sgTRUE: 'a subgoal  (* Another name for sgID *)

                 (* Always fails (backtracks) *)
val sgFALSE: 'a subgoal

                 (* Always succeeds but on backtracking it does not attempt 
                  * to find another solutions for all its previously 
                  * appearing sibling subgoals and for the parent *)
val sgCUT: 'a subgoal

                 (* Always succeeds but on backtracking it does not attempt 
                  * to find another solutions for all its previously 
                  * appearing sibling subgoals. It does attempt to find other 
                  * subgoals for the parent *)
val sgCOMMIT: 'a subgoal

        (* Depending on the current substitution decides which of the 
         * hypotheses to do. In the process it updates the accumulated 
         * result. The choice function can use unification *) 
val sgIF:     ('a program -> Logic.subst -> 'a -> bool * 'a) 
              -> 'a subgoal -> 'a subgoal -> 'a subgoal


        (* Checks that a condition holds and continues if yes. The test 
         * function can update the substitution. This is just an abbreviation 
         * for sgIF f sgTRUE sgFALSE *)
val sgCHECK: ('a program -> Logic.subst -> 'a -> bool * 'a) -> 'a subgoal


        (* A sequencing of tacticals *)
val sgSEQ: 'a subgoal list -> 'a subgoal

        (* A binary sequencing operator *)
val (&&&) : 'a subgoal -> 'a subgoal -> 'a subgoal

        (* A general choice tactical. The choice function can use unification 
         * *)
val sgCHOICE: ('a program -> Logic.subst -> 'a -> 'a subgoal * 'a) 
  -> 'a subgoal

        (* A disjunction *)
val sgOR: 'a subgoal list -> 'a subgoal


        (* Checks that the given atom is alredy known to be true. The atom 
         * shares unification  variables with the goal. In the process of 
         * searching the known assumptions it does instantiate some 
         * unification variables. The transform function CANNOT use 
         * unification *)
val sgASSUMPTION: Logic.pred -> ('a -> Proof.proof -> 'a) -> 'a subgoal


        (* Search for the proof of the atom. The unification variables in the 
         * atom are shared with the axiom conclusion. *)
val sgPROVE: Logic.pred -> ('a -> Proof.proof -> 'a) -> 'a subgoal


       (* Succeeds if the given expression is not a unification variable *)
val sgIsGround: Logic.exp -> 'a subgoal

       (* Succeeds if the given expression is an integer meeting the given 
        * condition *)
val sgIsInt: Logic.exp -> (int -> bool) -> 'a subgoal

       (* Suceeds if the given expression is a constant *)    
val sgIsConst: Logic.exp -> 'a subgoal


       (* Always succeeds but adds something to the accumulator *)
val sgACCUM: ('a program -> Logic.subst -> 'a -> 'a) -> 'a subgoal


       (* Searches for the proof of an atom. Uses an axiom selector *)
val prove: 'a program -> ('a axiom -> bool) 
                      -> Logic.spred -> Proof.proof option



                                        (* Create an empty program *)
val emptyProgram: 'a -> 'a program

val snapshot: 'a program -> unit
                                        (* Extend the program by asserting an 
                                         * atom, along with its proof *)
val assume: 'a program -> Logic.pred -> Proof.proof -> string -> unit

                                        (* Extend the program by adding a new 
                                         * axiom *)
val extendProgram: 'a program -> 'a axiom -> unit

val undo: 'a program ->unit

(*** UNIFICATION functions for use in tacticals ***)
val eunify : 'a program -> Logic.sexp -> Logic.sexp -> bool
val punify : 'a program -> Logic.spred -> Logic.spred -> bool


(* Like sgCHECK with success f the two terms unify *)
val sgUNIF : Logic.exp -> Logic.exp -> 'a subgoal
