

val addClauseOfPredicate: Logic.pred -> Proof.proof -> ingoal:bool -> unit
(** Given a predicate and its proof, transform the predicate into a clause 
 * and add it to the set of clauses to be used for case analysis. The clause 
 * will be retracted on an undo. *)


val addUnitClauseOfPredicate: Logic.pred -> 
                              (Util.symbol -> Proof.proof -> Proof.proof) -> 
                              ingoal:bool -> 
                              unit
(** Make a clause out of a predicate and a proof-by-contradiction generator. 
 * [addUnitClauseOfPredicate P] is just like [{!Engine.assume} P], the 
 * difference being just in how the proofs are handled. In the case of 
 * {!Engine.assume} a proof of [P] must be provided. In the case of 
 * [addUnitClauseOfPredicate] you must provide a proof of false hypothetical 
 * on a proof of [P]. If [D] is a proof of false that uses a hypothesis named 
 * [H] of type [P], then the function you provide must create a proof of 
 * false that binds [H] (it is not hypothetical anymore). *)

(** The case analysis needs to call the Engine to perform global operations. 
 * We package those operations as follows *)
type engineCallbacks = {
    egAboutToFail: unit -> bool; (** We call this function when we are about 
                                  * to fail. Return true if the is some 
                                  * progress done by this function. May raise 
                                  * Contradiction. *)

    egPushHyp: Util.symbol -> Logic.pred -> unit;
      (** We call this function to announce a new hypothesis. This is used 
       * for proof checking. The hypothesis should be popped on undo *)

    egBroadcast: Edag.messagebody -> Proof.proof -> string -> unit;
      (** We call this function to broadcast a message. This function may 
       * raise Contradiction *)

    egSnapshot: unit -> unit; (** Mark a state to which we must undo *)

    egUndo: unit -> unit; (* Do an undo *)
        
  }    

(** Perform the case analysis using the current clauses. In case of success 
 * (we can prove false), calls the provided continuation. In case of 
 * failure return *) 
val caseAnalysis: engineCallbacks -> (Proof.proof -> unit) -> unit



val convertToDNF: Logic.pred -> Logic.pred
(** This is a function that we use to convert a predicate to DNF, which is 
 * more convenient for case analysis. This function descends in conjunction 
 * and universal quantification *)

val init: unit -> unit
(** A function called when we start proving a file *)

val finish: unit -> unit
(** A function called when we are done proving a file *)


val debug: bool ref
(** A flag that can be set to turn on debugging of the case module *)

val options: Options.optionDescr list
(** A number of command-line options for the Cases module *)
