(* This module contains data types and functions necessary for parsing the 
 * CVC input syntax (Stanford's Cooperative Validity Checker), and for 
 * generating the Input data structures used by the rest of the prover *)
module I = Input
module L = Logic
module H = Hashtbl

module E = Errormsg
open Pretty


let preprocess = ref false

type typ = 
    TId of string           (* A declared type *)
  | TReal 
  | TBool 
  | TTuple of typ list
  | TRecord of (string * typ) list
  | TEnum of string list (* An enumeration type *)
  | TFun of (typ * typ)
  | TArray of typ * typ

type expr = 
    EId of string
  | ENId of string (* A negated boolean constant *)

  | ENum of int
  | EBool of bool

  | EPlus of expr * expr
  | EMinus of expr * expr
  | EMult of expr * expr
  
  | EEq of expr * expr
  | ENeq of expr * expr
  | EGeq of expr * expr
  | ELeq of expr * expr
  | EGt of expr * expr
  | ELt of expr * expr

  | ENot of expr
  | EAnd of expr * expr
  | EOr of expr * expr
  | EXor of expr * expr
  | EImpl of expr * expr
  | EIff of expr * expr

  | EIf of expr * expr * expr
  | ELet of (string * typ * expr) list * expr

  | ETuple of expr list
  | EApp   of string * expr list 
  | ENApp  of string * expr list (* A negated boolean predicate *)

  | EAll of string * typ * expr (* Universal quantification *)
  | EExist of string * typ * expr (* Existential quantification *)

  | EUpd of expr * expr list * expr (* Update in a multi-dimensional array *)
  | ESel of expr * expr (* Select *)

  | EAnnot of expannot * expr

  | ELambda of (string * typ) list * expr

and expannot =
  | ATrigger of expr

type input = 
    ConstDecl of string list * typ
  | ConstInitDecl of string * typ * expr * Input.loc (* A constant with initializer *)
  | TypeConstDecl of string list
  | TypeConstInitDecl of string * typ
  | Query of expr * Input.loc
  | QueryInvalid of expr * Input.loc
  | Assert of expr * Input.loc
  | Retract of Input.loc
  | Print of expr 
  | Echo of string
  | Block of (string * string) list  * Input.loc
  | EndBlock of Input.loc
  | Eof  


let rec d_typ () = function
    TId n -> text n
  | TReal -> text "REAL"
  | TBool -> text "BOOL"
  | TTuple tl -> 
      dprintf "[@[%a@]]" (docList ~sep:(chr ',' ++ break) (d_typ ())) tl
  | TRecord tl -> 
      dprintf "[#@[%a@]#]" 
        (docList ~sep:(chr ',' ++ break) 
           (fun (n, t) -> dprintf "%s : %a" n d_typ t)) tl
  | TFun (t1, t2) -> dprintf "[%a -> %a]" d_typ t1 d_typ t2
  | TEnum nl -> dprintf "{@[%a@]}" (docList ~sep:(chr ',' ++ break) text) nl
  | TArray (tidx, telem) -> dprintf "ARRAY(%a -> %a)" 
        d_typ tidx d_typ telem
        


let rec d_exp () (e: expr) = 
  let d_binop (op: string) (e1: expr) (e2: expr) = 
    dprintf "(@[%a %s@? %a@])" d_exp e1 op d_exp e2
  in
  match e with 
    EId n -> text n
  | ENId n -> text "NOT n"
  | EBool b -> dprintf "%b" b
  | ENum n -> num n
  | ENot e -> dprintf "(NOT %a)" d_exp e
  | EPlus (e1, e2) -> d_binop "+" e1 e2
  | EMinus (e1, e2) -> d_binop "-" e1 e2
  | EMult (e1, e2) -> d_binop "*" e1 e2
  | EEq (e1, e2) -> d_binop "=" e1 e2
  | ENeq (e1, e2) -> d_binop "<>" e1 e2
  | EGeq (e1, e2) -> d_binop ">=" e1 e2
  | EGt (e1, e2) -> d_binop ">" e1 e2
  | ELeq (e1, e2) -> d_binop "<=" e1 e2
  | ELt (e1, e2) -> d_binop "<" e1 e2

  | EAnd (e1, e2) -> d_binop "AND" e1 e2
  | EOr (e1, e2) -> d_binop "OR" e1 e2
  | EXor (e1, e2) -> d_binop "XOR" e1 e2

  | EImpl (e1, e2) -> d_binop "IMPL" e1 e2
  | EIff (e1, e2) -> d_binop "IFF" e1 e2
        
  | EIf (b, et, ef) -> 
      dprintf "(@[if %a@?THEN %a@?ELSE %a@])" d_exp b d_exp et d_exp ef
  | EUpd (a, idx, v) -> 
      dprintf "(%a WITH %a := %a)"
        d_exp a
        (docList ~sep:nil (fun i -> dprintf "[%a]" d_exp i)) idx
        d_exp v
  | ESel (a, idx) -> dprintf "%a[%a]" d_exp a d_exp idx

  | ELet (nl, body) -> 
      dprintf "(@[let @[%a@]@?in %a@])" 
        (docList ~sep:(text "," ++ break) 
           (fun (n, t, e) -> dprintf "%s : %a = %a" n d_typ t d_exp e))
        nl
        d_exp body

  | EApp (f, args) -> 
      dprintf "%s%a" f d_exp (ETuple args)
  | ENApp (f, args) -> 
      dprintf "NOT %s%a" f d_exp (ETuple args)
  | ETuple exprs -> 
      dprintf "(@[%a@])"
        (docList ~sep:(chr ',' ++ break) (d_exp ())) exprs
  | EAll (x, t, body) -> 
      dprintf "(FOR@[ALL %s:%a.%a@])"
        x d_typ t d_exp body
  | EExist (x, t, body) -> 
      dprintf "(EXI@[STS %s:%a.%a@])"
        x d_typ t d_exp body 
  | EAnnot (a, e) -> 
      dprintf "(ANN@[OT %a,@?%a@])"
        d_annot a d_exp e
  | ELambda (params, body) -> 
      dprintf "(LAM@[DA (@[%a@]),@?%a@])"
        (docList ~sep:(chr ',' ++ break) 
           (fun (n, t) -> dprintf "%s : %a" n d_typ t)) params
        d_exp body

and d_annot () (a: expannot) = 
  match a with 
    ATrigger e -> dprintf "TRIG(%a)" d_exp e


(* Count the queries here *)
let nextQueryId = ref 0

(* Keep here the constants *)
let constantDecls: (string, typ) H.t = H.create 113
let typeConstantDecls: (string, unit) H.t = H.create 113

(* Keep here the global names *)
let names: (string, typ) H.t = H.create 113
let typeNames: (string, typ) H.t = H.create 113

(** We keep track of scopes *)
type what = 
    VLet of expr * env * L.exp option ref * L.pred option ref
                 (* This is a LET-bound variable, with the original 
                  * expression, its environment, and two memo locations for 
                  * the conversion to expression and to predicate *)

  | VBound       (* This is a quantified variable *)

and env = (string * what) list

(** Lookup a bound name, to be used in either expression or predicate 
 * context. Give a function for what to do if bound, and what to do if let *)
let lookupVar (env: (string * what) list) 
              ~(handlebound: int -> 'a)
              ~(handleconst: 'a)
              ~(handlelet: expr -> env -> 
                             L.exp option ref -> L.pred option ref -> 'a)
              (n: string) : 'a = 
  let rec loop (idx: int) = function
      (n', what) :: _ when n' = n -> begin 
        match what with 
          VBound -> handlebound idx
        | VLet (e,env,eo,po) -> handlelet e env eo po
      end
    | (_, what) :: rest -> 
        loop (if what = VBound then (idx + 1) else idx) rest
    | [] -> handleconst
  in
  loop 0 env


(* Convert an arithmetic expression. Must also have the list of bound 
 * variables and names.  *)
let rec convArith (env: env) (e: expr) : L.exp = 
  let ca (e: expr) : L.exp = convArith env e in 
  match e with 
    ENum n -> L.integer n
  | EId n -> 
      lookupVar env
        ~handlebound:(fun idx -> L.Var idx)
        ~handleconst:(L.Fun(n,[]))
        ~handlelet:(fun e' env' eo po -> 
          match !eo with 
            Some e'' -> e''
          | None -> begin 
              let e'' = convArith env' e' in 
              eo := Some e'';
              e''
          end)
        n

  | EPlus (e1, e2) -> L.plus (ca e1) (ca e2)
  | EMinus (e1, e2) -> L.minus (ca e1) (ca e2)
  | EMult (e1, e2) -> L.times (ca e1) (ca e2)
  | ELet (nl, body) -> 
      let env' = convLet env nl in 
      convArith env' body

  | EIf _ -> 
      E.s (E.unimp "If in convArith")

  | EApp(n, args) -> L.Fun(n, List.map ca args)

  | ESel (a, idx) -> 
      L.Fun("sel", [ ca idx; ca a ])

  | EUpd (a, idx, v) -> 
      (* The last index is the one being updated. The previous ones are used 
       * to find the array being updated *)
      let rec loop (arr: L.exp) = function
          [ updidx ] -> arr, ca updidx
        | selidx :: rest -> 
            loop (L.Fun("sel", [ ca selidx; arr ])) rest
        | [] -> E.s (E.bug "convArith: empty indices")
      in
      let arr, updidx = loop (ca a) idx in
      L.Fun("upd", [updidx; ca v; arr])

  | e -> E.s (E.unimp "convArith: %a" d_exp e)

and convLet (env: env) (bl: (string * typ * expr) list) : env = 
  match bl with 
    [] -> env
  | (n, t, e) :: rest -> 
      convLet ((n, VLet (e, env, ref None, ref None)) :: env) rest


(* Convert a predicate. Must also have the list of bound variables. *)
let rec convPred (env: env) (p: expr) : L.pred = 
  let cp (p: expr) : L.pred = convPred env p in
  let ca (e: expr) : L.exp = convArith env e in 
  match p with 
  | EImpl (p1, p2) -> L.Imply (cp p1, cp p2)
  | EAnd (p1, p2) -> begin
      match cp p1, cp p2 with
        L.And orl1, L.And orl2 -> L.And (orl1 @ orl2)
      | L.And orl1, p2 -> L.And (orl1 @ [p2])
      | p1, L.And orl2 -> L.And (p1 :: orl2)
      | p1, p2 -> L.And [ p1; p2 ]
  end

  | EEq (e1, e2) -> L.eq (ca e1) (ca e2)
  | ENeq (e1, e2) -> L.ne (ca e1) (ca e2)
  | EAll (var, _, p) -> 
      let p' = convPred ((var, VBound) :: env) p in 
      L.All (var, p')

  | EExist (var, _, p) -> 
      let p' = convPred ((var, VBound) :: env) p in 
      L.Exist (var, p')

  | EId n -> begin (* This name must be a bool *)
      lookupVar env
        ~handlebound:(fun idx -> L.ne (L.Var idx) L.zero)
        ~handleconst:(L.pFun n [])
        ~handlelet:(fun e' env' eo po -> 
          match !po with 
            Some p'' -> p''
          | None -> begin 
              let p'' = convPred env' e' in 
              po := Some p'';
              p''
          end)
        n
  end
  | ENId n -> begin (* Negated boolean constant *)
      lookupVar env
        ~handlebound:(fun idx -> L.eq (L.Var idx) L.zero)
        ~handleconst:(L.eq (L.Fun(n,[])) L.zero)
        ~handlelet:(fun e' env' eo po -> 
          match !po with 
            Some p'' -> p''
          | None -> begin 
              let p'' = L.negatePred (convPred env' e') in 
              po := Some p'';
              p''
          end)
        n
  end

  | EApp (n, args) -> L.pFun n (List.map ca args)

  | ENApp (n, args) -> L.eq (ca p) L.zero
      

  | EBool true -> L.Bool true
  | EBool false -> L.Bool false

  | EGeq (e1, e2) -> L.geq (ca e1) (ca e2)
  | EGt  (e1, e2) -> L.gt (ca e1) (ca e2)
  | ELeq (e1, e2) -> L.leq (ca e1) (ca e2)
  | ELt  (e1, e2) -> L.lt (ca e1) (ca e2)


  | EIff (e1, e2) ->
      let e1' = cp e1 in
      let e2' = cp e2 in
      L.Or [L.And [ e1'; e2']; L.And [L.Not e1'; L.Not e2']]

  | EOr (e1, e2) -> begin
      match cp e1, cp e2 with
        L.Or orl1, L.Or orl2 -> L.Or (orl1 @ orl2)
      | L.Or orl1, p2 -> L.Or (orl1 @ [p2])
      | p1, L.Or orl2 -> L.Or (p1 :: orl2)
      | p1, p2 -> L.Or [ p1; p2 ]
  end

  | ELet (nl, body) -> 
      let env' = convLet env nl in 
      convPred env' body

  | EIf (b, et, ef) ->
     L.And [L.Imply(cp b, cp et);
            L.Imply(cp (ENot b), cp ef) ]

  (** Push negation down *)
  | ENot(p) -> (L.negatePred (cp p))

  | EAnnot(a, p) -> 
      L.Pannot (convAnnot env a, cp p)

  | _ -> E.s (E.unimp "convPred: %a" d_exp p)

and convAnnot (env: env) (a: expannot) : L.pannot = 
  let cp (p: expr) : L.pred = convPred env p in
  let ca (e: expr) : L.exp = convArith env e in 
  match a with 
    ATrigger p -> L.ATrigger (cp p)
    
(* Remember here the assert statements *)
let open_asserts: int ref = ref 0
let open_blocks: int ref = ref 0

let nextGoalName () = 
  let gname = "Query" ^ (string_of_int !nextQueryId) in
  incr nextQueryId;
  gname
  
let rec parseNextInput (theparser: unit -> input) : I.input list = 
  let res = theparser () in
  Parsing.clear_parser ();
  let doAssert (p: L.pred) (l: E.location) : I.input list = 
    (* If this is a top-level assert, start a new goal *)
    incr open_asserts;
    if !open_asserts = 1 then begin
      [ I.StartGoal (nextGoalName (), l); I.Assume (p, l) ]
    end else 
      [ I.Assume (p, l) ]
  in

  (* Keep parsing until we see a query *)
  let rec doit res = 
    match res with
      Eof ->  [ I.Eof ]
          
    | ConstDecl (clst, t) -> 
        List.iter (fun n -> H.add constantDecls n t) clst;
        parseNextInput theparser
          
    | ConstInitDecl (n, t, e, l) -> begin
        H.add names n t;
        (* There are two cases here: either t is a function, or is not *)
        match t with 
          TFun(_, tres) -> begin
            (* We are dealing with a function. We will create a matching 
             * rule. *)
            match e with 
              ELambda (params, body) -> 
                (* Get the list of parameters as expressions *)
                let eparams: expr list = 
                  List.fold_right
                    (fun (n, _) acc -> EId n :: acc)
                    params
                    []
                in
                (* Make the expression n(params) *)
                let ne = EApp(n, eparams) in
                (* Construct
                    Trigger(have(n(params))):
                    n(params) <=> body
                *)
                let axbody = 
                  EAnnot(ATrigger(EApp("have", [ ne ])),
                         EIff(ne, body))
                in
                (* Now add the FORALL on top *)
                let p' = List.fold_left
                    (fun acc (n, t) -> EAll(n, t, acc))
                    axbody
                    params
                in
                doAssert (convPred [] p') l

            | _ -> assert false
          end
        | TReal | TId _ -> 
            (* We are dealing with a scalar. Just assert the equality *)
            let p' = convPred [] (EEq(EId n, e)) in 
            doAssert p' l
              
        | TBool ->
            (* We are dealing with a predicate *)
            let p' = convPred [] (EIff(EId n, e)) in 
            doAssert p' l
            
        | _ -> E.s (E.unimp "Initializers not fully supported yet")
    end
          
    | TypeConstDecl (clst) -> 
        List.iter (fun n -> H.add typeConstantDecls n ()) clst;
        parseNextInput theparser
          
    | TypeConstInitDecl (n, t) -> 
        H.add typeNames n t;
        parseNextInput theparser
          
    | Print e -> ignore (E.log "%a" d_exp e); parseNextInput theparser
          
    | Echo s -> 
        ignore (E.log "%s" s); 
        parseNextInput theparser
          
    | Query (e, l) | QueryInvalid (e, l) -> 
        (* If we are at top-level, start a new goal *)
        let e' = convPred [] e in 
        let i = match res with 
          Query _ -> I.Prove (e', l)
        | _ -> I.ProveInvalid (e', l)
        in
        if !open_asserts = 0 then begin
          [ I.StartGoal (nextGoalName (), l); 
            i;
            I.EndGoal l ]
        end else 
          [ i ]
            
    | Assert (e, l) -> doAssert (convPred [] e) l
            
    | Retract l -> begin (* Retract the last assert *)
        if !open_asserts = 0 then 
          E.s (E.error "Found retract but with no asserts");
        decr open_asserts;
        if !open_asserts = 0 then 
          [ I.UnAssume l; I.EndGoal l]
        else
          [ I.UnAssume l ]
    end
    | Block (pars, l) -> 
        incr open_blocks;
        [ I.Block (pars, l) ]
          
    | EndBlock l ->
        if !open_blocks = 0 then 
          E.s (E.error "Found EndBlock but with no Block");
        decr open_blocks;
        [ I.UnBlock l ]
  in
  doit res
          
let initParser () = 
  nextQueryId := 0;
  H.clear constantDecls;
  H.clear names;
  open_asserts := 0;
  open_blocks := 0;
  ()
  

