module L = Logic
module E = Errormsg
module H = Hashtbl
module IH = Inthash 
module U = Util
module P = Proof
module CL = Clist
module BM = Proof.Bitmap

open Pretty

let debug = ref false (** set with -debug DAG *)

let debugStats = ref false

let log fmt = ignore (E.log "DAG: "); E.log fmt

      (* These are the function symbols that can occur in nodes. Note that we 
       * maintain the invariant that two fSym are pointer-equal iff. they are 
       * equal. Use {!Edag.uniqueHead} for this purpose. *)
type fSym =
  BINOP of L.binop |
  MOD | DIV | CEIL | FLOOR |
  NOT |
  INT of int |
  UNINT of Util.symbol |
  VAR of Util.symbol |
  BIND | BRUIJN of int

(* The position in the Simplex tableau *)
type simplPos = Row of int  | Col of int | NotSimpl

(** We work with proof thunks, to generate proofs lazily *)
type proofthunk = unit -> P.proof 

(* The node type *)
type node = { id: int; (* A unique ID *)
              exp: L.exp; (* The expression that corresponds to this node. 
                           * This is always obtained from the head and the 
                           * .exp of the args *)
              head: int;  (* The (code of the) head function symbol. Use 
                           * {!Edag.uniqueHead} to obtain the code *)

	      args: node list; (* List of arguments *)
	      
	      mutable active:  bool; (** If the node is part of some active 
                                      * assertion. Valid only on a root.  *)
              
	      mutable eqc      : node; (* the equivalence class, a circular 
                                        * list*)
	      mutable root     : node; (* the class representative. Points to 
                                        * itself if it is the root. *)
              
              mutable rootwhy: eqproof; 
              (** This explains why the current root is equal to this node. 
               * This is an equality proof root = n *)
 
              (*** The following fields are only meaningful for a root **)
	      mutable size     : int;  (* size of the equivalence class. Only 
                                        * meaningful for a root. *)


	      mutable parents    : node circlist option; 
              (** the parents: nodes that have at least one argument in the 
               * same equivalence class to this one. We have all such nodes, 
               * not just their equivalence classes. Only meaningful for a 
               * root. *)


              mutable newParentsCallbacks: (node -> unit) circlist option;
              (** This function is called for each new active parent node of 
               * this node (a parent is a node that has an argument in the 
               * same equivalence class with this node). This function is 
               * called when the EDAG is closed. Only meaningful for a root. *)

              mutable mergeCallbacks: (unit -> unit) circlist option;
              (** This function is called after the node was merged. This 
               * function is called when the EDAG is closed. Only meaningful 
               * for a root. There is also a global mergeCallbacks. *)
              
              mutable forbidCallbacks: (unit -> unit) circlist option;
              (** This function is called after the node acquires more 
               * elements in the forbid list. This function is called when 
               * the EDAG is closed. Only meaningful for a root. *)
              
	      (* circular list of nodes known to be not-equal to this one. 
               * With each node, we keep a tuple (nodea,nodeb,pf_a_neq_b) 
               * that witnesses this disequality *)
	      mutable forbid   : 
                (node * (node * node * P.proof)) circlist option;
              
              (* The size of the forbid set. We use this so that we search 
               * the smallest of two sets *)
              mutable forbidsize: int;

	      (* Next is the Simplex state. *)
	      spos     : simplPos ref;
	    }
      
      
      (** A circular list. Has the advantage that it is easy to concatenate 
       * two lists. This is always non-empty, so you must use it with option 
       * to allow for empty lists *)
and 'a circlist = { 
            car: 'a;
    mutable cdr: 'a circlist; 
  }

and eqproof = 
  | Refl                   (* A proof by reflexivity *)
  | One of bool * P.proof  (* A proof. If false, then we have the Sym *)
  | Sym of eqproof
  | Concat of eqproof * eqproof

and message = 
    { body    : messagebody;            (* The predicate being asserted. This 
                                         * is most often an atom, but it 
                                         * might be the left-hand side of an 
                                         * implication, when that is not a 
                                         * conjunction. *)
      owner   : Util.symbol;            (* Who is responsible for producing a 
                                         * proof of this  *)

      proof:   Proof.proof;             (* The proof of this message *)

      fromgoal: bool;                   (* If it is from a goal. *)
    } 
      

and messagebody = 
    Intern of Logic.binop * node * node (** The body of the message is 
                                              * an interned node *)
  | Extern of Logic.pred                (** A full predicate *)


let genProofs = ref true (** Whether to generate proofs or not *)

let noProof = Proof.Unimplemented "not genProofs"

(** Keep here a pointer for generating proofs of equivalence of two nodes *)
let genEqProof: (node -> node -> Proof.proof) ref = 
  ref (fun _ -> assert false)

type neqwitness = node * node * P.proof
let genNeqProof: (node -> node -> neqwitness -> Proof.proof) ref = 
  ref (fun _ -> assert false)


let eqproofToProof (e1: L.exp) (e2: L.exp) (eq: eqproof) : P.proof = 
  let rec eqproofToPargs (dir: bool) acc = function
      Refl -> acc
    | One (direct, p) -> P.I (if direct = dir then 1 else -1) :: P.P p :: acc
    | Sym eq' -> eqproofToPargs (not dir) acc eq'
    | Concat (e1, e2) -> 
        eqproofToPargs dir (eqproofToPargs dir acc e1) e2
  in
  let eqs = eqproofToPargs true [] eq in
  match eqs with 
    P.I dir :: P.P p :: [] -> 
      if dir = 1 then p else 
      if dir = -1 then P.Symm p else
      P.mkArithEq e1 e2 eqs
  | eqs -> P.mkArithEq e1 e2 eqs

let rec eqProofSymConcat (r_a: eqproof) (r_b: eqproof) = 
  match r_a, r_b with 
  | Concat (p1, r1), Concat (p1', r2) when p1 == p1' -> eqProofSymConcat r1 r2
  | _, _ -> Concat (Sym r_a, r_b)

(** Generate the eqproof for two nodes that belong to the same class *)
let eqProofEqualNodes (n1: node) (n2: node) = 
  assert (n1.root == n2.root);
  eqProofSymConcat n1.rootwhy n2.rootwhy


  
(** A major hack. Remember here some functions that can give us the vars and 
 * hypotheses *)
let varsAndHyps: (unit -> (L.var list * (U.symbol * L.pred) list)) ref 
    = ref (fun _ -> assert false)

let proofChecker (msg: doc) (prf: Proof.proof) (prd: Logic.pred) : bool = 
  if not !genProofs then 
    true
  else begin
    let old = !P.printMagicProofWarning in
    P.printMagicProofWarning := !P.checkProofs;
    let vars, hyps = (!varsAndHyps)() in 
    let res = 
      try Proof.check vars hyps prf prd; true
      with E.Error -> begin
        ignore (E.warn "Proof checking failed: %a\n" insert msg);
        false
      end
    in
    P.printMagicProofWarning := old;
    res
  end

let proofInferencer (prf: Proof.proof) : Logic.pred = 
  assert (!genProofs);
  let vars, hyps = (!varsAndHyps)() in 
  Proof.infer vars hyps prf

let proofWellFormed (msg: doc) (prf: Proof.proof) : bool = 
  if not !genProofs then 
    true
  else begin
    let old = !P.printMagicProofWarning in
    P.printMagicProofWarning := !P.checkProofs;
    let res = 
      try ignore (proofInferencer prf); true
      with E.Error -> begin
        ignore (E.warn "Proof well-formedness failed: %a\n" insert msg);
        false
      end
    in
    P.printMagicProofWarning := old;
    res
  end
    

let uniqueHeads: (fSym, int) H.t = H.create 13
let headToFSym: fSym IH.t = IH.create 13
let lastHead = ref 0 
let uniqueHead (h: fSym) : int = 
  try H.find uniqueHeads h 
  with Not_found -> begin
    incr lastHead;
    H.add uniqueHeads h !lastHead;
    IH.add headToFSym !lastHead h;
    !lastHead
  end
  
let fSymOfHead (h: int) : fSym = 
  try IH.find headToFSym h
  with Not_found -> assert false


  
(* Make a dummy node *)
let dummyNode : node = 
  let n = 
    { id = -1; 
      exp = L.Fun("<dummy>", []);
      active = false;
      head = uniqueHead (INT 0);
      args = []; 
      eqc = (Obj.magic ());
      size = 0;
      root = (Obj.magic ());
      rootwhy = Refl;
      parents = None;
      newParentsCallbacks = None;
      mergeCallbacks = None;
      forbidCallbacks = None;
      forbid = None;
      forbidsize = 0;
      spos = ref NotSimpl } in
  n.eqc <- n;
  n.root <- n;
  n


let trueNode = ref dummyNode
let falseNode = ref dummyNode

let predicateOfMessageBody (b: messagebody) : L.pred = 
  match b with 
    Intern (bop, n1, n2) -> L.PNez(L.BinOp(bop, n1.exp, n2.exp))
  | Extern p-> p
  

let d_messagebody () (b: messagebody) = L.d_pred () (predicateOfMessageBody b)


  (* Pretty print a message *)
let d_message () msg = 
  dprintf "%s(%s:%a)"
    (if msg.fromgoal then "Goal" else "Assert")
    (U.symbolName msg.owner)
    d_messagebody msg.body


(* proof trace support; sorry for the cruft *)
let doProofTrace: bool ref = ref false
let proofTraceDepth: int ref = ref 0
let proofTraceInd () : string =
  "trace: " ^ (String.make !proofTraceDepth ' ') ^ (String.make !proofTraceDepth ' ')
let matchid = Util.registerSymbolName "MA"


(** Accumulate here the new messages *)
let newMessages: message list ref = ref []
let newMessageWhy (msg: message) (why: string) : unit =
  if !doProofTrace && msg.owner!=matchid then (
    if (why = "goal") then
      ignore (E.logg "%s-------------------------------------------------------\n"
                     (proofTraceInd ()));
    ignore (E.logg "%s%a: %s (from %s)\n"
                   (proofTraceInd ())
                   d_messagebody msg.body
                   why
                   (U.symbolName msg.owner));
  );

  match msg.body with
    Extern (L.Bool true) -> ()
  | _ -> (* Force the generation of the proofs *)
      if !P.checkProofs then 
        assert 
          (let p = predicateOfMessageBody msg.body in
          proofChecker (text "newMessage") msg.proof p);
      newMessages := msg :: !newMessages

let newMessage (msg: message) : unit =
  (newMessageWhy msg "(unknown reason)")

(* 
 * a stack based structure for snapshots and undos 
 *)
type undoElem = 
  | UndoRun of (unit -> unit)           (** A function to run at undo *)

  | Snapshot of int


let d_event () = function 
  | Snapshot _ -> text "*" 
  | UndoRun f -> text "run"



let undoStack : undoElem list ref = ref [ ]

let registerUndoRun (f: unit -> unit) = 
  undoStack := UndoRun f :: !undoStack

(** Mark a snapshot, return a unique ID *)
let snapshotId = ref 0
let snapshot () : int = 
  incr snapshotId;
  undoStack := (Snapshot !snapshotId) :: !undoStack;
  let oldDepth = !proofTraceDepth in
  (registerUndoRun (fun _ -> proofTraceDepth := oldDepth));
  !snapshotId

let checkCurrentSnapshot (mark: int) : bool = 
  let rec loop = function
      [] -> assert false
    | UndoRun _ :: rest -> loop rest
    | Snapshot id :: _ -> id = mark
  in
  loop !undoStack

let pushWithUndo (l: 'a list ref) (e: 'a) = 
  let oldl = !l in 
  l := e :: !l;
  registerUndoRun (fun _ -> match !l with 
                              [] -> assert false
                            | e' :: l' -> assert (e == e' && l' == oldl);
                                          l := l')

let setWithUndo (r: 'a ref) (v: 'a) = 
  let old = !r in 
  r := v;
  registerUndoRun (fun _ -> assert (!r == v); r := old)


let incrWithUndo (r: 'a ref) = 
  incr r;
  registerUndoRun (fun _ -> decr r)

(** A proof memoizer. Use this to ensure that a proof for a message is not 
 * generated more than once *)
let memoizeProof (f: unit -> Proof.proof) : (unit -> Proof.proof) = 
  let pp: Proof.proof option ref = ref None in 
  (fun _ -> 
    match !pp with 
      Some p -> p
    | None -> 
        let p = f () in 
        pp := Some p;
        registerUndoRun (fun _ -> pp := None);
        p)


(* sm: I noticed that 'newNodeId', below, is the function that 
 * is actually responsible for node id creation, so I commented
 * this one to make it clear it is not being used.
let id = ref 0
let newId () =
  incr id;
  !id
*)


let findCirclist (f: 'n -> bool)
                 (l: 'n circlist option) : 'n = 
  match l with 
    None -> raise Not_found
  | Some first -> 
      let rec loop (next: 'n circlist) = 
        if f next.car then next.car else
        if next.cdr == first then 
          raise Not_found
        else loop next.cdr
      in
      loop first

let foldCirclist (f: 'acc -> 'n -> 'acc) 
                 (acc: 'acc) 
                 (l: 'n circlist option) : 'acc = 
  match l with 
    None -> acc
  | Some first -> 
      let rec loop (acc: 'acc) (next: 'n circlist) = 
        let acc' = f acc next.car in
        if next.cdr == first then 
          acc'
        else loop acc' next.cdr
      in
      loop acc first
        
        (* Scan a circular list *)
let iterCirclist (f: 'a -> unit)
                 (cl: 'a circlist option) = 
  foldCirclist (fun () n -> f n) () cl


exception FoundNeqWitness of neqwitness
let forbiddenEquality (n1: node) (n2: node) : (unit -> P.proof) option = 
  (* Each node will be in the other's node forbid list. We will search in the 
   * shorter forbid list, but remember the original n1 and n2 for proofs *)
  let n1', n2' = 
    if n1.forbidsize > n2.forbidsize then 
      n2, n1
    else
      n1, n2
  in
  (* n1' has the shortest forbid list *)
  let n2r = n2'.root in 
  try 
    iterCirclist 
      (fun (n', w) -> 
        if n'.root == n2r then 
          raise (FoundNeqWitness w)
        else
          ())
      n1'.root.forbid;
    None
  with FoundNeqWitness w -> 
    Some (fun _ -> !genNeqProof n1 n2 w)

      
        (** Concatenate two circular lists.  *)
let concatenateCirclist (l1: 'a circlist option) (l2: 'a circlist option) = 
  match l1, l2 with 
  | _, None -> l1
  | None, _ -> l2
  | Some c1, Some c2 -> 
      assert (c1 != c2);
      let t = c1.cdr in c1.cdr <- c2.cdr; c2.cdr <- t; l1


        (** Split two circular lists. 
            After
              l1 = concatenateCirclist l1 l2 
            call
              l1 = splitCirclist l1 l2 
         **)
let splitCirclist (l1: 'a circlist option) (l2: 'a circlist option) = 
  match l1, l2 with
    _, None -> l1			(* No change was made to l1 *)
  | Some c1, Some c2 ->
      if c1 == c2 then
	None
      else begin
        let t = c1.cdr in
	c1.cdr <- c2.cdr;
        c2.cdr <- t;
        l1
      end
  | _ -> E.s (E.bug "splitCirclist: l1 is None and l2 is Some")
  

    (* Find in an equivalence class of a node *)
let rec findEqc (f: node -> bool) 
                (n: node) : node = 
  (* Apply it to the node itself *)
  if f n then n else begin
    (* Apply to the rest *)
    let rec loop (next: node) = 
      if next == n then 
        raise Not_found
      else
        if f next then next else loop next.eqc
    in
    loop n.eqc
  end

        
    (* Scan the equivalence class of a node and apply a function to all of 
     * them. *)
let rec foldEqc (f: 'acc -> node -> 'acc)
                (acc: 'acc)  
                (n: node) : 'acc = 
  let acc1 = f acc n in (* Apply it to the node itself *)
  (* Apply to the rest *)
  let rec loop (next: node) (acc: 'acc) = 
    if next == n then 
      acc
    else
      loop next.eqc (f acc next)
  in
  loop n.eqc acc1

let iterEqc (f: node -> unit) (n: node) : unit = 
  foldEqc (fun _ n -> f n) () n


       (** Add an element to a circular list *)
let addToCirclist (old: 'a circlist option) (what: 'a) : 'a circlist option = 
  (* make a new element *)
  let nf = { car = what; cdr = Obj.magic () } in 
  (* fix it to make sure the GC likes us *)
  match old with 
    None -> nf.cdr <- nf; Some nf
  | Some f -> (* Splice *) 
      nf.cdr <- f.cdr;  f.cdr <- nf; old

       (** Remove an element from a circular list.
           After
              l1 = addToCirclist l1 x
           call
              l1 = removeFromCirclist l1
        **)
let removeFromCirclist 
    (old: 'a circlist option) 
    (no: 'a option) (** If not None we can do some sanity checking *)
    : 'a circlist option = 
  match old with
    None -> E.s (E.bug "removeFromCirclist")
  | Some l -> 
      (* Check that we are removing that element indeed *)
      assert (match no with 
        Some n -> if n != l.cdr.car then false else true
      | None -> true);
      if l == l.cdr then (* It is a singleton *)
        None 
      else begin
        l.cdr <- l.cdr.cdr;
        old
      end
        

let headToString (h: int) = 
  match fSymOfHead h with 
    BINOP bop -> L.binop_to_string bop
  |  MOD   ->  "MOD"
  |  DIV   ->  "DIV"
  |  CEIL  ->  "CEIL"
  |  FLOOR ->  "FLOOR"
  |  (INT n) -> string_of_int n
  |  (VAR v) -> U.symbolName v
  |  (UNINT s) ->  U.symbolName s
  |  (NOT) -> "NOT"
  |  (BIND) -> "BIND"
  |  (BRUIJN n) -> "BRUIJN(" ^ string_of_int n ^ ")"

let headArgsToExp (h: int) (args: L.exp list) : L.exp = 
  match fSymOfHead h, args with 
    BINOP bop, [a1; a2] -> L.BinOp(bop, a1, a2)
  | UNINT name, _ -> L.Fun(U.symbolName name, args)
  | VAR v, _ -> L.NVar (U.symbolName v)
  | INT n, _ -> L.integer n
  | _ -> E.s (E.unimp "headArgsToExp: %s" (headToString h))
        

let splitExpHeadArgs (e: L.exp) : int * L.exp list = 
  let fsym, (args: L.exp list) = 
    match e with 
    | L.Int n -> (INT (L.to_int n)), []
    | L.NVar v -> (VAR (U.registerSymbolName v)), []
    | L.Var b -> (BRUIJN b), []
    | L.BinOp(bop, e1,e2) -> (BINOP bop), [e1; e2]

    | L.Fun(name,args) ->
        (UNINT (U.registerSymbolName name)), args
    | _ -> E.s (E.unimp "insertNodeExp: %a" L.d_exp e)
  in
  uniqueHead fsym, args
  

(** Print the top-level of a node *)
let d_nodeTop () (n: node) : doc = 
  let rec nodelevel (l: int) (n: node) = 
    if l <= 0 then num n.id
    else begin 
      if n.args = [] then 
        text (headToString n.head) (* num n.id *)
      else
        dprintf "%s(%a)" 
          (headToString n.head) 
          (docList (nodelevel (l - 1))) n.args
    end
  in
  nodelevel 4 n


let d_sPos () = function
    Row n -> dprintf "Row(%d)" n
  | Col n -> dprintf "Col(%d)" n
  | NotSimpl -> nil

let d_nodeExp () (n: node) : doc = 
  let rec d_own (level: int) (n': node) : doc = 
    num n'.id ++ text ":" ++ 
    (if n'.args = [] then 
      text (headToString n'.head)
    else
      if level > 3 then 
        dprintf "N%d" n'.id
      else
        dprintf "%s(@[%a@])" (headToString n'.head)
          (docList ~sep:(chr ',' ++ break)
             (d_own (level + 1))) n'.args)
  in
  d_own 0 n
          
let d_node () (n: node) : doc = 
  dprintf "%a@!\tParents=@[%a@]@!\tRoot=%d (Eqc=@[%a@])@!\tSpos=%a  active=%b@!\tForbid: @[%a@]@!"
    d_nodeExp n
    (docList num)
    (foldCirclist (fun acc x -> x.id :: acc) [] n.parents)
    n.root.id
    (docList num) 
    (foldEqc (fun acc x -> x.id :: acc) [] n)
    d_sPos !(n.spos)
    n.active
    (docList num) 
    (foldCirclist (fun acc (x, _) -> x.id :: acc) [] n.forbid)



					(* This is the whole graph *)
let edag : node list ref = ref []


let printGraph () =
  List.iter (fun n -> ignore (E.log "%a" d_node n)) !edag
(*
  ignore (E.log "EGraph history: ");
  ignore (E.log "%a@!" 
            (docList (fun x -> d_event () x)) !history)
*)

(** Call this function when something related to this node changes **)
let watchNode (msg: doc) (n: node) : bool = 
  let containsNodeId (id: int) = 
    try ignore (findEqc (fun n' -> n'.id = id) n.root); true 
    with Not_found -> false
  in
  if false then begin
    ignore (log "Watch node %a (%a)\n" d_nodeTop n insert msg);
    printGraph ()
  end;
  true

    
(** We allow people to insert hooks for various events. People should not be 
 * doing a lot of work in these callbacks. Note that the EDAG might not be in 
 * a consistent state when this is called. Better save these nodes and handle 
 * them later in your ASSUME ***)


(** We record a series of delayed thunks that must be called when the EDAG is 
 * closed *)
let delayedCallbacks: (unit -> unit) list ref = ref []
let registerCallback (f: unit -> unit) = 
  delayedCallbacks := f :: !delayedCallbacks
let runCallbacks () : unit = 
  let l = !delayedCallbacks in 
  delayedCallbacks := [];
  List.iter (fun f -> f ()) l

   (** Called after a node's active bit is set or reset. This is called when 
    * a node is asserted, or is merged with an active equivalence class. This 
    * function is called when the EDAG is closed. This function is called 
    * only for the root of an equivalence class. *)
let activateCallbacks: (node -> unit) list ref = ref []

let addActivateCallback (f: node -> unit) : unit = 
  pushWithUndo activateCallbacks f



  (** Called after a node is merged. Called both for the root that changes 
   * and for the one that does not change. Called only for roots. *)
let mergeCallbacks: (node -> unit) list ref = ref []
let addGlobalMergeCallback (f: node -> unit) = 
  mergeCallbacks := f :: !mergeCallbacks;
  registerUndoRun
    (fun _ -> 
      match !mergeCallbacks with 
        f' :: rest when f == f' -> mergeCallbacks := rest
      | _ -> assert false)

let addNewParentsCallback (which: node) (f: node -> unit) : unit = 
  let whichr = which.root in 
  whichr.newParentsCallbacks <- addToCirclist whichr.newParentsCallbacks f;
  registerUndoRun 
    (fun _ -> 
      whichr.newParentsCallbacks 
      <- removeFromCirclist whichr.newParentsCallbacks (Some f))

let addMergeCallback (which: node) (f: unit -> unit) : unit = 
  let whichr = which.root in 
  whichr.mergeCallbacks <- addToCirclist whichr.mergeCallbacks f;
  registerUndoRun 
    (fun _ -> 
      whichr.mergeCallbacks 
      <- removeFromCirclist whichr.mergeCallbacks (Some f))
  

let addForbidCallback (which: node) (f: unit -> unit) : unit = 
  let whichr = which.root in 
  whichr.forbidCallbacks <- addToCirclist whichr.forbidCallbacks f;
  registerUndoRun 
    (fun _ -> 
      whichr.forbidCallbacks 
      <- removeFromCirclist whichr.forbidCallbacks (Some f))
  

(* Announce that n got a new active parent what. This should only be called 
 * for a root n, and for each new what that has arguments in the equivalence 
 * class of n *)
let callbackNewParent (n: node) (what: node) = 
  assert (n.root == n);
  if what.root.active then begin 
(*
    ignore (E.log "Node %d:%a got new parent %d:%a\n" 
              n.id L.d_exp n.exp what.id L.d_exp what.exp);
*)
    iterCirclist 
      (fun (f: node -> unit) -> 
        registerCallback (fun _ -> f what)) 
      n.newParentsCallbacks
  end
    
let addToParents (n: node) (what: node) = 
  assert (n == n.root);
  n.parents <- addToCirclist n.parents what;
  assert (watchNode (text "addToParents") n);
  if true || n == n.root then begin
    callbackNewParent n what
  end

let removeFromParents (n: node)  (what: node) = 
  n.parents <- removeFromCirclist n.parents (Some what);
  assert (watchNode (text "removeFromParents") n)


let iterParents (n: node) (f: node -> unit) = 
  iterCirclist f n.parents


let concatenateNewParentsCallbacks (n: node) (what: node) = 
  n.newParentsCallbacks <- 
    concatenateCirclist n.newParentsCallbacks what.newParentsCallbacks

let splitNewParentsCallbacks (n: node) (what: node) = 
  n.newParentsCallbacks <- 
    splitCirclist n.newParentsCallbacks what.newParentsCallbacks

let concatenateMergeCallbacks (n: node) (what: node) = 
  n.mergeCallbacks <- 
    concatenateCirclist n.mergeCallbacks what.mergeCallbacks

let splitMergeCallbacks (n: node) (what: node) = 
  n.mergeCallbacks <- 
    splitCirclist n.mergeCallbacks what.mergeCallbacks

let concatenateForbidCallbacks (n: node) (what: node) = 
  n.forbidCallbacks <- 
    concatenateCirclist n.forbidCallbacks what.forbidCallbacks

let splitForbidCallbacks (n: node) (what: node) = 
  n.forbidCallbacks <- 
    splitCirclist n.forbidCallbacks what.forbidCallbacks


let nodeId = ref 0 
let newNodeId () = 
  incr nodeId;
  !nodeId


    
(** Activate a node, and all of its subnodes. Call only for a root. *)
let rec activateNode (a: node) : unit = 
  assert (a == a.root);
  if not a.active then begin
    a.active <- true;
    assert(watchNode (text "activateNode") a);
    List.iter (fun f -> registerCallback (fun _ -> f a)) !activateCallbacks;

    registerUndoRun 
      (fun _ -> 
        a.active <- false;
        List.iter (fun f -> registerCallback (fun _ -> f a)) 
          !activateCallbacks);
    (* scan all elements of the equivalence class, and activate their 
     * arguments. Also tell them about the new parents *)
    iterEqc 
      (fun n -> 
        List.iter 
          (fun arg -> 
            callbackNewParent arg.root n;
            activateNode arg.root) 
          n.args)
      a
  end


      
let rec loopPairs f l1 l2 = 
  match l1, l2 with 
  | [], [] -> true
  | [], _ -> false
  | _, [] -> false
  | h1::t1, h2::t2 -> f h1 h2 && loopPairs f t1 t2


					(* Tests congruence. *)

let congruent (head: int) (args: node list) (n: node) = 
  n.head == head &&
  loopPairs (fun n1 n2 -> n1.root == n2.root) n.args args
    

	                                (* Makes a node, in a class by itself*)
let makeNewNode (sym: int) (args: node list)  =
  (* Convert a node into an expression *)
  let rec map f = function [] -> [] | h::t->let h'=f h in h':: map f t in 
  let theexp = headArgsToExp sym (map (fun n -> n.exp) args) in 

  let newNode = { id   = newNodeId ();
                  active = false;
                  exp  = theexp;
		  head = sym;
		  args = args;
		  parents = None;
		  eqc   = dummyNode; (* Temporary *)
		  size  = 1;
		  root  = dummyNode; (* Temporary *)
                  rootwhy = Refl;
                  newParentsCallbacks = None;
                  mergeCallbacks = None;
                  forbidCallbacks = None;
		  forbid = None;
                  forbidsize = 0;
		  spos  = ref NotSimpl} 
  in
  newNode.eqc  <- newNode;
  newNode.root <- newNode;
  newNode
      

(** Keep a mapping of heads to nodes *)
let headToNode: node IH.t = IH.create 113


let mkProofFromThunk (pt: P.proof option) (e: L.exp) = 
  match pt with 
    None -> P.Refl e
  | Some p -> p

    (* Insert a new node in the EGraph, if necessary. The args are the 
     * result of inserting the arguments. See the comments for InsertExpNode 
     * for details *)
let insertHeadArgs 
    (e: L.exp) (** The actual expression we are inserting this node for. Used 
                * only for the proofs *)
    (ihead : int) (** The head of the expression *)

    (args: (node * P.proof option) list) 
           (** The result of inserting the arguments *)

    : node * P.proof option =

  (** This actually adds the node to the EDAG *)
  let addNodeGraph (sym: int) (args: node list) : node =
    let n = makeNewNode sym args in
    (* ignore (E.log "added new node %a\n" L.d_exp n.exp); *)
    edag := n :: !edag;

    IH.add headToNode sym n;

    (* Adjust the parents of all arguments and of their roots *)
    List.iter
      (fun arg -> addToParents arg.root n)
      args;

    registerUndoRun (fun _ -> 
      if !debug then 
        ignore (log "Removing node %a\n" d_nodeExp n);
      (* Make sure this is the last one we have added *)
      assert (IH.mem headToNode n.head && n == IH.find headToNode n.head);
      IH.remove headToNode n.head;
      (* Delete from the edag *)
      (match !edag with
        n' :: rest when n == n' -> edag := rest
      | _ -> E.s (E.bug "Empty edag when retracting node"));
      
      (* Remove the parents, in the opposite order *)
      List.iter
        (fun arg -> removeFromParents arg.root n)
        (List.rev n.args));

    n
  in

  match args with 
    [] -> begin (** Is a leaf. No need to worry about congruences. *)
      try
        (IH.find headToNode ihead), None
      with Not_found -> begin
        let leaf = addNodeGraph ihead [] in
        leaf, None
      end
    end
        
  | (firstarg, _) :: _ -> begin

      (* Let's find if there is a congruent node in the EDAG already. 
       * Suffices to check among the parents of the first argument *)
      let (n: node), (identical: bool) = 
        try 
          (* we'll store here if we found a node whose arguments are identical 
           * to args *)
          let is_identical = ref false in 
          let n = 
            findCirclist 
              (fun p -> 
                (** See if p is identical or congruent with (ihead, args) *)
                let p_identical = ref true in 
                let congr: bool = 
                  p.head == ihead &&
                  loopPairs 
                    (fun pa (a, _) ->
                      p_identical := (pa == a) && !p_identical;
                      pa.root == a.root)
                    p.args args 
                in
                if congr then begin (* We found one *)
                  is_identical := !p_identical;
                  true
                end else
                  false)

              firstarg.root.parents
          in
          n, !is_identical
        with  Not_found -> begin
	  (* Create the node *)
          let rec map f = function [] -> [] 
                                 | h::t->let h'=f h in h':: map f t in 
	  let new_n = addNodeGraph ihead (map fst args) in
          new_n, true
        end
      in
      
      let prf = 
        if identical && (List.for_all (fun (_, po) -> po == None) args) then
          (* All arguments are identical also *)
          None
        else begin 
          (* Must build a congruence proof *)
          let rec prargs (args: (node * P.proof option) list)
                         (nargs: node list) : P.proof list = 
            match args, nargs with 
              [], [] -> []
            | (a, pe_a) :: args',  na :: nargs' -> 
                let pe_a: P.proof = mkProofFromThunk pe_a a.exp in
                let pa_na: P.proof = (!genEqProof) a na in 
                let pe_na = 
                  match pe_a, pa_na with 
                    P.Refl _, _ -> pa_na
                  | _, P.Refl _ -> pe_a
                  | _ -> P.Trans (a.exp, pe_a, pa_na)
                in
                pe_na :: (prargs args' nargs')
            | _ -> assert false
          in
          let p = if !genProofs then P.Cong(e, n.exp, prargs args n.args) 
                  else noProof in 
          Some p
        end
      in

      (* Do a sanity check *)
      assert (
      if !genProofs then 
        let prf' = mkProofFromThunk prf e in
        proofChecker (text "after insertHeadArgs") prf' (L.eq e n.exp)
      else
        true);
      n, prf
  end


(** Insert an expression into the EDAG. Return (n, None) if node n 
 * identically represents e (n.exp = e). Or, return (n, Some p), if n 
 * represents an expression congruent to e, and p is a proof for 
 * the equality e = n.exp (in this direction) *)
let rec insertExpNode (e: L.exp) : node * P.proof option =
  let (fsym: int), (args: L.exp list) = splitExpHeadArgs e in 
  let rec map f = function [] -> [] | h::t->let h'=f h in h':: map f t in 
  insertHeadArgs e fsym (map insertExpNode args)
 

					(* Undo stuff. Optionally, give the 
                                         * last element to undo, tested by 
                                         * pointer equality *)

let undo (mark: int option) =
  (* Scan the history backwards *)
  let rec loop = function 
    | [] -> E.s (E.bug "undo history is empty")
    | Snapshot id :: resth -> begin
        match mark with 
          Some id' when id <> id' -> (* Keep going *) loop resth
        | _ -> (* Done *) undoStack := resth
    end
    | UndoRun f :: resth -> f (); loop resth

  in
  loop !undoStack;

  runCallbacks ()

type proofTrans = (P.proof -> P.proof) option (** None, means the identity *)

let applyProofTrans (pt: proofTrans) (p: P.proof) = 
  match pt with 
    None -> p
  | Some pp -> pp p


let matchingDisabled: string option ref = ref None

(** This will be turned on when matching is disabled (to prevent matching cycles) 
 **)

(** Transform a predicate into a messagebody. Because when we insert nodes 
 * into the EDAG we might use congruent but not identical nodes, we produce 
 * also two proof transformers. The first one, given a proof of the 
 * predicate, produce a proof of the message body. The second one is the 
 * inverse. *)
let messageBodyOfPredicate (pred: L.pred)
    : messagebody * proofTrans * proofTrans = 
  match pred with 
  | L.PNez(L.BinOp(bop, e1, e2) as e) -> 
      let n1, pe1n1 = insertExpNode e1 in 
      let n2, pe2n2 = insertExpNode e2 in 
      runCallbacks ();
      let mb = Intern (bop, n1, n2) in
      let ptrans1, ptrans2 = 
        match pe1n1, pe2n2 with 
          None, None -> None, None
        | _ -> begin
            let pred_msg (pred_prf: P.proof) =
              let p1 = mkProofFromThunk pe1n1 e1 in 
              let p2 = mkProofFromThunk pe2n2 e2 in 
              P.PCong (pred_prf, 
                       [ P.Cong(e, 
                                L.BinOp(bop, n1.exp, n2.exp),
                                [ p1; p2 ]) ])
            in
            let msg_pred (msg_prf: P.proof) =
              let p1 = mkProofFromThunk pe1n1 e1 in 
              let p2 = mkProofFromThunk pe2n2 e2 in 
              P.PCong (msg_prf, 
                       [ P.Cong(L.BinOp(bop, n1.exp, n2.exp),
                                e, 
                                [ P.Symm p1; P.Symm p2 ]) ])
            in
            Some pred_msg, Some msg_pred
        end
      in
      mb, ptrans1, ptrans2
      
  | L.PFun (n, args) -> E.s (E.bug "The parser should never produce PFun")

  | _ -> Extern pred, None, None


let expSize (e: L.exp) : int =
  let psz = ref 0 in 
  ignore (L.visitLogicExp 
    (object
      inherit L.nopLogicVisitor
      method vexp (e: L.exp) = 
        incr psz;
        L.DoChildren
    end)
    e);
  !psz


(** Find the smallest representative of an equivalence class. Memoize the 
 * results and drop the memo table when you undo *)
let debugWitness = false

let witnessDepth = ref 8

let smallestMemoForClass: (L.exp * int) IH.t = IH.create 13
let smallestMemoForNode: (L.exp * int) IH.t = IH.create 13

let tempVarRegexp = Str.regexp_string "___"
let isTempName (s: string) : bool = 
  try ignore (Str.search_forward tempVarRegexp s 0); true
  with Not_found -> false

let depthExceededCount = ref 0 
let depthExceeded () : L.exp * int = 
  incr depthExceededCount;
  L.Fun(".." ^ string_of_int !depthExceededCount ^ "..", []), 1

    (** Find the smallest expression that uses this node's head and 
     * equivalent args *)
let rec findSmallestWithSizeForNode (recursive: int list) 
                                    (depth: int)
                                    (n: node) : L.exp * int = 
  (** Construct the list of smallest arguments, and the sum of their 
   * sizes *)
  if debugWitness then 
    ignore (E.log "(enter SmallestForNode %a\n" L.d_exp n.exp);
  let (e', sz') as res = 
    try IH.find smallestMemoForNode n.id
    with Not_found -> begin 
      let totsz = ref (getSizeForHead n.head n.args) in
      let args = 
        let rec map f = function [] -> [] | h::t->let h'=f h in h':: map f t in 
        map 
          (fun a -> 
            let a', a'sz = 
              findSmallestWithSizeForClass recursive (depth + 1) a.root in 
            (* make the size of arguments at least 0, to ensure monotony *)
            let a'sz = if a'sz < 0 then 0 else a'sz in
            totsz := a'sz + !totsz;
            a')
          n.args
      in
      let e = headArgsToExp n.head args in 
      let e' = finishSmallest e in 
      IH.replace smallestMemoForNode n.id (e', !totsz);
      (e', !totsz)
    end
  in
  if debugWitness then 
    ignore (E.log "  exit SmallestForNode %a is %a (sz=%d))\n" 
            L.d_exp n.exp L.d_exp e' sz');
  res

      (* For the head. We count constants as size -1, and the constants that 
       * do not have ___ in their name as size 0. The other constants are 
       * size 1. *)
and getSizeForHead (h: int) (args: node list) : int = 
  match fSymOfHead h with 
    INT _ -> assert (args == []); -1
  | (UNINT v) | (VAR v) when args == [] ->
      let vn = U.symbolName v in 
      if vn = "true" then 
        -1
      else
        if isTempName vn then 1 else 0
        
  | _ -> 1 (* 1 for the head *)
  

(** do some simplification for smallest nodes *)
and finishSmallest (e: L.exp) : L.exp = 
  match e with 
  | L.Fun(n, args) -> begin

      let rec map f = function [] -> [] | h::t->let h'=f h in h':: map f t in 
      match n, map finishSmallest args with 
        "sel", [a; m] -> begin
          let n = sprint 1000 (dprintf "%a.%a" L.d_exp a L.d_exp m) in
          L.Fun(n, [])
        end

      | "upd", [a; v; m] -> begin
          let n = 
            sprint 1000 (dprintf "%a[%a:=%a]" L.d_exp m L.d_exp a L.d_exp v) in
          L.Fun(n, [])
      end

      | _, args' -> L.Fun(n, args')
  end
  | _ -> e
  

  
    (** Fins the smallest expression for an entire class *)
and findSmallestWithSizeForClass (recursive: int list) 
                                 (depth: int) (* The depth at which we are *)
                                 (n: node) : L.exp * int = 
  assert (n == n.root);
  
  if debugWitness then 
    ignore (E.log "(enter SmallestForClass of %a at depth %d\n" 
              L.d_exp n.exp depth);
  (* See if this is recursive *)
  let (e, sz) as res = 
    if List.mem n.id recursive then begin
      (* There must be a leaf in each recursive class *)
      try 
        let n' = findEqc (fun n' -> n'.args == []) n in
        (* We found the leaf *)
        (finishSmallest n'.exp, getSizeForHead n'.head n'.args)
      with Not_found -> begin
(*
        ignore (E.warn "Cannot find a leaf in recursive class of %a\n" 
                  L.d_exp n.exp);
*)
        (finishSmallest n.exp, 100000)
      end
    end else begin 
      try 
        IH.find smallestMemoForClass n.id
      with Not_found -> begin
        (* See if we have exceeded the depth *)
        if depth >= !witnessDepth then begin 
          depthExceeded ()
        end else begin 
          (* Scan the entire equivalence class *)
          let res =
            foldEqc 
              (fun (old, oldsz) n' -> 
                let e', sz = 
                  findSmallestWithSizeForNode (n.id :: recursive) 
                    (depth + 1) n' in 
                if sz < oldsz then (e', sz) 
                else (old, oldsz))
              (finishSmallest n.exp, 10000000)
              n.eqc
          in
          IH.replace smallestMemoForClass n.id res;
          res
        end
      end
    end
  in
  if debugWitness then 
    ignore (E.log "  exit SmallestForClass of %a is %a (sz=%d))\n" 
            L.d_exp n.exp L.d_exp e sz);
  res


and findSmallestForClass (n: node) : L.exp = 
  let e, _ = findSmallestWithSizeForClass [] 0 n.root in 
  e

and findSmallestForNode (n: node) : L.exp = 
  let e, _ = findSmallestWithSizeForNode [] 0 n in 
  e


(** Save here the witness *)
let theWitness: doc list ref = ref []
let clearWitness () = theWitness := []
let accumWitnessPred (p: L.pred) = 
  match p with 
    L.PNez(L.BinOp(L.SetEq,e1,e2)) when e1 = e2 -> ()
  | e -> 
      let d = L.d_pred () p in 
      theWitness := d :: !theWitness

let accumWitness (d: doc) = 
  theWitness := d :: !theWitness

exception Interrupted of string


type testResult = 
    NotDetermined
  | Determined of bool * Proof.proof (* Determined(true,p) means that the 
                                      * result is "true" and the proof is p. 
                                      * Determined(false, p) means that the 
                                      * result is "false" and the proof of 
                                      * the negated predicate is p. *)

let messageTesters: (messagebody -> testResult) list ref = ref []

(** Register here a list of message testers *)
let registerMessageTester (f: (messagebody -> testResult)) : unit = 
  messageTesters := f :: !messageTesters
                           

  (** Do a quick test of a messagebody. Return (false, None) if it is not 
   * determined. Return (true, None) if it is determined to be true. Return 
   * (true, Some f) when it is determined to be false, where f is a function 
   * that given a proof of the message, will produce a proof of falsehood *)
let testMessageBody (msg: messagebody) : testResult = 
  let rec loopTestMsg = function
      [] -> NotDetermined
    | f :: rest -> 
        let r = f msg in if r != NotDetermined then r else loopTestMsg rest
  in
  loopTestMsg !messageTesters



let useFreshHyps = ref false (** Whether to use always fresh hyps *)

(* We have here some code that tries to generate a compact range of 
 * hypothesis names. This is only active if useFreshHyps is false.  *)
let firstHypothesis: U.symbol ref = ref 0 
let lastHypothesis: U.symbol ref = ref 0 
let nextHypothesis: U.symbol ref = ref 0
let maxHypothesisUsed: U.symbol ref = ref 0

let setUseFreshHyps (b: bool) = 
  useFreshHyps := b;
  if b then begin
    firstHypothesis := 0; lastHypothesis := (-1)
  end else begin
    let nrHypotheses = 10000 in
    firstHypothesis := 
      U.registerSymbolRange nrHypotheses (fun i -> "H" ^ (string_of_int i));
    lastHypothesis := !firstHypothesis + nrHypotheses - 1;
    nextHypothesis := !firstHypothesis;
    maxHypothesisUsed := !firstHypothesis - 1;
  end


let genNewHypothesisSymbol (base: string) : U.symbol = 
  if !useFreshHyps then 
    U.registerSymbolName (L.newVar base)
  else begin
    if !nextHypothesis > !lastHypothesis then 
      E.s (E.bug "Too many hypotheses\n");
    incrWithUndo nextHypothesis;
    let r = !nextHypothesis - 1 in 
    if r > !maxHypothesisUsed then 
      maxHypothesisUsed := r;
    r
  end

(* This function computes the free hypotheses in a proof, if not already 
 * computed *)
let rec computeFreeHyps (rs: P.relset) (p: P.proof) : unit = 
  if not rs.P.relcomputed then begin
    (** Keep a set of bound hypotheses *)
    let bound = BM.make (BM.size rs.P.relbm) in 
    P.scanHyps
      ~pushBound:(fun h -> 
        let h = h - rs.P.relfirst in 
        assert (h >= 0 & h < BM.size rs.P.relbm);
        let wasset = BM.testAndSetTo bound h true in 
        assert (not wasset);
        true)
      ~popBound:(fun h -> 
        let h = h - rs.P.relfirst in 
        assert (h >= 0 & h < BM.size rs.P.relbm);
        BM.setTo bound h false)
      ~seenHyp:(fun h -> 
        let h' = h - rs.P.relfirst in 
        if not (BM.test bound h') then begin
          assert (h' >= 0 & h' < BM.size rs.P.relbm);
          BM.setTo rs.P.relbm h' true
        end)
      ~seenFreeHyps:(fun rs' p' -> 
        (* Compute the free hyps recursively *)
        computeFreeHyps rs' p'; 
        ignore (BM.union_except rs.P.relbm rs'.P.relbm bound);
        (* Do not scan further *)
        false)
      
      p;
    
    rs.P.relcomputed <- true;
  end


(* Generate a FreeHyps node *)
let mkFreeHyps (p: P.proof) : P.proof = 
  let rs = { 
    P.relfirst = !firstHypothesis;
    P.relcomputed = false;
    P.relbm = BM.make (!maxHypothesisUsed + 1 - !firstHypothesis)
  } 
  in
  (* precompute the bitmap *)
  if not !P.checkProofs then begin 
    computeFreeHyps rs p;
    (* Now drop the proof *)
    P.FreeHyps(rs, P.Magic ("dropped due to FreeHyps", L.Bool false))
  end else 
    P.FreeHyps(rs, p)
  
      
     
let hasHyp (prf: P.proof) (h: U.symbol) : bool = 
  let res = 
    try
      P.scanHyps 
        ~pushBound:(fun i -> 
          if i = h then 
            false (* No need to scan the rest, we know h is not free *)
          else
            true)
        ~popBound:(fun i -> ())
        ~seenHyp:(fun s -> 
          if s = h then raise Not_found)
        
        ~seenFreeHyps:(fun rs p -> 
          computeFreeHyps rs p;
          if BM.test rs.P.relbm (h - rs.P.relfirst) then 
            raise Not_found;
          (* Anyway, do not scan anymore *)
          false)

        prf;
      false
    with Not_found -> true
  in
  res


(** Create a new alpha-unique name. *)
let paramAlphaTable: (string, unit Alpha.alphaTableData ref) H.t = H.create 13
let newAlphaName (old: string) : string = 
  let undo = ref [] in
  let f = 
    fst (Alpha.newAlphaName paramAlphaTable (Some undo) old ()) in
  registerUndoRun (fun _ -> Alpha.undoAlphaChanges paramAlphaTable !undo);
  f
  



let initialize () = 
  edag := [];
  matchingDisabled := None;

  newMessages := [];
  activateCallbacks := [];
  delayedCallbacks := [];
  mergeCallbacks := [];
  undoStack := [];
  IH.clear smallestMemoForClass;
  IH.clear smallestMemoForNode;
  IH.clear headToNode;

  messageTesters := [];

  H.clear paramAlphaTable;

  ()
  
  
