
module L = Logic
module P = Proof
module E = Errormsg
module H = Hashtbl

open Pretty

let debug = ref false
let debugJDK = false
let mustPrint n = debugJDK || 
               (String.length n >= 5 && String.sub n 0 5 <> "java/")


let objectOverhead = 8
let vtOverhead = 16
let arrayLenOffset = 16
let arrayDataOffset = 20

(***** Class paths *****)
let classPath : string list ref = ref []
let sepChar, dirSep =
  if Sys.os_type = "Win32" || Sys.os_type = "Cygwin" then
    ';', "\\"
  else
    ':', "/"

      
let setClassPath classpath = 
  let rec loop start = 
    try
      let sepPos = String.index_from classpath start sepChar  in 
      (String.sub classpath start (sepPos - start)) :: loop (sepPos + 1)
    with Not_found -> 
      [String.sub classpath start (String.length classpath - start)]
  in
  classPath := loop 0;
  if !debug then
    ignore (E.log "JV:setting classpath to @[%a@]\n"
              (docList text) !classPath)

  (* given the internal name of a Java class file, try to open it. Raise 
   * Not_found if the file does not exists  *)
let openClassFile name : (string * in_channel) = 
  let cname = name ^ ".class" in
  let rec loop = function
      [] -> raise Not_found
    | cp :: rest -> 
        let fname = cp ^ dirSep ^ cname in
        try
          (fname, open_in_bin fname)
        with _ -> loop rest
  in
  loop !classPath

(***** IO functions *****)
let fetch1 chan : int = 
  int_of_char (input_char chan)
let fetch2 chan : int = 
  let h = fetch1 chan in
  let l = fetch1 chan in
  l + (h lsl 8)

let fetch4 chan : int = 
  let h = fetch2 chan in
  let l = fetch2 chan in
  if h lsr 15 != 0 then 
    E.s (E.unimp "Trying to fetch a u4 that does not fit in an int")
  else
    l + (h lsl 16)

let skip4 chan : unit = 
  ignore (input_char chan);
  ignore (input_char chan);
  ignore (input_char chan);
  ignore (input_char chan)

  
(**** Classes ****)
type loadState = 
  | NotLoaded
  | BeingLoaded
  | Loaded


type jclass = 
    { filename : string;                   (* The file name *)
      cname  : string;                   (* The internal name of the class *)
      csymbol : string;                 (* The symbol name of the class *)

      mutable loadState : loadState;   (* The fields that follow are only 
                                        * guaranteed to be correct when 
                                        * loadState = Loaded *)

      mutable cflags     : int;
      mutable super      : jclass option;
      mutable interfaces : jclass array;
      mutable methods    : method_info array;
      mutable fields     : field_info array;

      mutable instanceSize : int;
      mutable vTableSize   : int;

    } 

and field_info = 
    { fname   : string;
      mutable fty     : jtype;
      ftydesc : string;
      fflags  : int;
      foffset : int;                    (* Offset into the instance, or 0 if 
                                         * not an instance field *)
    } 
      
and method_info = 
    { mname   : string;
      mutable mty  : jtype option * jtype list; (* Result, arguments *)
      mtydesc : string;
      mflags  : int;
      moffset : int;                    (* Offset in the virtual table, or 0 
                                         * if not a virtual method *)
      mindex  : int;                    (* Index into the method table of the 
                                         * class *)
    } 
      
and jtype = 
    Bool 
  | Int
  | Double
  | Long
  | Float
  | Short
  | Byte
  | Char
  | Array of jtype
  | Class of jclass
  | String                              (* Special case for string
					   initializers *)


type constantEntry = 
    CP_None                             (* Most entries are not interesting. 
                                         * So we just use CP_None for them *)
  | CP_Utf8 of string
  | CP_Class of jclass
  | CP_ClassRef of int


let dummyMethod = 
  { mname        = "<dummy method>";
    mty          = None, [];
    mtydesc      = "";
    mflags       = 0;
    moffset      = 0;
    mindex       = 0;
  }


let dummyField = 
  { fname        = "<dummy field>";
    fty          = Int;
    ftydesc      = "";
    fflags       = 0;
    foffset      = 0;
  } 

let dummyClass =     
    { filename   = "<dummy filename>";
      cname      = "<dummy class>";
      csymbol    = "dummy_symbol";
      loadState  = NotLoaded;
      cflags     = 0;
      super      = None;
      interfaces = Array.create 0 (Obj.magic 0);
      methods    = Array.create 0 dummyMethod;
      fields     = Array.create 0 dummyField;
      instanceSize = 0;
      vTableSize = 0;
    } 

(*** Flags ***)
let isStatic f      = (f land 0x0008) <> 0
let isInterface f   = (f land 0x0200) <> 0
let isPublic f      = (f land 0x0001) <> 0

let isConstr mname  = mname = "<init>"
let isStaticInit mname  = mname = "<clinit>"


(***** Symbol manipulation ****)
type symbol = 
    Constr of jclass * jtype list
  | Method of jclass * int   (* The index into the method array *)
  | Field  of jclass * int   (* The index into the field array *)
  | Const  of jclass * jtype * int * int  
                             (* The type and how many and the index in the 
                                class descriptor *)                     
  | StaticInit of jclass
(*  | ClassDesc of jclass *)
  | VirtualMT of jclass

let dsym () = function
    Constr (jc, _) -> dprintf "CO(%s)" jc.cname
  | Method (jc, i) -> dprintf "M(%s,%s)" jc.methods.(i).mname jc.cname
  | Field (jc, i) -> dprintf "F(%s,%s)" jc.fields.(i).fname jc.cname
  | Const (jc, jt, i, _) -> dprintf "C(%d,%s)" i jc.cname
  | StaticInit jc -> dprintf "SI(%s)" jc.cname
(*  | ClassDesc jc -> dprintf "CDT(%s)" jc.cname *)
  | VirtualMT jc -> dprintf "VT(%s)" jc.cname



   (* The parser *)

let loadedSymbols : (string, symbol) H.t = H.create 123

let rec typeToString = function
    Byte -> "B"
  | Char -> "C"
  | Double -> "D"
  | Float -> "F"
  | Int -> "I"
  | Long -> "J"
  | Short -> "S"
  | Bool -> "Z"
  | Array t -> "A" ^ typeToString t
  | Class c -> "L" ^ c.csymbol ^ "X"
  | String -> "T"

and argsToString = function
    [] -> ""
  | args -> 
      let rec loop = function
          [] -> ""
        | t :: rest -> typeToString t ^ loop rest
      in
      loop args

and classToString n = 
    let rec loop i = 
      try
        let ei = String.index_from n i '/' in
        let len = ei - i in
        let comp = String.sub n i len in
        string_of_int len ^ comp ^ loop (ei + 1)
      with Not_found -> begin
        let comp = String.sub n i (String.length n - i) in
        string_of_int (String.length comp) ^ comp
      end
    in
    loop 0

and symbolToString = function
    Constr (c, args) -> "_" ^ c.csymbol ^ "_CO" ^ argsToString args
  | Method (c, idx) -> 
      assert (idx < Array.length c.methods);
      let m = c.methods.(idx) in
      let ret, args = m.mty in
      "_" ^ c.csymbol ^ "_M" ^ m.mname ^ argsToString args
  | Field (c, idx) -> 
      assert (idx >= 0 && idx < Array.length c.fields);
      let f = c.fields.(idx) in
      assert (isStatic f.fflags);
      "_" ^ c.csymbol ^ "_S" ^ f.fname
  | Const (c, t, count, index) -> 
      "_" ^ c.csymbol ^ "_CT" ^ string_of_int index ^ "_" ^ 
      string_of_int count ^ typeToString t
  | StaticInit c -> 
      "_" ^ c.csymbol ^ "_I"
(*  | ClassDesc c ->
      "_" ^ c.csymbol ^ "_C" *)
  | VirtualMT c ->
      "_" ^ c.csymbol ^ "_vt"

let addSymbol c s = 
  let sname = symbolToString s in
  H.add loadedSymbols sname s; 
  if !debug && mustPrint c.cname then 
    ignore (E.log "JV:Defining symbol %s for %a\n" sname dsym s)


(*** Calculate the type of a symbol ****)
let binary n e1 e2 = L.Fun(n, [e1;e2])
let nullary n = L.Fun(n, [])
let unary n e = L.Fun(n, [e])

let rec typeToLF = function
    Bool -> nullary "jbool"
  | Byte -> nullary "jbyte"
  | Char -> nullary "jchar"
  | Int -> nullary "jint"
  | Short -> nullary "jshort"
  | Long -> nullary "jlong"
  | Float -> nullary "jfloat"
  | Double -> nullary "jdouble"
  | Array t -> unary "jarray" (typeToLF t)
  | Class c -> unary "jinstof" (nullary (symbolToString (VirtualMT c)))
  | String -> unary "jinstof" (nullary "_4java4lang6String_vt")

let typeOfSymbol = function
  | Field (c, idx) -> 
      let f = c.fields.(idx) in
      let t = typeToLF f.fty in
      L.Fun("jstatic", [t]), P.G("jStatic", [P.E t])

  | (VirtualMT c) as s -> 
      let cname = symbolToString s in
      L.Fun("jclasstype", []), P.G("jClasstype", [P.E (L.Fun(cname, []))])

  | Const (c,ty,count,idx) -> 
      let t = typeToLF ty in
      L.Fun("jinit", [t; L.integer count]), P.G("jInit", [])

  | s -> E.s (E.unimp "typeOfSymbol for %a\n" dsym s)


(**** Search the class hierarchy for override *)

 
   (* A table of classes indexed by their internal name *)
let javaClasses : (string, jclass) H.t = H.create 17


   (* Look for a class file, but do not load it *)
let rec findClass (n : string) : jclass = 
  try
    H.find javaClasses n
  with Not_found -> begin
    try 
      let (filename, chan) = openClassFile n in 
      let magich = fetch2 chan in
      let magicl = fetch2 chan in
      if magich <> 0xCAFE || magicl <> 0xBABE then 
        E.s (E.error "Invalid .class file: %s\n" filename);
      close_in chan;
      let this = 
        { filename   = filename;
          cname      = n;
          csymbol    = classToString n;
          loadState  = NotLoaded;
          cflags     = 0;
          super      = None;
          interfaces = Array.create 0 (Obj.magic 0);
          methods    = Array.create 0 (Obj.magic 0);
          fields     = Array.create 0 (Obj.magic 0);
          instanceSize = 0;
          vTableSize = 0;
        }  in
      H.add javaClasses n this;
      addSymbol this (VirtualMT this);
      if !debug && mustPrint n then
        ignore (E.log "JV: added %s to table\n" n);
      this
    with Not_found -> 
      E.s (E.error "Cannot find class file for %s.\nCLASSPATH is %a.\nMake sure the class has the right package name\n" 
             n (docList text) !classPath)
  end

    (* Like findClass but also load it *)
and findAndLoadClass (n : string) : jclass = 
  let this = findClass n in
  match this.loadState with
    NotLoaded -> begin 
      this.loadState <- BeingLoaded;
      loadClassFile this; 
      this 
    end
  | Loaded -> this
  | BeingLoaded -> 
      E.s (E.bug "Cycle in the class hierarchy while loading %s\n" n)

   (* Now we actually load the class *)
and loadClassFile (this : jclass) : unit = 
  let chan = 
    try open_in_bin this.filename with _ -> E.s (E.bug "Cannot open %s\n"
                                                   this.filename) in
  let magich = fetch2 chan in
  let magicl = fetch2 chan in
  if magich <> 0xCAFE || magicl <> 0xBABE then 
    E.s (E.error "Invalid .class file: %s\n" this.filename);
  let minor = fetch2 chan in
  let major = fetch2 chan in

  (* Load the constant pool *) 
  let cpCount = (fetch2 chan) - 1 in
  if cpCount < 0 then 
    E.s (E.error "Empty constant pool in %s\n" this.cname);
  (* Create the constant pool *)
  let cPool = Array.create cpCount CP_None in
  let rec loop i = 
    if i >= cpCount then ()
    else
      let tag = fetch1 chan in
      let entry, skipOne = 
        match tag with
          1 ->                              (* UTF8 *)
            let len = fetch2 chan in
            let res = String.create len in
            let rec loop i left = 
              if left <= 0 then () 
              else begin
                let chr = fetch1 chan in
                let thechr, left' = 
                  if chr > 0 && chr < 128 then
                    chr, left - 1
                  else if chr land 0xE0 = 0xC0 then begin (* Two-char *)
                    let chr2 = fetch1 chan in
                    if chr2 land 0xC0 <> 0x80 then
                      E.s (E.unimp "Found invalid char in UTF8 in %s at #%d"
                             this.cname i);
                    ((chr land 0x1F) lsl 6) + (chr2 land 0x3F), left - 2
                  end else if chr land 0xF0 = 0xE0 then begin (* Three char *)
                    let chr2 = fetch1 chan in
                    let chr3 = fetch1 chan in
                    if chr2 land 0xC0 <> 0x80 || chr3 land 0xC0 <> 0x80 then
                    E.s (E.unimp "Found invalid char in UTF8 in %s at #%d"
                           this.cname i);
                    ((chr land 0x0F) lsl 12) + 
                      ((chr2 land 0x3F) lsl 6) + (chr3 land 0x3F), left - 3
                  end else
                    E.s (E.unimp "Found invalid char in UTF8 in %s at #%d"
                           this.cname i)
                in
                let thechr' = 
                  if thechr >= 0 && thechr < 256 then
                    char_of_int thechr else '_'
                in
                String.set res i thechr';
                loop (i + 1) left'
              end
            in
            loop 0 len;
            CP_Utf8 res, false

        | 3 | 4 | 12 -> skip4 chan; CP_None, false
        | 5 | 6 -> skip4 chan; skip4 chan; CP_None, true
        | 7 -> 
            let cidx = fetch2 chan in
            if cidx < 0 && cidx >= cpCount then
              E.s (E.error "Found invalid index %d when reading %s"
                     cidx this.cname);
            CP_ClassRef cidx, false

        | 8 -> ignore (fetch2 chan); CP_None, false
        | 9 | 10 | 11 -> skip4 chan; CP_None, false
        | _ -> E.s (E.error "Found unexpected cpool tag %d when reading %s"
                      tag this.cname)
      in
      assert (i < cpCount);
      cPool.(i) <- entry;
      if skipOne then 
        loop (i + 2)
      else
        loop (i + 1)
  in
  loop 0;

  (* FLAGS *)
  let cflags = fetch2 chan in
  this.cflags <- cflags;

  (* THIS *)
  let thisNameIdx = fetch2 chan in
    
  (* A function to resolve names *)
  let resolveName idx = 
    if idx <= 0  || idx > cpCount then
      E.s (E.error "Found index out of bounds while resolving class in %s"
             this.cname);
    match cPool.(idx - 1) with
      CP_Utf8 n -> n
    | _ -> E.s (E.error "Name at #%d in %s is not Utf8\n"
                  idx this.cname)
  in
    
  (* We need a function to resolve classes from the contant pool *)
  let resolveAndLoadPoolClass idx = 
    if idx <= 0  || idx > cpCount then
      E.s (E.error "Found index out of bounds while resolving class in %s"
             this.cname);
    match cPool.(idx - 1) with
      CP_ClassRef ci -> 
        let n = resolveName ci in
        let c = findAndLoadClass n in
        cPool.(idx - 1) <- CP_Class c;
        c
    | CP_Class c -> c
    | _ -> E.s (E.error "Name of class at #%d in %s is not Utf8\n"
                  idx this.cname)
  in

  (* A function to skip attributes *)
  let skipAttributes () =
    let aCount = fetch2 chan in
    for i = 0 to aCount - 1 do 
      ignore (fetch2 chan);
      let len = fetch4 chan in
      for j = 0 to len - 1 do 
        ignore (input_char chan)
      done
    done
  in
  (* SUPER *)

  let superIdx = fetch2 chan in
  if superIdx = 0 then begin
    if this.cname <> "java/lang/Object" then 
      E.s (E.error "Class %s does not have a superclass\n" this.cname)
  end else 
    this.super <- Some (resolveAndLoadPoolClass superIdx);

  (* INTERFACES *)
  let iCount = fetch2 chan in
  assert (iCount >= 0);
  this.interfaces <- Array.create iCount dummyClass;
  for i = 0 to iCount - 1 do
    let iIdx = fetch2 chan in
    this.interfaces.(i) <- resolveAndLoadPoolClass iIdx
  done;
  
  (* FIELDS*)
  if !debug && mustPrint this.cname then
    ignore (E.log "JV: Start to load fields of %s\n" this.cname);
  let fCount = fetch2 chan in
  let fOffset = 
    ref (match this.super with
      None -> objectOverhead
    | Some s -> s.instanceSize)
  in
  assert (fCount >= 0);
  assert (!fOffset >= objectOverhead);
  this.fields <- Array.create fCount dummyField;
  for i = 0 to fCount - 1 do
    let fflags = fetch2 chan in
    let name = resolveName (fetch2 chan) in
    let descr = resolveName (fetch2 chan) in
    if !debug && mustPrint this.cname then
      ignore (E.log "JV:Loading field %s(desc=%s) of %s. " 
                name descr this.cname);
    skipAttributes ();
    let foff = 
      if isStatic fflags then 0
      else begin
        let res = !fOffset in
        let instSizeOf td = 
          try
            let c = String.get td 0 in
            if c = 'J' || c = 'D' then 8 else 4
          with _ -> E.s (E.bug "invalid type descriptor")
        in
        fOffset := !fOffset + instSizeOf descr;
        res
      end
    in
    if !debug && mustPrint this.cname then
      ignore (if foff > 0 then (E.log " (at offset %d)\n" foff) else 
              (E.log " (static field %x, foff=%d, fOffset=%d)\n" 
                 fflags foff !fOffset));
    this.fields.(i) <- { fname = name;
                         fflags  = fflags;
                         ftydesc = descr;
                         fty     = Int;
                         foffset = foff; };

  done;
  this.instanceSize <- !fOffset;
  if !debug && mustPrint this.cname then
    ignore (E.log "JV: Instance of %s has %d bytes\n" 
              this.cname this.instanceSize);

  (* METHODS *)
  if !debug && mustPrint this.cname then
    ignore (E.log "JV: Start to load methods of %s\n" this.cname);
  let mCount = fetch2 chan in
  let mOffset = 
    ref (match this.super with
      None -> vtOverhead
    | Some s -> s.vTableSize)
  in
  assert (mCount >= 0);
  this.methods <- Array.create mCount dummyMethod;
  for i = 0 to mCount - 1 do
    let mflags = fetch2 chan in
    let name = resolveName (fetch2 chan) in
    let descr = resolveName (fetch2 chan) in
    if !debug && mustPrint this.cname then
      ignore (E.log "JV:Loading method %s(desc=%s) of %s. " 
                name descr this.cname);
    let moff = 
      if isStatic mflags || isConstr name || isStaticInit name then
        0
      else begin
        let rec isOverride cls name (typedescr : string) = 
          let compareArgTypes td1 td2 = 
            try
              String.sub td1 0 (String.index td1 ')') = 
              String.sub td2 0 (String.index td2 ')')
            with Not_found -> false
          in
          match cls.super with
            None -> None
          | Some s -> 
              let rec search i = 
                if i >= Array.length s.methods then
                  isOverride s name typedescr
                else 
                  let mth = s.methods.(i) in
                  if not (isStatic mth.mflags) && 
                    name = mth.mname && 
                    compareArgTypes mth.mtydesc typedescr 
                  then
                    Some (s, i)
                  else
                    search (i + 1)
              in
              search 0
        in
        (* See if this is an override *)
        match isOverride this name descr with
          None -> begin
            let theOffset = !mOffset in
            mOffset := !mOffset + 4;
            theOffset
          end
        | Some (cls, idx) -> 
            cls.methods.(idx).moffset
      end
    in
    if !debug && mustPrint this.cname then
      ignore (E.log " at index %d and offset %d@!" i moff);
    let m = { mname = name;
              mty   = None, [];
              mtydesc = descr;
              mflags = mflags;
              moffset = moff;
              mindex = i;
            } in
    this.methods.(i) <- m;
    skipAttributes ();
  done;
  this.vTableSize <- !mOffset;

  skipAttributes ();

  this.loadState <- Loaded;
  close_in chan;

  (* Now that we have finalized computing the virtual table size and the 
   * field offsets we can process the types of the fields and methods. We do 
   * this now to avoid that we end-up loading a subclass before we finished 
   * loading this class *)

  (**** FIELDS - again ****)
  Array.iteri (fun i f -> 
    f.fty <- typeOfDescr f.ftydesc;
    if isStatic f.fflags then begin
      addSymbol this (Field (this, i))
    end) this.fields;

  (*** METHODS - again ***)
  Array.iteri (fun i m -> 
    let res, args = methodTypeOfDescr m.mtydesc in
    m.mty <- res, args;
    let sym = 
      if isConstr m.mname then 
        Constr (this, args)
      else if isStaticInit m.mname then
        StaticInit this
      else
        Method (this, i)
    in
    addSymbol this sym) this.methods;

  if !debug && mustPrint this.cname then
    ignore (E.log "JV:Finished loading %s\n" this.cname)

and typeOfDescrHelp t (i: int) : (jtype * int) = 
  if String.length t <= i then
    E.s (E.error "Wrong type descriptor %s (at %d)@!" t i);
  let c = String.get t i in
  match c with
    'B' -> Byte, i + 1
  | 'C' -> Char, i + 1
  | 'D' -> Double, i + 1
  | 'F' -> Float, i + 1
  | 'I' -> Int, i + 1
  | 'J' -> Long, i + 1
  | 'S' -> Short, i + 1
  | 'T' -> String, i + 1
  | 'Z' -> Bool, i + 1
  | '[' -> 
      let (ty, ei) = typeOfDescrHelp t (i + 1) in
      Array ty, ei 
  | 'L' -> 
      begin
        try
          let ei = String.index_from t (i + 1) ';' in
          let cname = String.sub t (i + 1) (ei - i - 1) in
          let c = findClass cname in
          Class c, ei + 1
        with Not_found ->
          E.s (E.error "Wrong type descriptor (unfinished class): %s @@%d" t i)
       end
  | _ -> E.s (E.error "Wrong type descriptor (unknown character): %s @@%d" t i)

and typeOfDescr t = 
  let ty, _  = typeOfDescrHelp t 0 in
  ty

and methodTypeOfDescr t = 
  if String.get t 0 <> '(' then
    E.s (E.error "Wrong type descriptor for method: %s@" t);
  let rec loop i = 
    if String.get t i = ')' then
      let res = 
        if String.get t (i + 1) = 'V' then
          None
        else
          let (res, _) = typeOfDescrHelp t (i + 1)  in
          Some res
      in
      (res, [])
    else
      let (ty, nexti) = typeOfDescrHelp t i in
      let (res, tys) = loop nexti in
      (res, ty ::tys)
  in
  loop 1
  

  (* Raises Not_found if this is not a symbol name *)
let parseAndLoadSymbolClass (s : string) : jclass = 
  let len = String.length s in
  if !debug then
    ignore (E.log "JV: parsing symbol name %s\n" s);
  let parseChar c i = 
    if i < len && String.get s i = c then
      i + 1
    else
      raise Not_found
  in
  let parseString str i : int = 
    let n = String.length str in
    if i + n > len then
      raise Not_found;
    for j = i to i + n - 1 do
      if String.get str (j - i) <> String.get s j then 
        raise Not_found
    done;
    i + n
  in
  let parseInt i : (int * int) = 
    let rec loop i sum = 
      if i < len then
        let c = int_of_char (String.get s i) - 48 in
        if c >= 0 && c <= 9 then
          loop (i + 1) (sum * 10 + c)
        else
          (i, sum)
      else
        (i, sum)
    in
    loop i 0
  in
  let parseLenComp i : (int * string) = 
    let (i, n) = parseInt i in
    if n > 0 && i + n <= len then 
      let res = String.create n in
      String.blit s i res 0 n;
      (i + n, res)
    else
      raise Not_found
  in
  let parseClassName i : (int * string) = 
    let rec loop i = 
      try 
        let (i', n) = parseLenComp i in
        let (ei, acc) = loop i' in
        (ei, if acc = "" then n else n ^ "/" ^ acc)  
      with Not_found -> 
        (i, "")
    in 
    let (ei, cname) = loop i in
    if cname = "" then 
      raise Not_found
    else
      (ei, cname)
  in
  (* And now the parsing *)
  let (after, cname) = parseClassName (parseChar '_' 0) in
  (* Now load the class *)
  if !debug then
    ignore (E.log "JV: parse succesful: loading class %s\n" cname);
  let cls = findAndLoadClass cname in
  (* Maybe it is the symbol for a constant *)
  begin 
    try
      let (ei, index) = parseInt (parseString "_CT" after) in
      let (ei, count) = parseInt (parseString "_" ei) in
      let ty, _ = typeOfDescrHelp s ei in
      addSymbol cls (Const(cls, ty, count, index))
    with Not_found -> ()
  end;
  cls
  
let getSymbol s = 
  try
    try
      H.find loadedSymbols s
    with Not_found -> begin
      ignore (parseAndLoadSymbolClass s);
      H.find loadedSymbols s
    end
  with Not_found -> begin
    if !debug then
      ignore (E.log "Cannot find Java symbol for %s\n" s);
    raise Not_found
  end
    

let unparseSymbol : symbol -> string = 
  function s -> 
    E.s (E.unimp "unparsing symbol")


(***** Find the type of a constant. Load enough classes to find it out *****)

let findTypeConstant n = 
  if !debug then
    ignore (E.log "JV: find type of constant %s\n" n);
  try
    let sym = getSymbol n in
    if !debug then 
      ignore (E.log "JV: found symbol %a\n" dsym sym);
    let ty, proof = typeOfSymbol sym in
    Some (ty,proof)

  with Not_found -> None


let getAndLoadClassOfSymbol n = 
  try
    let smb = getSymbol n in
    let cls = 
      match smb with
        Constr (x, _) -> x
      | Method (x, _) -> x
      | Field (x, _) -> x
      | Const (x, _, _, _) -> x
      | StaticInit x -> x
      | VirtualMT x -> x
    in
    begin
      match cls.loadState with
        NotLoaded -> loadClassFile cls
      | BeingLoaded -> E.s (E.bug "Cycle in the class hierarchy at %s\n" n)
      | Loaded -> ()
    end;
    Some smb
  with Not_found -> None
      
(**** Check that two constants are in the subclass relationship *)
let checkSubtype n1 n2 = 
  if !debug then
    ignore (E.log "JV: checkSubtype %s <= %s\n" n1 n2);
  match getAndLoadClassOfSymbol n1, getAndLoadClassOfSymbol n2 with
    Some(VirtualMT c1), Some(VirtualMT c2) -> 
      let rec findSuperClass f c = 
        if f c then true else
        match c.super with 
          None -> false
        | Some sup -> findSuperClass f sup
      in
      let rec findSuperInterface c = 
        if c == c2 then true else
        let rec loop i = 
          if i >= Array.length c.interfaces then 
            false
          else
            findSuperInterface c.interfaces.(i) || loop (i + 1)
        in
        loop 0  
      in
      if isInterface c2.cflags then begin
        findSuperClass findSuperInterface c1
      end else
        findSuperClass (fun c -> c == c2) c1
          
  | _ -> false


(**** Get the type of an instance field *)
let getFieldType c off = 
  if !debug then
    ignore (E.log "JV: getFieldType %s[%d]\n" c off);
  if off < objectOverhead || off land 3 <> 0 then 
    None
  else
    match getAndLoadClassOfSymbol c with
    | Some (VirtualMT cls) -> 
        let findInstanceField cls = 
          if !debug then 
            ignore (E.log "JV: searching for field in %s\n" cls.cname);
          let fields = cls.fields in
          let rec search i = 
            if i >= Array.length fields then
              None
            else
              let fld = cls.fields.(i) in
              if fld.foffset = off then begin
                let ty = typeToLF fld.fty in
                if !debug then 
                  ignore (E.log "JV: type of %s[%d] is %a\n"
                            c off L.d_exp ty);
                Some (typeToLF fld.fty)
              end else
                search (i + 1)
          in
          search 0
        in
        let rec searchClass cls = 
          match cls.super with
            None -> findInstanceField cls
          | Some s when s.instanceSize <= off -> 
              findInstanceField cls
          | Some s -> searchClass s
        in
        searchClass cls
    | _ -> None
          



(**** Check the signature of a virtual method *)
let checkSignatureVirtual ~host ~off ~signame = 
  if !debug then
    ignore (E.log "JV: checkSignatureVirtual %s[%d] = %s\n" 
              host off signame);
  try 
    if off < vtOverhead || off land 3 <> 0 then 
      raise Not_found;
    let hostclass, sghost, sgmth = 
      match getAndLoadClassOfSymbol host, getAndLoadClassOfSymbol signame with
        Some (VirtualMT host), Some (Method (sghost, sgidx)) -> 
          host, sghost, sghost.methods.(sgidx)
      | _ -> raise Not_found
    in
    if !debug then
      ignore (E.log "JV: found the classes for %s and %s\n"
                host signame);
    (* The offsets must be the same *)
    if off <> sgmth.moffset then begin
      if !debug then
        ignore (E.log "JV: check failed because offsets %d <> %d\n" 
                  off sgmth.moffset);
      raise Not_found
    end;

    (* And the host must be a subtype of sghost *)
    let rec checkHost c = 
      if c == sghost then true
      else
        match c.super with
          None -> raise Not_found
        | Some s -> checkHost s
    in
    checkHost hostclass
  with Not_found -> begin
    if !debug then
      ignore (E.log "JV: check did not succeed\n");
    false
  end

(**** Get the signature of a virtual method. This and the check version 
    * should be merged *)
let getSignatureVirtual ~host ~off = 
  if !debug then
    ignore (E.log "JV: getSignatureVirtual %s[%d]\n" host off);
  if off < vtOverhead || off land 3 <> 0 then 
    None
  else
    match getAndLoadClassOfSymbol host with
    | Some (VirtualMT cls) -> 
        let findVirtualMethod cls = 
          if !debug then 
            ignore (E.log "JV: searching for method in %s\n" cls.cname);
          let methods = cls.methods in
          let rec search i = 
            if i >= Array.length methods then
              None
            else
              let mth = cls.methods.(i) in
              if mth.moffset = off then begin
                if !debug then 
                  ignore (E.log "JV: found method %s in %s\n" 
                            mth.mname cls.cname);
                Some (symbolToString (Method (cls, i)))
              end else
                search (i + 1)
          in
          search 0
        in
        let rec searchClass cls = 
          match findVirtualMethod cls, cls.super with
            Some m, _ -> Some m
          | None, Some s -> searchClass s
          | None, None -> None
        in
        searchClass cls
    | _ -> None          


let options = [
    "-classpath", Arg.String setClassPath, 
                  "the search paths for .class files";
] 

