(** The representation of logical elements *)

(** {b API} *)

type location = Errormsg.location

(** {b Expressions} *)

type var = string

(** The main type for symbolic expressions. Each expression can have one of a 
 * few types: 
 - Int: any integer value
 - Sign32: a value that can be represented in 2's complement on 32 bits 
   (ranging from -2^31..2^31-1)
 - Unsign32: a value ranging from 0..2^32-1
 - Other: most notably the values of the "upd" memory constructor
*)

type exp =                             
  | Int of Int32.t                      (** Integer constants (of type 
                                         * Sign32). If you want to convert to 
                                         * an int, use {!Logic.to_int}, not 
                                         * In32.to_int because the latter 
                                         * silently drops a bit. *)
  | Float of float                      (** A floating point value **)
  | UnOp of unop * exp                  (** A unary operation. Type of 
                                         * operand and result depends on the 
                                         * operation. *)
  | BinOp of binop * exp * exp          (** A binary operation. Type of 
                                          * operand and result depends on the 
                                          * operation. **)

  | Fun of string * exp list		(** A constructed expression, with 
                                         * the constructor name and a number 
                                         * of arguments. *)

  | Var of int (*de Bruijn*)            (** A logical variable. This 
                                         * expression can only occur within 
                                         * the scope of a forall or exist 
                                         * predicate. The integer is a 
                                         * deBruijn index. *)
  | NVar of var                         (** A named variable *)

  | UVar of int                          (** A unification variable *)

  | Ite of exp * exp * exp              (** IfThenElse(b,e1,e2). If b is not 
                                         * zero then b1, else e2 *)
and unop = 
    Mod32                               (** Computes modulo 2^32. Of type Int 
                                         * -> Unsign32. *)
  | BNot                                (** Bitwise complement *)
  | RepS32                              (** Tests whether the argument (an 
                                         * Int) is in the range of Sign32. 
                                         * Has a value of zero or one *)
  | RepU32                              (** Tests whether the argument (an 
                                         * Int) is in the range of Unsign32. 
                                         * Has a value of zero or one *)

and binop = 
    Plus                                (** Addition (the real one). Two 
                                         * operands of type Int or Sign32 and 
                                         * result of type Int.  *)
  | Minus                               (** Subtraction (the real one). Two 
                                         * operands of any integer type and 
                                         * result of type Int. *)
  | Times                               (** Multiplication (the real one ). 
                                         * Two operands of type Int or Sign32 
                                         * and result of type Int. *)
 
  (** The following operators work on 32-bit integers *)
  | Add                                 (** Addition modulo 2^32. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)
  | Sub                                 (** Subtraction modulo 2^32. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)
  | Mult                                (** Multiplication modulo 2^32. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)
  | Div                                 (** Integer Division. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)

  | BAnd                                (** Bitwise AND *)
  | BOr                                 (** Bitwise OR *)
  | BXor                                (** Bitwise XOR *)
      
  | Sll                                 (** Shift left logical *)
  | Srl                                 (** Shift right logical *)
  | Sra                                 (** Shift right arithmetic *)

  (** The following operators compare two 32-bit values and result in either 
    * one or zero *)
  | SetEq                               (** equal, printed "=" when used with 
                                         * PNez. *)
  | SetNe                               (** not equal, printed "<>" when used 
                                         * with PNez. *)
  | SetGt                               (** signed greater, printed ">" when 
                                         * used with PNez. *)
  | SetGe                               (** signed greater or equal, printed 
                                         * ">=" when used with PNez. *)
  | SetLt                               (** signed less, printed "<" when 
                                         * used with PNez. *)
  | SetLe                               (** signed less or equal, printed 
                                         * "<=" when used with PNez. *)
  | SetAb                               (** unsigned greater (above), printed 
                                         * ">u" when used with PNez. *)
  | SetAe                               (** unsigned greate equal, , printed 
                                         * ">=u" when used with PNez.  *)
  | SetBl                               (** unsigned less (below), printed 
                                         * "<u" when used with PNez. *)
  | SetBe                               (** unsigned less or equal , printed 
                                         * "<=u" when used with PNez. *)
  | SetPe                               (** parity even *)
  | SetPo                               (** parity odd *)

  | SignExt                             (** signext(e1,e2) we sign-extend the 
                                         * least-significant e2 bytes of e1, 
                                         * to a full 32-bit word  *)

(** A symbolic predicate *)
type pred =
    Bool of bool                    (** A boolean value of True or False *)  
  | PNez of exp                     (** A shorthand for e <> 0 *)
  | PFun of string * exp list       (** Constructed predicate *)
  | NegPFun of string * exp list    (** Negation of a predicate *)

  | And of pred list                (* A conjunction of predicates. *)
  | Or of pred list                 (* A disjunction of predicates. *)

  | Not of pred                     (* A negated predicate *)

  | Imply of pred * pred            (** An implication *)
  | All of string * pred            (** A universal quantification of 
				      * the variable and predicate. Give the 
                                      * name of the quantified variable. *)
  | Exist of string * pred          (** An existential quanitification of the 
                                      * variable and predicate. Give the name 
                                      * of the quantified variable. *)
  | Pannot of pannot * pred        (** This is like the second predicate, but 
                                    * it can be used to attach various 
                                    * annotations to predicates. The 
                                    * annotations do not have logical meaning. *)

and pannot = 
  | ATrigger of pred               (** The trigger is used to control the 
                                    * instantiation of the quantifiers *)



exception CannotNegate of pred

(** Negate a binary operation. Raises CannotNegate if it cannot do it. *)
val negateBinopRaise: binop -> binop

(** Negate an expression. This is the same as BinOp(SetEq,e,zero) but in most 
 * cases it yields a better expression. *)
val negateExp: exp -> exp

(** Negate a predicate. Predicated of the form PFun cannot 
 * be negated. Raise CannotNegate in that case.  *)
val negatePredRaise: pred -> pred

(** Like {!Logic.negatePredRaise} except that it generates an error for 
 * predicates that cannot be negated. *)
val negatePred: pred -> pred


(** Compare two predicates, ignoring annotations *)
val equalPred: pred -> pred -> bool

(** A constructor for making Int expressions *)
val integer: int -> exp

exception IntegerTooLarge
(** Converts an int32 into an integer. Raises IntegerTooLarge if the 
  * conversion cannot be made soundly. *)
val to_int: int32 -> int
 

val newVar: string -> var
val resetVars: unit -> unit  (** Reset the newVar count to the value that it 
                             * has the first time when we called this 
                             * function *)
val varName: string -> string

val isAtomPred: pred -> bool


(** A few useful constants *)
val zero: exp
val one: exp 

val minus: exp -> exp -> exp 
val plus: exp -> exp -> exp 
val times: exp -> exp -> exp 

val eq: exp -> exp -> pred
val ne: exp -> exp -> pred
val geq: exp -> exp -> pred
val leq: exp -> exp -> pred
val gt: exp -> exp -> pred
val lt: exp -> exp -> pred

val usePFun: bool ref (** whether the parser will produce PFun. If false, it 
                       * produces only uninterpreted expressions equal to 
                       * true *)
val trueExp: exp
    (** This is the expression that denotes TRUE. This is used for predicate 
     * constructors. In order to asert the predicate [f(x)], introduce an 
     * expression constructor [f] and say that [f(x) = true]. *)

val pFun: string -> exp list -> pred (** make a PFun, depending on the 
                                     * {!Logic.usePFun} flag *)

(** Construct a normalized conjunction of predicates.  Uses the And
 *  constructor only if there are two or more predicates.
 *)
val pand: pred list -> pred
(** Construct a normalized disjunction of predicates.  Uses the Or
 *  constructor only if there are two or more predicates.
 *)
val por: pred list -> pred

(** We have integer maps built-in. Make an integer map *)
val makeMap: (int * exp) list -> exp

(** Lookup an index into the map. Raises Not_found *)
val findMap: exp -> int -> exp

(** Insert a new entry into the map: [insertMap map idx e] *)
val insertMap: exp -> int -> exp -> exp

(** Substitution on predicates, given a function that can substitute a 
 * subexpression. The argument function is also passed the abstraction depth. 
 * If the function returns Some e then that is used as a replacement for the 
 * subexpression. If the function returns None then the subexpressions of the 
 * current expression are processed recurisively. If no substitutions are 
 * performed this function returns the exact same object.*)
val substPred: ?depth:int -> (int -> exp -> exp option) -> pred -> pred
val substExp:  ?depth:int -> (int -> exp -> exp option) -> exp -> exp



(* A substitution is a vector of expressions *)
type soption =  Unif | Done of sexp 

and subst = soption array

and sexp = exp * subst

and spred = pred * subst

(** Unify two expressions. The first argument is a function that can be used 
 * to update the substitution. If it tries to unify two unification 
 * variables, the first one gets instantiated. *)
val eunify : (subst -> int -> sexp -> bool) 
               -> (exp * subst) -> (exp * subst) -> bool
val punify : (subst -> int -> sexp -> bool) 
               -> (pred * subst) -> (pred * subst) -> bool


(** Tries to unroll the unification variables completely, but if there are 
 * still unresolved unification variables it leaves the expression unchanged 
 * (so that the original subsitution still applies) *)
val unrollUVars: subst -> exp -> exp

(** Like {!Logic.unrollUVars} but raises HasUVar if there are unification 
 * variables left *)
val unrollUVarsRaise: subst -> exp -> exp

val unrollUVarsPredRaise: subst -> pred -> pred

(** Unroll only superficially *)
val unrollExp: sexp -> sexp

exception HasUVar 

(** Like {!Logic.unrollUVars} but gives an error and stops the execution of 
 * there are unification variables left *)
val unrollUVarsError: subst -> exp -> exp 

(** Create a substitution for the given number of unification variables. *)
val newSubst: int -> subst

val emptySubst: subst

val newSubstFromList: exp list -> subst

val d_sexp: unit -> sexp -> Pretty.doc
val d_spred: unit -> spred -> Pretty.doc


(* Print a substitution *)
val d_subst: unit -> subst -> Pretty.doc

(** Print an explicit substitution without trying to perform the substitution *)
val d_sexpNoUnroll: unit -> sexp -> Pretty.doc


(** Try to simplify an expression as much as possible, using the rules of 
 * arithmetic (even to an integer) *)
val simplifyArith: exp -> exp

(** Build an expression from a linear combination of expressions. Tries to be 
 * smart when the factors are 0, 1 and -1 *)
val buildExpLinearComb: (int * exp) list -> exp


(******************
 ******************  VISITOR
 ******************)
(** {b The visitor} *)

(** Different visiting actions. 'a will be instantiated with [exp], [pred] *)
type 'a visitAction = 
    SkipChildren                        (** Do not visit the children. Return 
                                            the node as it is. *)
  | DoChildren                          (** Continue with the children of this 
                                            node. Rebuild the node on return 
                                            if any of the children changes 
                                            (use == test) *)
  | ChangeTo of 'a                      (** Replace the expression with the 
                                            given one *)
  | ChangeDoChildrenPost of 'a * ('a -> 'a) (** First consider that the entire 
                                           exp is replaced by the first 
                                           parameter. Then continue with 
                                           the children. On return rebuild 
                                           the node if any of the children 
                                           has changed and then apply the 
                                           function on the node *)



(** A visitor interface for traversing Logic expressions and predicates. 
 * Create instantiations of this type by specializing the class 
 * {!Logic.nopLogicVisitor}. *)
class type logicVisitor = object

  method vexp: exp -> exp visitAction          
    (** Invoked on each expression occurrence. The subtrees are the 
     * subexpressions *)

  method vpred: pred -> pred visitAction          
    (** Invoked on each predicate occurrence. The subtrees are the 
     * subexpressions *)

end

(** Default Visitor. Traverses the SAL tree without modifying anything *)
class nopLogicVisitor: logicVisitor

(* Visit an expression *)
val visitLogicExp: logicVisitor -> exp -> exp

(* Visit a predicate *)
val visitLogicPred: logicVisitor -> pred -> pred




(* ------ Logic pretty printer ------ *)

(** {b Pretty Printing} *)

(** Logic has a fairly easy to use mechanism for printing error messages. This 
 * mechanism is built on top of the pretty-printer mechanism (see 
 * {!Pretty.doc}) and the error-message modules (see {!Errormsg.error}). 

 Here is a typical example for printing a log message: {v 
ignore (Errormsg.log "Expression %a is not positive (at %s:%i)\n"
                        d_exp e loc.file loc.line)
 v}

 and here is an example of how you print a fatal error message that stop the 
* execution: {v 
Errormsg.s (Errormsg.bug "Why am I here?")
 v}

 Notice that you can use C format strings with some extension. The most 
useful extension is "%a" that means to consumer the next two argument from 
the argument list and to apply the first to [unit] and then to the second 
and to print the resulting {!Pretty.doc}. For each major type in SAL there is 
a corresponding function that pretty-prints an element of that type:
*)

(** Convert a unary operation to a string *)
val unop_to_string: unop -> string

(** Convert a binary operation to a string *)
val binop_to_string: binop -> string

(** See if we can print this binary operator in a special way when it occurs 
 * in a PNez *)
val pbinop_to_string: binop -> string option

(** A printer interface for Logic trees. Create instantiations of 
 * this type by specializing the class {!Logic.defaultLogicPrinterClass}. Each 
 * method returns None if it wants the default printing (actually, that of 
 * the superclass) *)
class type logicPrinter = object

  method pExp: exp -> Pretty.doc option
    (** Invoked for each expression. If it return None, it defaults to the 
     * default printer. *)

  method pPred: pred -> Pretty.doc option
    (** Invoked for each predicate *)

  method pAnnot: pannot -> Pretty.doc option
    (** Invoked for each annotation *)

end


class defaultLogicPrinterClass: logicPrinter
val defaultLogicPrinter: logicPrinter

class infixLogicPrinterClass: logicPrinter
val infixLogicPrinter: logicPrinter



(** Print an expression given a pretty printer *)
val printExp: logicPrinter -> exp -> Pretty.doc
  
(** Print a predicate given a pretty printer *)
val printPred: logicPrinter -> pred -> Pretty.doc

(** Convert a predicate to a {!Pretty.doc} *)
val d_pred: unit -> pred -> Pretty.doc

(** Convert a predicate annotation to a {!Pretty.doc} *)
val d_pannot: unit -> pannot -> Pretty.doc

(** Convert an expression to a {!Pretty.doc} *)
val d_exp: unit -> exp -> Pretty.doc

(** Convert an expression list to a {!Pretty.doc} *)
val d_explist: unit -> exp list -> Pretty.doc

val d_loc: unit -> location -> Pretty.doc

class uvarLogicPrinterClass: subst -> logicPrinter

(** rename a variable in a predicate *)
val renameVariable: pred -> string(*old*) -> string(*new*) -> pred

(** evaluate a binary operation, if possible *)
val evalBinop : binop -> int -> int -> int option

