(* matchrule.ml *)
(* handle universally-quantified assumptions via matching rules *)

(* NOTE: This module is obsolete.  Use 'matching.ml' instead. *)

module E = Errormsg       (* E.log *)
module P = Proof          (* P.proof *)
open Satproc              (* message, etc. *)
open Pretty               (* dprintf *)
open Logic                (* pred, exp, logicVisitor, etc. *)
open Edag                 (* node, etc. *)


let matchid = Util.registerSymbolName "MA"        (* satproc id *)

let debug = ref false     (* debug flag, settable from command line: "-debug MA" *)


(* a matching rule and associated axiom *)
type rule = {
  (* This is the pattern set against which e-dag nodes are matched.  Each
   * pattern expression contains one or more unification variables, numbered 
   * sequentially from 0 up to numUnivVars-1.  For a match, *each* pattern
   * expression must match.
   *
   * Note that multiple pattern expressions really only works when there is
   * just one unification variable (hopefully this will be fixed).
   *
   * This list cannot be empty.
   *)
  rpattern: exp list;

  (* number of unification variables in 'rpattern' *)
  numUnifVars: int;

  (* This is the entire body of the rule; it also contains unification
   * variables, the same ones as in 'pattern'. *)
  rbody: pred;
}

(* list of active matching rules *)
let rules : rule list ref = ref []


(* in debug mode, print a log string *)
let debugLog (s: string) : unit =
begin
  if !debug then
    (ignore (E.log "MA: %s\n" s));
end


(* ---------------------------- matching ---------------------------- *)
(* The code for this section is based closely on:
 *
 *   David Delefs, Greg Nelson and James B. Saxe
 *   Simplfy: A Theorem Prover for Program Checking
 *   Hewlett-Packard Systems Research Center
 *   Technical Report HPL-2003-148
 *   http://www.hpl.hp.com/techreports/2003/HPL-2003-148.html
 *
 * Note that the tech report describes the process using coroutine
 * iterators, whereas I just explicitly accumulate sets.
 *)


(* A substitution is a map from unification variable index to the e-dag
 * node for that variable, or None if it hasn't been mapped yet. *)
type substitution = node option array

(* The way the algorithm is written in HPL-2003-148, operations
 * on subtitutions are not destructive.  However, I'm using OCaml arrays,
 * whose updates *are* destructive.  Therefore the approach I have taken
 * is to use this function, which makes a copy (only) just before modifying. *)
let bindUVar (sub: substitution) (idx: int) (node: node) : substitution =
begin
  let ret:substitution = (Array.copy sub) in
  ret.(idx) <- Some(node);     (* bind in the copy *)
  ret
end

(* true if two substitutions are equal *)
let equalSubst (s1: substitution) (s2: substitution) : bool =
begin
  let len:int = (Array.length s1) in

  if (len <> (Array.length s2)) then
    false      (* different array sizes (shouldn't happen..) *)

  else (
    let rec loop (i: int) : bool =
    begin
      (* test elements 'i' for equality *)
      let localMatch:bool =
        (match (s1.(i), s2.(i)) with
        | (None, None) -> true
        | (Some(n1), Some(n2)) -> n1.root == n2.root    (* equivalent nodes? *)
        | (_, _) -> false)
      in

      if (not localMatch) then
        false                       (* this element differs *)
      else if (i < (len-1)) then
        (loop (i+1))                (* check next element *)
      else
        true                        (* all elements equal *)
    end in
    (loop 0)
  )
end


(* A set of substitutions, which is what the matching algorithm yields.
 * I should probably use the binary-tree version of lists, since union
 * is a frequent operation. *)
type substSet = substitution list

let emptySubstSet:substSet = []

let isEmptySubstSet (s: substSet) : bool =
begin
  match s with
  | _::_ -> false
  | [] -> true
end

let singletonSubstSet (sub: substitution) : substSet =
begin
  [sub]
end

let unionSubstSet (s1: substSet) (s2: substSet) : substSet =
begin
  s1 @ s2
end

let substSetContains (s: substSet) (e: substitution) : bool =
begin
  (* yes, I could avoid using 'fun' by relying on currying... *)
  (List.exists (fun sElt -> (equalSubst sElt e)) s)
end

let intersectSubstSet (s1: substSet) (s2: substSet) : substSet =
begin
  (* dumb, O(n^2) algorithm *)
  (List.filter (fun s1Elt -> (substSetContains s2 s1Elt)) s1)
end

(* If 's' is {x,y,z}, this computes (f (f (f init x) y) z).   Of course,
 * 'f' should be commutative and associative, since the element order is 
 * not determined (it is a set). *)
let foldLeftOverSet (f: 'a -> substitution -> 'a) (init: 'a) (s: substSet) : 'a =
begin
  (List.fold_left f init s)
end

let iterOverSet (f: substitution -> unit) (s: substSet) : unit =
begin
  (List.iter f s)
end

let existsOverSet (f: substitution -> bool) (s: substSet) : bool =
begin
  (List.exists f s)
end


(* Given a term pattern and an initial substitution, yield all the
 * substitutions that match the pattern with a node in the e-dag. *)
let rec matchTermWithAny (pat: exp) (sub: substitution) : substSet =
begin
  (* loop over e-dag nodes *)
  let attemptNode (acc: substSet) (n: node) : substSet =
  begin
    (* match 'pat' with 'n', yielding a set of substitutions; prepend
     * this set to 'acc' *)
    (unionSubstSet (matchTermWithNode pat n sub 0) acc)
  end in
  (List.fold_left attemptNode emptySubstSet !Edag.edag)
end


(* Match a term pattern with a specific e-dag node.  'nesting' should
 * initially be 0; it's used for debugging. *)
and matchTermWithNode (pat: exp) (node: node) (sub: substitution)
                      (nesting: int) : substSet =
begin
  let debug:bool = false in        (* static debug flag for this function *)

  if debug then
    (ignore (E.log "matchTerm[%d]:\n  pat: %a\n  node: %a\n"
                   nesting
                   d_exp pat
                   d_exp node.exp));

  (* for convenient recursive calls *)
  let recurse (pat: exp) (node: node) : substSet =
  begin
    (matchTermWithNode pat node sub (nesting+1))
  end in

  (* yield 'sub' as-is, if the bool is true *)
  let yieldSubIf (b: bool) : substSet =
  begin
    if (b) then
      (singletonSubstSet sub)     (* yield copy of 'sub' *)
    else
      emptySubstSet               (* yield nothing, doesn't match *)
  end in

  let ret:substSet = match pat with
  | UVar(idx) -> (
      (* is this UVar already bound? *)
      match sub.(idx) with
      | None -> (
          (* not already bound, so bind it *)
          if debug then
            (ignore (E.log "matchTerm[%d]: bound UVar(%d) to %a\n"
                           nesting
                           idx
                           d_exp node.exp));
          (singletonSubstSet (bindUVar sub idx node))
        )
      | Some(node2) -> (
          (* bound to 'node2', they must be equal *)
          (yieldSubIf (node.root.id = node2.root.id))
        )
    )
  | _ -> (
      (* We must match the topmost constructor of 'pat' with the function
       * symbol in 'node', but there is a catch: *any* of the nodes equivalent
       * to 'node' are candidates to make this work.  The accumulation here
       * is additive. *)
      let rec loop (acc: substSet) (n: node) : substSet =
      begin
        (* attempt to match 'n.head' with 'pat' *)
        (* This follows edag.ml, 'insertExpNode()', for translation between
         * Logic world and Edag world. *)
        let localSet:substSet = (
          match (pat, n.head, n.args) with
          | (NVar(v1), Edag.VAR(v2), []) -> (yieldSubIf (v1 = v2))
          | (Int(i1), Edag.INT(i2), []) -> (yieldSubIf ((to_int i1) = i2))
          | (BinOp(bop,e1,e2), BINOP bop', [_;_]) when bop = bop' ->
              (matchLists [e1;e2] n.args sub (nesting+1))
          | (Fun(name1,args1), Edag.UNINT(name2), args2) -> (
              if (name1 = name2) then
                (matchLists args1 args2 sub (nesting+1))
              else
                emptySubstSet
            )
          | (_,_,_) -> emptySubstSet    (* no match *)
        ) in

        (* concatenate accumulator and local substitution set *)
        let ret:substSet = (unionSubstSet localSet acc) in

        if (n.id = node.id) then (
          (* wrapped around, we're done iterating over equivalent nodes *)
          ret
        )
        else (
          (* continue with next node in list *)
          (loop ret n.eqc)
        )
      end in
      (loop emptySubstSet node.eqc)
      
      (* NOTE: The code above simply iterates over 'node.eqc'.  However,
       * as HPL-2003-148 points out, if two nodes are congruent then they
       * will yield the same set of matches.  Therefore one should only
       * consider one node from each congruence set (the congruence relation
       * is a subset of the equivalence relation).  But since the Kettle
       * e-dag does not maintain congruence classes, I can't do this, and
       * therefore may in general yield some substitutions more than once.
       *)
    )
  in

  (* note that we only print mismatches when no node in the entire set
   * of equivalent nodes matches; individual mismatches in the set are not
   * reported *)
  if (debug && (isEmptySubstSet ret)) then
    (ignore (E.log "matchTerm[%d]: mismatch\n" nesting));

  ret
end


(* Match a list of patterns with a list of e-dag nodes.  The accumulation
 * here is multiplicative, which leads to the exponential worst-case
 * behavior. *)
and matchLists (plist: exp list) (nlist: node list) (sub: substitution)
               (nesting: int) : substSet =
begin
  match (plist, nlist) with
  | ([], []) -> (
      (* reached the end of the lists, the initial substitution is
       * the final substitution *)
      (singletonSubstSet sub)
    )

  | ((pat :: ptail), (node :: ntail)) -> (
      (* match 'pat' with 'node' *)
      let headSet:substSet = (matchTermWithNode pat node sub nesting) in

      (* Now here it gets tricky: *each* substitution yielded from the
       * call just above gets used as an initial substitution for the
       * subsequent matching along this list.  That is how bindings from
       * earlier subtrees are communicated to later subtrees.  All of the
       * resulting substitution sets are unioned. *)
      let loop (acc: substSet) (headSub: substitution) : substSet =
      begin
        (unionSubstSet (matchLists ptail ntail headSub nesting) acc)
      end in
      (foldLeftOverSet loop emptySubstSet headSet)
    )
        
  | _ -> (
      (* This can happen when user-defined function symbols are used
       * with inconsistent numbers of arguments, since we don't actually
       * type check the input predicate.  Just say they are different. *)
      emptySubstSet
    )
end
(* ------------------------ END of matching ------------------------ *)


(* replace UVars with edag node exps *)
class substVisitor (sub: substitution) : logicVisitor = object
  inherit nopLogicVisitor
  method vexp (e:exp) : exp visitAction =
  begin
    match e with
    | UVar(idx) -> (
        match sub.(idx) with
        | Some(node) -> ChangeTo(node.exp)      (* replace *)
        | None -> (failwith "incomplete substitution")
      )
    | _ -> DoChildren
  end
end

(* Given a substitution created by 'unifyPattern', apply it to the rule predicate
 * to yield an instantiated fact.  There is a little wasted work here since
 * I am creating a Logic.exp out of Edag.nodes, which will then get converted
 * back into an Edag.node, but oh well. *)
let applySubst (rule: pred) (sub: substitution) : pred =
begin
  let vis:logicVisitor = (new substVisitor sub) in
  (visitLogicPred vis rule)
end

let applySubstToExp (e: exp) (sub: substitution) : exp =
begin
  let vis:logicVisitor = (new substVisitor sub) in
  (visitLogicExp vis e)
end


(* add a node to the e-dag *)
let internNode (state: Engine.state) (e: exp) : Edag.node =
begin
  (* this function exists because the interface of Edag.insertExpNode 
   * used to be more complicated; now it is just a wrapper *)
  (Edag.insertExpNode e)
end


(* use the equalities in rewrite mode only: match complete LHSs *)
(* (copied/modified from tryMatchRules, below) *)
let applyRewriteRules
  (state: Engine.state)
  (excludeRule: rule) (excludeSub: substitution)   (* avoid this pair *)
  : unit =
begin
  (debugLog "applyRewriteRules");

  (* one round of matching; return true if new fact added *)
  let roundOfMatching () : bool =
  begin
    (* loop over all the matching rules; return true if new fact added *)
    let attemptRule (r: rule) : bool =
    begin
      (* use only the equality rules *)
      match (r.rbody) with
      | PNez(BinOp(SetEq, lhs, rhs)) -> (     (* lhs,rhs are L.exp *)
          (* initial, empty substitution *)
          let emptySub: substitution = (Array.create r.numUnifVars None) in

          (* try to unify r's LHS with nodes in the e-dag *)
          (*                  ^^^     difference from tryMatchRules *)
          let subs: substSet = (matchTermWithAny lhs emptySub) in

          (* loop over the resulting substitutions, applying them *)
          let applySub (sub: substitution) : bool =
          begin
            (* avoid the exclusion pair *)
            if ((r == excludeRule) && (equalSubst sub excludeSub)) then
              false

            else (
              (* the unifying substitution is in 'sub'; apply it *)
              let slhs:exp = (applySubstToExp lhs sub) in
              let srhs:exp = (applySubstToExp rhs sub) in

              (* get corresponding e-dag nodes *)
              let nlhs:Edag.node = (internNode state slhs) in
              let nrhs:Edag.node = (internNode state srhs) in
              
              (* are these already equal? *)
              if (nlhs.root == nrhs.root) then
                false
                
              else (
                (* print the new fact *)
                (if !debug then
                  (ignore (E.log "MA: rewriting generates: %a = %a\n" 
                                 d_exp slhs d_exp srhs)));

                (* add this new fact *)
                (* inefficiency: this re-interns the nodes *)
                (Engine.addFact state
                  (PNez(BinOp(SetEq, slhs, srhs))));

                (* this may expose additional matching opportunities *)
                true
              )
            )
          end in
          (existsOverSet applySub subs)
        )
      | _ -> false
    end in
    (List.exists attemptRule !rules)
  end in

  (* do matching rounds until no new facts *)
  while (roundOfMatching ()) do
    (debugLog "next round of rewriting...");
  done
end


(* Given a symbolic predicate, return true if this predicate reduces
 * to a triviality.  Here, 'reduces' means in the Knuth-Bendix
 * sense, i.e. applying matching rules with full LHS triggers.
 *
 * The given rule 'r' and substitution 'sub' are the ones that gave
 * rise to the fact 'p', and hence the subsequent rewriting test must
 * explicitly avoid matching that pair again.
 *)
let reducesToTriviality
  (state: Engine.state)           (* so can call snapshot/undo/assume *)
  (p: L.pred)
  (r: rule) (sub: substitution)   (* match pair to avoid *)
  : bool =
begin
  (* for now, only simple equalities are reduced *)
  (match p with
  | PNez(BinOp(SetEq, t1, t2)) -> (     (* t1,t2 are L.exp *)
      (* 1. remember the current node generation count *)
      (Engine.snapshot state);

      (* 2. add both terms to the e-dag *)
      let n1:Edag.node = (internNode state t1) in
      let n2:Edag.node = (internNode state t2) in

      (* 3. apply full LHS matching to reduce the terms, sans (r,sub) *)
      (applyRewriteRules state r sub);

      (* 4. did the terms get merged? if so they reduce to triviality *)
      let triv:bool = (n1.root == n2.root) in

      (* 5. remove everything added to e-dag since step 1 *)
      (Engine.undo state);

      (* 6. report result *)
      triv
    )
  | _ -> (
      false
    )
  )
end


(* right now, 'reducesToTriviality' only handles equalities, so this
 * checks for such equalities so we can do different things for
 * forms that are not handled *)
let isEquality (p: L.pred) : bool =
begin
  (match p with
  | PNez(BinOp(SetEq, _, _)) -> true
  | _ -> false)
end


(* Kick off rule instantiation.  I considered doing this from within
 * 'assert' but decided that I neither aboutToFail==true nor ==false
 * has the right timing.  So the Engine has a hook to call this
 * directly. *)
let tryMatchRules (state: Engine.state) : unit =
begin
  (debugLog "tryMatchRules");

  (* one round of matching; return true if new fact added *)
  let roundOfMatching () : bool =
  begin
    (* loop over all the matching rules; return true if new fact added *)
    let attemptRule (r: rule) : bool =
    begin
      (* initial, empty substitution *)
      let emptySub: substitution = (Array.create r.numUnifVars None) in

      (* try to unify r's pattern with nodes in the e-dag, in general
       * yielding a set of substitutions *)
      let subs: substSet = 
        match r.rpattern with
        | [] -> (failwith "empty pattern list")
        | [onePat] -> (matchTermWithAny onePat emptySub)
        | firstPat :: rest -> (
            (* for now, just do the naive, inefficient thing: run each
             * match pattern independently, and intersect the results *)
            let initSubs: substSet = (matchTermWithAny firstPat emptySub) in
            (List.fold_left
              (fun (ss:substSet) (pat:exp) -> 
                (intersectSubstSet ss (matchTermWithAny pat emptySub)))
              initSubs rest)
          )
      in

      (* loop over the resulting substitutions, applying them *)
      let applySub (sub: substitution) : bool =
      begin
        (* the unifying substitution is in 'sub'; apply it *)
        let substituted:pred = (applySubst r.rbody sub) in

        (* print the fact *)
        (if !debug then
          (ignore (E.log "MA: new fact: %a\n" d_pred substituted)));

        (* test the fact for reducibility *)
        (if (reducesToTriviality state substituted r sub) then (
          (debugLog "the new fact reduces to a triviality; not adding");
          false       (* no new fact *)
        )
        else (
          (* add this new fact *)
          (Engine.addFact state substituted);

          (* for now, regard this is a new fact (that will causes us to
           * loop again), only if we tested it for triviality..
           *
           * 12/01/04: That's a little too clever for my own good.  It
           * makes it harder to predict when the prover will go into an
           * infinite loop.  I will make this just 'true', meaning that
           * we get KB-style behavior for pure equality rules and 
           * Simplify-style (potentially infinite looping) behavior for
           * rules that have more than just an equality.
           *
           * Ug, but then I make the same non-equality facts over and
           * over because I don't know they're redundant.... reverting.
           *)
          (isEquality substituted)
        ))
      end in
      (existsOverSet applySub subs)
    end in
    (List.exists attemptRule !rules)
  end in

  (* do matching rounds until no new facts *)
  while (roundOfMatching ()) do
    (debugLog "next round of matching...");
  done
end


(* initialize this satproc module *)
let init () : unit =
begin
  (debugLog "initialize");
  Engine.tryMatchRules := (Some tryMatchRules);
end


(* take a snapshot *)
let snapshot () : unit = ()


(* undo to state of most recent snapshot *)
let undo () : unit = ()


(* replace references to a variable with UVars *)
(* also, find the trigger pattern *)
class replaceVisitor (map: (string, int) Hashtbl.t) = object(self)
  inherit nopLogicVisitor

  (* this is empty until we find at least one trigger pattern *)
  val mutable trigger: exp list = []

  (* call this after using the visitor *)
  method getTrigger () : exp list =
  begin
    trigger
  end

  method vexp (e:exp) : exp visitAction =
  begin
    match e with
    | NVar(s) -> begin
        if (Hashtbl.mem map s) then
          ChangeTo(UVar(Hashtbl.find map s))    (* replace *)
        else
          DoChildren                            (* nop, really *)
      end

    | Fun("pat", args) -> begin
        (* trigger pattern *)

        (* 'pat' is a unary function *)
        let a:exp = match args with
          [a] -> a
        | _ ->
            ignore (E.warn "Trigger pattern must have exactly one argument");
            raise E.Error
        in

        (* must explicitly visit 'a' since the 'ChangeTo' action does not *)
        let a:exp = (visitLogicExp (self :> logicVisitor) a) in

        (* accumulate triggers *)
        trigger <- a :: trigger;

        (* rewrite the input so as to elide the use of 'pat' *)
        ChangeTo(a)
      end

    | _ -> DoChildren
  end
end


(* take note of a fact *)
let assume (acc: message list)   (* facts other procs have discovered this round *)
           (msg: message)        (* fact being asserted *)
           : message list =      (* 'acc', plus any facts I discover *)
begin
  if !debug then
    ignore (E.log "MA: assume: %a\n" d_message msg);

  (* remember universally-quantified facts *)
  (match msg.body with
  | Extern (All(_,_) as body) -> begin
      try 
        (* loop: as long as the fact is quantified, collect the quantified
         * variable names; eventually yield an unquantified predicate *)
        let rec loop (fact: pred) (nextIndex: int) (map: (string, int) Hashtbl.t) 
            : (pred * int) =
          begin
            match fact with
            | All(qvar,qbody) -> begin
                (* ensure uniqueness of the quantified variable name *)
                if (Hashtbl.mem map qvar) then begin 
                  ignore (E.warn "MA: duplicate quantified variable");
                  raise E.Error
                end;

                (* bind 'qvar' to 'nextIndex' *)
                (Hashtbl.add map qvar nextIndex);
                
                (* continue accumulating variables *)
                (loop qbody (nextIndex + 1) map)
            end

            | _ -> (fact, nextIndex)    (* not quantified; yield it *)
          end in
        
        (* use the loop to extract the variables *)
        let map: (string, int) Hashtbl.t = (Hashtbl.create 3) in
        let ((unquantFact:pred), (numUVs:int)) = (loop body 0 map) in
        
        (* finished collecting variables; now replace with UVars *)
        let vis:replaceVisitor = (new replaceVisitor map) in
        let substFact:pred = (visitLogicPred (vis :> logicVisitor) unquantFact) in
        
        (* get the trigger pattern *)
        let pats : exp list = (vis#getTrigger ()) in

        (* print the rule *)
        if !debug then (
          ignore (E.log "MA: patterns: %a\n" (d_list ", " d_exp) pats);
          ignore (E.log "MA: numUnifVars: %d\n" numUVs);
          ignore (E.log "MA: body: %a\n" d_pred substFact);
        );

        if ((List.length pats) = 0) then (
          (* ignore the rule... *)
          (ignore (E.warn "MA: no pattern specified"));
        )
        else (
          (* remember this rule *)
          rules := { rpattern = pats;
                     numUnifVars = numUVs;
                     rbody = substFact } :: !rules;
        );
      with E.Error -> ()
  end
  | _ -> ());     (* ignore non-quantified facts *)

  (* return the same set we started with *)
  acc
end


let witness (w: L.pred list) = w
                       
(* list of entry points exported as a satproc *)
let entries: Satproc.entries = {
  init      = init;
  snapshot  = snapshot;
  undo      = undo;
  assume    = assume;
  debug     = debug;
  ident     = matchid;
  witness   = witness;
  descr     = "match-based handling of universally-quantified assumptions";
}


(* EOF *)
