open Logic
open Options
open Proof
open Pretty

module E = Errormsg
module H = Hashtbl
module A = Ati

(* This can be set when only parsing, so that missing externs don't error *)
let ignoreExterns = ref false

let quietFlag = ref false

type coq_mode =
    CoqNone
  | CoqTerse
  | CoqVerbose

let coqFlag = ref CoqNone

let doShowCandidates = 2
let doShowUndo = 3
let debug = ref 0

let specialPrinter = ref defaultLogicPrinter

let d_exp () e = printExp !specialPrinter e
let d_pred () p = printPred !specialPrinter p

let spyTable : (string, unit) H.t = H.create 17
let setSpy yes what = 
  if yes then
    H.add spyTable what ()
  else
    try H.remove spyTable what with _ -> ()

     (* We will keep enough information to allow us to undo unification *)
type undoElem = 
    Snapshot of int
  | DoneUnif of subst * int
  | AddRule of rule
  | OverwriteContra of proof option (* The old contents of program.contra *)
  | Table of string


and program =                            (* A collection of rules  *)
    {         id: int;                   (* Unique identifier *)
      mutable undo: undoElem list;
      rules: (string, rule * int) H.t;   (* The rules, indexed by their head. 
                                          * For each rule we also store the 
                                          * size of its conclusion. We want 
                                          * to match first the rules with a 
                                          * bigger head *)
      rulesByName: (string, rule * int) H.t;
                                         (* The rules, indexed by their name *)

      tabled: (string, unit) H.t;        (* The set of tabled goal heads *)  
      tables: (pred, proof * subst) H.t; (* Map a tabled goal into the 
                                          * substitution and the proof *)
        
      mutable contra: proof option;      (* Remember if we can prove false *) 
      ati: rule A.state option;

      mutable goalId: int;
      mutable level: int;  (* The level we are at. Starts at 0 on the first 
                            * invocation and then increases with each nested 
                            * invocation *)
      mutable verify: in_channel -> out_channel -> bool;
	(* Pass channels for a coqtop process to verify that all constants and
	 * lemmas have Coq counterparts. *)

    } 


(* A DSL for specifying rules for Prolog-style proof search *)
and  rule = 
    { name:    string;                  (* For debugging. Also used for 
                                         * selecting rules  *)
      maxUnif: int;                     (* The maximal value of a 
                                         * unification variable index. *)
      conc:    Logic.pred;              (* The conclusion, typically 
                                         * containing unification variables.  *)
      sub:     subgoal;        
      init:    proof list;              (* The initial accumulator *)
                                        (* A proof constructor that is given 
                                         * the program, the current 
                                         * substitution and name of the rule 
                                         * and the accumulator *)
      isassume: bool;                   (* If it is an assumption *)
      mkprf:    program -> subst -> proof list -> proof
}

and  subgoal = 
               program -> (* The current program *)
               subst -> (* The substitution for this rule *)
               cont -> (* Success continuation. Pass on the accumulator + 
                          * everything new *)
              proof list ->      (* The accumulated results *)
              back ->    (* Backtracking function *)
              result 

    (* The success continuation of a subgoal *)
and  cont = proof list -> back -> result

    (* Retry means to retry the last successful subproof, independent of its 
     * level *)
    (* Commit i means to retry the last successful subproof at level lower 
     * than i. This means that a different clause for the parent might be 
     * tried *)
    (* Cut i means to skip all subgoals at level >= i and the parent (the 
     * last success at level i - 1). This means that no more clauses of the 
     * parent are retried  *)
and backinfo = Retry | Cut of int | Commit of int

    (* The failure continuation of a subgoal. Used for implementing 
     * backtraching. If the argument is Some i then it means that we must not 
     * backtrack on all successful goals at levels > i and also on the last 
     * successful subgoal at level i *)
and back = backinfo -> result

and result = Failure of backinfo | Success of proof * back


let level p = p.level
let goalId p = p.goalId
let setLevel p i = p.level <- i
let setGoalId p i = p.goalId <- i

let copyProgram (p: program) = 
  { p with level = p.level }

(**************** PRINTING **************)
let dax () ax = 
  dprintf "Rule %s(@[%a@])"
    ax.name d_pred ax.conc


let printRules (prg: program) : unit = 
  H.iter (fun _ (r, c) -> 
    ignore (E.log "%a@!" dax r))
    prg.rules

(************** COLLECTING STATS *********)
let prologStatsFile: string = 
  try Sys.getenv "PROLOG_STATS"
  with Not_found -> ""

let prologStats: bool = prologStatsFile <> ""

(** Keep for each rule some counters **)
type ruleStats = {
    mutable rsTryToMatch: int;
    mutable rsMatched: int;
    mutable rsFailed: int; 
  } 

let ruleStatsTable: (string, ruleStats) H.t = H.create 113

let initStats () = 
  H.clear ruleStatsTable;
  if prologStats then begin
    ignore (E.log "Will collect Prolog statistics\n");
    (try 
      let cin = open_in_bin prologStatsFile in
      ignore (E.log "  Loading previous Prolog stats from %s\n" prologStatsFile);
      let all: (string * ruleStats) list = Marshal.from_channel cin in
      close_in cin;
      List.iter (fun (n, rs) -> H.add ruleStatsTable n rs) all;
    with _ -> ignore (E.warn "Cannot read Prolog stats file %s\n" prologStatsFile))
  end
  
let getRuleStats (ax: rule) : ruleStats = 
  try H.find ruleStatsTable ax.name
  with Not_found -> begin
    let rs = { rsTryToMatch = 0; rsMatched = 0; rsFailed = 0 } in
    H.add ruleStatsTable ax.name rs;
    rs
  end

let statsTryToMatch (ax: rule) = 
  let rs = getRuleStats ax in
  rs.rsTryToMatch <- rs.rsTryToMatch + 1

let statsMatched (ax: rule) = 
  let rs = getRuleStats ax in
  rs.rsMatched <- rs.rsMatched + 1

let statsFailed (ax: rule) = 
  let rs = getRuleStats ax in
  rs.rsFailed <- rs.rsFailed + 1

let dumpStats (prg: program) = 
  if prologStats then begin
    ignore (E.log "Prolog stats\n");
    (* A list of name, stats, and frequency of matching *)
    let all : (string * ruleStats * float) list ref = ref [] in
    H.iter 
      (fun n (r, _) ->
        let rs = getRuleStats r in
        let freq = 
          if rs.rsTryToMatch = 0 then 1.0 
          else (float_of_int rs.rsMatched) /. (float_of_int rs.rsTryToMatch)
        in
        all := (n, rs, freq) :: !all;)
      prg.rulesByName;
    
    (* sort in increasing order of frequency *)
    let sorted = List.sort (fun (_, _, a) (_, _, b) -> compare a b) !all in
    List.iter
      (fun (n, rs, freq) -> 
        ignore (E.log "Rule %-30s Tried: %5d, Matched: %5d (%5.2f), Failed: %5d\n"
                  n rs.rsTryToMatch rs.rsMatched (100.0 *. freq) rs.rsFailed))
      sorted;
    (* Marshal the list out *)
    (try 
      let cout = open_out_bin prologStatsFile in 
      ignore (E.log "Saving the Prolog stats to %s\n" prologStatsFile);
      Marshal.to_channel cout
        (List.map (fun (n, rs, _) -> (n, rs)) sorted) [];
      close_out cout
    with _ -> 
      ignore (E.warn "Cannot save the Prolog stats to %s\n" prologStatsFile));
  end


(**************** THE UNDO STACK ************)
let snapshotId   = ref 0 

let snapshot prg = incr snapshotId;
                   if !debug >= doShowUndo then
                     ignore (E.log "KT:snapshot %d\n" !snapshotId);
                   prg.undo <- Snapshot !snapshotId :: prg.undo

let getPredIndex = function
    PFun(n, _) -> n
  | PNez(BinOp(bop, _, _)) -> binop_to_string bop
  | PNez(UnOp(unop, _)) -> unop_to_string unop
  | Bool b -> string_of_bool b
  | And ps -> "/\\"
  | Or ps -> "\\/"
  | Exist _ -> "EXISTS"
  | All _ -> "FORALL"
  | a -> E.s (E.unimp "getPredIndex: %a" d_pred a)


  (* Undo all the unifications *)
let undo ?(keepSubst=false) prg = 
  let rec loop = function
      [] -> E.s (E.bug "KT:empty undo stack")
    | Snapshot id :: rest -> 
        if !debug >= doShowUndo then
          ignore (E.log "KT:Finished undoing to snapshot %d\n" id);
        rest
    | DoneUnif (s, i) :: rest -> begin
        if not keepSubst then begin
          assert(i >= 0 && i < Array.length s);
          assert (s.(i) != Unif);
          s.(i) <- Unif;
        end;
        loop rest
    end
    | AddRule a :: rest -> begin
        if !debug >= doShowUndo then
            ignore (E.log "Retracting rule %s\n" a.name);
        let name = getPredIndex a.conc in
        (* Remove the last rule that we have added for this name *)
        H.remove prg.rules name;
        if a.name <> "" then 
          H.remove prg.rulesByName a.name;
        (match prg.ati with 
          Some ati -> A.deleteLastClause ati
        | _ -> ());
        loop rest
    end
    | Table s :: rest -> begin
        if !debug >= doShowUndo then
          ignore (E.log "Retracting tabling of %s\n" s);
        H.remove prg.tabled s;
        (* Now remove all the goals that have been memoized for it *)
        List.iter
          (fun p -> 
            match p with 
              PFun(n, _) when n = s -> 
                H.remove prg.tables p
            | _ -> ())
          (Util.keys prg.tables);

        loop rest
    end
    | OverwriteContra old :: rest ->
        if !debug >= doShowUndo then 
          ignore (E.log "Retracting contradition\n");
        prg.contra <- old;
        loop rest
  in
  prg.undo <- loop prg.undo

  (* Change a substitution and record it *)
let changeSub prg s i es = 
  prg.undo <- DoneUnif (s, i) :: prg.undo;
  assert (i >= 0 && i < Array.length s);
  s.(i) <- Done es;
  true

let undoBack (prg:  program) (back: back) : back = 
  fun cut -> undo prg; back cut


let eunify prg se1 se2 = Logic.eunify (changeSub prg) se1 se2
let punify prg sp1 sp2 = Logic.punify (changeSub prg) sp1 sp2

let dbinfo () = function
    Retry -> text "Retry"
  | Cut i -> dprintf "Cut %d" i
  | Commit i -> dprintf "Commit %d" i

(** Interface to ATI *)
class kettleClauseData (rule: rule) = object
  method name = rule.name
  method conc = rule.conc
  method hyps: pred list = []
  method maxUnif = rule.maxUnif
  method clause = rule
end

(****************** PROGRAMS *************)
let lastProgramId = ref 0 
let emptyProgram (useATI: bool) :  program = 
  { id        = (incr lastProgramId; !lastProgramId);
    goalId    = 0;
    level     = 0;
    undo      = [];
    rules     = H.create 17; (* The rules indexed by the getPredIndex of their 
                             * conclusion *)
    rulesByName = H.create 17; 

    tabled    = H.create 17;
    tables    = H.create 17;
    contra    = None;
    ati       = if useATI then begin
      let dummyRule = 
        { name = "<dummy rule>";
          maxUnif = -1;
          conc = Bool false;
          sub = (fun p s succ acc back -> E.s (E.bug "<dummy rule>"));
          init = [];
          isassume = false;
          mkprf = (fun p s acc -> E.s (E.bug "<dummy rule>"));
        } 
      in
      Some (A.make (new kettleClauseData dummyRule)) 
    end else 
      None;
    verify = fun _ _ -> false
  } 

let programId prg = prg.id

let rec findExpSize (e: exp) : int =
  match e with 
    Int _ | Float _ | Var _ | UVar _ | NVar _ -> 1
  | BinOp(_, e1, e2) -> 1 + findExpSize e1 + findExpSize e2
  | UnOp(_, e) -> 1 + findExpSize e
  | Fun(_, args) -> 
      List.fold_left (fun acc a -> acc + findExpSize a) 1 args
  | Ite(b,e1,e2) -> 1 + findExpSize b + findExpSize e1 + findExpSize e2

and findPredSize (p: pred) : int = 
  match p with 
    Bool _ -> 1
  | PFun(_, args) | NegPFun (_, args) -> 
      List.fold_left (fun acc a -> acc + findExpSize a) 1 args
  | PNez e -> findExpSize e
  | And ps | Or ps ->
      List.fold_left (fun acc p -> acc + findPredSize p) 1 ps
  | Not p -> 1 + findPredSize p
  | Exist (_, p) | All (_, p) -> 1 + findPredSize p
  | Imply (p1, p2) -> 1 + findPredSize p1 + findPredSize p2
  | Pannot (_, p) -> findPredSize p



  (** Create a special rule to announce that a constant is tabled *)
let tableConstantRuleName = "@@tableConstantRule"

  (* Add a new rule to the program *)
let extendProgram (prg: program) (ax: rule) = 
  if ax.name = tableConstantRuleName then begin
    let totable: string = 
      match ax.conc with 
        PFun(x, _) -> x
      | _ -> E.s (E.bug "Found tableConstantRule bug wrong conclusion")
    in
    if !debug > 0 then
      ignore (E.log "KT:Will table %s@!" totable);
    (* Record it in the undo list *)
    prg.undo <- Table totable :: prg.undo;
    H.add prg.tabled totable ();
    ()
  end else begin
    if !debug > 0 then
      ignore (E.log "KT:Adding %a@!" dax ax);
    let name = getPredIndex ax.conc in
    prg.undo <- AddRule ax :: prg.undo;
    let weight = 
      if ax.isassume then 
        1000 (* very high weight for assumptions, so that we try them first *)
      else
        findPredSize ax.conc
    in
    H.add prg.rules name (ax, weight);
    if ax.name <> "" then 
      H.add prg.rulesByName ax.name (ax, weight);
    (match prg.ati with 
      Some ati -> (* Add it to the ATI as well *) 
        A.addClause ati (new kettleClauseData ax)
          
    | _ -> ())
  end
      

  (* Do nothing tactical *)
let sgID = 
  fun prg s cont acc back -> cont acc back

let tableConstantRule (n: string) : rule = 
  { name    = tableConstantRuleName;
    conc    = PFun(n, []);
    maxUnif = -1;             (* No free unification variables *)
    sub     = sgID;
    init    = [];
    isassume = true;
    mkprf   = fun _ _ _ -> E.s (E.bug "mkPrf for tableConstantRule");
  } 

  (* Add an assumption to the program. Model it as a trivial rule *)
let assume prg (at: pred) (prf: proof) (name: string) = 
  if at = Bool false then begin (* Add a contradition *)
    prg.undo <- OverwriteContra prg.contra :: prg.undo;
    prg.contra <- Some prf
  end else begin
    extendProgram prg { name    = name;
                        conc    = at;
                        maxUnif = -1;             (* No free unification variables *)
                        sub     = sgID;
                        init    = [];
                        isassume = true;
                        mkprf   = fun _ _ _ -> prf;
                      } 
  end

(** We can select the rules to use, as the first step in the proof. *)
type ruleSelector = 
    AllRules                                (** Use all rules *)
  | OneRule of string                       (** Use only the given rule *)
  | OneRuleAmong of string list             (** Use one of several *)
  | AllRulesMatching of (rule -> bool)      (** Use all rules that match a 
                                             * given predicate *)



let getCandidateRules prg (axselector:  ruleSelector) 
                          (a : pred) 
                          (s: subst) : ( rule * int) list = 
  (* Now sort the axioms in the decreasing order of goal size *)
  let sortCandidates (c: (rule * int) list) : (rule * int) list =
    List.sort (fun (_,w1) (_,w2) -> - (compare w1 w2)) c
  in
  let showCandidates () axlist = 
    (docList ~sep:(chr ',' ++ break) (fun (ax, _) -> text ax.name)) () axlist
  in
  match axselector with 
    OneRule n -> begin
      if !debug >= doShowCandidates then 
        ignore (E.log "Candidates (OneRule): %s\n" n);
      try  [ H.find prg.rulesByName n ]
      with Not_found -> 
        ignore (E.warn "You wanted to use only non-existent rule %s\n" n);
        []
    end
  | OneRuleAmong nl -> 
      let candidates = 
        List.fold_left
          (fun acc n -> 
            try (H.find prg.rulesByName n) :: acc
            with Not_found -> acc)
          []
          nl
      in
      (* Sort them *)
      let sorted = sortCandidates candidates in
      if !debug >= doShowCandidates then 
        ignore (E.log "Candidates (OneRuleAmong ...): %a\n" 
                  showCandidates sorted);
      sorted
        
  | AllRules | AllRulesMatching _ -> begin
      let all_candidates = 
        match prg.ati with
          None -> 
            let aname: string = getPredIndex a in
            let all_candidates = H.find_all prg.rules aname in
            all_candidates
        | Some ati -> 
            let choices: (rule * int) list = 
              List.map 
                (fun c -> 
                  let r = c#clause in
                  (r, 1)) 
                (A.findChoices ati (a, s)) 
            in
            choices
      in
      let weighted_axioms: (rule * int) list = 
        match axselector with
          AllRules -> all_candidates
        | AllRulesMatching f -> 
            List.filter (fun (ax, _) -> f ax) all_candidates 
        | _ -> E.s (E.bug "Prolog.getCandidates")
      in
      (* Now sort the axioms in the decreasing order of goal size *)
      let sorted = sortCandidates weighted_axioms in
      if !debug >= doShowCandidates then begin
        ignore (E.log "Candidates: @[%a@]\n"
                  showCandidates sorted);
        if axselector <> AllRules then
          ignore (E.log "  selected from:@[%a@]\n" 
                    showCandidates all_candidates);
      end;
      sorted
  end

let checkMatchRule (prg: program) (s: subst) (ax: rule) (a: pred) : subst option = 
  (* Create a subtitution appropriate for rule *)
  let axs = newSubst (1 + ax.maxUnif) in
  if Logic.punify (changeSub prg) (ax.conc, axs) (a, s) then
    Some axs
  else begin

    (*if !debug >= 1 then
      ignore (E.log "Failed to unify %a with@! %a\n" 
		d_pred ax.conc
		d_pred a);*)

    None
  end

(************ TACTICALS **************)

  (* Always succeeds with a given result. *)
let sgTRUE (prg :  program) 
           (s : subst) (cont:  cont) (acc: proof list) (back: back) = 
       (* Add the res to the accumulator *)
  sgID prg s cont acc back

 (* Always fails *)
let sgFALSE prg s cont acc back = back Retry

 (* Always succeeds but on backtracking initiates a CUT *)
let sgCUT prg s cont acc back = 
  let thisId = prg.goalId in
  prg.goalId <- 1 + thisId;
  let thisLevel = prg.level in
  let id = dprintf "%d.%d:" thisId thisLevel in
  if !debug > 0 then
    ignore (E.log "KT:%a Install Cut@!" insert id);
  let back' binfo = 
    let newbinfo = match binfo with
      Retry -> Cut thisLevel
    | Cut i -> if i < thisLevel then binfo else Cut thisLevel
    | Commit i -> if i < thisLevel then binfo else Cut thisLevel
    in
    if !debug > 0 then
      ignore (E.log "KT:%a Processing cut. Propagating %a@!" 
                insert id dbinfo newbinfo);
    back newbinfo
  in
  cont acc back'

let sgCOMMIT prg s cont acc back = 
  let thisId = prg.goalId in
  prg.goalId <- 1 + thisId;
  let thisLevel = prg.level in
  let id = dprintf "%d.%d:" thisId thisLevel in
  if !debug > 0 then
    ignore (E.log "KT:%a Install Commit@!" insert id);
  let back' binfo = 
    let newbinfo = match binfo with
      Retry -> Commit thisLevel
    | Commit i -> if i < thisLevel then binfo else Commit thisLevel
    | Cut i -> if i <= thisLevel then binfo else Commit thisLevel
    in
    if !debug > 0 then
      ignore (E.log "KT:%a Processing commit. Propagating %a@!" 
                insert id dbinfo newbinfo);
    back newbinfo
  in
  cont acc back'

 (* Conditional *)
let sgIF (cond:  program -> subst -> proof list -> bool * proof list ) 
         (tsg:  subgoal) 
         (fsg:  subgoal) prg s cont acc back = 
  snapshot prg;  (* Take a snapshot because the conditional can have 
                  * side-effects *)
  let which, res = cond prg s acc in
  let back' = undoBack prg back in
  if which then 
    tsg prg s cont res back'
  else
    fsg prg s cont res back'

 (* Sequencing *)
let rec sgSEQ = function
  [] -> sgID
  | sg :: sglist -> 
      fun prg s cont acc back ->
        sg prg s (fun acc1 back1 -> sgSEQ sglist prg s cont acc1 back1)
          acc
          back
    
let (&&&) sg1 sg2 = 
  fun prg s cont acc back -> 
    sg1 prg s (fun acc1 back1 -> sg2 prg s (fun acc2 back2 -> cont acc2 back2)
                                           acc1
                                           back1)
              acc
              back
                                        
 (* A selection tactical. The selection function can use unification *)
let sgCHOICE (choose:  program -> subst -> proof list ->  subgoal * proof list)
    :  subgoal = 
  fun prg s cont acc back -> 
    snapshot prg;
    let sg', acc' = choose prg s acc in
    sg' prg s cont acc' (undoBack prg back)

 (* A disjunction. Consumes CUT *)
let rec sgOR :  subgoal list ->  subgoal = function
    [] -> sgFALSE
  | sg :: sglist -> 
      fun prg s cont acc back -> 
        sg prg s (fun acc1 back1 -> cont acc1 back1)
                 acc
                 (function Retry -> sgOR sglist prg s cont acc back
                   | cut -> back cut)

   (*** MAIN ENTRY POINT ***)
let rec prove (prg:  program) 
              ?(subst=emptySubst)
              ?(startRule=AllRules) 
              (a: pred) : proof option = 
  Stats.time "Prolog proving"
    (fun _ -> 
      prg.goalId <- 0;
      prg.level  <- 0;
      (* Record the undo stack and then snapshot immediately after *)
      let oldUndo = prg.undo in
      snapshot prg;

      let res = proveIntern prg startRule (a, subst) in
      (* Now revert back to the last state *)
      (* I want to preserve the substitutions. *)

      while prg.undo != oldUndo do
        undo ~keepSubst:true prg
      done; 

      match res with
        Success (x, _) -> 
          if !debug >= 3 then 
            ignore (E.log "Proof = @[%a@]\n" d_prf x);
          Some x
      | Failure _ -> None)
    ()

   (*** INTERNAL PROVER ****)
and proveIntern 
          (prg:  program) 
          (axselector: ruleSelector)
          ((a, s): spred) : result = 
  let thisId = prg.goalId in
  let thisLevel = prg.level in
  prg.level <- 1 + thisLevel;  (* Prepare the level for the subgoals *)
  let id = 
    if !debug > 0 then 
      dprintf "%d.%d:" thisId thisLevel 
    else
      nil
  in
  prg.goalId <- 1 + thisId;
  if !debug > 0 then
    ignore (E.log "KT:%a Call %a@!" insert id d_spred (a, s));
  (* Check the rules in turn *)
  let rec checkRules = function
      [] -> begin
        if !debug > 0 then
          ignore (E.log "            no match. Fail@!");
        Failure Retry
      end
    | (ax, _) :: rest -> begin
        snapshot prg;
        if prologStats then statsTryToMatch ax;
        match checkMatchRule prg s ax a with
          None -> (* This one does not actually match *)
            undo prg;
            checkRules rest
        | Some axsub ->   (* This is a match. Get the substitution for the 
                           * rule  *)
            if prologStats then statsMatched ax;
            if !debug > 0 then
              ignore (E.log "            match with %s@!" ax.name);
            let back binfo =
              undo prg; 
              let retry, newbinfo = match binfo with
                Retry -> true, Retry
              | Cut i -> false, if i > thisLevel then Retry else binfo
              | Commit i -> 
                  if i > thisLevel then true, Retry else false, binfo
              in
              if retry then begin
                if !debug > 0 then
                  ignore (E.log "KT:%a Redo %a@!"
                            insert id d_spred (a, s));
                checkRules rest 
              end else begin
                if !debug > 0 then
                  ignore (E.log "KT:%a Fail/Cut through with %a. Now=%a@!" 
                            insert id dbinfo binfo dbinfo newbinfo);
                if prologStats then statsFailed ax;
                Failure newbinfo
              end
            in
                (* Run the subgoals *)
            match ax.sub prg axsub 
                  (fun acc back -> 
		    (* RS 9/26/2003: This "mkprf" will blow up if 
		       we've succeeded but some variable isn't instantiated.
		       This means that every successful (sub)proof must be
		       fully instantiated!  Surely something more general
		       is possible, where some sibling goal instantiates
		       a variable? *)
		    begin try
		      Success (ax.mkprf prg axsub acc, back)
		    with
		      Invalid_argument "instantiation" ->
			Failure Retry
		    end)
                  ax.init 
                  (fun cut -> Failure cut) 
            with
                    (* Report success but be ready to continue scanning the 
                     * assumptions if somebody backtracks here *)
              Success (prf, back') -> begin
                if !debug > 0 then
                  ignore (E.log "KT:%a Success %a@!" 
                            insert id d_spred (a, s));
                (* Chain two backtracking functions together *)
                let rec prependToBack back1 = 
                  fun cut -> 
                    match back1 cut with
                      Failure cut -> back cut
                    | Success(pf, back') -> 
                        Success (pf, prependToBack back')
                in
                Success (prf, 
                         prependToBack back')
              end
            | Failure cut -> back cut
    end
  in
  (* See if we have a contradiction *)
  match prg.contra with 
    Some pf_false -> 
      if !debug > 0 then
        ignore (E.log "KT: proved with false elimination@!");
      prg.level <- prg.level - 1;
      Success (Unimplemented "falsee"(* Falsee pf_false *), 
               fun cut -> E.s (E.unimp "Prolog.back for falsee"))
  | _ -> begin
      (* Define the actual proof engine *)
      let proveit (a: pred) (s: subst) : result = 
        (* Get the list of rules that match this goal *)
        let axlist = getCandidateRules prg axselector a s in
        let res: result = checkRules axlist in
        prg.level <- prg.level - 1;
        res
      in
      (* See if we are tabling this goal *)
      match a with 
        PFun(n, _) when H.mem prg.tabled n -> 
          proveWithTables prg id a s proveit
      | _ ->  (* Not tabled. Just prove it *)
          proveit a s
  end

(** Proves a goal using tables *)
and proveWithTables (prg: program) (id: doc) 
                    (a: pred) (s: subst) 
                    (proveit: pred -> subst -> result) : result = 
  (* Apply the substitution completely. Rename the residual unification 
   * variables with UVar 0, ... Keep a list of already found unification 
   * variables. The first one we gave the index 0, ... *)
  let foundUnif: (int * subst) list ref = ref [] in
  let rec doOneExp (s': subst) (depth: int) e : exp option =
    match e with 
      UVar i -> begin
        assert (i >= 0 && i < Array.length s');
        match s'.(i) with 
        | Done (e'', s'') -> 
            Some (substExp ~depth:depth (doOneExp s'') e'')
        | Unif -> (* An exposed unification variable. See if we 
                   * already have it *)
            let rec loop (idx: int) = function
                [] -> (* We don't *)
                  foundUnif := !foundUnif @ [ (i, s') ];
                  Some (UVar idx)
              | (i', s'') :: _ when i = i' && s == s'' -> 
                  Some (UVar idx)
              | _ :: rest -> loop (idx + 1) rest
            in
            loop 0 !foundUnif
      end
    | _ -> None
  in   
  let a' = substPred (doOneExp s) a in
  (* Now make the substitution *)
  let s' = Array.of_list 
      (List.map (fun (i, s) -> Done (UVar i, s)) !foundUnif) in
  if !debug > 0 then 
    ignore (E.log "KT:%a Checking for tabled %a@!" 
              insert id d_spred (a', s'));
  try 
    let prf, sorig = H.find prg.tables a' in
    if !debug > 0 then 
      ignore (E.log "KT:%a Found in the table as %a@!" 
                insert id d_spred (a', sorig));
    (* Now apply the sorig to s'. We do this by unification, knowing that s' 
     * contains only unification variables which get instantiated. Thus, 
     * sorig will not get changed !!! *)
    assert (Array.length s' = Array.length sorig);
    for j = 0 to Array.length s' - 1 do
      if not (Logic.punify (changeSub prg) (a', s') (a', sorig)) then 
        E.s (E.bug "proveWithTables: unification");
    done;
    Success (prf, 
             fun cut -> 
               if not !quietFlag then 
                 ignore (E.warn "Backtracking through tabled goal %a"
                           d_spred (a', sorig));
               if !debug > 0 then 
                 ignore (E.log "KT:%a Failure (tabled backtrack) %a@!" 
                           insert id d_spred (a', sorig));
               Failure cut)
  with Not_found -> begin
    if !debug > 0 then 
      ignore (E.log "KT:%a Not found in the table@!" insert id);
    (* Now try to prove the original *)
    let res = proveit a' s' in
    match res with 
      Success (prf, back) -> 
        if !debug > 0 then 
          ignore (E.log "KT:%a Table %a@!" insert id d_spred (a', s'));
        H.add prg.tables a' (prf, s');
        res
                  
          (* Memoize the result *)
    | Failure _  -> res
  end
            
    
  (* Search for the proof of an atom. Uses a selector to rule out some 
   * rules. The selector CANNOT do unification, or must undo. Consume CUTS *)
let sgPROVE_sel (axselector: ruleSelector)
                (a: pred) 
                (trans: proof list -> proof -> proof list)  
                 : subgoal = 
  fun prg s cont acc back ->
    match proveIntern prg axselector (a, s) with
      Failure binfo -> back binfo
    | Success (prf, back') ->
        let rec backtrack (getNext : back) : back = 
          fun binfo -> begin
            match getNext binfo with
              Failure binfo' -> back binfo'
            | Success (prf, back') -> 
                cont (trans acc prf) (backtrack back')
          end
        in
        cont (trans acc prf) (backtrack back')
  
let sgPROVE a trans = sgPROVE_sel AllRules a trans

let sgASSUMPTION a trans = 
  sgPROVE_sel (AllRulesMatching (fun ax -> ax.maxUnif = -1)) a trans


let sgIsGround e = 
  fun prg s cont acc back ->
    match unrollExp (e, s) with
      (UVar i, s') when s'.(i) = Unif -> back Retry
    | _ -> cont acc back


let sgIsInt e f = 
  fun prg s cont acc back ->
    match unrollExp (e, s) with
      (Int i, _) when f (to_int i) -> cont acc back
    | _ -> back Retry

let sgIsConst e = 
  fun prg s cont acc back ->
    match unrollExp (e, s) with
      (Fun(_, []), _) -> cont acc back
    | _ -> back Retry

let sgDO f = 
  fun prg s cont acc back -> 
    f s acc;
    cont acc back

let sgCHECK f = 
  sgIF f sgTRUE sgFALSE
      
let sgCHECK1 f1 = 
  sgCHECK (fun p s acc -> if f1 s then true, acc else false, acc)
    
let sgACCUM f =
  fun prg s cont acc back ->
    cont (f prg s acc) back

let sgUNIF t1 t2 = sgCHECK (fun prg s acc -> (eunify prg (t1, s) (t2, s), acc))
      
 
(** For the external functions we must pass in some ML functions. All such 
 * functions must take a program argument, a substitution and a number of 
 * expressions. The expression arguments are already substituted before being 
 * passed to the function. The result of an external function is either 
 * "None", meaning that the external function failed, or "Some (s, po)", 
 * where "s" is some string that identifies the success kind and "po" is an 
 * optional proof. *)
type externFunc = 
  | ExtList of (program -> subst -> exp list 
             -> (string * proof option) option)

  | Ext1 of (program -> subst -> exp 
             -> (string * proof option) option)
  | Ext2 of (program -> subst -> exp -> exp 
             -> (string * proof option) option)
  | Ext3 of (program -> subst -> exp -> exp -> exp 
            -> (string * proof option) option)
  | Ext4 of (program -> subst -> exp -> exp -> 
                          exp -> exp 
            -> (string * proof option) option)
  | Ext5 of (program -> subst -> exp -> exp -> 
                             exp -> exp -> exp 
            -> (string * proof option) option)

(* Put here the external functions to be used by the parser *)
let externs: (string, externFunc) Hashtbl.t = Hashtbl.create 13

(* A pointer to the program to be used by the parser *)
let theProgram: program ref = ref (emptyProgram false)

type coqVerifier = in_channel -> out_channel -> bool

(* This is the pointer to the parser. It is set from Prolog_lex.mll *)
let thePrologParser : (Lexing.lexbuf -> 
                         rule list * lemma list * coqVerifier) ref = 
  ref (fun _ -> E.s (E.bug "thePrologParser was not initialized"))

let externFunctions: (string, externFunc) H.t = H.create 17

(* This is the main lexer entry point that dispatches to the right lexer *)
let parseRules ?(externs:(string * externFunc) list = []) 
               (fname: string) (prg:program)
               : (unit -> unit) * lemma list = 
  Stats.time "parsing Prolog rules"
    (fun _ -> 
      let lexbuf = E.startParsing fname in
      if not !quietFlag then
        ignore (E.log "Start parsing rules from file %s\n" fname);
      Parsing.clear_parser ();
      H.clear externFunctions;
      List.iter (fun (n, f) -> H.add externFunctions n f) externs;
      theProgram := prg; (* Set the program *)
      let rules, lemmas, ver = !thePrologParser lexbuf in
      H.clear externFunctions;
      E.finishParsing ();
      prg.verify <- ver;
      (fun () -> List.iter (fun r -> extendProgram prg r) rules), lemmas)
    ()
  

(* Parse rules from a string *)
let parseRulesFromString ?(externs:(string * externFunc) list = [])
                         (str: string) 
                         ?(file:string="<nofile>") 
                         ?(line:int=1)
                         (prg: program) : 
                         (unit -> unit) * lemma list = 
  Stats.time "parsing Prolog rules"
    (fun _ -> 
      let lexbuf = 
        E.startParsingFromString ~file:file ~line:line str in
      if not !quietFlag then
        ignore (E.log "Start parsing rules from string at %s:%d\n" file line);
      Parsing.clear_parser ();
      H.clear externFunctions;
      List.iter (fun (n, f) -> H.add externFunctions n f) externs;
      theProgram := prg; (* Set the program *)
      let rules, lemmas, ver = !thePrologParser lexbuf in
      H.clear externFunctions;
      E.finishParsing ();
      prg.verify <- ver;
      (fun () -> List.iter (fun r -> extendProgram prg r) rules), lemmas)
    ()


let options : optionDescr list = 
  let isDebug = ref false in
  [
    { optInUI = "debug Prolog";
      optRestart = false;
      optKind = OBool isDebug;
      optExtra  = (fun () -> debug := (if !isDebug then 1 else 0));
      optCommandLine = "";
      optHelp = "Turn on debugging for the Prolog interpreter" } ;

    { optInUI = "debug level Prolog";
      optRestart = false;
      optKind = OInt debug;
      optExtra  = (fun () -> isDebug := (!debug <> 0));
      optCommandLine = "-kettle-debug";
      optHelp = "Set the debugging level for the Prolog interpreter" } ;

 ]

let countRules prg = Hashtbl.fold (fun _ _ n -> n + 1) prg.rules 0

let verifyCoq prg inch outch = prg.verify inch outch

let varTypes = H.create 1000
let getVarTypes n = H.find varTypes n
let setVarTypes n vs = H.replace varTypes n vs

