
(** This module contains the implementation of a small but powerful logic 
 * interpreter. A program is a set of rules. Each rule has a name, a 
 * conclusion, a subgoal, and a proof generator that can produce a proof of 
 * the conclusion if the subgoal is satisfied. Both the conclusion and the 
 * subgoal can include unification variables. The power of this interpreter 
 * lies in the kinds of subgoals that are allowed. Like Prolog, the subgoals 
 * include hypotheses, cut, but also allow the execution of arbitrary Ocaml 
 * code to decide whether a subgoal is satisfied and to extend the current 
 * substitution. The language of subgoals also include conjunction and 
 * disjunction. *)

type program                            (** A collection of rules (has 
                                         * internal state) *)

                                        (** Create an empty program. Specify 
                                         * whether to use ATI for this 
                                         * program. *)
val emptyProgram: bool -> program


val copyProgram: program -> program     (** Make a copy of the mutable parts 
                                         * of the program (its execution 
                                         * state, such as the goal depth). 
                                         * This does not include the rules. 
                                         * If you want to retract the rules 
                                         * use the "snapshot" and "undo". *)

val programId: program -> int           (** A unique ID for each program *)

(**********************************************************************)
(**{1 External interface for Prolog} 

   The easiest way to use Prolog is to 
 * write the rules in a Prolog-style language in a file and then to load them 
 * into a program. The rules can refer to functions written in Ocaml. These 
 * functions are called "external" and they must be declared before parsing 
 * the rules. 

   An external function must take a program argument, a substitution and a 
   number of 
 * expressions. The expression arguments are already substituted before being 
 * passed to the function. The result of an external function is either 
 * "None", meaning that the external function failed, or "Some (s, po)", 
 * where "s" is some string that identifies the success kind and "po" is an 
 * optional proof.

*)

(** This can be set when only parsing, so that missing externs don't error *)
val ignoreExterns : bool ref

type externFunc = 
  | ExtList of (program -> Logic.subst -> Logic.exp list 
             -> (string * Proof.proof option) option)
            (** This is the general type of a external function *)
  | Ext1 of (program -> Logic.subst -> Logic.exp 
             -> (string * Proof.proof option) option)
            (** This is an external function that takes only one expression 
             * argument *)
  | Ext2 of (program -> Logic.subst -> Logic.exp -> Logic.exp 
             -> (string * Proof.proof option) option)
            (** This is an external function that takes two expression 
             * arguments *)
  | Ext3 of (program -> Logic.subst -> Logic.exp -> Logic.exp -> Logic.exp 
            -> (string * Proof.proof option) option)
  | Ext4 of (program -> Logic.subst -> Logic.exp -> Logic.exp -> 
                          Logic.exp -> Logic.exp 
            -> (string * Proof.proof option) option)
  | Ext5 of (program -> Logic.subst -> Logic.exp -> Logic.exp -> 
                             Logic.exp -> Logic.exp -> Logic.exp 
            -> (string * Proof.proof option) option)

(** Parse a set of rules specified in the given file, using the given list of 
 * external functions. Use the declarations from the given program. Returns a 
 * closure that can be used to add the parsed rules into the program, and a 
 * list of corresponding lemmas. *)
val parseRules: ?externs:(string * externFunc) list -> 
                string -> 
                program -> 
                (unit -> unit) * Proof.lemma list

(** Parse a set of rules specified in the given string, using the given list 
 * of external functions. Use the declarations from the given program. 
 * Returns a closure that can be used to add the parsed rules into the 
 * program, and a list of corresponding lemmas. *)
val parseRulesFromString: ?externs:(string * externFunc) list -> 
                          string -> 
                          ?file:string -> 
                          ?line:int -> 
                          program -> (unit -> unit) * Proof.lemma list

(** Extend by asserting a predicate, along with a proof and a name. This 
 * creates and adds to the program a rule with the given conclusion and no 
 * subgoal. *)
val assume: program -> Logic.pred -> Proof.proof -> string -> unit


(** Tell the {!Prolog.program} to remember this state. You can use this 
 * before you use {!Prolog.assume} to add an assumption that you might want 
 * to retract later. Similarly you can use this before adding to a program a 
 * set of parsed rules that you might want to retract. *)
val snapshot: program -> unit      

(** Revert the state of the program to the last un-matched 
 * {!Prolog.snapshot}. You can use this to retract assertions. *)
val undo: ?keepSubst:bool -> program -> unit


(** An external function can use unification on expressions and on predicates *)
(** Unification for expressions *)
val eunify: program -> Logic.sexp -> Logic.sexp -> bool

(** Unification for predicates *)
val punify: program -> Logic.spred -> Logic.spred -> bool



type subgoal                           (** The type of subgoals *)

(** The internal type of a rule. You should not need to use this, unless you 
 * want to use the direct interface to Prolog *)
type rule = 
    { name:    string;                  (* For debugging. Also used for 
                                         * selecting rules.  *)
      maxUnif: int;                     (* The maximal value of a 
                                         * unification variable index.  *)
      conc:    Logic.pred;                (** The conclusion, typically 
                                         * containing unification variables. *)
      sub:     subgoal;                 (** The subgoal *)
      init:    Proof.proof list;        (** The initial accumulator *)

                                        (** A proof constructor that is given 
                                         * the program, the current 
                                         * substitution and name of the rule 
                                         * and the accumulator. *)
      isassume: bool;                   (* If it is an assumption *)
      mkprf:   program -> Logic.subst -> Proof.proof list -> Proof.proof
}




(** We can select the rules to use, as the first step in the proof. *)
type ruleSelector = 
    AllRules                                (** Use all rules *)
  | OneRule of string                       (** Use only the given rule *)
  | OneRuleAmong of string list             (** Use one of several *)
  | AllRulesMatching of (rule -> bool)      (** Use all rules that match a 
                                             * given predicate *)


(******************************************************************)
(** {2 Invoking Prolog} *)



(** Prove a predicate given the program. A substitution can be given if the 
 * predicate to be proved contains unification variables. If a [startRule] is 
 * provided then the prover uses only that rule to start the proof. For the 
 * proofs of the subgoal any rule in the program can be used. Returns a proof 
 * if it was successful and fills in the substitution. *)
val prove: program -> ?subst: Logic.subst 
                   -> ?startRule: ruleSelector
                   -> Logic.pred -> Proof.proof option

(******************************************************************)
(** {2 Debugging Prolog Rules} 
  If you turn on debugging and you will see a series of lines:

    [KT:n.l: operation goal]

    where [n] is the invocation identifier. Every time {!Prolog.prove} 
    is invoked internally a 
    new invocation identifier is generated. The invocation identifier 
    is reset at the 
    external invocation of {!Prolog.prove}. 

    [l] is the level. It starts at 0 and as prove is invoked recursively it 
    grows. 

    operation is one of: [Call], [Success], [Redo], [Fail] or [Cut]
    an invocation is created by prove and the first operation is a [Call]. If 
    one clause matches then the next operation on the invocation is [Success] 
    and the subgoals are processed. If one of the subgoals backtracks the 
    [Redo] operation is initiated and the next clause is tried followed by
    [Success] and subgoals, etc. When [Call] or [Redo] fail the Fail operation 
    * is 
    initiated. [Cut] is a special case of [Redo] which does not attempt 
    any other  
    matches but [Fail]s immediately. 


    This debugger model is inspired by Prolog debuggers. See for example the 
    Sicstus Prolog manual online:
     http://www.cs.bham.ac.uk/~pjh/prolog_course/sicstus_manual_v3_5/sicstus_9.html#SEC75
    
      
*)

(** Turn on debugging. Value of 0 turns off debugging, 1 shows the call 
 * stack, 2 also shows the candidates for matching, 3 shows the proofs. *)
val debug : int ref

(** Print the names of the active rules in the program *)
val printRules: program -> unit


(** Initialize the collection of statistics. Statistics are only collected if 
 * the environment variable PROLOG_STATS is set *)
val initStats: unit -> unit

val dumpStats: program -> unit

(********************************************************************)
(** {2 Profiling Prolog Rules}

   Not yet *)

(*******************************************************************)
(** {1 Internal interface to Prolog} It is also possible to specify the rules 
 * directly in Ocaml. There are two advantages of only doing this: you do not 
 * have to create a separate file and the interface with the external 
 * functions is somewhat more direct. However, this is a lower level 
 * interface and you it can take as much as 10 times more code to accomplish 
 * the same thing. *)



                                        (** Extend the program by adding a new 
                                         * rule *)
val extendProgram: program -> rule -> unit

 (** Create a special rule that announces that a constant is tabled *)
val tableConstantRule: string -> rule


(*** TACTICALS for subgoals ***)
    (** Always succeeds. *)
val sgTRUE: subgoal 

    (** Always fails (backtracks) *)
val sgFALSE: subgoal


    (** Always succeeds but on backtracking it does not attempt to find other 
     * solutions for all its previously appearing sibling subgoals nor for 
     * the conclusion *)
val sgCUT: subgoal

    (** Always succeeds but on backtracking it does not attempt to find 
     * another solutions for all its previously appearing sibling subgoals. 
     * It does attempt to find other subgoals for the parent.*)
val sgCOMMIT: subgoal


    (** A general choice tactical. The choice function can use unification *)
val sgCHOICE: (program -> Logic.subst -> Proof.proof list -> subgoal * Proof.proof list) 
               -> subgoal

    (** Depending on the current substitution decides which of the hypotheses 
     * to do. In the process it updates the accumulated result. The choice 
     * function can use unification. "sgIF f sgtrue sgfalse" is like 
     * "sgCHOICE (fun p s acc -> let which, acc' = f p s acc in if which then 
     * sgtrue, acc' else sgfalse, acc')" *) 
val sgIF:     (program -> Logic.subst -> Proof.proof list -> bool * Proof.proof list) 
              -> subgoal -> subgoal -> subgoal


    (** Checks that a condition holds and continues if it does. The test 
     * function can update the substitution. This is just an abbreviation for 
     * sgIF f sgTRUE sgFALSE. *)
val sgCHECK: (program -> Logic.subst -> Proof.proof list -> bool * Proof.proof list) -> subgoal


   (** A simpler version of sgCHECK where the function returns the same accumulator. *)
val sgCHECK1: (Logic.subst -> bool) -> subgoal

    (** Like {!Prolog.sgCHECK} with success f the two terms unify *)
val sgUNIF: Logic.exp -> Logic.exp -> subgoal

    (** A sequencing of tacticals *)
val sgSEQ: subgoal list -> subgoal

    (** A binary sequencing operator *)
val (&&&) : subgoal -> subgoal -> subgoal

    (** A disjunction. The goals are tried in order. *)
val sgOR: subgoal list -> subgoal


    (** Checks that the given predicate is already known to be true (without 
     * having to prove any more subgoals). The predicate shares unification 
     * variables with the goal. In the process of searching the known 
     * assumptions it does instantiate some unification variables. *)
val sgASSUMPTION: Logic.pred -> (Proof.proof list -> Proof.proof -> Proof.proof list) -> subgoal


    (** Search for the proof of the atom. The unification variables in the 
     * predicate are shared with the rule conclusion. The given function 
     * should collect the proof. *)
val sgPROVE: Logic.pred -> (Proof.proof list -> Proof.proof -> Proof.proof list) -> subgoal


    (** A function that just processes the accumulator *)
val sgACCUM: (program -> Logic.subst -> Proof.proof list -> Proof.proof list) -> subgoal

    (** Succeeds if the given expression is not a unification variable *)
val sgIsGround: Logic.exp -> subgoal

    (** Succeeds if the given expression is an integer meeting the given 
     * condition *)
val sgIsInt: Logic.exp -> (int -> bool) -> subgoal

    (** Suceeds if the given expression is a constant *)    
val sgIsConst: Logic.exp -> subgoal

type coqVerifier = in_channel -> out_channel -> bool

(** This is a pointer that is set to point to the parsing function. You 
 * should not use this one *)
val thePrologParser: (Lexing.lexbuf -> 
                         rule list * Proof.lemma list * coqVerifier) ref

(** A pointer to the current program, during parsing *)
val theProgram: program ref

(** This is where we communicate to the parser the current externs *)
val externFunctions: (string, externFunc) Hashtbl.t


(** Export a few global references for printing debugging information from 
 * kettle_parse *)
val goalId: program -> int
val level: program -> int
val setGoalId: program -> int -> unit
val setLevel: program -> int -> unit

val debug: int ref

val options: Options.optionDescr list

(** If you set this flag then Prolog will be quiet *)
val quietFlag: bool ref

(* Control the level of checking against Coq implementations *)
type coq_mode =
    CoqNone
  | CoqTerse
  | CoqVerbose
val coqFlag : coq_mode ref

(* Count how many rules a program has *)
val countRules: program -> int

(* Database of lemma variable types *)
val getVarTypes : string -> string list
val setVarTypes : string -> string list -> unit

(* Check that the definitions for this program are validated by a coqtop
 * process with the given channels *)
val verifyCoq: program -> in_channel -> out_channel -> bool


(** The user of this module can override the printer for expressions and 
 * predicates *)
val specialPrinter: Logic.logicPrinter ref

