(* This program is free software.  See LICENSE for more information *)

(* smtlib.ml *)

module L = Logic

(** contains all the types, exceptions and helper functions used in the 
 * parsing process *)

(** signifies that a boolean connective was attempted to be used with an 
 * unaccepted number of arguments. i.e (implies x) or (not x y) *)
exception Connective_rank_mismatch
exception Validation_error of string

(* TO-DO get list of valid logic names *)
let logic_names = Hashtbl.create 100;;
Hashtbl.add logic_names "---" "---";;


(* create tables used in parser *)
  let symbol_table = Hashtbl.create 10;;
  let extrasorts = Hashtbl.create 100;;
	let extrafuns = Hashtbl.create 100;;
	let extrapreds = Hashtbl.create 100;;

(* initialize tables so they have a type *)
Hashtbl.replace symbol_table "---" "---";
Hashtbl.replace extrasorts "---" "---";
Hashtbl.replace extrafuns "---" "---";
Hashtbl.replace extrapreds "---" "---";

(* normally, these will be gotten from somewhere else *)

Hashtbl.replace extrasorts "Int" "Int";
Hashtbl.replace extrasorts "Real" "Real";
Hashtbl.replace extrasorts "Array" "Array";
Hashtbl.replace extrapreds "<=" "<=";
Hashtbl.replace extrapreds ">=" ">=";
Hashtbl.replace extrapreds ">" ">";
Hashtbl.replace extrapreds "<" "<";
Hashtbl.replace extrapreds "select" "select";
Hashtbl.replace extrafuns "+" "+";
Hashtbl.replace extrafuns "-" "-";
Hashtbl.replace extrafuns "*" "*";
Hashtbl.replace extrafuns "select" "select";
Hashtbl.replace extrafuns "store" "store";
Hashtbl.replace extrafuns "~" "~";


(** type of a variable's id *)
type var_id = string

(** the type of an annotation. *)
type annotation = string * string option

(** contains a list of annotations*)
type annotations = annotation list

(** represents a quantifier *)
type quant =  FORALL 
            | EXISTS

(** the type that sorts are stored as *)
type sort_name = string

(** associates a variable id to its corresponding sort*)
type binder = var_id * sort_name

(** the type used to represent a sort declaration along with any annotations 
 * that it carries*)
type annotated_sort = ANNOTATED_SORT of sort_name * annotations

(** the type used to represent a predicate declaration, along with any 
 * annotations it carries.*)
type pred_sig = PRED_SIG of string * sort_name list * annotations

(** the type used to represent a function declaration, along with any 
 * annotations it carries.*)
type func_sig = FUNC_SIG of string * sort_name list * sort_name * annotations

type term = Logic.exp
(*
    T of term2 (** term without annotations*)
  | TA of term2 * annotations (** term with annotations*)

(** the datatype that is used to store a term*)
and term2 = Logic.
  | VAR of var_id (** a variable(could be a bound variable, or a variable 
                   * from an earlier let formula*)
  | NUMERAL of int32 (** a numeral *)
  | FUN of string * term list (** A function*)
  | ITE of formula * term * term (** An if then else term.*)
*)

(** the datatype used to store the type of term being used*)
and formula =
    F of formula2 (** a formula without annotations*)
  | FA of formula2 *annotations (** an annotated formula*)

(** the datatype that is used to store a formula *)
and formula2 =
    FLET of string * formula * formula (** a flet formula *)
  | LET of string * term * formula (** a let formula *)
  | FVAR of string (** a variable that corresponds to a previous flet formula*)
  | AND of formula list (** a and b and c and d ... where a b c d ... are 
                         * formulas*)
  | OR of formula list (** a or b or c or d or ... where a b c d ... are 
                        * formulas*)
  | NOT of formula (** not a where a is a formula*)
  | IMPLIES of formula*formula (** a-> b where a and b are formulas*)
  | XOR of formula list (** a xor b xor c xor d ... where a b c d ... are 
                         * formulas*)
  | IFF of formula list (** a <--> b <--> c ... where a b c d ... are formulas*)
  | DISTINCT of term list (** t1 is distinct from t2 is distinct from t3 ... 
                           * where t1, t2, t3 ,... are terms *)
  | EQUALS of term list (** t1=t2=t3=... where t1 t2 t3 ... are terms*)
  | IF_THEN_ELSE of formula * formula * formula (** Store the formula If Then 
                                                 * Else construct *)
  | PRED of string * term list (** Store a predicate with a list of its 
                                * arguments. Note that a propositional 
                                * predicate would simply use the empty list 
                                * for terms *)
  | TRUE_PRED (** Stores the predicate for true *)
  | FALSE_PRED (** Stores the predicate for false *)
  | QUANT of quant * binder list * formula (** Stores the type of quantifier 
                                            * in a, the list of (variable, 
                                            * sort) pairs in b, and the 
                                            * quantified subformula in c *)
(** the datatype used to store the type of formula being used*)


(** Used to store the status of a benchmark*)
type status = SAT 
            | UNSAT
            | UNKNOWN



        
(** Stores the name of a benchmark and a list of all the benchmark attributes 
 * that it has *)
type benchmark = 
    { mutable name: string;
      mutable assumptions: Logic.pred list;
      mutable status: status;
      mutable formula: Logic.pred;
      mutable logic: string;
      mutable extrasorts: sort_name list;
      mutable extrafuns: func_sig list;
      mutable extrapreds: pred_sig list;
      mutable notes: string list;
      mutable annotations: annotation list;
    }

(* Will update in place here the attributes *)
let currentBenchmark = { 
    name = "";
    assumptions = [];
    formula = L.Bool true;
    status = UNKNOWN;
    logic = "";
    extrasorts = [];
    extrapreds = [];
    extrafuns = [];
    notes = [];
    annotations = [];
  } 
let formulaDefined = ref false
let statusDefined = ref false
let logicDefined = ref false


(** Keep here some information about the new names that were added *)
let parserWitness: Pretty.doc list ref = ref []  

let initParser () = 
  currentBenchmark.assumptions <- [];
  currentBenchmark.formula <- L.Bool true;
  currentBenchmark.status <- UNKNOWN;
  currentBenchmark.logic <- "";
  currentBenchmark.extrasorts <- [];
  currentBenchmark.extrafuns <- [];
  currentBenchmark.extrapreds <- [];
  currentBenchmark.notes <- [];
  currentBenchmark.annotations <- [];
  
  formulaDefined := false;
  statusDefined := false;
  logicDefined := false;

  parserWitness := [];
  
  ()
  

