%{
(* This program is free software.  See LICENSE for more information *)
module I = Input
module E = Errormsg
module H = Hashtbl

open Smtlib
open Pretty


let parse_error msg : unit =
  E.parse_error msg


let checkTermAnnotations (a: annotations) : unit = 
  if a != [] then 
    ignore (E.warn "Found term annotations")

let addFormulaAnnotations (f: L.pred) (a: annotations) : L.pred = 
  if a != [] then begin
    ignore (E.warn "Found formula annotations\n");
  end;
  f

let letBindings: (string, L.exp) H.t = H.create 113
let fletBindings: (string, L.pred) H.t = H.create 113

let debugNewNames = false

let pushBinding (b: (string, 'a) H.t) (n: string) (v: 'a) : unit = 
  H.add b n v
let popBinding (b: (string, 'a) H.t) (n: string) : unit = 
  H.remove b n

let foundBound (b: (string, 'a) H.t) (n: string) : 'a = 
  try
    H.find b n
  with Not_found -> 
    E.s (E.bug "Cannot find bound variable %s" n)

let checkDecls = false

let negatePred (p: L.pred) =
  match p with
    L.Bool _ | L.PNez _ -> L.negatePred p
  | _ -> L.Not p

let accumIff (p1: L.pred) (p2: L.pred) (acc: L.pred list) : L.pred list = 
  match p1, p2 with 
    L.PNez(L.BinOp((L.SetEq|L.SetNe) as op1, v1, t1)),
    L.PNez(L.BinOp((L.SetEq|L.SetNe) as op2, v2, t2)) 
    when t1 == L.trueExp && t2 == L.trueExp -> 
      if op1 = op2 then 
        (L.eq v1 v2) :: acc
      else
        (L.ne v1 v2) :: acc

  | _ -> L.Imply (p1, p2) :: L.Imply (p2, p1) :: acc

let nextIteId = ref 0 
let mkExpIte (p: L.pred) (e1: L.exp) (e2: L.exp) = 
  let iteexp = L.Fun("ite_exp" ^ string_of_int !nextIteId, []) in
  incr nextIteId;
  if debugNewNames then 
    ignore (E.log "introduced %a = if %a\n  then %a\n  else %a\n"
              L.d_exp iteexp L.d_pred p L.d_exp e1 L.d_exp e2);
  (* Remember the witness *)
  parserWitness :=
    (dprintf "%a = if %a then %a else %a"
       L.d_exp iteexp L.d_pred p L.d_exp e1 L.d_exp e2) ::
    !parserWitness;

  currentBenchmark.assumptions <- 
    (L.Or [ L.And [ p; L.eq iteexp e1];
            L.And [ negatePred p; L.eq iteexp e2] ]) 
    :: currentBenchmark.assumptions;
  iteexp

let nextPredicateId = ref 0
let newPredicateName () = 
  let r = L.pFun ("pred" ^ string_of_int !nextPredicateId) [] in
  incr nextPredicateId;
  r

%}

%token <string> NUMERAL_TOK
%token <string> VAR_TOK
%token <string> SYM_TOK
%token <string> STRING_TOK
%token <string> AR_SYMB
%token <string> USER_VAL_TOK
%token TRUE_TOK
%token FALSE_TOK
%token ITE_TOK
%token NOT_TOK
%token IMPLIES_TOK
%token IF_THEN_ELSE_TOK
%token AND_TOK
%token OR_TOK
%token XOR_TOK
%token IFF_TOK
%token EXISTS_TOK
%token FORALL_TOK
%token LET_TOK
%token FLET_TOK
%token NOTES_TOK
%token SORTS_TOK
%token FUNS_TOK
%token PREDS_TOK
%token EXTENSIONS_TOK
%token DEFINITION_TOK
%token AXIOMS_TOK
%token LOGIC_TOK
%token COLON_TOK
%token LPAREN_TOK
%token RPAREN_TOK
%token SAT_TOK
%token UNSAT_TOK
%token UNKNOWN_TOK
%token ASSUMPTION_TOK
%token FORMULA_TOK
%token STATUS_TOK
%token BENCHMARK_TOK
%token EXTRASORTS_TOK
%token EXTRAFUNS_TOK
%token EXTRAPREDS_TOK
%token LANGUAGE_TOK
%token DOLLAR_TOK
%token EQUALS_TOK
%token DISTINCT_TOK
%token SEMICOLON_TOK
%token EOF_TOK

%start benchmark
%type <Smtlib.benchmark> benchmark
%type <Logic.pred> an_formula prop_atom an_atom
%type <Logic.exp> basic_term

%%
benchmark:
    LPAREN_TOK BENCHMARK_TOK SYM_TOK bench_attributes RPAREN_TOK
    {
      
     currentBenchmark.name <- $3;
     if checkDecls then begin
       if not !formulaDefined then 
         raise (Validation_error "Formula not defined.");
       if not !statusDefined then 
         raise (Validation_error "Status not defined.");
       if not !logicDefined then 
         raise (Validation_error "Logic not defined.");
       
       Hashtbl.clear symbol_table;
       Hashtbl.clear extrasorts;
       Hashtbl.clear extrafuns;
       Hashtbl.clear extrapreds;
       
       Hashtbl.replace symbol_table "---" "---";
       Hashtbl.replace extrasorts "---" "---";
       Hashtbl.replace extrafuns "---" "---";
       Hashtbl.replace extrapreds "---" "---";
       
       Hashtbl.replace extrasorts "Int" "Int";
       Hashtbl.replace extrasorts "Real" "Real";
       Hashtbl.replace extrasorts "Array" "Array";
       Hashtbl.replace extrapreds "<=" "<=";
       Hashtbl.replace extrapreds ">=" ">=";
       Hashtbl.replace extrapreds ">" ">";
       Hashtbl.replace extrapreds "<" "<";
       Hashtbl.replace extrapreds "select" "select";
       Hashtbl.replace extrafuns "+" "+";
       Hashtbl.replace extrafuns "-" "-";
       Hashtbl.replace extrafuns "*" "*";
       Hashtbl.replace extrafuns "select" "select";
       Hashtbl.replace extrafuns "store" "store";
       Hashtbl.replace extrafuns "~" "~";
     end;
     currentBenchmark

   }

bench_attributes:
    bench_attribute bench_attributes { () }
  | bench_attribute                  { () }

bench_attribute:
      COLON_TOK ASSUMPTION_TOK an_formula {
         currentBenchmark.assumptions <- $3 :: currentBenchmark.assumptions
    } 

  | COLON_TOK FORMULA_TOK an_formula {
         currentBenchmark.formula <- $3;
         if checkDecls && !formulaDefined then 
           raise (Validation_error "More than one formula found.");
         formulaDefined := true;
            
    }

  | COLON_TOK STATUS_TOK status {
         currentBenchmark.status <- $3;
         if checkDecls && !statusDefined then 
           raise (Validation_error "More than one status found.");
         statusDefined := true;
     }
  | COLON_TOK LOGIC_TOK SYM_TOK {
         currentBenchmark.logic <- $3;
         if checkDecls && !logicDefined then 
           raise (Validation_error "More than one logic definition found.");
         logicDefined := true;
     }

  | COLON_TOK EXTRASORTS_TOK LPAREN_TOK sort_symb_decls RPAREN_TOK {
        currentBenchmark.extrasorts <- currentBenchmark.extrasorts @ $4
   }

  | COLON_TOK EXTRAFUNS_TOK LPAREN_TOK fun_symb_decls RPAREN_TOK  {
        currentBenchmark.extrafuns <- currentBenchmark.extrafuns @ $4
   }

  | COLON_TOK EXTRAPREDS_TOK LPAREN_TOK pred_symb_decls RPAREN_TOK {
        currentBenchmark.extrapreds <- currentBenchmark.extrapreds @ $4
   }

  | COLON_TOK NOTES_TOK STRING_TOK {
        currentBenchmark.notes <- $3 :: currentBenchmark.notes
   }
  | annotation {
        currentBenchmark.annotations <- $1 :: currentBenchmark.annotations
   }



status:
    SAT_TOK {SAT}
  | UNSAT_TOK {UNSAT}
  | UNKNOWN_TOK {UNKNOWN}

sort_symbs:
    sort_symb sort_symbs
    { 
      (* returns a list of sort names *)
      $1::$2
    }
  | sort_symb 
    { 
      (* returns a singleton list of one sort name *)
      [$1]
    }

sort_symb_decls:
    sort_symb_decls sort_symb_decl
    { 
      (* returns a list of sort names *)
      $2::$1
    }
  | sort_symb_decl 
    { 
      (* returns a singleton list of one sort name *)
      [$1]
    }


fun_symb_decls:
    fun_symb_decls fun_symb_decl { $2::$1 }
  | fun_symb_decl { [$1] }



fun_symb_decl:
    LPAREN_TOK fun_symb sort_symb sort_symbs annotations RPAREN_TOK
    {
      (* returns the signature of a function with annotations *)
        if checkDecls && Hashtbl.mem extrafuns $2 then
          raise (Validation_error ("function "^$2^" already defined."))
        else (
       	  if checkDecls then Hashtbl.add extrafuns $2 $2;
	  FUNC_SIG($2,$4, $3,$5))
    }

  | LPAREN_TOK fun_symb sort_symb annotations RPAREN_TOK
    {
      (* returns the signature of a function with annotations *)
        if checkDecls && Hashtbl.mem extrafuns $2 then
          raise (Validation_error ("function "^$2^" already defined."))
        else (
          if checkDecls then Hashtbl.add extrafuns $2 $2;
          FUNC_SIG($2,[],$3,$4))
    }

pred_symb_decls:
    pred_symb_decls pred_symb_decl {$2::$1}
  | pred_symb_decl {[$1]}

pred_symb_decl:
    LPAREN_TOK pred_symb sort_symbs annotations RPAREN_TOK
    {
      (* returns the signature of a predicate with annotations *)
        if checkDecls && Hashtbl.mem extrapreds $2 then
          raise (Validation_error ("predicate "^$2^" already defined."))
        else (
	  if checkDecls then Hashtbl.add extrapreds $2 $2;
	  PRED_SIG($2,$3,$4))
    }


  | LPAREN_TOK pred_symb  RPAREN_TOK
    {
      (* returns the signature of a predicate without annotations *)
      if checkDecls && Hashtbl.mem extrapreds $2 then
        raise (Validation_error ("predicate "^$2^" already defined."))
      else (
      	if checkDecls then Hashtbl.add extrapreds $2 $2;
        PRED_SIG($2,[],[]))
    }

an_formulas:
    an_formula an_formulas { $1::$2 }
  | an_formula             { [$1] }

an_formula:
    an_atom {$1}

  | LPAREN_TOK NOT_TOK an_formula annotations RPAREN_TOK 
      { 
        addFormulaAnnotations (negatePred $3) $4
      } 
 
  | LPAREN_TOK IMPLIES_TOK an_formula an_formula annotations RPAREN_TOK 
      { 
        addFormulaAnnotations (L.Imply ($3, $4)) $5
      } 


  | LPAREN_TOK IF_THEN_ELSE_TOK an_formula an_formula an_formula annotations RPAREN_TOK 
      { 
        let b = $3 in 
        let pt = $4 in 
        let pf = $5 in 
        let p = 
          match pt with 
            L.Bool true ->
              (* b OR pf *)
              (match pf with 
                L.Bool pfb -> if pfb then pf else b
              | _ -> L.Or [ b; pf ])
          | L.Bool false -> 
              (* (not b) AND pf *)
              let nb = negatePred b in 
              (match pf with 
                L.Bool pfb -> if pfb then nb else pf
              | _ -> L.And [ nb; pf ])
          | _ -> begin
              match pf with 
                L.Bool true -> 
                  (* (not b) OR pt *)
                  L.Or [ negatePred b; pt ]
              | L.Bool false -> 
                  (* b AND pt *)
                  L.And [ b; pt ]
              | _ -> 
                  L.Or [ L.And [ $3; $4 ];
                         L.And [ negatePred $3; $5 ] ] 
          end
        in
        addFormulaAnnotations p $6
      } 


  | LPAREN_TOK AND_TOK an_formulas annotations RPAREN_TOK 
      { 
        let p = L.And $3 in 
        addFormulaAnnotations p $4
      } 

  | LPAREN_TOK OR_TOK an_formulas annotations RPAREN_TOK 
      { 
        let p = L.Or $3 in 
        addFormulaAnnotations p $4
      } 


  | LPAREN_TOK XOR_TOK an_formulas annotations RPAREN_TOK 
      { 
        (* This computes a parity of the predicates *)
        E.s (E.unimp "XOR");
        let p = L.And $3 in 
        addFormulaAnnotations p $4
      } 

  | LPAREN_TOK IFF_TOK an_formulas annotations RPAREN_TOK 
      { 
        (* All predicates are equivalent *)
        let rec mkIff (acc: L.pred list) (p: L.pred) (rest: L.pred list) 
            : L.pred list = 
          match rest with 
            [] -> acc
          | p' :: rest' -> 
              mkIff (accumIff p p' acc) p' rest'
        in
        let p = 
          match $3 with 
          | [] -> L.Bool true
          | p :: rest -> L.And (mkIff [] p rest) 
        in
        addFormulaAnnotations p $4
      } 


  | LPAREN_TOK quant_symb quant_vars an_formula annotations RPAREN_TOK
    {
      (* returns a quantified formula with annotations *)
      (* Now construct the formula *)
      let isAll = ($2 = FORALL) in
      let rec mkBind (bl: (string * 'a) list) : L.pred = 
        match bl with 
          [] -> $4
        | (b, _) :: bl' -> 
            if isAll then L.All (b, mkBind bl')
            else L.Exist (b, mkBind bl')
      in
      addFormulaAnnotations (mkBind $3) $5
    }

  | LPAREN_TOK LET_TOK letbinding an_formula annotations RPAREN_TOK
    {
      (* returns a let formula with annotations *)
      popBinding letBindings $3;
      addFormulaAnnotations $4 $5
    }

  | LPAREN_TOK FLET_TOK fletbinding an_formula annotations RPAREN_TOK
    {
      popBinding fletBindings $3;
      (* returns a flet formula with annotations *)
      addFormulaAnnotations $4 $5
    }



letbinding: 
      LPAREN_TOK VAR_TOK an_term RPAREN_TOK 
      { pushBinding letBindings $2 $3;
        $2 
      }
 
fletbinding: 
      LPAREN_TOK fvar an_formula RPAREN_TOK 
      { 
        (* Take a look at the formula. If it is simple, then just substitute 
         * it *)
        (match $3 with 
          L.Bool _ | L.PNez _ -> (* a simple one *)
            pushBinding fletBindings $2 $3
        | _ -> (* a more complex one. Push some global assumptions and then 
                * replace with a simple predicate *)
            let np: L.pred = newPredicateName () in
            if debugNewNames then 
              ignore (E.log "introduced %a iff %a\n"
                        L.d_pred np L.d_pred $3);
            (* Remember the new name for witnesses *)
            parserWitness :=
              (dprintf "%a <=> %a"
                 L.d_pred np L.d_pred $3) ::
              !parserWitness;
            
            currentBenchmark.assumptions <- 
              accumIff np $3 currentBenchmark.assumptions;

            pushBinding fletBindings $2 np);
        $2
      } 

quant_vars:
    quant_vars quant_var 
    {
      (* returns a list of binders *)
      $2::$1
    }
  | quant_var 
    {
      (* returns a singleton list with one binder in it *)
      [$1]
    }

quant_var: 
    LPAREN_TOK VAR_TOK sort_symb RPAREN_TOK
    {
      (* returns the name of the variable along with the name of its sort *)
      ($2,$3)
    }

quant_symb:
    EXISTS_TOK {EXISTS}
  | FORALL_TOK {FORALL}

an_atom:
    prop_atom 
    {
      (* returns a formula for a propositional atom with no annotations *)
      $1
    }
  | LPAREN_TOK prop_atom annotations RPAREN_TOK 
    {
      (* returns a formula for a propositional atom with annotations *)
      addFormulaAnnotations $2 $3
    }
  | LPAREN_TOK pred_symb an_terms annotations RPAREN_TOK
    {
      (* returns a formula for a predicate that has 1 or more arguments with 
       * annotations*)
     if checkDecls && not (Hashtbl.mem extrapreds $2) then
       raise (Validation_error ("predicate "^$2^" not yet defined."));
     let p = 
       match $2, $3 with 
         "<=", [ e1; e2] -> L.leq e1 e2
       | "<", [e1; e2] -> L.lt e1 e2
       | ">=", [e1; e2] -> L.geq e1 e2
       | ">", [e1; e2] -> L.gt e1 e2
       | _ -> L.pFun $2 $3
     in
     addFormulaAnnotations p $4
    }

  | LPAREN_TOK EQUALS_TOK an_term an_terms annotations RPAREN_TOK
    {
      (* returns a formula for equality with annotations
         Note that it requires at least two terms *)
      let rec mkEqual (acc: L.pred list) (tl: L.exp list) = 
        match tl with 
          [] -> acc
        | t :: tl' -> mkEqual ((L.eq $3 t) :: acc) tl'
      in
      let p = 
        match mkEqual [] $4 with 
          [] -> L.Bool false
        | [ p ] -> p
        | pl -> L.And pl
      in
      addFormulaAnnotations p $5
    }

  | LPAREN_TOK DISTINCT_TOK an_term an_terms annotations RPAREN_TOK
    {
      (* returns a formula for distinct with annotations
         Note that it requires at least two terms*)
      let rec mkNeq (acc: L.pred list) (t: L.exp) (tl: L.exp list) 
          : L.pred list = 
        match tl with 
          [] -> acc
        | t' :: tl' -> 
            mkNeq ((L.ne t t') :: acc) t tl'
      in
      let rec mkDistinct (acc: L.pred list) (tl: L.exp list) : L.pred list = 
        match tl with 
          t :: tl' -> mkDistinct (mkNeq acc t tl') tl'
        | [] -> acc
      in
      let p = 
        match mkDistinct [] ($3::$4) with 
          [] -> L.Bool false
        | [ p ] -> p
        | pl -> L.And pl
      in
      addFormulaAnnotations p $5
    }


prop_atom:
  | TRUE_TOK
    {
      (* returns a formula for true *)
      L.Bool true
    }
  | FALSE_TOK
    { 
      (* returns a formula for false *)
      L.Bool false

    }
  | fvar
    {
      (* returns a formula variable that should be bound with a flet formula 
       * higher in the parse tree *)
       foundBound fletBindings $1
    } 

  | pred_symb 
    {
      (* returns a predicate with no arguments (a propositional variable) *)
      (
        if checkDecls && not (Hashtbl.mem extrapreds $1) then
          raise (Validation_error ("predicate "^$1^" not yet defined."));
       L.pFun $1 []
      )
    }
  

an_terms:
    an_term an_terms 
    {
      (* returns a list of terms*)
      $1::$2
    }
  | an_term 
    { 
      (* returns a singleton list of terms *)
      [$1]
    }

an_term:
    basic_term 
    { 
      (* returns a basic term with no annotations *)
      $1
    }
  | LPAREN_TOK basic_term annotations RPAREN_TOK 
    {
      (* adds annotations to a basic term and returns an annotated term *)
      checkTermAnnotations $3;
      $2
    }
  | LPAREN_TOK fun_symb an_terms annotations RPAREN_TOK
    {
      (* returns a function term with at least one argument with annotations *)
      (
        if checkDecls && not (Hashtbl.mem extrafuns $2) then
          raise (Validation_error ("function "^$2^" not yet defined."));
       checkTermAnnotations $4;
       let e = 
         match $2, $3 with
           "+", [e1; e2] -> L.plus e1 e2
         | "-", [e1; e2] -> L.minus e1 e2
         | "*", [e1; e2] -> L.times e1 e2
         | "store", [em;ea;ev] -> L.Fun("upd", [ea; ev; em])
         | "select", [em;ea] -> L.Fun("sel", [ea; em]) 
         | "~", [e1] -> begin 
             match e1 with 
               L.Int i -> L.Int (Int32.neg i)
             | _ -> L.minus L.zero e1
         end
         | _ -> L.Fun($2, $3)
       in
       e
      )
    }

  | LPAREN_TOK ITE_TOK an_formula an_term an_term annotations RPAREN_TOK
    {
      (* returns an ITE term with annotations *)
      checkTermAnnotations $6;
      mkExpIte $3 $4 $5
    }


basic_term:
    VAR_TOK 
    { 
      (* lookup the variable to see if it is bound or not *)
      foundBound letBindings $1
    }
  | NUMERAL_TOK 
    {
      (* returns a numeral term*)
      let i = 
        try Int32.of_string $1
        with (Failure _ as e) -> 
          Printf.printf "Invalid integer %s\n"  $1;
          raise e
      in
      L.Int i
    }
  | fun_symb 
    {
      (* returns a function term with no arguments (a constant)*)
      (
        if checkDecls && not (Hashtbl.mem extrafuns $1) then
          raise (Validation_error ("function "^$1^" not yet defined."));
        L.Fun($1,[])
      )
    }

annotations:
    /* (* empty *) */     { [] }
|   annotation annotations { $1 :: $2 }
  
annotation:
    attribute 
    {
      (* returns the attribute's name and None because it doesn't have a corresponding value *)
      ($1,None)
    }
  | attribute user_value 
    {
      (* returns the attribute's name along with its corresponding value*)
      ($1,Some $2)
    }

user_value:
    USER_VAL_TOK
    {
    	$1
    }

sort_symb_decl:
    SYM_TOK 
    { 
      (* returns the name of the sort symbol*)
      (
        Hashtbl.add extrasorts $1 $1;
        $1 
      )
    }

sort_symb:
    SYM_TOK 
    { 
      (* returns the name of the sort symbol*)
      
      (
        if checkDecls && not (Hashtbl.mem extrasorts $1) then
	  raise (Validation_error ("sort "^$1^" is not defined."));
        $1
      )
    }

pred_symb:
    SYM_TOK 
    {
        (* returns the name of the predicate *)
        $1
    }
  | AR_SYMB 
    {
        (* returns the name of the predicate  *)
        $1
    }

fun_symb:
    SYM_TOK
    {
        (* returns the name of the function *)
        $1
    }
  | AR_SYMB 
    { 
        (* returns the name of the function *)
        $1
    }


attribute:
    COLON_TOK SYM_TOK { ":" ^ $2}

fvar:
    DOLLAR_TOK SYM_TOK { "$" ^ $2}
