/* 
 * Copyright (c) 1999 Cedilla Systems Incorporated.  All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cedilla Systems Incorporated. ("Confidential Information").  You shall
 * not disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cedilla Systems.
 * 
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hash.h"
#include "input.h"
#include "main.h"


 
#ifdef SMALLMEM
#define  BUCKETS_SHIFT 5
#define  BUCKET_SIZE 4
#else
#define  BUCKETS_SHIFT 8
#define  BUCKET_SIZE 8
#endif

#define  NR_HASH_BUCKETS (1 << (BUCKETS_SHIFT))  /* These must be 2^k */




			      /* A bucket is a list of clusters of 
                               * HASH_ENTRIES. It behaves like a list of 
                               * entries but it is optimized by allocating 
                               * a cluster of entries at a time */
typedef struct BucketData {
  struct BucketData *next;
  HASH_ENTRY entries[BUCKET_SIZE];
} BUCKET_DATA;


typedef struct {
  int size;          /* The number of valid entries in this bucket. This 
                      * includes the entries with a NULL key */
  BUCKET_DATA *data; /* A number of entries */
} HASH_BUCKET;

typedef struct hash {
  HASH_BUCKET *hash;
  key2Hash  k2h;
  keyCompare kc;
} * PCONCRETE_HASH;


                              /* A hash table is actually an array of 
                               * NR_HASH_BUCKETS * HASH_BUCKET */



                              /* Keep a list of pre-allocated buckets */
#define BUCKET_CACHE_SIZE      (2 * NR_HASH_BUCKETS)

#ifdef SMALLMEM
#define BUCKET_CACHE_PREALLOC  (BUCKET_CACHE_SIZE >> 2)
#else
#define BUCKET_CACHE_PREALLOC  BUCKET_CACHE_SIZE
#endif

static  BUCKET_DATA *bucketCache[BUCKET_CACHE_SIZE];
static  int nextFreeBucket = 0;


static  BUCKET_DATA *acquireHashBucket(void) {
  if(nextFreeBucket == 0) {
    BUCKET_DATA *buck;
    buck = (BUCKET_DATA*)calloc(1, sizeof(BUCKET_DATA));
    if(! buck) {
      fprintf(stderr, "Cannot allocate hash bucket\n");
      exit(1);
    }
    return buck;
  } else {
    return bucketCache[-- nextFreeBucket];
  }
}

static int releaseHashBucket(BUCKET_DATA *buck) { 
  if(nextFreeBucket < BUCKET_CACHE_SIZE) {
    bucketCache[nextFreeBucket ++] = buck;
  } else {
    free(buck);
  }
  return 0;
}

int     preallocateHashes(void) {
  int i;
  nextFreeBucket = 0; /* So that acquire does not steal our buckets  */
  _ASSERT(BUCKET_CACHE_PREALLOC <= BUCKET_CACHE_SIZE);
  for(i=0;i<BUCKET_CACHE_PREALLOC;i++) {
    bucketCache[i] = acquireHashBucket();
  }
  nextFreeBucket = i;
  return 0;
}

int releaseHashes(void) {
  int i;
  for(i=0;i<nextFreeBucket;i++) {
    free(bucketCache[i]);
  }
  nextFreeBucket = 0;
  return 0;
}

/**************** NewHash *******************/
PHASH NewHash(key2Hash k2h, keyCompare kc) {
  /* Allocate a hash */
  PCONCRETE_HASH res;
  /* Allocate room for the hash table and pointers to buckets */
  res = (struct hash*)calloc(1, sizeof(struct hash) +
                             NR_HASH_BUCKETS * sizeof(HASH_BUCKET));
  if(! res) {
    fprintf(stderr, "Cannot allocate hash table\n");
    exit(1);
  }
  res->k2h = k2h;
  res->kc  = kc;
  res->hash = (HASH_BUCKET*)(res + 1);
  return (PHASH)res;
}


int stringToInt(HASH_KEY key) {
  int code = 0;
  char *str = (char*)key;
  for(code=0; *str; str ++) {
    code += *str;
  }
  return code;
}

int stringCompare(HASH_KEY str1, HASH_KEY str2) {
  // Do pointer equality first
  return str1 == str2 || !strcmp((char*)str1, (char*)str2);
}

int stringPointerCompare(HASH_KEY str1, HASH_KEY str2) {
  return (str1 == str2); 
}

PHASH NewStringHash(void) {
  return NewHash(stringToInt, stringCompare);
}

PHASH NewStringPointerHash(void) {
  return NewHash(stringToInt, stringPointerCompare);
}

 
void FreeHash(PHASH ah) {
  int i;
  PCONCRETE_HASH hin = (PCONCRETE_HASH)ah;
  
  HASH_BUCKET *h = (HASH_BUCKET*)hin->hash;
  for(i=0;i<NR_HASH_BUCKETS;i++) {
    HASH_BUCKET *buck = & h[i];
    BUCKET_DATA *bdata = buck->data;
    while(bdata != NULL) {
      BUCKET_DATA *t_bdata = bdata;
      bdata = bdata->next;
      releaseHashBucket(t_bdata);
    }
  }
  free(hin);
}

typedef enum {LOOKUP, LOOKUP_ADD} HashOper;

/* Find the key in the table. If LOOKUP then returns a pointer to where the 
 * data for the key is stored, or NULL if the key is not found. If 
 * LOOKUP_ADD then returns a pointer to the data (adds the entry if none 
 * found). If DELETE then it returns NULL. */
static HASH_ENTRY * ProcessHash(PHASH ah, HASH_KEY key,
                                int *found, HashOper oper) {
  int bucket_no, i, k;
  unsigned int t;
  PCONCRETE_HASH hin = (PCONCRETE_HASH)ah;
  
  BUCKET_DATA *buck = NULL;
  BUCKET_DATA **next = NULL;
  HASH_ENTRY *target = NULL;
  HASH_BUCKET *h = (HASH_BUCKET*)hin->hash;
  
  if(! key) {
    _ERROR0("ProcessHash called on the NULL key");
  }
  
  _ASSERT(h);

  			      /* Converts a hash key to a bucket index */
  
  for(bucket_no = 0,
      t = (unsigned int)hin->k2h(key);

      t;

      t >>= BUCKETS_SHIFT) {
    bucket_no ^= (t & (NR_HASH_BUCKETS - 1));
  }

  next = & h[bucket_no].data;

  i = BUCKET_SIZE;
  for(k=h[bucket_no].size;k > 0;) { /* Look for the data */
    HASH_ENTRY *e;
    buck = *next;             /* Get the next cluster */ 
    next = &(buck->next);     /* Move one to next cluster */
    e = buck->entries;        /* This is the current entry */
    for(i=0;i < BUCKET_SIZE && k > 0; k--, i++, e++) {
      if(! e->key) {
        if(!target) target = e;
      } else {
        if(key == e->key || hin->kc(key, e->key)) {
          if(found) *found = 1;
          return e;
        }
      }
    }
    if(k == 0)  /* Not in the bucket, hence not in table */
      break;
    _ASSERT(i == BUCKET_SIZE);
  }
  _ASSERT(k == 0);
  if(found) *found = 0;		      /* Here if not found */
  if (oper != LOOKUP_ADD) {
    return NULL;
  }
  if(! target) {
			      /* Must create a new entry */
    if(i == BUCKET_SIZE) {		     
      if(! next) {
        next = &(h[bucket_no].data);
      }
      _ASSERT(*next == NULL);
      buck = acquireHashBucket();
      *next = buck;
      buck->next = NULL;
      i = 0;
    }
    target = &buck->entries[i];
    h[bucket_no].size ++;
  }
  target->key  = key; // Set the key
  target->data = NULL;
  return target;
}

			      /* Lookup a hash key. */
HASH_ENTRY* HashLookup(PHASH h, HASH_KEY key) {
  return ProcessHash(h, key, NULL, LOOKUP);
}

			      /* Extend the hash. If the data already exists 
                               * then replace it*/
HASH_ENTRY* HashLookupOrAdd(PHASH h, HASH_KEY key, int *found) {
  return ProcessHash(h, key, found, LOOKUP_ADD);
}



int MapHash(PHASH ah, void* (*f)(HASH_KEY, void*, void*), void* closure) {
  int i;
  PCONCRETE_HASH h = (PCONCRETE_HASH)ah;
  HASH_BUCKET *pBucket = (HASH_BUCKET*)h->hash;

  for(i=0;i<NR_HASH_BUCKETS;i++, pBucket ++) {
    int sz = pBucket->size;
    BUCKET_DATA *pData = pBucket->data;
    HASH_ENTRY *pEntry;
    int k = 0;
    if(sz == 0) continue; 
    pEntry = pData->entries;
    for(;sz > 0;sz --, k++, pEntry++) {
      if(k == BUCKET_SIZE) {
        k = 0;
        pData  = pData->next;
        pEntry = pData->entries;
      }
      if(pEntry->key == NULL)
        continue;
      pEntry->data = (*f)(pEntry->key, pEntry->data, closure);
    }
  }
  return 0;
}

int IterHash(PHASH ah, void (*f)(HASH_KEY, void*, void*), void* closure) {
  int i;
  PCONCRETE_HASH h = (PCONCRETE_HASH)ah;
  HASH_BUCKET *pBucket = (HASH_BUCKET*)h->hash;

  for(i=0;i<NR_HASH_BUCKETS;i++, pBucket ++) {
    int sz = pBucket->size;
    BUCKET_DATA *pData = pBucket->data;
    HASH_ENTRY *pEntry;
    int k = 0;
    if(sz == 0) continue; 
    pEntry = pData->entries;
    for(;sz > 0;sz --, k++, pEntry++) {
      if(k == BUCKET_SIZE) {
        k = 0;
        pData  = pData->next;
        pEntry = pData->entries;
      }
      if(pEntry->key == NULL)
        continue;
      (*f)(pEntry->key, pEntry->data, closure);
    }
  }
  return 0;
}






