module E = Errormsg
module I = Input
module P = Proof
module S = Satproc
module H = Hashtbl
module T = Ics
open Pretty
exception Failure

(* George's data structures *)
let ccid = ref ""
let proveIt _ = E.s (E.unimp "Ics.proveIt")
type undoelem = SnapShot  (* This is the last snapshot *)
let undoList: undoelem list ref = ref []
(* ------------- *)



let axioms = ref ([])
let disEqualities = ref ([])

let debug = ref false
let repetitions = 30    
let timing = true
let failure = ref(false)

type expr = int * (int * baseElt) list (* constant * (coefficient * term) list *)
and baseElt = Var of string
  | Func of string * expr list

let rec exp2str (e: I.exp): string = 
  match e with
    I.Var v -> v
  | I.Int i -> Printf.sprintf "%d" i
  | I.Plus (e1,e2) -> Printf.sprintf "(%s + %s)" (exp2str e1) (exp2str e2)  
  | I.Minus (e1,e2) -> Printf.sprintf "(%s - %s)" (exp2str e1) (exp2str e2)  
  | I.Times (e1,e2) -> Printf.sprintf "(%s * %s)" (exp2str e1) (exp2str e2)  
  | I.Fun (f,el) -> 
      begin match el with
	[] -> f
      |	[e] -> Printf.sprintf "%s(%s)" f (exp2str e) 
      | e::el' -> 
	  let arguments = List.fold_left (fun temp e' -> Printf.sprintf "%s,%s" temp (exp2str e')) (exp2str e) el'
	  in
	  Printf.sprintf "%s(%s)" f arguments
      end
  | _ -> E.s (E.unimp "evalExp: Unimplemented")


let exp2term (e: I.exp): T.term = 
  let s = exp2str e in
  let t = T.term_of_string s in
  t
  

let docTerm () t = dprintf "%s" (T.term_to_string t)
let eqAtom t = T.atom_mk_equal t (T.term_of_string "0")
let diseqAtom t = T.atom_mk_diseq t (T.term_of_string "0")

let c: T.context ref = ref (T.context_empty ())

let initialize_structures () = 
  failure := false;
  T.reset ();
  c := T.context_empty ()

let process_assumptions () = 
  List.iter (fun a -> 
    let str = T.atom_to_string a in
    if (!debug) then ignore (eprintf "ICS: Processing Assumption %s\n" str);

    let s = T.process !c a in
    if (T.is_inconsistent s) then begin
      if (not timing) then ignore (eprintf "ICS: Axiom %s is inconsistent\n" str);
      raise Failure;
    end
    else if (T.is_consistent s) then c := T.d_consistent s
    else begin
      if (not timing) then ignore (eprintf "ICS: Axiom %s is redundant\n" str);
    end ) !axioms


let verify_disEqualities () =
  List.iter (fun a -> 
    let str = T.atom_to_string a in
    if (!debug) then ignore (eprintf "ICS: Processing DisEquality %s\n" str);
    let s = T.process !c a in
    if (T.is_inconsistent s) then begin
      if (not timing) then ignore (eprintf "ICS: DisEquality %s is inconsistent\n" str);
      raise Failure;
    end ) !disEqualities


let verify_goal a = 
  let str = T.atom_to_string a in
  if (!debug) then ignore (eprintf "ICS: Processing Goal %s\n" str);

    let s = T.process !c a in
    if (T.is_redundant s) then begin
      if (not timing) then ignore (eprintf "ICS: Proved Goal %s\n" str);
    end
    else begin
      if (not timing) then ignore (eprintf "ICS: Unable to Prove Goal %s\n" str);
      raise Failure;
    end


let assume acc msg = 
  match msg.S.body with
    S.Equality(e1, e2) -> 
      let e = I.Minus (e1, e2) in
      let t = exp2term e in
      if msg.S.goal then begin
	for i = 1 to repetitions do
	  let task () =
	    initialize_structures ();
	    try
	      process_assumptions ();
	      verify_disEqualities ();
	      verify_goal (eqAtom t);
	    with Failure -> failure := true;
	  in
	  Stats.time "Total time after parsing" task ()
	done;
	Stats.print stdout "Timings:\n";
	if (!failure) then raise (S.Contradiction (P.G("inconsistent", [])))
	else raise (S.Proved (P.G("eqid", [P.E e1])));
      end
      else begin
	axioms := List.append !axioms [eqAtom t];
      end;
      acc
  | S.Literal a -> 
      begin match a with
	I.Eq (e1, e2) -> 
	  let e = I.Minus (e1,e2) in
	  let t = exp2term e in
	  axioms := List.append !axioms [eqAtom t];
      |	I.Neq (e1, e2) ->
	  let e = I.Minus (e1,e2) in
	  let t = exp2term e in
	  disEqualities := (diseqAtom t) :: !disEqualities; (* collect disEqualities *)
      |	I.FalseAtom -> if (not timing) then ignore (E.log "RAND: FalseAtom \n"); assert (false); 
      |	I.TrueAtom -> () 
      |	_ -> ()
      end;
      acc


let initialize id = 
  ccid := id;
  undoList := [];
  { S.snapshot = (fun () -> ());
    S.undo = (fun () -> ());
    S.assume = assume;
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = id;
    S.descr  = "integrated canonizer and solver";
  } 

