(* Kettle Graphical User Interface
 * Original Implemented with LablTk ToolKit, by Kun Gao
 * Adapted to Kettle by George Necula
 *)

(* --- Supporting Code --- *)
open Logic            (* ntermediate language *)
open Options
open Pretty           (* pretty printing *)
open Tk               (* open the Tk interface *)


(* --- Support Modules --- *)
module E = Errormsg   (* Error messaging module *)
module G = Engine     (* Engine definitions *)
module ST  = Stree    (* State tree *)
module IH = Inthash
module GL = Guilib
module SW = Sourcewin
module RI = Reginfo

module U = Util
module GA = GrowArray
module P = Proof

let debug = false

(* --- Type declarations --- *)
type tree = ST.tree               (* tree type *)
type treeNode = ST.treeNode       (* node of tree *)
type canvas = Widget.canvas Widget.widget     (* canvas type *)
type toplevel = Widget.toplevel Widget.widget (* toplevel type *)
type frame = Widget.frame Widget.widget       (* frames *)
type regGroup = { mutable gName: string;      (* name of group *)
		  mutable gRegs: RI.registerInfo list; (* registers in group *)
		  mutable gChange: int list;  (* registers that changed *)
		}


let showSource = ref true
let showInterm = ref true

let sourceColor = ref "black"
let intermColor = ref "blue"

let showFullProofs = ref false

(* --- Gui Global variables and Flags --- *)

(* string of current source file, empty string means no reloading *)
let nextSrcName: string ref = ref ""

(* are we finished or not *)
let finished: bool ref = ref false

(* this keeps track of whether we have stopped or not *)
let isStopped: bool ref = ref false  (* initially we are not stopped *)

(* Breakpoints *)
let breakPts: int list ref = ref []


(* Global Register tracking used by Registers *)
(* a list of all active register groups *)
let regGroups: (string * frame) list ref = ref []

(* --- GUI Options Variable --- *)
let breakOnErrors: bool ref = ref true
let stepBackNum: int ref = ref 0



(* ----- Class Interface for TK GUI ----- *)
(* UI is composed of components managed by a main window. *)
(* Each component knows how to initialize and update itself *)
(* when given the state tree. It is up to the main manager window *)
(* to place each component in the correct location. *)
(* furthermore, the manager window provide services that *)
(* each component can utilize *)


type goalResult = 
    Proved of P.proof   (** Proved with a given proof *)
  | Failed of doc list  (** Failed with a given witness *)
  | InProgress          (** Goal in progress *)

type kettleState = {
    (** The goals in this block, in reverse order. Those without proofs are 
     * either failed or not yet finished *)
    mutable goals: (pred * goalResult) list;

    (** The assumptions in this block, in reverse order *)
    mutable assumes: (U.symbol * pred) list;


    (** Whether there are any not yet finished proofs, or in nested blocks *)
    mutable hasUnfinishedProofs: bool;
  }


exception IsKettleState of kettleState
let castToKettleState (state: ST.state) : kettleState = 
(*
  ignore (E.log "Calling cast for node %s:%s\n" 
            state#name state#infoline);
*)
  try state#downcast;
      E.s (E.error "downcast did not raise an Exception")
  with IsKettleState s -> s
    | _ -> E.s (E.error "Failure casting state to Kettle state")

class kstate (assumptions: (U.symbol * pred) list)
             (blockpars: (string * string) list) = 
  let s = { goals = []; assumes = []; hasUnfinishedProofs = false } in
object (self)
  val s = s 
  method private getParam (n: string) : string = 
    match List.filter (fun (n', _) -> n' = n) blockpars with 
      (_, what) :: _ -> what
    | [] -> raise Not_found

  method name = 
    try self#getParam "name" with Not_found -> ""

  method infoline = 
    self#name ^ 
    (match self#filename with 
      None -> ""
    | Some f -> " at " ^ f ^ 
        (match self#line with 
          None -> ""
        | Some l -> " line " ^ string_of_int l))

  method filename = 
    try Some (self#getParam "file") with Not_found -> None

  method line = 
    try Some (int_of_string (self#getParam "line")) with _ -> None

  method downcast = raise (IsKettleState s)

  method displayAssumptions = 
    List.map (fun (n, p) -> dprintf "%s: %a" (U.symbolName n) G.d_pred p) 
      assumptions


  method displayState = 
    (* We will display the goals *)
    let acc = ref [] in
    let id = ref 0 in
    List.iter (fun (g, po) -> 
      acc := 
        { RI.rName = "goal" ^ string_of_int !id;
          RI.rGroup = "ThisBlock";
          RI.rOneLineVal = G.d_pred () g;
          RI.rExpandedVal = None } :: !acc;
      acc := 
        { RI.rName = "proof" ^ string_of_int !id;
          RI.rGroup = "ThisBlock";
          RI.rExpandedVal = None;
          RI.rOneLineVal = 
            (match po with 
              Failed w -> 
                let idx = ref 0 in 
                dprintf "Failure witness:\n%a\n"
                  (docList ~sep:nil 
                     (fun p -> incr idx; 
                       dprintf "%d: %a\n" !idx insert p)) w
            | InProgress -> text "InProgress"
            | Proved p -> 
                if !showFullProofs then 
                  P.d_prf () p
                else begin 
                  let h = IH.create 13 in 
                  P.getFreeHyp h p;
                  docList (fun (n, _) -> text (U.symbolName n)) ()
                    (IH.tolist h)
                end)
        } :: !acc;
      incr id)

      s.goals;

    List.iter (fun (n, p) -> 
      acc := { RI.rName = "assume " ^ U.symbolName n;
               RI.rGroup = "ThisBlock";
               RI.rExpandedVal = None;
               RI.rOneLineVal = dprintf "%a" G.d_pred p } :: !acc)
      s.assumes;

    List.rev ! acc

  method copy  = ({< s = { s with goals = s.goals} >} :> ST.state)

end  


let src_view_opt: GL.uiComponent option ref = ref None


let changedSourceView () = 
  
  match !src_view_opt with 
    None -> ()
  | Some v -> v#init; 
              GL.updateUI true

class kettleSource
    (top: 'a Widget.widget) : GL.uiComponent
    = 

  (* A function to insert source *)
  let srcTag = "s" in
  let intermTag = "i" in
  let insertSrcLine
      (isinterm: bool)
      _
      (info: SW.instrInfo)
      (txt: Widget.text Widget.widget) 
      (where: textIndex) =
    let str = info.SW.text ^ "\n" in
    (* insert the text *)
    Text.insert txt ~index:where
      ~tags:[ (if isinterm then intermTag else srcTag) ]
      ~text:(string_of_int info.SW.lineno ^ ":\t" ^ 
             (if isinterm then "    " ^ str else str));
    1  (* Always one line *)
  in                       

  (** We will load all source files into one stream, for now. *)
  let sourceStrm = 
    new SW.concatenatedSourcesStream
      ~name:"Source"
      ~show:showSource
      ~insertInstr:(insertSrcLine false)
  in      

  (** We allow some intermediate files as well *)
  let intermStrm = 
    new SW.concatenatedSourcesStream
      ~name:"Interm"
      ~show:showInterm
      ~insertInstr:(insertSrcLine true)
  in


  object (self)
    inherit SW.sourceWindow ~saveID:"sourcecode" top
    method name = "Source code"


    method update (force: bool) = begin
      let nd = GL.currNode () in
      (match nd.ST.state#filename with 
        Some f -> 
          self#setFileName f;
          (match nd.ST.state#line with 
            Some l -> self#setLineNumber l
          | _ -> ())

      | _ -> ())
    end

    (** Set the current file name *)
    val mutable fileStartIndex = 0
    method private setFileName (f: string) = 
      (** See if we can find this file *)
      let startlno, endlno, justLoaded = 
        intermStrm#sourceFileRange f  in
      if justLoaded && startlno < endlno then begin
        (* ignore (E.log "We have just loaded %s\n" f); *)
        (* Scan the newly loaded intermediate file and see if we have to load 
         * some source files for it *)
        let info = intermStrm#info in
        let filerexp = Str.regexp ".*file=\\([/a-zA-Z0-9_.-]+\\).*" in
        let linerexp = Str.regexp ".*line=\\([0-9]+\\).*" in
        let sfile = ref "" in 
        let sourcestartlno = ref 0 in (* The start index of sfile *)
        let sourceendlno = ref 0 in
        let sline = ref 0 in
        for i = startlno to endlno do 
          let iinfo = GA.get info i in
          let l = iinfo.SW.text in 
          (* search for file=... *)
          if Str.string_match filerexp l 0 then begin
            sfile := Str.matched_group 1 l;
            sline := 1; (* Just in case we do not have a line *)
(*
            ignore (E.log "Found file=%s in %s at line %d\n" 
                      !sfile f (i - startlno + 1));
*)
            (* Make sure the source is loaded *)
            let sstart, send, _ = sourceStrm#sourceFileRange !sfile in
            if sstart < send then begin
              sourcestartlno := sstart;
              sourceendlno := send
            end else
              sfile := "" (* We could not load it *)
          end;
          (* See if we have a line as well *)
          if Str.string_match linerexp l 0 then begin
            sline := int_of_string (Str.matched_group 1 l);
(*
            ignore (E.log "Found line=%d in %s at line %d\n" 
                      !sline f (i - startlno + 1));
*)
            if !sfile = "" then 
              ignore (E.warn "Kettle.gui: Found source line but not source file in %s"
                        f)
            else begin
              if !sline + !sourcestartlno - 1 > !sourceendlno then begin
                ignore (E.warn "Found source line %d too big in %s"
                          !sline f);
                sline := 0
              end else
                iinfo.SW.supLine <- !sline + !sourcestartlno - 1
            end;
          end else begin
            (* No line number. Use the previous *)
            if !sfile <> "" && !sline <> 0 then 
              iinfo.SW.supLine <- !sline + !sourcestartlno - 1
          end
        done;
        self#init
      end;

      fileStartIndex <- startlno


      

    method private setLineNumber (l: int) = 
      let l' = fileStartIndex + l - 1 in
      if l' > 0 && l' <= GA.max_init_index intermStrm#info then 
        self#selectLine 1 l'
      else
        if GA.max_init_index intermStrm#info > fileStartIndex then 
          ignore (E.warn "Trying to select line number %d (max = %d)\n" 
                    l (GA.max_init_index intermStrm#info))

      
    method init =
      (* The order of these is important. We use 1 as the index of intermStrm 
       * in other places *)
      self#setSources [ (sourceStrm :> SW.sourceStream); 
                        (intermStrm :> SW.sourceStream) ] (fun () -> ());

      (* Color the display *)
      let font = "Helvetica -12" in
      
      Text.tag_configure self#textbox 
        ~tag:srcTag 
        ~foreground:(`Color !sourceColor)
        ~font:font;
      
      Text.tag_configure self#textbox 
        ~tag:intermTag 
        ~foreground:(`Color !intermColor)
        ~font:font;
      
end



let options : optionDescr list =
  SW.sourcewinOptions	@
  [
   { optInUI   = "Show Intermediate in Code Window";
     optRestart = false;
     optKind   = OBool showInterm;
     optExtra  = changedSourceView;
     optCommandLine = "-showInterm";
     optHelp  = "Show Interm"; };
   
   
   { optInUI   = "Show Source in Code Window";
     optRestart = false;
     optKind   = OBool showSource;
     optExtra  = changedSourceView;
     optCommandLine = "-showSource";
     optHelp  = "Show Source in GUI"; };

    { optInUI   = "Source Color";
      optRestart = false;
      optKind   = OString sourceColor;
      optExtra  = changedSourceView;
      optCommandLine = "";
      optHelp = "changes source code color"; };
    
    { optInUI   = "Interm Color";
      optRestart = false;
      optKind   = OString intermColor;
      optExtra  = changedSourceView;
      optCommandLine = "";
      optHelp = "changes intermediate code color"; };
    
    { optInUI   = "Show full proofs";
      optRestart = false;
      optKind   = OBool showFullProofs;
      optExtra  = changedSourceView;
      optCommandLine = "";
      optHelp = "shows the full proofs"; };
    

 ]


(* --- Class For Menu System --- *)
class menu (top: 'a Widget.widget) : GL.uiComponent =
  (* widgets used for menu *)
  let mbar = Frame.create ~relief:`Raised ~width:220 top in
  (* menu items *)
  let m_file = Menubutton.create ~text:"File" ~underline:0 mbar in
  let m3 = Menubutton.create ~text:"Options" ~underline:0 mbar in
  let m4 = Menubutton.create ~text:"Registers" ~underline:0 mbar in

  let m5 = Menubutton.create ~text:"Help" ~underline:0 mbar in
  let group_menu = Menu.create m4 in
  let _ = (* pack the main menu items *)
    pack [m_file; m3; m4 ] ~side:`Left;
    pack [m5] ~side:`Right;
    Menubutton.configure ~menu:group_menu m4;
  in
    (* Function that takes a menu, and returns a function
       while takes one option, and places it in the menu *)
  let mkOption (menu: Widget.menu Widget.widget)
      (opt: optionDescr) = 
    GL.mkOption top menu opt 
      (fun () -> ())
  in
      
  let _ = 
    (* File Menu *)
    let m1a = Menu.create m_file in
    Menubutton.configure ~menu:m1a m_file;
    
    (* -- separator -- *)
    Menu.add_separator m1a;
    
    (* menu options - quit command *)
    Menu.add_command ~label:"Quit" 
      ~command:GL.closeUI 
      m1a;
    
    
    (* The UI Option Submenu *)
    let gui_options = Menu.create m3 in
    List.iter
      (mkOption gui_options)
      options;
    
    (* Options Menu *)
    (* We add one item for the architecture and one for the vcgen. For now we 
     * put in there only those things that can be changed without restarting *)
    let option_menu = Menu.create m3 in
    
    (* add GUI submenu *)
    Menu.add_cascade ~label:"GUI Options" ~underline:0
      ~menu:gui_options option_menu;

    Menubutton.configure ~menu:option_menu m3;
  in


  object (self)
    inherit GL.defaultUIComponent ~name:"Menu"
                                  ~expand:false
                                  ~fill:`X
                                  ~dockable:false
                                  ~saveID:""
                                  mbar

    method copy (tl: toplevel) = 
      new menu tl


    (* --- My Public Methods --- *)
    method init = begin (* init menus *)
    end

    (* update *)
    method update (force: bool) = begin
    end
  end


(* --- Class for Buttons --- *)
class button_line top : GL.uiComponent =
  (* widgets used for button bar *)
  let bank1 = Frame.create ~borderwidth:1 ~width:80 ~height:20 top in
  
  let srcImage = GL.makeImage "source.gif" in
  let nosrcImage = GL.makeImage "nosource.gif" in
  let source = 
    Button.create 
      ~takefocus:false
      ~image:srcImage
      ~width:16 bank1 in
  let _ = GL.addTooltip "Toggle showing of source" source in

  let intermImage = GL.makeImage "interm.gif" in
  let nointermImage = GL.makeImage "nointerm.gif" in
  let interm = 
    Button.create 
      ~takefocus:false
      ~image:intermImage
      ~width:16 bank1 in
  let _ = GL.addTooltip "Toggle showing of intermediate code" interm in


  let _ =

    pack [ interm; source] ~side:`Left;
  in
  
  object (self)
    inherit GL.defaultUIComponent 
        ~name:"Action Bar"
        ~saveID:""
        ~dockable:false
        ~fill:`X
        ~expand:false
        bank1
        
    method copy (tl: toplevel) = new button_line tl

    method init = begin
      Button.configure source
	~command: (fun () ->
          showSource := not !showSource;
          changedSourceView ());
      Button.configure interm
	~command: (fun () ->
          showInterm := not !showInterm;
          changedSourceView ());
    end


    method update (_: bool) = begin
      Button.configure source
         ~image:(if!showSource then nosrcImage else srcImage);
      Button.configure interm
         ~image:(if!showInterm then nointermImage else intermImage);
    end
  end

(** A class to customize the GUI *)
class kettleCustomGui = object

    (* This is used for showing the node label *)
    method showNodeLabel (nd: ST.treeNode) (can: GL.canvas) 
                         ~(x:int) ~(y:int) = 
      let ks = castToKettleState nd.ST.state in
      let hasChildren = nd.ST.children <> [] in
      (* Now add the label *)
      let newText = 
	Canvas.create_text can 
          ~x:x ~y:y
          ~anchor:`W
          ~fill:
          (if nd == ST.currNode (GL.theTree ()) then `Blue 
          else if ks.hasUnfinishedProofs then `Red
          else `Black)
          ~text: nd.ST.state#name;
      in
      if nd == ST.currNode (GL.theTree ()) then 
        Canvas.configure_text ~font:"Helvetica -12 bold" can newText;

      (* For nodes that are not done, and are leaves, we want to be able 
       * to select them *)
      Canvas.bind 
        ~events:[`ButtonPress] 
        ~fields:[`MouseX; `MouseY]
        ~action:
        (fun _ -> 
          ST.setCurrent (GL.theTree ()) nd;
          GL.withoutTreeScrolling GL.updateUI true)
        can newText;

      newText
      
end

let run () : unit =
  ignore (E.log "Initialize VCGen\n");

  let configFile = "kettle_gui.cfg" in
  
  (* Load the configuration *)
  U.loadConfiguration configFile;

  U.useConfigurationInt "updateInterval" (fun i -> GL.updateInterval := i);
  U.useConfigurationString "sourceColor" (fun s -> sourceColor := s);
  U.useConfigurationString "intermColor" (fun s -> intermColor := s);
  U.useConfigurationBool "showSource" (fun b -> showSource := b);
  U.useConfigurationBool "showInterm" (fun b -> showInterm := b);
  U.useConfigurationBool "showFullProofs" (fun b -> showFullProofs := b);

  U.useConfigurationList "sourceDirectories"
    (fun l -> 
      SW.searchDirectories := 
        List.map 
          (fun s -> 
            match s with 
              U.ConfString s -> s
            | _ -> 
                ignore (E.warn 
                          "Found invalid configuration in searchDirectories");
                "")
          l);


  GL.initGui ();
  
  
  (* Configure the Main and Execution TK Window *) 
  Wm.title_set (GL.tkTop ()) "Visual VCGen Interface";
  
  let top = GL.tkTop () in

  let menu: GL.uiComponent = new menu top in
  let buttons: GL.uiComponent = new button_line top in
   
  (* instantiate source *)
  let src_view = new kettleSource top in
  src_view_opt := Some src_view;

  (* instantiate infobar *)
  let status = new GL.info_line top in
  
  let stateTree = new GL.stateTreeGui top (new kettleCustomGui) in

  (* instantiate assumptions *)
  let assumptions = 
    new GL.registers 
      ~title:"Inherited assumptions"
      ~saveID:"Assumptions"
      ~popupTitle:(fun ri -> "Assumption")
      ~getState:(fun _ -> 
        List.map
          (fun d -> 
            { RI.rName = ""; RI.rGroup = "";
              RI.rExpandedVal = None; RI.rOneLineVal = d })
          (GL.currNode ()).ST.state#displayAssumptions)
      top 
  in

  (* Reset my settings *)
  nextSrcName := "";
  finished := false;
  isStopped := false;
  breakPts := [];
  regGroups := [];
  
  (* instantiate groups *)
  let goals = 
    new GL.registers 
      ~title:"Assumptions and goals in this node"
      ~saveID:"ThisBlock"
      ~getState:(fun _ -> (GL.currNode ()).ST.state#displayState) 
      top
  in

  (* make a container for everything *)
  let cont = new GL.container top in
  (* Must pack the container frame *)
  pack [cont#frame] ~inside:top ~expand:true ~fill:`Both;

  cont#setPanes
    (`Col [ `Row [ `Win menu; `Win buttons;];
            `Row [ `Col [ `Win stateTree;
                          `Win goals ];
                   `Col [ `Win src_view;
                          `Win assumptions; ] ];
            `Win status ]);

  cont#init; (* Init the whole container *)
  
  GL.theUIComponents := [ (cont :> GL.uiComponent) ];

  (* Set the focus on the tree *)
  stateTree#focus;

  GL.updateUI true; (* update UI for the first time *)
  
  (* ----- Invoke the Main Loop ----- *)
  mainLoop();
  
  (* Save some configuration *)
  U.setConfiguration "sourceColor" (U.ConfString !sourceColor);
  U.setConfiguration "intermColor" (U.ConfString !intermColor);
  U.setConfiguration "showInterm" (U.ConfBool !showInterm);
  U.setConfiguration "showSource" (U.ConfBool !showSource);
  U.setConfiguration "updateInterval" (U.ConfInt !GL.updateInterval);
  U.setConfiguration "sourceDirectories" 
    (U.ConfList (List.map (fun s -> U.ConfString s) !SW.searchDirectories));
  U.setConfiguration "showFullProofs" (U.ConfBool !showFullProofs);

  (* Now save the configuration *)
  U.saveConfiguration configFile;
  

  ()


(* ----- Interface to G.uiType ----- *)
class ui : G.uiType = 
  (* we make a new gtkui object *)

  (* make the new execution tree *)
  let _ = GL.initTree [] in
  let currNode : ST.treeNode ref = ref (ST.root (GL.theTree())) in

  (** Remember here the assumptions *)
  let assumptions : (U.symbol * pred) list ref = ref [] in 

  object (self)
    method name = "gui"
    method options = options

    (** We generate our own blocks, if we are not inside a user specified 
     * block *)
    val mutable userBlockDepth = 0
    method private checkEnterBlock (name: string) = 
      if userBlockDepth = 0 then begin 
        self#enterBlock [] [ ("name", name); 
                             ("file", (!G.currLoc).E.file);
                             ("line", string_of_int (!G.currLoc).E.line) ];
        userBlockDepth <- 0; (* undo *)
      end

    method private checkExitBlock = 
      if userBlockDepth = 0 then begin
        self#exitBlock;
        userBlockDepth <- 0; (* undo *)
      end

    method enterBlock (assumptions_in: (U.symbol * pred) list) 
                      (params: (string * string) list) = 
      let t = GL.theTree () in
      userBlockDepth <- userBlockDepth + 1;
(*
      ignore (E.log "Entering block %a with %d assumptions\n"
                (docList (chr ';') (fun (n,p) -> dprintf "%s=%s" n p))
                params
                (List.length assumptions));
*)
      currNode := ST.newNode t (new kstate !assumptions params) !currNode

    method exitBlock = 
      userBlockDepth <- userBlockDepth - 1;
      let p = !currNode in 
      if p.ST.parent = None then 
        E.s (E.error "Gui.exitBlock called outside any block!");

      (* If this block does not have unfinished proofs, we collapse it *)
      let thiss = castToKettleState p.ST.state in
      if not thiss.hasUnfinishedProofs then 
        p.ST.collapsed <- true;

      (* Set the new current node to the parent of this one. *)
      currNode := 
         (match p.ST.parent with 
           None -> E.s (E.bug "Found exit block at top level")
         | Some p' -> p')


    method private getCurrentState = 
      castToKettleState (!currNode).ST.state
      
    method start_goal (g: pred) = 
      self#checkEnterBlock "goal";
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.start_goal called outside any block!");

      if debug then 
        ignore (E.log "start_goal\n");
      (* Add this goal to the current block *)
      let s = self#getCurrentState in
      assert (match s.goals with 
        (_, InProgress) :: _ -> ignore (E.warn "start_goal on a InProgress");
                                true
      | _ -> true);
      s.goals <- (g, InProgress) :: s.goals

    method goal_success (g: pred) (p: P.proof) = begin
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.goal_success called outside any block!");

      if debug then 
        ignore (E.log "goal_success\n");
      let s = self#getCurrentState in
      (* Get the last goal *)
      (match s.goals with 
        (_, InProgress) :: rest -> s.goals <- (g, Proved p) :: rest
      | _ -> E.s (E.bug "goal_success but no goal_start"));

      self#checkExitBlock
    end
   
    method goal_failure (loc: E.location) 
                        (why: string) (g: pred) (witness: doc list) = 
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.goal_failure called outside any block!");
      (* mark this node and all parents with hasUnfinishedProofs *)
      let p = !currNode in

      if debug then 
        ignore (E.log "goal_failure\n");
      let s = self#getCurrentState in
      (* Get the last goal *)
      (match s.goals with 
        (_, InProgress) :: rest -> s.goals <- (g, Failed witness) :: rest
      | [] -> E.s (E.bug "goal_failure but no goal_start. empty")
      | _ -> E.s (E.bug "goal_failure but no goal_start"));

      s.hasUnfinishedProofs <- true;
      
      ST.iterParents 
        (fun p -> 
          let ps = castToKettleState p.ST.state in 
          ps.hasUnfinishedProofs <- true)
        p;

      self#checkExitBlock

    method assume (mkname: (unit -> U.symbol)) (p: pred) = 
      let name = mkname () in
      self#checkEnterBlock "assume";
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.assume called outside any block!");
      let s = self#getCurrentState in
      s.assumes <- (name, p) :: s.assumes;
      assumptions := (name, p) :: !assumptions

    method unassume = 
      self#checkExitBlock;
      (match !assumptions with 
        _ :: rest -> assumptions := rest
      | _ -> ignore (E.warn "Kettle GUI: assumptions empty"));
      ()


    method run = 

      (* Set the current node *)
      let t = GL.theTree () in
      (match (ST.root t).ST.children with 
        [] -> ignore (E.warn "There are no root nodes to show\n")
      | r :: _ -> ST.setCurrent t r; 
          run ());

      (* We might have opened some blocks but have not closed them. This 
       * means that the *)
      ()
  end
