(* A library of functions for the VCGen Graphical User Interface
 * Implemented with LablTk ToolKit
 * COOLAID Project: 2003
 *)

(* --- Supporting Code --- *)
open Options
open Pretty           (* pretty printing *)
open Tk               (* open the Tk interface *)


(* --- Support Modules --- *)
module E = Errormsg   (* Error messaging module *)
module ST  = Stree    (* State tree *)
module H = Hashtbl    (* hashtable *)
module IH = Inthash
module U = Util
module RI = Reginfo

(* --- Type declarations --- *)
type mousePos = { mutable cursorX: int;   (* type of mouse position *)
		  mutable cursorY: int } 
type canvas = Widget.canvas Widget.widget     (* canvas type *)
type frame = Widget.frame Widget.widget       (* frames *)
type widget = Widget.any Widget.widget
type label = Widget.label Widget.widget
type menu = Widget.menu Widget.widget
type menubutton = Widget.menubutton Widget.widget
type message = Widget.message Widget.widget

type toplevel = Widget.toplevel Widget.widget (* toplevel type *)


(* --- Gui Global variables and Flags --- *)

let myTree: ST.tree option ref = ref None
      
let theTree () : ST.tree =
  match !myTree with
    None -> E.s(E.error "my Tree is not set\n");
  | Some t -> t
        

let myTop = ref None
  
let tkTop () : toplevel =
  match !myTop with
    None -> E.s(E.bug "No Tk Top found\n");
  | Some t -> t

(* tooltip background color *)
let popupColor = "yellow"
	
(* tooltip top *)
let myToolTipTop = ref None
let myPopup = ref None

let toolTipTop () : toplevel =
  match !myToolTipTop with
    None -> E.s(E.bug "No ToolTip Top found\n");
  | Some t -> t
let popUp () : message = 
  match !myPopup with
    None -> E.s(E.bug "No Message Popup found\n");
  | Some m -> m

let currNode () = 
  ST.currNode (theTree ())
let visibleNode () = 
  ST.visibleNode (theTree ())

(* whether to use bitmaps *)
let useTkBitmaps = 
  begin try
    "" = Sys.getenv "NO_TK_BITMAPS"
  with
  | Not_found -> true
  end


(* --- GUI Options Variable --- *)
let guiMsg: bool ref = ref false

(* turns on/off docking *)
let dockEnabled: bool ref = ref true

(* ----- Class Interface for TK GUI ----- *)
(* UI is composed of components managed by a main window. *)
(* Each component knows how to initialize and update itself *)
(* when given the state tree. It is up to the main manager window *)
(* to place each component in the correct location. *)
(* furthermore, the manager window provide services that *)
(* each component can utilize *)


(* --- Interface for UI Components --- *)
class type uiComponent =
  object
    method init: unit   (* initialization method *)
    method frame: frame  (* The Tk.frame for this component *)
    method expand: bool  (* Whether to expand this one *)
    method fill: Tk.fillMode (* How to fill it *)
    method update: bool -> unit (* update method. Say whether we must do a 
                                 * complete job. Otherwise this is likely to 
                                 * be a transient state. *)

    method focus: unit    (* Set the focus to this component *)

    method dockable: bool (* Whether this can be docked *)
    method name: string   (* Name of this window *)
    method copy: toplevel -> uiComponent (* make this a child of some top *)

    method saveConfig: unit (** Save configuration into Guilib.windowConfig *)
    method restoreConfig: unit (** Restore the configuration. This is done 
                                * automatically when the components are 
                                * created, but you can do it at a later time 
                                * as well *)
  end
    
(* Remember here the top level uiComponents *)
let theUIComponents: uiComponent list ref = ref []

   

let saveUISizeConfig (saveID: string) (f: frame) = 
  U.setConfiguration (saveID ^ "_width") (U.ConfInt (Winfo.width f));
  U.setConfiguration (saveID ^ "_height") (U.ConfInt (Winfo.height f));
  ()

let restoreUISizeConfig (saveID: string) (f: frame) = 
  U.useConfigurationInt (saveID ^ "_width")
    (fun w -> Frame.configure f ~width:w);
  U.useConfigurationInt (saveID ^ "_height")
    (fun h -> Frame.configure f ~height:h);
  ()
  


class defaultUIComponent  ?(saveID="") 
                          ?(expand=true)
                          ?(fill=`Both)
                          ?(dockable=true)
                          ?(name="Default UI Component")
                          (frame: frame) : uiComponent = 
  let restoreMe () = 
    (* Just when we get created, we restore the size configuration *)
    if saveID <> "" then 
      restoreUISizeConfig saveID frame;
  in
  let _ = restoreMe () in
  
  object
    method init = ()
    method update _ = ()
    method expand = expand
    method fill: Tk.fillMode = fill
    method frame = frame  
        
    method dockable = dockable
    method name = name
        
    method focus = Focus.set frame

    method copy (tl: toplevel) = 
      new defaultUIComponent ~saveID:saveID frame
        
        (* Save the window size *)
    method saveConfig = 
      if saveID <> "" then 
        saveUISizeConfig saveID frame
          
          
          (* Restore the window size *)             
    method restoreConfig = restoreMe ()
          
end


let updateInterval: int ref = ref 1000 (* In milliseconds *)
let lastUpdate: float ref = ref 0.0  (* In seconds *)

(*** UPDATE UI ***)

let updateUI (force: bool) = 
  (* See if we must update *)
  let now: float = Unix.gettimeofday () in
  if force || now >= !lastUpdate +. (float_of_int !updateInterval /. 1000.0) 
  then begin
    lastUpdate := now;

    (* Now go over the global update list *)
    List.iter (fun f -> f#update force) !theUIComponents

  end;
  (* do a tk draw by yielding to gtk *)
  Tk.update (); 
  ()


let initUI () = 
  List.iter (fun f -> f#init) !theUIComponents



(* --- Class for Register State --- *)
class registers 
                ?(title="Registers")
                ?(saveID="Registers")
                ?(width=40)
                ?(popupWidth=300)
                ?(popupHeight=250)
                ?(popupTitle=(fun ri -> "Register " ^ ri.RI.rName))
                ~(getState: unit -> RI.registerInfo list)
                (top: 'a Widget.widget) : uiComponent =
  (* widgets used for registers *)
  let reg_body = Frame.create ~relief:`Groove top in
  let reg_name = Label.create ~text:title reg_body in
  let reg_list = Listbox.create reg_body
      ~width:width ~height:12 ~selectforeground:`Red ~takefocus:true 
      ~selectmode:`Extended ~highlightthickness:1
  in
  let reg_scrl = Scrollbar.create reg_body ~takefocus:false in
  let reg_scrl2 = Scrollbar.create ~orient:`Horizontal reg_body 
      ~takefocus:false in
  let scroll_link sb sb2 lb =
    Listbox.configure ~yscrollcommand:(Scrollbar.set sb) lb;
    Scrollbar.configure ~command:(Listbox.yview lb) sb;
    Listbox.configure ~xscrollcommand:(Scrollbar.set sb2) lb;
    Scrollbar.configure ~command:(Listbox.xview lb) sb2
  in

  (** Keep a mapping of line numbers to register data *)
  let lineContents: (int, RI.registerInfo) H.t = H.create 13 in

  (* pack main widgets *)
  let _ = 
    pack [reg_name] ~side:`Top ~expand:false ~fill:`Y;
    pack [reg_scrl] ~side:`Right ~fill:`Y ;
    pack [reg_scrl2] ~side:`Bottom ~fill:`X;
    pack [reg_list] ~side:`Top ~fill:`Both ~expand:true;
    scroll_link reg_scrl reg_scrl2 reg_list;

  in
  (* If we click in this window, then we take the keyboard focus as well *)
  let _ = 
    Tk.bind ~events:[`ButtonPress]
      ~action:(fun _ -> Focus.set reg_list)
      reg_list
  in
  (* If we get the focus, select the element closest to mouse cursor *)
  let _ = 
    Tk.bind ~events:[`FocusIn] ~fields:[`MouseY]
      ~action:(fun (e: Tk.eventInfo) -> 
        (* If nothing is selected, then select one *)
        if Listbox.curselection reg_list = [] then begin
          let nearest = Listbox.nearest reg_list e.ev_MouseY in
          Listbox.selection_set
            reg_list ~first:nearest ~last:nearest
        end)
      reg_list
  in
  
      
      
  object (self)
    inherit defaultUIComponent ~saveID:saveID ~name:title reg_body

    method copy (tl: toplevel) = 
      new registers ~title:title ~saveID:saveID ~width:width
	~popupWidth:popupWidth ~popupHeight:popupHeight ~popupTitle:popupTitle
	~getState: getState
	tl


    (* create a new window for single register value *)
    method private makeRegWindow (ri: RI.registerInfo) =
      (* make a new top level window to present this value *)
      let regWindow = 
        Toplevel.create ~height:popupHeight ~width:popupWidth top in
      (* Set window heading *)
      Wm.title_set regWindow (popupTitle ri);
      
      (* Construct the Tk window with some stuff *)
      let regview_body = Frame.create ~relief:`Sunken regWindow in
      let reg_buttons = Frame.create ~relief:`Sunken regWindow in
      let regview_name = Label.create ~relief:`Raised regWindow in
      let regview_text = Text.create 
	  ~font:"Helvetica -12 bold" ~width:60 ~height:16 regview_body in
      let regview_scrl = Scrollbar.create regview_body ~takefocus:false in
      let regview_link sb tb =
	Text.configure ~yscrollcommand:(Scrollbar.set sb) tb;
	Scrollbar.configure ~command:(Text.yview tb) sb 
      in
      regview_link regview_scrl regview_text;
      (* nice buttons to go with tk window *)
      let regview_close = 
        Button.create ~text:"Close Window" ~width:16 reg_buttons 
          ~takefocus:true
	  ~command:(fun () -> Tk.destroy regWindow)
      in

      let txtvar = Textvariable.create () in

      pack [ regview_name ] ~side:`Top ~fill:`X;
      pack [ regview_body ] ~side:`Top ~fill:`Both ~expand:true;
      pack [ reg_buttons ] ~side:`Bottom ~fill:`X;
      pack [ regview_scrl ] ~side:`Right ~fill:`Y;
      pack [ regview_text ] ~side:`Left ~fill:`Both ~expand:true;
      
      pack [ regview_close ] ~side:`Left;

      let data = 
        match ri.RI.rExpandedVal with 
          None -> ri.RI.rOneLineVal
        | Some f -> begin 
            docList ~sep:line (fun r -> r.RI.rOneLineVal) () (f ())
        end
      in
      (* update the text with values *)
      Label.configure regview_name ~text:(popupTitle ri);
      let data = 
        if ri.RI.rName = "" then 
          data
        else
          text (ri.RI.rName ^ "  ") ++ data
      in
      let width = Winfo.width regview_text in
      let fontsz = 6 in (* I have no idea how to best find how many 
                         * characters fit in a line *)
      (* ignore (E.log "Formating register %s for width %d\n"
                ri.U.rName (popupWidth / fontsz)); *)
      Text.insert regview_text ~index:(`End,[]) 
        ~text:(sprint (popupWidth / fontsz) data);

        (* Turn off editing, so that we cannot delete and insert. *)
      Text.configure ~state:`Disabled regview_text;

      (* Take focus on the text window *)
      Focus.set regview_text;

      (* If we press Esc or Enter, then we close. The problem with this is 
       * that the event is also sent to the regular handler, which dies. To 
       * solve this we must break from this event handler. *)
      Tk.bind ~events:[`KeyPress] ~extend:false ~breakable:true
        ~fields:[`KeySymString]
        ~action:
        (fun (e: Tk.eventInfo) -> 
          (* ignore (E.log "Pressed keysym %s\n" e.ev_KeySymString); *)
          let yscroll ~(scroll: scrollValue) = 
            Text.yview regview_text scroll
          in
          match e.ev_KeySymString with 
            "Return" | "Escape" -> Tk.destroy regWindow; Tk.break ()
          | "Next" -> yscroll ~scroll:(`Page 1)
          | "Prior" -> yscroll ~scroll:(`Page (-1))
          | "Up" -> yscroll ~scroll:(`Unit (-1))
          | "Down" -> yscroll ~scroll:(`Unit 1)
          | "Home" -> yscroll ~scroll:(`Moveto 0.0)
          | "End" -> yscroll ~scroll:(`Moveto 1.0)
          | _ -> ())
        regview_text


    method init = begin (* init registers *)
      H.clear lineContents;

      (* bind action to view single registers. Bind it to ENTER and to 
       * DoubleClick *)
      let showPopup (e: Tk.eventInfo) = 
	(* let (`Num reg_line) = Listbox.nearest reg_list e.ev_MouseY in *)
        List.iter
          (fun (`Num reg_line) -> 
            (* Find the register info for that line *)
            (try 
              let r = H.find lineContents reg_line in
	      self#makeRegWindow r
            with Not_found -> 
              ignore (E.warn "listbox: nothing on line %d\n" reg_line)))
          (Listbox.curselection reg_list)
      in
      Tk.bind 
	~events:([`Modified ([`Double], `ButtonPress) ]) 
	~action:showPopup
	reg_list;

      Tk.bind
        ~events:[`KeyPress] ~fields:[`KeySymString;]
        ~action:(fun (e: Tk.eventInfo) -> 
          if e.ev_KeySymString = "Return" then 
            showPopup e)
        reg_list;
          
    end

    method update (_: bool) = begin (* update registers *)
      (* save location *)
      let savedX, _ = Listbox.xview_get reg_list in
      let savedY, _ = Listbox.yview_get reg_list in
      let width = Winfo.width reg_list in
      let fontsz = 6 in (* I have no idea how to best find how many 
                         * characters fit in a line *)

      (* clear box *)
      Listbox.delete reg_list ~first:(`Num 0) ~last:`End;
      Listbox.selection_clear reg_list ~first:(`Num 0) ~last:`End;
      H.clear lineContents;
      
      (* fill box *)
      let lineNum = ref 0 in
      List.iter
	(fun r ->
	  let (res: string) = 
            let toshow: doc = r.RI.rOneLineVal in
            (* Find out how many columns we have in the pane *)
            sprint (width / fontsz) toshow
          in
	  if res <> "" then begin
            H.add lineContents !lineNum r;
            incr lineNum;
	    Listbox.insert reg_list 
              ~index:`End ~texts:[(r.RI.rName ^ "  " ^ res)];
         end)
	(getState ());
      (* restore location *)
      Listbox.xview reg_list (`Moveto savedX);
      Listbox.yview reg_list (`Moveto savedY);
    end
  end


(** Loading GIF buttons *)
let makeImage (name: string) : [>`Photo of (string)] =
  (* See if we can load it from the resident bitmaps *)
  try
    let _, str = List.find (fun (n, _) -> n = name) Config_base.bitmaps in
    Imagephoto.create ~format:"gif"
      ~data:str ()
  with _ -> begin
    (* Try to load it from the file *)
    Imagephoto.create ~format:"gif"
      ~file:(Config_base.libdir ^ "/" ^ name) ()
  end

exception FoundNode of ST.treeNode

(* --- Class for Execution Path --- *)

(** Set this to prevent the tree from scrolling *)
let dontScrollTree = ref false

let withoutTreeScrolling (f: 'a -> 'b) (arg: 'a) : 'b = 
  dontScrollTree := true;
  U.tryFinally f 
    (fun _ -> dontScrollTree := false)
    arg

class stateTreeGui 
    ?(saveID="stateTreeGui")
    (top: 'a Widget.widget)
    (custom: < 
        showNodeLabel: ST.treeNode -> canvas -> x:int -> y:int -> tagOrId 
            >) : uiComponent 
                   = 

  let debug = false in

  (* default settings for sizes *)
  let deltaY = 14 in       (* The height of a line *)
  let deltaX =  8 in       (* The indentation for the next level *)
  let handleWidth = 6 in
  let handleHeight = 8 in

  let handleOffsetX = 1 in (* How many pixels to the right to move the handle*)
  let handleOffsetY = 1 in (* How many pixels down to move the handle *)

  (* How many pixels to the left does the handle end *)
  let handleExtentL = 6 / 2 - handleOffsetX in
  (* How many pixels to the left does the handle end *)
  let handleExtentR = 2 + handleOffsetX in
  (* How many pixels down the left does the handle end *)
  let handleExtentD = 3 + handleOffsetY in
	

  (* --- Widgets used for exec path --- *)
  let execpath = Frame.create ~relief:`Groove top in
  let scrollRegionX = ref 500 in
  let scrollRegionY = ref 1000 in
  let can = Canvas.create execpath ~width:250 
                                   ~height:190 
                                   ~highlightthickness:1
                                   ~scrollregion:(0,0,
                                                  !scrollRegionX,
                                                  !scrollRegionY) in
  let mapscrly = 
    Scrollbar.create ~orient:`Vertical ~takefocus:false execpath in
  let mapscrlx = Scrollbar.create ~orient:`Horizontal 
      ~takefocus:false execpath in

  (* pack the main widgets *)
  let _ =
    pack [mapscrly] ~side:`Right ~fill:`Y;
    pack [mapscrlx] ~side:`Bottom ~fill:`X;
    pack [can] ~side:`Top ~fill:`Both ~expand:true;

  in

    (* my bitmap images *)
  let collapsedImg = makeImage "collapsed.gif" in
  let expandedImg = makeImage "expanded.gif" in

    (** A function for drawing nodes on the canvas *)
    (*  
       |--- Node (1) without children
       | 
       |-O- Expanded node (2) with children
       | |
       | |-X- Collapsed node (3) with children
       | 
       |  
       ^
       node2_x == node1_x
         ^
         node3_x = node2_x + deltaX
    *)

  (** Make sure a node is visible.  *)


  let viewNode (nd: ST.treeNode) : unit =
    let x = nd.ST.x in
    let y = nd.ST.y in

    let xview = (-2, 0.7) in (* nodeX - 2*deltaX should be at 0.7 in window *)
    let yview = ( 4, 1.0) in
    
    (* See if we have enough scroll region *)
      
    (* Make sure we see the current label in the middle *)
    let x_frac = 
      (* Get what fractions we can see *)
      let x_frac_l, x_frac_r = Canvas.xview_get can in
      let x_raw = 
        (float_of_int (x + (fst xview) * deltaX)) 
          /. (float_of_int !scrollRegionX) 
          -. (x_frac_r -. x_frac_l) *. (snd xview) in
      if x_raw < 0.0 then 0.0 
      else if x_raw > 1.0 then 1.0
      else x_raw
    in
    
    let y_frac = 
      let y_frac_t, y_frac_b = Canvas.yview_get can in
      let y_raw = 
        (float_of_int (y + (fst yview) * deltaY)) 
          /. (float_of_int !scrollRegionY) 
          -. (y_frac_b -. y_frac_t) *. (snd yview) in
      if y_raw < 0.0 then 0.0 
      else if y_raw > 1.0 then 1.0
      else y_raw
    in

    if not !dontScrollTree then begin
      Canvas.xview can (`Moveto x_frac);
      Canvas.yview can (`Moveto y_frac);
    end;

(*
   ignore (E.log "Setting x_frac=%f, y_frac=%f\n" x_frac y_frac);
 *)
    ()
  in


    (* Show the collapsed/expanded handle of the node *)
  let showHandle (nd: ST.treeNode) (update: bool -> unit) : unit = 
    let x = nd.ST.x in
    let y = nd.ST.y in
    
    let handleX = x + handleOffsetX in
    let handleY = y + handleOffsetY in
    
    (* Get the old handle and the text, if any *)
    let handle = 
      Canvas.create_image can
	~x:handleX
	~y:handleY
	~anchor: `Center
	~image: 
	(if nd.ST.collapsed then 
	  collapsedImg else expandedImg)
    in
    Canvas.bind ~events:[`ButtonPress] 
      ~fields:[`MouseX; `MouseY]
      ~action:
      (fun _ -> 
        nd.ST.collapsed <- not nd.ST.collapsed;
        withoutTreeScrolling update true)
      can handle;
    
    ()
  in



      (* Draws a single node *)
  let rec drawNode 
      (nd: ST.treeNode) 
      (prevy: int) (* The y coordinate of the previous sibling or the parent *)
      (update: bool -> unit) 
       : int (* Return the next y *) = 
        
    let x = nd.ST.x in
    let y = nd.ST.y in
    (* Draw a line from the previous sibling or parent to this one *)
    ignore
      (Canvas.create_line can 
         ~xys:[(x - deltaX, prevy); (x - deltaX, y)]
         ~fill:`Black);
    let hasChildren = nd.ST.children != [] in
    if not hasChildren then begin
      (* Draw a longer horizontal line *)
      ignore
        (Canvas.create_line can 
           ~xys:[(x - deltaX, y); (x + deltaX, y)]
           ~fill:`Black);
    end else begin
      showHandle nd update;
      (* Draw a horizontal line before and after it *)
      ignore
        (Canvas.create_line can 
           ~xys:[(x - deltaX, y); (x - handleExtentL, y)]
           ~fill:`Black);
      ignore
        (Canvas.create_line can 
           ~xys:[(x + handleExtentR, y); (x + deltaX, y)]
           ~fill:`Black);
    end;
    
    let label = 
      if nd == ST.root (theTree ()) then 
        Canvas.create_text can ~x:(x + deltaX) ~y:y
          ~anchor:`W
          ~text:"Root"
      else
        custom#showNodeLabel nd can (x + deltaX) y 
    in
    (** See that we have enough scroll region *)
    let changed = ref false in
    if x + 100 > !scrollRegionX then begin
      scrollRegionX := !scrollRegionX + !scrollRegionX / 2;
      changed := true;
    end;
    if y + 100 > !scrollRegionY then begin
      scrollRegionY := !scrollRegionY + !scrollRegionY / 2;
      changed := true
    end;
    if !changed then begin
(*
          ignore (E.log "Changed scroll region to %dx%d\n"
                !scrollRegionX !scrollRegionY);    
*)
      Canvas.configure 
        ~scrollregion:(0,0,!scrollRegionX,!scrollRegionY) 
        can;
    end;

    (* If we double click on a handle, we toggle its collapsed status *)
    Canvas.bind ~events:[`Modified ([`Double], `ButtonPress) ]
      ~fields:[`MouseX; `MouseY]
      ~action:
      (fun _ -> 
        (* Redraw the tree, with this node visible. Do this before we swap 
         * the collapsed. *)
        ST.setVisible (theTree()) nd;
        nd.ST.collapsed <- not nd.ST.collapsed;
        withoutTreeScrolling update true)
      can label; 

    (* Now draw the children *)
    if not nd.ST.collapsed then 
      drawSiblings nd.ST.children (x + deltaX) y update
    else
      y + deltaY
	
        (* Draws all children of a node *)
  and drawSiblings
      (nds: ST.treeNode list) 
      (x: int)        (* The x coordinate of all the children *)
      (parenty: int)  (* The y coordinate of the parent *)
      (update: bool -> unit)
      : int (* Return the next y *) = 
    let (lasty, nexty) = 
      List.fold_left
        (fun (lasty, nexty) nd -> 
          nd.ST.x <- x;
          nd.ST.y <- nexty;
          let nexty' = drawNode nd lasty update in
          (nexty, nexty'))
        (parenty + handleExtentL, parenty + deltaY)
        nds
    in
    nexty
      
  in (* drawNode/drawSiblings *)

  (** For each node index, remember the line offset from its parent *)
  let nodeYOffset: int IH.t = IH.create 113 in

  (** For each node index, remember the number of lines it takes, along with 
   * all its subtrees *)
  let nodeYExtent: int IH.t = IH.create 113 in

  (** Compute the layout (offset and extent) for all subtrees. Return the 
   * extent of this one *)
  let rec computeLayout (n: ST.treeNode) : int = 
    try 
      IH.find nodeYExtent n.ST.index
    with Not_found -> begin
      (* Layout all the children *)
      let ext = 
         List.fold_left 
          (fun acc c -> 
            (* Acc is the Y offset from the parent node for this child *)
            IH.add nodeYOffset c.ST.index acc;
            if debug then 
              ignore (E.log "Found offset %d for node: %s\n" 
                        acc c.ST.state#infoline);
            acc + computeLayout c)
          1 (* The node itself takes one line *)
          
          (* Do not consider the children if the node is collapsed *)
          (if n.ST.collapsed then [] else n.ST.children)
      in
      IH.add nodeYExtent n.ST.index ext;
      if debug then 
        ignore (E.log "Found extent %d for node: %s\n" 
                  ext n.ST.state#infoline);
      ext
    end
  in
  

  (** Given a node, compute layouts of its parents, until we find one that 
   * extends at least a given number of lines before and after the line of 
   * the node itself. Return the found root, and how many lines before our 
   * node it starts *)
  let computeLayoutAround (n: ST.treeNode) 
                          ~(above: int) 
                          ~(below: int) : ST.treeNode = 
    (* Clean the hash tables *)
    IH.clear nodeYOffset;
    IH.clear nodeYExtent;
    if debug then 
      ignore (E.log "Compute layout -%d:+%d centered on %s\n"
                above below n.ST.state#infoline);

    (* We try successively n and its parents. p is n or a parent. *)
    let rec tryParents (p: ST.treeNode) = 
      (* First, layout p *)
      let p_ext = computeLayout p in
      (* Compute n's offset from p *)
      let rec findNodeOffset (n': ST.treeNode) = 
        if p.ST.index = n'.ST.index then 0 else
        match n'.ST.parent with 
          None -> E.s (E.bug "statetree: findNodeOffset for Root")
        | Some n'' -> 
            (* Get n's offset from p *)
            let offn' = 
              try IH.find nodeYOffset n'.ST.index 
              with Not_found -> 
                E.s (E.bug "statetree: findNodeOffset for node %s"
                    n'.ST.state#infoline)
            in
            offn' + findNodeOffset n''
      in
      let n_off_p = findNodeOffset n in
      if n_off_p >= above && p_ext - n_off_p >= below then 
        (* We are done *)
        p
      else begin
        if debug then 
          ignore (E.log "  Parent occupies -%d:+%d (%s)\n"
                    n_off_p (p_ext - n_off_p)
                    p.ST.state#infoline);
        (* Try the parent *)
        match p.ST.parent with 
          None -> (* We are at the root *) p
        | Some p' -> tryParents p'
      end
    in
    tryParents n
  in

  let rec update (complete: bool) = 
    let visible = ST.visibleNode (theTree ()) in
    (* Make sure that all parents are expanded. Find the top-most collapsed 
     * one. This is going to be the center of the view. *)
    let rec loop (last: ST.treeNode) (po: ST.treeNode option) = 
      match po with 
        None -> last
      | Some p -> 
          loop (if p.ST.collapsed then p else last) p.ST.parent
    in
    let nd = loop visible visible.ST.parent in 

    Tk.update (); (* To make sure Canvas.yview works *)
    
    (* Find out how much is visible *)
    let y_frac_t, y_frac_b = Canvas.yview_get can in
    let first_line = 
      int_of_float (y_frac_t *. (float_of_int !scrollRegionY) 
                      /. (float_of_int deltaY)) in
    let last_line = 
      int_of_float (y_frac_b *. (float_of_int !scrollRegionY) 
                      /. (float_of_int deltaY)) in
    if debug then 
      ignore (E.log "Will layout from %d-%d (%f, %f)\n" 
                first_line last_line y_frac_t y_frac_b);
    
    let p = 
      if complete then 
        ST.root (theTree ())
      else
        computeLayoutAround 
          nd
          ~above:((last_line - first_line) - 4)
          ~below:4
    in
    
    (* Delete everything from the canvas *)
    let all = Canvas.find can ~specs:[`All] in
    Protocol.remove_callbacks can; 
    Canvas.delete can all;

    Tk.bind ~events:[`ButtonPress] 
      ~action:(fun _ -> 
        Focus.set can)
      can;
    Tk.bind ~events:[`FocusIn] 
      ~action:(fun _ -> 
        Canvas.configure ~relief:`Raised can)
      can;
    Tk.bind ~events:[`FocusOut] 
      ~action:(fun _ -> 
        Canvas.configure ~relief:`Sunken can)
      can;

    Tk.bind ~events:[`KeyPress]
      ~fields:[`KeySymString]
      ~action:
      (fun (e: Tk.eventInfo) -> 
        (* ignore (E.log "Pressed keysym %s\n" e.ev_KeySymString); *)
        let curr = ST.currNode (theTree ()) in
        match e.ev_KeySymString with 
          "Left" -> (** collapse, or goto parent *)
            (* Get the current node *)
            if curr.ST.collapsed then begin
              (* go to the parent *)
              match curr.ST.parent with 
                Some p when p != ST.root (theTree ()) -> 
                  ST.setCurrent (theTree ()) p;
                  updateUI true
              | _ -> ()
            end else begin
              curr.ST.collapsed <- true;
              updateUI true
            end

        | "Right" -> (** expand, or go to first sibling *)
            if curr.ST.collapsed then begin
              curr.ST.collapsed <- false;
              updateUI true
            end else begin
              (* go to the first child *)
              match curr.ST.children with 
                c :: _ -> 
                  ST.setCurrent (theTree ()) c;
                  updateUI true
              | _ -> ()
            end
            
        | "Down" | "Up" -> begin (* Go to the next visible node *)
            let nexty = 
              if e.ev_KeySymString = "Down" then 
                curr.ST.y + deltaY 
              else
                curr.ST.y - deltaY
            in 
            (* Search for a node with the coordinate. Do not search into 
             * collapsed *)
            try 
              let rec search (n: ST.treeNode) = 
                if nexty = n.ST.y then raise (FoundNode n)
                else if n.ST.collapsed then ()
                else List.iter search n.ST.children
              in
              search (ST.root (theTree ()));
              (* Do nothing if no node below us *)
            with FoundNode n -> begin
              ST.setCurrent (theTree ()) n;
              updateUI true
            end
        end

        | "equal" | "minus" -> (* Expand/collapse all subtree *)
            let collapse = e.ev_KeySymString = "minus" in
            ST.iterNodeAndChildren (fun n -> n.ST.collapsed <- collapse) curr;
            updateUI true

        | _ -> ())
      can;
    
    (* Draw the root of the roots *)
    let rooty = 10  in
    let rootx = 4 in
    
    p.ST.x <- 4 + deltaX * p.ST.depth;
    p.ST.y <- 10;
    
    (* Now draw the children *)
    let nexty = 
      drawNode
        p
        rooty 
        update
    in

    viewNode nd;
    ()
  in

  (** Update the view and make sure that a certain node is visible *)
  (* Now configure the scroll bars *)
  let _ =
    Canvas.configure 
      ~yscrollcommand:(fun ~first ~last -> 
        (* ignore (E.log "Scroll to %f-%f\n" first last); *)
        Scrollbar.set mapscrly first last) can;
    Scrollbar.configure 
      ~command:(Canvas.yview can) 
      mapscrly;

    Canvas.configure ~xscrollcommand:(Scrollbar.set mapscrlx) can;
    Scrollbar.configure ~command:(Canvas.xview can) mapscrlx;
  in


  object (self)
    inherit defaultUIComponent ~saveID:saveID 
                               ~name:"Verification Path"
                               execpath

    method focus = Focus.set can 

    method copy (tl: toplevel) = 
      new stateTreeGui ~saveID:saveID tl custom
        
    method update (complete: bool) = update complete

	
    (* My public methods *)
    method init = begin 
      IH.clear nodeYOffset;
      IH.clear nodeYExtent;
    end

  end




(* --- Class for Info Box --- *)
class info_line ?(saveID="") (top: 'a Widget.widget) : uiComponent =
  (* widgets used for info box *)
  let infoframe = Frame.create ~relief:`Flat  ~height:18 top in
  let label = Label.create infoframe in
  (* pack main widgets *)
  let _ =
    pack [infoframe] ~side:`Top ~fill:`X;
    pack [label] ~side:`Left ~fill:`X;
  in
  
  object (self)
    inherit defaultUIComponent ~name:"Information Bar"
                               ~saveID:saveID
                               ~expand:false
                               ~fill:`X
                               ~dockable:false
                               infoframe  
    (* initialize *)
    method init = ()

    method copy (tl: toplevel) = 
      new info_line ~saveID:saveID top

    (* update *)
    method update (_: bool) = begin
      let node = visibleNode () in
      Label.configure label ~text:node.ST.state#infoline
    end
  end


 (*** Buttons & Menus with tooltips ***)

(* Generic addTip Routine *)
let addTip (isOkay: unit -> bool) 
           (tooltip: string)
           ?(delay=500) 
           (w: 'a Widget.widget) : unit =
  (** Have a timer, to delay the  appearance of the tooltip *)
  let t: Timer.t option ref = ref None in
  (** the tooltip top and message *)
  let topw = (toolTipTop ()) in
  let popupw = (popUp ()) in

  (* function to reset popup *)
  let clear_popup () =
    begin 
      match !t with
	Some t -> Timer.remove t
      | _ -> ()
    end;
    if Winfo.exists topw && Wm.state topw <> "withdrawn" then
      Wm.withdraw topw
  in

  (* function to raise popup then do the callback *)
  let raise_popup ev =
    t := Some (Timer.add ~ms:delay 
		 ~callback: (fun () ->
		   t := None;
		   update_idletasks();
		   if isOkay () then begin
		     Message.configure popupw ~text:tooltip;
		     Tk.raise_window topw;
		     Wm.geometry_set topw 
		       ("+"^(string_of_int (ev.ev_RootX + 9))^
			"+"^(string_of_int (ev.ev_RootY + 8)));
		     Wm.deiconify topw
		   end))
  in
  List.iter 
    (fun events -> bind w ~events ~extend:true 
        ~action:(fun _ -> clear_popup ()))
    [[`Leave]; [`ButtonPress]; [`ButtonRelease]; [`Destroy];
     [`KeyPress]; [`KeyRelease]];
  List.iter 
    (fun events ->
      bind w ~events ~extend:true ~fields:[`RootX; `RootY]
	~action:(fun ev -> clear_popup(); raise_popup ev))
    [[`Enter]; [`Motion]]

(* Add a tooltip message *)
(*
let addTooltip (tooltip: string)
               ?(delay=500)
               (w: 'a Widget.widget) : unit =
  addTip (fun () -> true) tooltip ~delay:delay w    
  
(* Special purpose popup for menu *)
let addMenuTooltip (tooltip: string)
                   ?(delay=500)
                   (w: Widget.menu Widget.widget) 
                   (idx: int) : unit =
  let isMenu =
    (fun () ->
      idx = Menu.index w `Active)
  in
  addTip isMenu tooltip ~delay:delay w *)


let addTooltip (tooltip: string)
               ?(delay=500)
               (widget: 'a Widget.widget) : unit = 
  (** Have a timer, to delay the  appearance of the tooltip *)
  let timer: Timer.t option ref = ref None in
  
  (** We store here the tooltip toplevel, so that we can destroy it later *)
  let top_ref: toplevel option ref = ref None in
  
  let show (e: Tk.eventInfo) = 
    let top = Toplevel.create ~borderwidth:1 widget in
    (* Place this window close to the mouse pointer *)
    Wm.geometry_set top 
      ("+" ^ (string_of_int (e.ev_RootX + 5)) ^
       "+" ^ (string_of_int (e.ev_RootY + 5)));
    (* Tell the window manager not to decorate these windows *)
    Wm.overrideredirect_set top true;
    
    let lab = Label.create ~text:tooltip 
                           ~background:(`Color "white") 
                           ~relief:`Raised
                           top in
    top_ref := Some top;
    pack [lab] ~side:`Top ~expand:true ;
  in
    
  Tk.bind
    ~events:[`Enter]
    ~fields:[`MouseX; `MouseY; `RootX; `RootY]
    ~action:(fun (e: Tk.eventInfo) -> 
      if delay > 0 then begin
        timer := Some (Timer.add ~ms:delay
                         ~callback:(fun _ -> show e))
      end else 
        show e)
    widget;

  Tk.bind
    ~events:[`Leave]
    ~fields:[`MouseX; `MouseY]
    ~action:(fun (e: Tk.eventInfo) -> 
      (match !timer with 
        Some t -> Timer.remove t; timer := None
      | _ -> ());
      (match !top_ref with
        Some t -> Wm.withdraw t; Tk.destroy t; top_ref := None
      | _ -> ());
      ())
    widget      


let addMenuTooltip (tooltip: string)
               ?(delay=500)
               (widget: Widget.menu Widget.widget) (idx: int) : unit = 
  (** Have a timer, to delay the  appearance of the tooltip *)
  let timer: Timer.t option ref = ref None in
  
  (** We store here the tooltip toplevel, so that we can destroy it later *)
  let top_ref: toplevel option ref = ref None in
  
  let show (e: Tk.eventInfo) = 
    if idx = Menu.index widget `Active then begin
      let top = Toplevel.create ~borderwidth:1 widget in
      (* Place this window close to the mouse pointer *)
      Wm.geometry_set top 
	("+" ^ (string_of_int (e.ev_RootX + 5)) ^
	 "+" ^ (string_of_int (e.ev_RootY + 5)));
      (* Tell the window manager not to decorate these windows *)
      Wm.overrideredirect_set top true;
      
      let lab = Label.create ~text:tooltip 
          ~background:(`Color "white") 
          ~relief:`Raised
          top in
      top_ref := Some top;
      pack [lab] ~side:`Top ~expand:true ;
    end
  in
    
  Tk.bind
    ~events:[`Enter]
    ~fields:[`MouseX; `MouseY; `RootX; `RootY]
    ~action:(fun (e: Tk.eventInfo) -> 
      if delay > 0 then begin
        timer := Some (Timer.add ~ms:delay
                         ~callback:(fun _ -> show e))
      end else 
        show e)
    widget;

  Tk.bind
    ~events:[`Leave]
    ~fields:[`MouseX; `MouseY]
    ~action:(fun (e: Tk.eventInfo) -> 
      (match !timer with 
        Some t -> Timer.remove t; timer := None
      | _ -> ());
      (match !top_ref with
        Some t -> Wm.withdraw t; Tk.destroy t; top_ref := None
      | _ -> ());
      ())
    widget      

    
    (* Function that takes a menu, and returns a function
       while takes one option, and places it in the menu *)
let mkOption 
    (top: 'a Widget.widget) 
    (menu: Widget.menu Widget.widget)
    (opt: optionDescr) 
    (reload: unit -> unit) = 
  (* get the correct underline index and label *)
  let underlinePos: int = 
    try String.index opt.optInUI '&' with Not_found -> -1 in
  let myLabel: string = 
    if underlinePos = -1 then (* no index, return original label *) 
      opt.optInUI
    else (* there is an index, rip out the '&' *)
      let newlabel1: string = String.sub opt.optInUI 0 underlinePos in
      let newlabel2: string = String.sub opt.optInUI (underlinePos + 1) 
	  ((String.length opt.optInUI) - underlinePos - 1)
      in
      newlabel1 ^ newlabel2
  in

  let reloadColor: string = "Red" in

    (* creates a reload dialog, returns true if reloading *)
  let reloadWindow (top: 'a Widget.widget) (optName: string) : bool = 
    let msg: string = "The " ^ optName ^ 
      " option requires restart. Are you certain?" in
    let reload_p = Dialog.create ~parent:top 
        ~title:"Are you sure?"
        ~message:msg
        ~buttons:["Reload!"; "Cancel"]
        ~name:"reloading"
    in
    if reload_p () = 0 then true
    else false
  in

      (* Function that creates a input dialog, optname is name of option
       optvalue is current option value, and setfun is the function that takes
       an input and sets the variable appropriately *)
  let queryWindow 
      (top: 'a Widget.widget)
      (optName: string) 
      (optValue: string) 
      (setfun : string -> unit) : unit =
    let usr_input = Textvariable.create () in
    let query_p = Toplevel.create top in
    let moretop = Frame.create query_p in
    let toppart = Frame.create query_p in
    let botpart = Frame.create query_p in
    (* even more top of query box *)
    let info = Label.create 
        ~text:("Current value of " ^ optName ^ ": ") moretop in
    let infoval = Label.create 
        ~relief:`Groove ~text:optValue moretop in
    (* top of the query box *)
    let optlabel = Label.create 
        ~text:"Enter new value: " toppart in
    let optentry = Entry.create 
        ~textvariable:usr_input toppart in
    (* bottom of query box *)
    let optset = Button.create ~text:"Set!" botpart 
        ~command:(fun _ ->
	  (* set button, evoke the set function *)
	  setfun (Textvariable.get usr_input);
	  (* close this window *)
	  Tk.destroy query_p;
	         )
    in
    let optcancel = Button.create ~text:"Cancel" botpart 
        ~command:(fun _ ->
	  (* cancel button, just close window *)
	  Tk.destroy query_p;
	         )
    in
    (* pack widgets *)
    Wm.title_set query_p "Option Value:";
    pack [ moretop; toppart; botpart ] ~side:`Top ~padx:6 ~pady:4;
    pack [ info; infoval ] ~side:`Left ~padx:1 ~pady:5;
    pack [ optlabel ] ~side:`Left ~padx:1 ~pady:2;
    pack [ optentry ] ~side:`Left ~padx:1 ~pady:2;
    pack [ optset; optcancel ] ~side:`Left ~padx:1 ~pady:4
      
  in
  (* The foreground color of the menu item *)
  let foreground = 
    if opt.optRestart then (`Color reloadColor) else (`Color "")
  in
  

  match opt.optKind with
  | OUnit -> 
      Menu.add_command ~label:myLabel menu;
      let myIndex: int = Menu.index menu (`Pattern myLabel) in
      
      let getIt () = 
        opt.optExtra ();
	if !guiMsg then 
          ignore(E.log "Unit Option \"%s\" now set\n" 
		   myLabel);
        updateUI true; (* To reflect other options *)
      in
      (* let the menu know about its value and set function *)
      Menu.configure_command menu (`Num myIndex)
	~label:myLabel
	~foreground:foreground
	~command: (fun _ -> getIt ());
      
      (* add the correct text underlining *)
      if underlinePos >= 0 then
	Menu.configure_checkbutton menu (`Num myIndex) 
	  ~underline:(underlinePos);
	  
      (* add menu tooltip *)
      addMenuTooltip opt.optHelp menu myIndex
      
  | OBool(boolean_ref) -> (* Add a checkbuttoned menu item *)
      Menu.add_checkbutton ~label:myLabel menu;
      let myIndex: int = Menu.index menu (`Pattern myLabel) in
      (* create the checkbutton support *)
      let option_var: Textvariable.textVariable = Textvariable.create () in
      (* check status of bool, and set the var appropriately *)
      if !boolean_ref then Textvariable.set option_var "1";
      
      let getIt () = 
	boolean_ref := (Textvariable.get option_var = "1");
        opt.optExtra ();
	if !guiMsg then 
          ignore(E.log "Boolean Option \"%s\" now set to: %b\n" 
		   myLabel !boolean_ref)
      in
      Menu.configure_checkbutton menu (`Num myIndex)
	~variable:option_var
	~foreground:foreground
	~command: 
        (fun _ -> 
          if opt.optRestart then
	    if reloadWindow top myLabel then begin
              getIt ();
	      (* reload source *)
	      reload ();
	    end else begin
	      (* click on menu item toggles the display ... we should reset 
               * this *)
	      Textvariable.set option_var
                (if !boolean_ref then "1" else "0");
            end
          else begin
            (* No reloading *)
            getIt ();
	    (* refresh ourselves *)
	    updateUI true;
          end);
      
      (* add the correct text underlining *)
      if underlinePos >= 0 then
	Menu.configure_checkbutton menu (`Num myIndex) 
	  ~underline:(underlinePos);

      (* add menu tooltip *)
      addMenuTooltip opt.optHelp menu myIndex

  | OInt(integer_ref) -> (* add command button support *)
      Menu.add_command ~label:myLabel menu;
      let myIndex: int = Menu.index menu (`Pattern myLabel) in
      (* function to set the new integer value *)
      let getIt (str: string) = 
	integer_ref := 
          (try int_of_string(str)
          with _ -> 0);
        opt.optExtra ();
	if !guiMsg then 
          ignore(E.log "Integer Option \"%s\" now set to: %d\n" 
		   myLabel !integer_ref );
      in
      let setInt (str: string) : unit =
	if opt.optRestart then begin
	  if reloadWindow top myLabel then begin (* will reload the source *)
            getIt str;
	    (* reload source *)
	    reload ()
	  end
	end else begin (* don't need to reload source *)
          getIt str;
	  (* refresh ourselves *)
	  updateUI true;
	  (* now reconfigure my menu to say new value *)
	  Menu.configure_command menu (`Num myIndex)
	    ~label:(myLabel ^ ": " ^ string_of_int(!integer_ref));
	end
      in
      (* let the menu know about its value and set function *)
      Menu.configure_command menu (`Num myIndex)
	~label:(myLabel  ^ ": " ^ string_of_int(!integer_ref))
	~foreground:foreground
	~command: (fun _ ->
	  queryWindow top myLabel (string_of_int !integer_ref) setInt;
	          );
      
      (* add the correct text underlining *)
      if underlinePos >= 0 then
	Menu.configure_command menu (`Num myIndex) 
	  ~underline:(underlinePos);

      (* add menu tooltip *)
      addMenuTooltip opt.optHelp menu myIndex
      
  | OStringList(sep, string_list_ref) -> (* add string support *)
      Menu.add_command ~label:myLabel menu;
      let myIndex: int = Menu.index menu (`Pattern myLabel) in
      (* function to set the new string value *)
      
      let theOption () : string = 
        sprint 80 (docList ~sep:(chr sep) text () !string_list_ref)
      in
      let getIt (str: string) = 
	string_list_ref := !string_list_ref @ splitStringList sep str;
        opt.optExtra ();
	if !guiMsg then 
          ignore(E.log "String list Option \"%s\" now set to: \"%s\"\n" 
		   myLabel 
                   (theOption ()))
      in
      let setStr (str: string) : unit =
	if opt.optRestart then begin
	  if reloadWindow top myLabel then begin (* will reload the source *)
            getIt str;
	    (* reload source *)
	    reload (); 
          end
	end else begin (* don't need to reload source *)
          getIt str;
	  (* now reconfigure my menu to say new value *)
	  Menu.configure_command menu (`Num myIndex)
	    ~label:(myLabel ^ ": \"" ^ theOption () ^ "\"");
          updateUI true;
	end
      in
      (* let the menu know about its value and set function *)
      Menu.configure_command menu (`Num myIndex)
	~label:(myLabel  ^ ": \"" ^ theOption () ^ "\"")
	~foreground:foreground
	~command: ( fun _ ->
	  queryWindow top myLabel (theOption ()) setStr;
	           );
      
      (* add the correct text underlining *)
      if underlinePos >= 0 then
	Menu.configure_command menu (`Num myIndex) 
	  ~underline:(underlinePos);

      (* add menu tooltip *)
      addMenuTooltip opt.optHelp menu myIndex

  | OString(string_ref) -> (* add string support *)
      Menu.add_command ~label:myLabel menu;
      let myIndex: int = Menu.index menu (`Pattern myLabel) in
      (* function to set the new string value *)
      
      let getIt (str: string) = 
	string_ref := str;
        opt.optExtra ();
	if !guiMsg then 
          ignore(E.log "String Option \"%s\" now set to: \"%s\"\n" 
		   myLabel !string_ref )
      in
      let setStr (str: string) : unit =
	if opt.optRestart then begin
	  if reloadWindow top myLabel then begin (* will reload the source *)
            getIt str;
	    (* reload source *)
	    reload (); 
          end
	end else begin (* don't need to reload source *)
          getIt str;
	  (* now reconfigure my menu to say new value *)
	  Menu.configure_command menu (`Num myIndex)
	    ~label:(myLabel ^ ": \"" ^ !string_ref ^ "\"");
          updateUI true;
	end
      in
      (* let the menu know about its value and set function *)
      Menu.configure_command menu (`Num myIndex)
	~label:(myLabel  ^ ": \"" ^ !string_ref ^ "\"")
	~foreground:foreground
	~command: ( fun _ ->
	  queryWindow top myLabel !string_ref setStr;
	           );
      
      (* add the correct text underlining *)
      if underlinePos >= 0 then
	Menu.configure_command menu (`Num myIndex) 
	  ~underline:(underlinePos);

      (* add menu tooltip *)
      addMenuTooltip opt.optHelp menu myIndex



let initTree (roots: ST.state list) : unit = 
  let stateTree: ST.tree = ST.newTree roots in 
  
  (* set the tree *)
  myTree := Some stateTree;
  
  (* Set the current node, if the node is not empty *)
  (match (ST.root stateTree).ST.children with 
    r :: _ -> ST.setCurrent stateTree r
  | _ -> ());

  ()


(**************************************************************)
(* Support for stepping forward and backward and running until some condition 
 * is met *)
(**************************************************************)


(* For each snapshot we store some info *)
type snapshotInfo = {
            snTree: ST.tree; (** The state of the tree at the snapshot *)
            snRestore: unit -> unit; (** A function to restore all other 
                                      * state *)
            snKind: snapshotKind;
    mutable snDelta: int; (** How many steps since the previous snapshot *)
  }

and snapshotKind = 
  | SnapEnd     (** The end of a history. The snDelta field is as usual. May 
                 * not be compressed away. *)
  | SnapStart   (** The start of a history. The snDelta field is not 
                 * relevant. May not be compressed away. This is either the 
                 * first one, or immediately after a SnapEnd *)
  | SnapIntermediate  (** An intermediate state in a history. May be 
                       * compressed away *)

class stepper    
                ?(maxPrevStates = 32)   (** Remember this many previous 
                                         * states *)
                ?(snapshotInterval = ref 40) (** How often to take snapshots *)
                ~(snapshot: unit -> unit -> unit) (** Take a snapshot and 
                                                   * return a function that 
                                                   * erstores the snapshot *)
                (singleStep: unit -> bool) = 

  (** Save here the previous states. *)
  let previousStates: snapshotInfo option array = 
                      Array.create maxPrevStates None  
  in
  let previousStatesNextFree = ref 0 in (* Next free element in the array *)


  (** How many steps we have taken since the last snapshot in 
   * "previousStates" *)
  let stepsSinceLastSnapshot: int ref = ref 0 in

  let finished = ref false in

  let debug = false in

  let stripSome (so: snapshotInfo option) : snapshotInfo = 
    match so with 
      None -> E.s (E.bug "stepper: None")
    | Some si -> si
  in

  let dumpState () = 
    ignore (E.log "   Stepper stats: ");
    for i = 0 to !previousStatesNextFree - 1 do 
      match previousStates.(i) with 
        None -> E.s (E.bug "stepper: None in previous states")
      | Some si -> begin
          match si.snKind with 
            SnapStart -> ignore (E.log " [")
          | SnapEnd -> ignore (E.log " +%d]" si.snDelta)
          | SnapIntermediate -> ignore (E.log " +%d" si.snDelta)
      end
    done;
    ignore (E.log "\n")
  in
  
     (* Check the invariants *)
  let checkInvariants (where: string) = 
    let stop x = 
      dumpState ();
      E.s x
    in

    (* At least the initial start state *)
    assert (!previousStatesNextFree > 0);
    assert (!previousStatesNextFree <= Array.length previousStates);
    (* The first one is a Start. Then some Intermediate. After each End we 
     * have a Start. *)
    let expectStart = ref true in
    for i = 0 to !previousStatesNextFree - 1 do 
      match previousStates.(i) with 
        None -> stop (E.bug "None inside the array")
      | Some si -> begin
          if !expectStart <> (si.snKind = SnapStart) then 
            E.s (E.bug "stepper %s: expecting a start at index %d\n" 
                   where i);
          if si.snKind = SnapStart && si.snDelta <> 0 then 
            E.s (E.bug "stepper %s: start has a non-zero delta at index %d\n" 
                   where i);
          if si.snKind <> SnapStart && si.snDelta <= 0 then 
            stop (E.bug "stepper %s: non-positive delta %d at index %d\n" 
                   where si.snDelta i);
          (* After an End we expect a start *)
          expectStart := si.snKind = SnapEnd;
      end
    done;
    if !expectStart then 
      E.s (E.bug "stepper %s: we ended while expecting a start" where);
    (* the rest must be None, to please the GC. This is important because 
     * these snapshots carry a lot if info in them. *)
    for i = !previousStatesNextFree to Array.length previousStates - 1 do 
      if previousStates.(i) != None then 
        E.s (E.bug "stepper %s: non-None at index %d\n" where i)
    done;
    
    ()
  in
  let makeSnapshot (kind: snapshotKind) : snapshotInfo = 
    { snTree = ST.deepCopy (theTree ());
      snRestore = snapshot ();
      snDelta = !stepsSinceLastSnapshot;
      snKind = kind }
  in

  object (self)

    method init = 
      finished := false;
      previousStatesNextFree := 0;
      stepsSinceLastSnapshot := 0;
      Array.fill previousStates 0 (Array.length previousStates) None;
      (* Push a start state *)
      self#pushState SnapStart;
      if debug then checkInvariants "after init";

      ()


    (* Push the current state *)
    method private pushState (kind: snapshotKind) = 
      (* A function to compress the array. Called only if the array is full *)
      let compressArray () = 
        (* We have filled the array. Compress it. Two adjacent Intermediate 
         * nodes, are replaced with the latter one, with a delta equal to the 
         * sum of the deltas. Same if we have an Intermediate followed by an 
         * End. *)
        previousStatesNextFree := 0;
        let keep = ref true in (* Whether to keep the next one, if an 
                                * Intermediate *)
        Array.iteri 
          (fun i sio -> 
            let si = stripSome sio in
            if debug then ignore (E.log "Compressing index %d\n" i);
            (* Overwrite it on top of the previous, if not Start, and if 
             * not keep *)
            if si.snKind <> SnapStart && not !keep then begin
              if debug then ignore (E.log "Moving %d over\n" i);
              assert (!previousStatesNextFree > 0);
              (* Get the previous delta *)
              let previ = 
                stripSome previousStates.(!previousStatesNextFree - 1) in
              assert (previ.snKind = SnapIntermediate);
              si.snDelta <- si.snDelta + previ.snDelta;
              previousStates.(!previousStatesNextFree - 1) <- Some si;
              keep := true (* We keep the next one *)
            end else begin
              if debug then ignore (E.log "Copying %d\n" i);
              (* We must move it over *)
              if i <> !previousStatesNextFree then 
                previousStates.(!previousStatesNextFree) <- sio;
              incr previousStatesNextFree;
              (* We'll drop the next Intermediate, if this is an Intermediate*)
              keep := (si.snKind <> SnapIntermediate) 
            end)
          previousStates;

        (* Now make sure that these is enough space left. At least one 
         * quarter of the array *)
        if !previousStatesNextFree + (Array.length previousStates) / 4 
            >= Array.length previousStates then 
          E.s (E.unimp "stepper: too many incomplete histories");

        for i = !previousStatesNextFree to (Array.length previousStates - 1) 
        do 
          previousStates.(i) <- None;
        done;
        if debug then checkInvariants "after compress";
      in
      
      (* No need to save a snapshot if we have taken no steps from it. Or, 
       * perhaps we are saving the first entry. *)
      if kind = SnapStart || !stepsSinceLastSnapshot > 0 then begin
        if !previousStatesNextFree >= Array.length previousStates then 
          compressArray ();

        (* Save a deep copy of the tree, and a snapshot of the state of VCGen*)
        previousStates.(!previousStatesNextFree) <- 
          Some (makeSnapshot kind);
        
        incr previousStatesNextFree;
        if debug then begin
          ignore (E.log "push state %d (%d since last snapshot)\n" 
                    (!previousStatesNextFree - 1) !stepsSinceLastSnapshot);
          dumpState ()
        end;

        stepsSinceLastSnapshot := 0;
      end;

      ()

    (** Push a state that marks the End of a thread of execution *)

    method resetCurrentState (nd: ST.treeNode) = 
      (* If the last entry is a start and the delta is 0, then just replace 
       * it *)
      if debug then ignore (E.log "Stepper: adding a new start\n");
      if !stepsSinceLastSnapshot = 0 then begin
          assert ((stripSome previousStates.(!previousStatesNextFree - 1)).snKind = SnapStart); 
          if debug then ignore (E.log "   replaced top of the stack\n");
          ST.setCurrent (theTree ()) nd;
          previousStates.(!previousStatesNextFree - 1) <- 
            Some (makeSnapshot SnapStart);
        end else begin
          self#pushState SnapEnd;
          ST.setCurrent (theTree ()) nd;
          self#pushState SnapStart;
        end;

      if debug then checkInvariants "reset state";
      ()


        (* Restore the top state *)
    method private popState : snapshotKind = 
      if debug then begin
        ignore (E.log "pop state %d (%d since last snapshot)\n" 
                  (!previousStatesNextFree - 1) !stepsSinceLastSnapshot);
      end;
      if !previousStatesNextFree = 0 then 
        E.s (E.bug "The Guilib undo stack is empty");
      let si = stripSome previousStates.(!previousStatesNextFree - 1) in
      myTree := Some si.snTree;
      si.snRestore ();
      stepsSinceLastSnapshot := si.snDelta;
      decr previousStatesNextFree;
      previousStates.(!previousStatesNextFree) <- None;
      if debug then dumpState ();
      si.snKind


    (** Step forward until the function returns true. The function is queried 
     * after each single step, that does not raise an exception. *)
    method stepForwardUntil (f: unit -> bool) = 
      let rec loop () = 
        if debug then checkInvariants "before step forward";
        (* save a snapshot only once in a while, or if there are no previous 
         * states saved *)
        let wePushed = ref false in
        if  not !finished &&   
               (!stepsSinceLastSnapshot >= !snapshotInterval 
                || (!previousStatesNextFree = 0)) then begin
          self#pushState SnapIntermediate;
          wePushed := true;          
        end;
        (* do actual step *)
        finished := 
          (try singleStep ()
          with e -> begin
            (* We might have an error as we take the step. We must undo the 
             * push if we did any *)
            if !wePushed then 
              ignore (self#popState);
            updateUI true;
            raise e
          end);
            (* We promissed we'll run f after every step *)
        let mustStop = f () in
        if not !finished then begin
          incr stepsSinceLastSnapshot; 
          assert (!stepsSinceLastSnapshot >= 1);
          if debug then 
            ignore (E.log "after singleStep (%d steps since last snapshot) at line %s\n" 
                      !stepsSinceLastSnapshot 
                      (currNode()).ST.state#infoline);
          if not mustStop then loop ()
        end
      in
      loop ();
      if debug then checkInvariants "after step forward is done"

    method canStepBack = 
      !stepsSinceLastSnapshot >= 1 || !previousStatesNextFree > 1

    (** Step backward until the functionr returns true. The function is 
     * queried after each single step backwards. It should not be called if 
     * self#canStepBack returns false.  *)
    method stepBackwardUntil (f: unit -> bool) = 
      let rec loop () = 
        if debug then begin
          ignore (E.log "before step back once (%d since last snapshot)\n" 
                    !stepsSinceLastSnapshot);
          checkInvariants "before step backward"
        end;
        assert (!stepsSinceLastSnapshot >= 0);
        finished := false; (* Just in case we are stepping back from the end *)

        let emptyUndo = ref false in

        (* If we have only 0 steps since the last save, then we are on a 
         * Start. Unless this is the first start, pop twice *)
        if !stepsSinceLastSnapshot = 0 then begin
          (* Must be Start,...,End,Start *)
          assert (!previousStatesNextFree >= 3); 
          ignore (self#popState);  (* the start *)
          ignore (self#popState)   (* the end *)
        end else if !stepsSinceLastSnapshot = 1 then begin
          let kind = self#popState in
          if kind = SnapStart then 
            self#pushState kind
        end else begin
          (* We must restore the last snapshot. We pop it and then push it 
           * back *)
          let oldStepsSinceLastSnapshot = !stepsSinceLastSnapshot in
          assert (oldStepsSinceLastSnapshot > 1);
          self#pushState self#popState;

	  (* run for x times, but change temporarily the saveInterval to a 
           * geometric progression *)
          let oldSnapshotInterval = !snapshotInterval in
          let steps_to_take = ref (oldStepsSinceLastSnapshot - 1) in
          snapshotInterval := (!steps_to_take + 3) / 2;
          self#stepForwardUntil 
            (fun _ -> 
              decr steps_to_take; 
              snapshotInterval := (!steps_to_take + 3) / 2; (* At least 1 *)
              !steps_to_take <= 0) ;
          snapshotInterval := oldSnapshotInterval;
        end;

        if debug then checkInvariants "after step backward done";

        if not (f ()) && not !emptyUndo then loop ()
      in
      loop ()
      
end

      
  
let setWindowConfiguration () = 
  if !theUIComponents <> [] then begin
    (* Prompt everybody to save their configuration data *)
    List.iter (fun u -> u#saveConfig) !theUIComponents;
    U.setConfiguration "main_window_geom" 
      (U.ConfString (Wm.geometry_get (tkTop())));
  end


let getWindowConfiguration () = 

  let main_geom = 
    try 
      let geom = U.findConfigurationString "main_window_geom" in
      (* Sometimes the geometry string is 1x1 ??? *)
      if Str.string_match (Str.regexp "^\\([0-9]+\\)x\\([0-9]+\\)")
          geom
          0 &&
        50 < int_of_string (Str.matched_group 1 geom) &&
        50 < int_of_string (Str.matched_group 2 geom) then 
        
        geom
      else
        raise Not_found
    with _ -> "700x600+200+30"
  in
  Wm.geometry_set (tkTop ()) main_geom;
  ()


let initGui () = 

  lastUpdate := Unix.gettimeofday ();

  myTop := Some (openTk ());
  
  (* initialize the tooltips top once and just hide it *)
  myToolTipTop := Some (Toplevel.create (tkTop ()));
  let topw = (toolTipTop ()) in
  Wm.overrideredirect_set topw true;
  Wm.withdraw topw;
  myPopup := Some (Message.create topw ~name:"balloon"
			 ~background: (`Color popupColor) ~aspect:300);

  (* restoring window configuration *)
  getWindowConfiguration ();

  (* bind action to update the window settings *)
  let alreadySaved = ref false in

  Tk.bind
    ~events:([`Destroy])
    ~action:(fun _ ->
      (* save geometry settings into variable *)
      if not !alreadySaved then begin
        alreadySaved := true; 
        try setWindowConfiguration ()
        with Protocol.TkError _ -> () 
      end;
      ())
    (tkTop ());

  pack [(popUp ())]


let closeUI () = 
  if !guiMsg then ignore(E.log "\nClosing Graphical UI... ");
  (* close tk *)
  closeTk ();
  if !guiMsg then ignore(E.log "[ done ]\n")

let guiOptions: optionDescr list = 
  [ 

    { optInUI   = "&Update Interval (ms)";
      optRestart = false;
      optKind   = OInt updateInterval;
      optExtra  = (fun () -> ());
      optCommandLine = "-updateInterval";
      optHelp = "Number of miliseconds between GUI updates"; };
   
  ] 

(** Support for resizeable and dockable windows *)
type pane = [
  | `Row of pane list
  | `Col of pane list
  | `Win of uiComponent
  ]

let debug = false
    
(* debug print out pane *)
let rec prnPane (windows: pane) : doc = 
  match windows with
    `Win w -> text w#name
  | `Row l -> text "[ Row: " ++ prnPaneList l ++ text "]"
  | `Col l -> text "[ Col: " ++ prnPaneList l ++ text "]"
and prnPaneList (windowslist: pane list) : doc =
  seq (chr ';' ++ Pretty.break) prnPane windowslist

let d_pane () = prnPane


let rec iterPane (f: uiComponent -> unit) (windows: pane) : unit = 
  match windows with 
    `Win w -> f w
  | `Row pl 
  | `Col pl -> List.iter (iterPane f) pl

(** Layout a number of windows into a top level *)
let layout ?(separatorWidth=1)
           ?(separatorBackground=`Black)
           ?(makeIntermFrame=(fun top -> Frame.create ~borderwidth:0
                                                        top))
           ~(top:'a Widget.widget)
           ~(host: 'b Widget.widget)
           (windows: pane) : unit = 

  (* Ought to clean everything in the top widget *)
  (* Pack.forget (Pack.slaves top); *)

  (* my window *)
  let myWindows: pane ref = ref windows in
  let frameList = ref [] in
  (* Create this frame, just so that we have something to raise everybody 
   * over ??? *)
  let topFrame = Frame.create top in

  (* do the layout *)
  let rec doLayout () = begin
    (* Pack a list of uiComponents *)
    let packFrameList 
        (fl: (frame * bool * Tk.fillMode) list) 
        (horiz: bool) : frame = 
      (* This list must have at least two elements. Otherwise you should not 
       * be packing anything *)
      let master : frame = 
        let r = makeIntermFrame top in
        (* add to frame list *)
        frameList := r :: !frameList;
        raise_window ~above:topFrame r;
        Pack.propagate_set r false;
        r
      in
      
      (* Compute the total requested size *)
      let totReqX : int ref = ref 0 in
      let totReqY : int ref = ref 0 in


      let addReq (f: frame) = 
        (* At this point, we ensure that the requested widths and heights are 
         * never 1. That seems to be the default *)
        if Winfo.reqwidth f = 1 then Frame.configure ~width:10 f;
        if Winfo.reqheight f = 1 then Frame.configure ~height:10 f;

        if horiz then begin
          totReqX := !totReqX + Winfo.reqwidth f;
          totReqY := max !totReqY (Winfo.reqheight f)
        end else begin
          totReqX := max !totReqX (Winfo.reqwidth f);
          totReqY := !totReqY + Winfo.reqheight f 
        end
      in
      
      let packSide = if horiz then `Left else `Top in

      (* Go over the list and pack them all. We remember the last expandable 
       * component *)
      let rec loop (prevexpando: frame option) = function
          [] -> () (* We are done *)

        | (f,exp,fill) :: rest -> 
            raise_window ~above:master f;
            addReq f;

            (* We put a separator if there is an expandable element before 
             * this one *)
            if exp && prevexpando <> None then begin
              let prevexpand: frame = 
                match prevexpando with Some f -> f
                | None -> E.s (E.bug "guilib: packFrameList")
              in
              let sep = 
                if horiz then begin
                  totReqX := !totReqX + separatorWidth;
                  Canvas.create 
                    ~width:separatorWidth 
                    ~cursor:(`Xcursor "sb_h_double_arrow")
                    ~background:separatorBackground top;
                end else begin
                  totReqY := !totReqY + separatorWidth;
                  Canvas.create 
                    ~height:separatorWidth 
                    ~cursor:(`Xcursor "sb_v_double_arrow")
                    ~background:separatorBackground top;
                end
              in
              
              pack [sep] ~side:packSide ~inside:master 
                ~fill:(if horiz then `Y else `X);

              (* Watch when we drag the separator. The packer will first give 
               * the windows in a frame the widths they request. Then, it will 
               * distribute the remaining stuff equally among the windows.  *)

              let resizeEvent ~(force:bool) (e: Tk.eventInfo) = 
                (* We must use the RootX and RootY (instead of MouseX and 
                 * MouseY) because the latter are relative to the separator 
                 * position, which changes at the same time as the mouse 
                 * moves *)

                 (* Get the coordinates of the separator *)
   
                if horiz then begin
                  (* Get the total requested width on the thing to our left*)
                  let prev_req = 
                    Winfo.pixels f (`Pix (Winfo.reqwidth prevexpand)) in
                  (* And on the right *)
                  let f_req = 
                    Winfo.pixels f (`Pix (Winfo.reqwidth f)) in
                  
                  let deltaX = 
                    e.ev_RootX - (Winfo.rootx sep) - separatorWidth in 
                  
                  (* Never reduce the requested widths under 10 *)
                  let offset = 
                    max (10 - (prev_req + deltaX))
                      (10 - (f_req - deltaX)) in
                  let offset = max 0 offset in 
                  
                  Frame.configure 
                    ~width:(prev_req + deltaX + offset) prevexpand;
                  Frame.configure 
                    ~width:(f_req - deltaX + offset) f;
                end else begin
                  let prev_req = 
                    Winfo.pixels prevexpand 
                        (`Pix (Winfo.reqheight prevexpand)) in
                  let f_req = 
                    Winfo.pixels f (`Pix (Winfo.reqheight f)) in
                  
                  let deltaY = 
                    e.ev_RootY - (Winfo.rooty sep) - separatorWidth in
                  
                  (* Never reduce the requested widths under 10 *)
                  let offset = 
                    max (10 - (prev_req + deltaY))
                      (10 - (f_req - deltaY)) in
                  let offset = max 0 offset in 
                  
                  Frame.configure 
                    ~height:(prev_req + deltaY + offset) prevexpand;
                  Frame.configure 
                    ~height:(f_req - deltaY + offset) f;
                end;
              in
              
              (* Try a slighlty delayed resize event. if timerSet is Some 
               * then we have set a timed event. *)
              let timerSet : Timer.t option ref = ref None in

              Tk.bind 
                ~events:[`Modified([`Button1], `Motion)] 
                ~fields:[`RootX; `RootY ]
                ~action:(resizeEvent ~force:false)
                sep;
              (* Resize for the button release as well *)
              Tk.bind 
                ~events:[`ButtonRelease] 
                ~fields:[`RootX; `RootY]
                ~action:(resizeEvent ~force:true)
                sep;
            end;

            (* Now pack this frame *)
            Pack.propagate_set f false;
            pack [f] ~side:packSide
              ~inside:master ~expand:exp ~fill:fill;
            
            (* Go an pack the rest *)
            loop (if exp then Some f else prevexpando) rest
              
      in
      loop None fl;
      Frame.configure ~width:!totReqX ~height:!totReqY master;
      master
    in
    
    (* remove frame from pack list *)
    let removeFromPacking (ui: uiComponent) =
      let rec cleanPane (windows: pane) : pane =
        match windows with
	  `Row [ `Win p; p2 ] when p = ui ->  p2
        | `Row [ p2; `Win p ] when p = ui ->  p2
        | `Col [ `Win p; p2 ] when p = ui ->  p2
        | `Col [ p2; `Win p ] when p = ui ->  p2
        | `Row lst -> `Row (cleanPaneList lst)
        | `Col lst -> `Col (cleanPaneList lst)
        | `Win x -> `Win x (* can't remove last guy *)
      and cleanPaneList (windowslist: pane list) : pane list =
        let res = List.filter 
	    (fun p -> match p with 
	      `Win w -> ui <> w
	    | _ -> true) 
	    windowslist
        in
        List.map
	  cleanPane
	  res;
      in
      (* before *)
      ignore(E.log "%a\n" d_pane !myWindows);
      let res = cleanPane !myWindows in
      myWindows := res;
      ignore(E.log "%a\n" d_pane !myWindows);
    in
    
    (* create new top *)
    let makeUndocked (ui: uiComponent) x y = 
      let undocked = Toplevel.create (tkTop ()) in
      let newUI = ui#copy undocked in
      (* initialize *)
      newUI#init;
      theUIComponents := newUI :: !theUIComponents ;
      newUI#update true;
      (* setup the window nicely *)
      Wm.title_set undocked newUI#name;
      let width = Winfo.width ui#frame in
      let height = Winfo.height ui#frame in
      let geom = ((string_of_int width) ^ "x" ^ (string_of_int height) ^ "+" ^
		  (string_of_int x) ^ "+" ^ (string_of_int y)) in
      Wm.geometry_set undocked geom;
      (* pack *)
      pack [newUI#frame] ~inside:undocked ~fill:`Both ~expand:true;
      (* return the undock *)
      undocked
    in
    
    (* make handles for docking purposes *)
    let makeHandle (ui: uiComponent) : frame = 
      (* mouse track *)
      let myMouse: mousePos option ref = ref None in
      let docked: bool ref = ref true in
      let cleaned: bool ref = ref false in
      let undocked: toplevel option ref = ref None in
      (* create handles for this component *)
      let handle = Frame.create top in
      Pack.propagate_set handle false;
      let titlebar = Button.create ~text:(ui#name) handle in
      raise_window ~above:topFrame handle;
      raise_window ~above:handle ui#frame;
      pack [titlebar] ~inside:handle ~side:`Top ~fill:`X ~expand:false;
      pack [ui#frame] ~inside:handle ~fill:`Both ~expand:true;
      
      let trackMouseDrag () = 
        (* reset dragging *)
        Tk.bind
	  ~events:[`ButtonReleaseDetail 1]
	  ~action:
	  (fun e ->
	    myMouse := None;
	    if not !docked && not !cleaned then begin
	      (* remove window from set *)
	      removeFromPacking ui;
	      (* clear the list *)
	      List.iter (fun f -> Tk.destroy f)
	        !frameList;
	      frameList := [];
	      doLayout ();
	      updateUI true;
	      (* done cleaning *)
	      cleaned := true;
	    end
	  )
	  titlebar;
        (* watch dragging *)
        Tk.bind
	  ~events:[`Modified([`Button1], `Motion)]
	  ~fields:[`MouseX; `MouseY]
	  ~action:
	  (fun (e: Tk.eventInfo) ->
	    match !myMouse with
	      None ->
	        myMouse := Some { cursorX = e.ev_MouseX;
				  cursorY = e.ev_MouseY; }
	    | Some m ->
	        let threshold = 300 in
	        let deltaX = abs (e.ev_MouseX - m.cursorX) in
	        let deltaY = abs (e.ev_MouseY - m.cursorY) in
	        let rootx = (Winfo.rootx ui#frame) + e.ev_MouseX in
	        let rooty = (Winfo.rooty ui#frame) + e.ev_MouseY in
	        if deltaX + deltaY > threshold && !docked then begin
		  (* need to undock this ui *)
		  undocked := Some 
		      (makeUndocked ui rootx rooty);
		  (* remove from packing list *)
		  docked := false;
	        end;
	        begin match !undocked with
		  Some d -> 
		    let geom = 
		      let width = Winfo.width ui#frame in
		      let height = Winfo.height ui#frame in
		      ((string_of_int width) ^ "x" ^ (string_of_int height) ^ "+" ^
		       (string_of_int rootx) ^ "+" ^ (string_of_int rooty))
		    in
		    Wm.geometry_set d geom;
	        | None -> ()
	        end;
	        ())
	  titlebar;
      in
      
      (* enable/disable docking *)
      if !dockEnabled then 
        trackMouseDrag ();
      
      handle
    in
    
    let rec createWindows (p: pane) : frame * bool * Tk.fillMode = 
      (* Do the children first *)
      let horizontal, children = 
        match p with 
          `Win w -> 
            (* Load the configuration *)
	    let frame = 
	      if w#dockable && false then makeHandle w
	      else w#frame
	    in
            true, [ (frame, w#expand, w#fill) ]
              
        | `Row pl -> true, 
            List.map createWindows pl
        | `Col pl -> false, 
            List.map createWindows pl
      in
      match children with 
        [] -> E.s (E.bug "Layout: no children")
      | [ c ] -> c
      | _ -> (packFrameList children horizontal, true, `Both)
    in
    let all, exp, fill = createWindows !myWindows in
    
    pack [all] ~inside:host  ~expand:exp ~fill:fill;
    
    ()
   end
  in
  (* do the layout *)
  doLayout ()

(** A container for some windows *)
class container 
    ?(expand=true)
    ?(fill:Tk.fillMode=`Both)
    (top: 'a Widget.widget) = 
  let containerFrame = Frame.create ~width:100 ~height:100 top in
  object (self)
    inherit defaultUIComponent   ~name:"Container"
                                 ~expand:expand
                                 ~fill:fill
                                 containerFrame

    val mutable panes : pane option = None

    method private iterPanes (f: uiComponent ->unit) : unit = 
      match panes with 
        None -> ()
      | Some p -> 
          iterPane f p
      
    method copy (tl: toplevel) = 
      (* we do this to get pass the type checking.. even
	 though a container will never be called to copy *)
      E.s(E.error "Can't copy container");
      new info_line tl

    method saveConfig = 
      self#iterPanes (fun f -> f#saveConfig);
      (* And now save our own configuration *)
      
      ()
      
    method restoreConfig = self#iterPanes (fun f -> f#restoreConfig)

      
    method setPanes (p: pane) = 
      panes <- Some p;
      (* Do the layout. Delete some slaves  *)
      let slaves = Pack.slaves containerFrame in
      if slaves <> [] then 
        Pack.forget slaves;

      if debug then 
        ignore (E.log "container: setPanes\n");
      layout ~top:top ~host:containerFrame p

    method getPanes = panes

    method clearPanes = 
      (* We should forget now *)
      panes <- None

    method init = self#iterPanes (fun f -> f#init) 

    method update (complete: bool) = 
      self#iterPanes (fun f -> f#update complete)

  end

(** An UI component that adjusts to several register groups **)
class registerGroups 
    ?(title="")
    ?(width=40)
    ?(popupWidth=300)
    ?(popupHeight=250)
    ?(popupTitle=(fun ri -> "Register " ^ ri.RI.rName))
    ?(saveID="registerGroups")
    ?(columns=ref 1)
    ~(getState: unit -> RI.registerInfo list)
    (top: 'a Widget.widget) : uiComponent = 

  let cont = new container top in
  let _ = 
    U.useConfigurationInt (saveID ^ "columns")
      (fun c -> columns := c)
  in
    
  object (self)
    inherit defaultUIComponent ~name:"Registers"
                               ~saveID:saveID
                               cont#frame  as super

    method copy (tl: toplevel) = 
      new registerGroups
	~title:title ~width:width ~popupWidth:popupWidth
	~popupHeight:popupHeight ~popupTitle:popupTitle
	~saveID:saveID ~getState:getState tl

    method saveConfig = 
      super#saveConfig;
      if saveID <> "" then begin
        U.setConfiguration (saveID ^ "columns") (U.ConfInt !columns);
      end

    method restoreConfig = 
      super#restoreConfig;
      if saveID <> "" then begin
        U.useConfigurationInt (saveID ^ "columns")
          (fun c -> columns := c);
      end

    (* The current groups, in the order in which they will be shown *)
    val mutable currentGroups: (string * uiComponent) list = []
          
    val mutable currentColumns : int = !columns

    method init = 
      cont#init;
      currentGroups <- []
          
    method update (complete: bool) = 
      (* Get the state and see what groups we have *)
      let regs = getState () in
      (* make any new groups we need *)
      let oldGroups = currentGroups in
      List.iter (fun ri -> 
        if not (List.exists (fun (n, _) -> n = ri.RI.rGroup) currentGroups) 
        then begin
          (* Make a register file for just this group *)
          let regs = new registers 
              ~width:width 
              ~popupWidth:popupWidth
              ~popupTitle:popupTitle
              ~popupHeight:popupHeight
              ~getState:(fun () -> 
                List.filter (fun ri' -> ri.RI.rGroup = ri'.RI.rGroup)
                  (getState ()))
              ~title:ri.RI.rGroup 
              ~saveID:("register_" ^ ri.RI.rGroup)
              top in
          (* We must also initialize it *)
          regs#init;
          if debug then 
            ignore (E.log "registerGroups: adding group %s\n" ri.RI.rGroup);
          currentGroups <- currentGroups @ [(ri.RI.rGroup, regs)];
        end)
        regs;

      if oldGroups != currentGroups || !columns <> currentColumns then begin
        (* make a list of panes *)
        let panes = List.map (fun (_, u) -> `Win u) currentGroups in
        (* Pack them by !colums *)
        if !columns < 1 || !columns > 3 then begin
          ignore (E.warn "You cannot use %d columns in the register group"
                    !columns);
          columns := 1
        end;
        let rec loop (pl: pane list) = 
          if pl = [] then 
            []
          else begin
            (* Get the next row *)
            let rec getNextRow (c: int) = function
                [] -> [], []
              | h :: rest when c > 0 -> 
                  let row', rest' = getNextRow (c - 1) rest in
                  h :: row', rest'
              | rest -> [], rest
            in
            let row, rest = getNextRow !columns pl in
            `Row row :: loop rest
          end
        in
        cont#setPanes (`Col (loop panes));

        currentColumns <- !columns;
      end;
          
      (* Now update everybody *)
      cont#update complete
      
end




let reset () =
  theUIComponents := [];
  ()



(********* Testing stuff *******)
class testUIComponent ~(name:string)
                      ?(expand=true)
                      (top:toplevel) : uiComponent = 
  let fr = Frame.create ~borderwidth:2 ~relief:`Raised top in
  (* We use the Tk name of the frame as the name *)
  let name = "name (" ^ Widget.name fr ^ ")" in
  let label = Label.create ~text:name ~width:10 ~height:2 fr in
  let _ = 
    Frame.configure ~width:(Winfo.reqwidth label)
      ~height:(Winfo.reqheight label) fr;
    pack [label] ~inside:fr ~expand:true ~side:`Left ~fill:`Both;
  in

  object (self)
    inherit defaultUIComponent ~saveID:name
                               ~expand:expand
                               ~dockable:false
                               ~name:"Test UI Component"
                               fr
    method copy (tl: toplevel) = 
      new testUIComponent ~name:name ~expand:expand tl

  end




let testUI () = 

  let configFile = "testui_gui.cfg" in

  U.loadConfiguration configFile;

  initGui ();
  let top = tkTop () in

  (* make a container for everything *)
  let cont = new container top in
  (* Must pack the container frame *)
  pack [cont#frame] ~inside:top ~expand:true ~fill:`Both;


  let c1 = new testUIComponent ~name:"test1" top in
  let c2 = new testUIComponent ~expand:false ~name:"test2" top in
  let c3 = new testUIComponent ~name:"test3" top in
  let c4 = new testUIComponent ~name:"test4" top in
  let c5 = new testUIComponent ~name:"test5" top in

  let windows = 
    `Row [ `Col [ `Row [ `Win c1;
                         `Win c2;
                         `Win c5;
                       ];
                  `Win c3; 
                ];
           `Win c4
         ] 
  in
  Protocol.debug := true;

  cont#setPanes windows;

  cont#init; (* Init the whole container *)
  
  theUIComponents := [ (cont :> uiComponent) ];

  (* And restore the configuration *)
  List.iter (fun ui -> ui#restoreConfig) !theUIComponents;

  updateUI true;

  mainLoop(); 

  ()

