(** Tree Data Structure used to store Execution Path
 * of the VCGen
 * 
 * Node Structure: -> Stub Node (start of execution)
 *             or  -> Running Node (middle of execution)
 *             or  -> End Node  (finished execution)
 *
 * This tree representation assumes the following:
 * The state of execution is stored at the nodes
 * as a State (Vcgen.State.mli), or as a list of nodes
 *
 * The external world assess the State tree with
 * the following api:
 *)

(** Each element of the state tree must implement at least these methods *)
type state = <
  name: string;
  (** A name for the state, to be used in the UI *)

  infoline: string;
  (** A string to show on the info line *)

  filename: string option;
  (** A file to display *)

  line: int option;
  (** A line to set as current *)

  displayAssumptions: Pretty.doc list; 
  (** A list of assumptions **)

  displayState: Reginfo.registerInfo list;
  (** A list of registers *)

  downcast: unit;
  (** Call this method to throw an exception that allows you to downcast this 
   * object *)

  copy: state;
  (** Make a copy of the state, for later use *)

>

(** The type of a node parameterized by the type of the element *)
and treeNode =  
    { mutable parent   : treeNode option; (* None means root *)
      mutable state    : state;
      mutable children : treeNode list;
      mutable doneNode : bool;     (* This node is fully explored *)
      mutable index    : int;      (* index number to keep track of unique 
                                    * nodes  *)
      mutable collapsed: bool;  (* Whether this node is shown collapsed or 
                                 * not. This makes sense only if it has 
                                 * children *)

      mutable depth: int;       (** The depth in the tree. The root is at 
                                 * depth 0 *)
      mutable x: int;           (* The x coordinate of the canvas. This is 
                                 * the coordinate of the | *)
      mutable y: int;           (** The y coordinate of the canvas *)
    } 

(* The type of a tree *)
type tree

exception Empty

(** {b API} *)
val newTree : state list -> tree
    (** Make a new tree based on some roots. Note that the tree will have a 
     * special root node.  *)

val advanceCurrentToNextUnfinishedLeaf : tree -> unit
    (** Advance the next current node *)
 
val branchCurrentNode : tree -> state list -> unit


(** Iterate over all parents of a node, but ignore the real root, since that 
 * should not be visible by clients. *)
val iterParents: (treeNode -> unit) -> treeNode -> unit


(** Iterate over all the children of a node *)
val iterChildren: (treeNode -> unit) -> treeNode -> unit

(** Iterate over the node itself AND all the children of a node *)
val iterNodeAndChildren: (treeNode -> unit) -> treeNode -> unit

(** Each tree has a root node *)
val root: tree -> treeNode

(** Add a node to a given parent *)
val newNode: tree -> state -> treeNode -> treeNode

(** The current node. This is the node where we continue the verification 
 * from.  *)
val currNode: tree -> treeNode

(** Set the current node. Will also set the visible node. *)
val setCurrent: tree -> treeNode -> unit

(** This is the node to display. It is set whenever the current node is set, 
 * but might be set separately as well. *)
val visibleNode: tree -> treeNode

val setVisible: tree -> treeNode -> unit

(** Take a deep copy of the tree. *)
val deepCopy: tree -> tree




