# A regression tester for fast
#
require 5.000;

# Packages to import.
use Getopt::Long;           # Command-line option processing
use File::Basename;         # File name parsing
use Cwd;                    # Directory navigation
use strict;
use Data::Dumper;
use FindBin;
use lib "$FindBin::Bin";

use RegTest;

print "Test infrastructure for Fast\n";

# Create our customized test harness
my $TEST = FastRegTest->new(AvailParams => { 'PROVE' => 1,
                                             'POINTS' => 1,
                                             'ITERS' => 1,
                                             'SUCCESS' => 0},
                            LogFile => "fast.log",
                            CommandName => "testfast");

###
###
###
### Specialize RegTest
###
package FastRegTest;

use strict;
use Data::Dumper;

BEGIN {
    use RegTest;
    @FastRegTest::ISA = qw(RegTest);        # Inherit from RegTest
}

# The constructor
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@_);

    # print Dumper($self);

    return $self;
}



sub extraOptions {
    my($self) = @_;
    return ("--test");
}

sub extraHelpMessage {
    my($self) = @_;
    
    my ($scriptname, $extra) = $self->SUPER::extraHelpMessage();
    return ("testfast",
            $extra . << "EOF");

Additional arguments for Fast test harness
  --test       Test only, do not time.

  Default log file is fast.log
EOF
}

sub errorHeading {
    my($self, $err) = @_;
    return "Not executed" if $err == -1;
    return "Success" if $err == 0;
    return "Initial" if $err == 1000;
    return "Parse error" if $err == 1001;
    return "Proving error" if $err == 1002;
    return $self->SUPER::errorHeading($err);
}

sub startParsingLog {
    my($self, $tst) = @_;
    $tst->{instage} = 1000;
    $tst->{ErrorCode} = 0;
}


sub availableParameters {
    my($self) = @_;
    return %::availpars;
}

# We watch the log and we remember in what stage we are (so that we can
# interpret the error)

# Stages:
#  1000 - Start
#  1001 - Parsing
#  1002 - Proving

my %commonerrors = 
    ("^Parsing file" => sub { $_[1]->{instage} = 1001; },

     "^Proving goal with ([\\d]+) points" 
     => sub { $_[1]->{instage} = 1002; 
              $_[1]->{POINTS}  = $_[2]; },
# We are seeing an error from make. Try to classify it based on the stage
# in which we are
     "^make: \\*\\*\\*" => 
     sub { 
         if($_[1]->{ErrorCode} == 0) {
             $_[1]->{ErrorCode} = $_[1]->{instage};
         }},

    "^Segmentation fault" =>
     sub {
        if($_[1]->{ErrorCode} == 0) {
             $_[1]->{ErrorCode} = $_[1]->{instage};
         }},    

     "^Ran ([\\d.]+) iterations" => sub { $_[1]->{ITERS} = $_[2]; },
     
     "^Time\\s+([\\d.]+) ms" => sub { $_[1]->{PROVE} = $_[2]; },
     "^Error .+:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; },
    );

                                         
# Start with a few tests that must be run first
my $cong = "../src/src_rand/c/cong.exe";

sub addRandTest {
    my($name, # The name of the file (withouth the extension)
       $points, # The number of points
       %args, # An optional list of key => data
              #   Group => ['grp1', 'grp2' ]
              #   Enabled => 0
              #   ExtraArgs => "something to add to command line"
              #   Comm => "a comment to be printed if it fails"
              #   ExpectPattern => "something to appear on output"
              #   MustFail => 1
       ) = @_;
    my $extra = ($TEST->{option}->{test} ? "--test=1" : "");

    my $tst = 
        $TEST->newTest(
                  Name => "$name.$points",
                  Dir => ".",
                  Cmd => "$cong --points=$points $extra $args{ExtraArgs} $name.p",
                  Group => [],
                  ExtraArgs => 
                       ($TEST->{option}->{test} ? "--test" : ""),
                  Patterns => \%commonerrors);
    delete $args{ExtraArgs};
    foreach my $k (keys %args) {
        $tst->{$k} = $args{$k};
    }
    return $tst;
}


#####################################################


########### PAPER #############
my %papertests = ( 
                   "arith-dense" => 30,
                   "arith-sparse" => 20,
                   "both-dense" => 35,
                   "both-sparse" => 50,
                   "uf-single" => 21,
                   "uf-copies" => 21,
                   "uf-wrong" => 80
                   );
foreach my $tst (keys %papertests) {
    &addRandTest($tst, $papertests{$tst},
             Group => ["paper"], 
             ExpectPattern => "Proved Goal");
}

######### ARITH TRUE ##############
my %arithtrue = (
                 "arith1-true" => 5,
                 "arith2-true" => 5,
                 "arith3-true" => 5,
                 "arith4-true" => 5,
                 "arith5-true" => 5,
                 "arith6-true" => 5,
                 "arith7_20DenseEq-true" => 25,
                 "arith8_26DenseEq-true" => 30,
                 );

foreach my $tst (keys %arithtrue) {
    &addRandTest($tst, $arithtrue{$tst},
             Group => ["arith"], ExpectPattern => "Proved Goal");
}

######## ARITH FALSE ############
my %arithfalse = (
		  "arith1-false" => 5,
		  "arith2-false" => 5,
		  "arith3-false" => 5,
		  "arith4-false" => 5,
		  "arith5-false" => 5,
		  "arith6-false" => 5,
		  "arith7-false" => 25,
		  "arith8-false" => 30,
                 );

foreach my $tst (keys %arithfalse) {
    &addRandTest($tst, $arithfalse{$tst},
             Group => ["arith-false"], 
             MustFail => 1,
             ExpectPattern => "Unable to prove goal");
}


####### BOTH TRUE ############
my %bothtrue = (
		"both1-true" => 5,
		"both2-true" => 35,
		"both3-true" => 40,
		"both4-true" => 40,
		"both5-true" => 45,
		"both6-true" => 40,
		"both7_sparse-true" => 80,
		"both9_sparse-true" => 50,
		"both10_sparse-true" => 25,
		"both11_sparse-true" => 20,
		"both12_sparse-true" => 20,
		"both13_sparse-true" => 20,
		"both14_sparse-true" => 15,
		"both15_sparse-true" => 10,
		"both16_sparse-true" => 15,
		"both17_sparse-true" => 15,
		"both18_sparse-true" => 20,
		);

foreach my $tst (keys %bothtrue) {
    &addRandTest($tst, $bothtrue{$tst},
             Group => ["both"], ExpectPattern => "Proved Goal");
}

###### BOTH INCONSISTENT ########
my %both_incon = (
                  "both1-incon" => 5,
                  "both2-incon" => 40,
		  "both3-incon" => 10,
                  );

foreach my $tst (keys %both_incon) {
    &addRandTest($tst, $both_incon{$tst},
             Group => ["both-incon"], 
             ExpectPattern => "Inconsistent");
}


###### UF TRUE #################
my %uf_true = (
	       "uf0-true" => 5,
	       "uf1-true" => 5,
	       "uf3-true" => 10,
	       );

foreach my $tst (keys %uf_true) {
    &addRandTest($tst, $uf_true{$tst},
             Group => ["uf"], 
             ExpectPattern => "Proved Goal");
}



###### UF FALSE #################
my %uf_false = (
		"uf1-false" => 5,
		"uf2-false" => 5,
		"uf3-false" => 5,
	       );

foreach my $tst (keys %uf_false) {
    &addRandTest($tst, $uf_false{$tst},
             Group => ["uf-false"], 
             MustFail => 1,
             ExpectPattern => "Unable to prove goal");
}

###### BAD #################
my %bad = (
		"both19_sparse250EqCrashes" => 500,
	       );

foreach my $tst (keys %bad) {
    &addRandTest($tst, $bad{$tst},
             Group => ["bad"], 
             Enabled => 0,
             MustFail => 1,
             ExpectPattern => "Unable to prove goal");
}


####### EQUALITY
sub mkTest {
    my($name, # The name of the file (withouth the extension)
       %args, # An optional list of key => data
              #   Group => ['grp1', 'grp2' ]
              #   Enabled => 0
              #   ExtraArgs => "something to add to command line"
              #   Comm => "a comment to be printed if it fails"
              #   ExpectPattern => "something to appear on output"
              #   MustFail => 1
       ) = @_;

    my $tst = 
        $TEST->newTest(
                  Name => "$name",
                  Dir => ".",
                  Cmd => "make $args{ExtraArgs} $name",
                  Group => [],
                  ExtraArgs => "",
                  Patterns => \%commonerrors);
    delete $args{ExtraArgs};
    foreach my $k (keys %args) {
        $tst->{$k} = $args{$k};
    }
    return $tst;
}

&mkTest("equality/0", 
        Group => ['cc'],
        ExtraArgs => "EXTRAARGS=-checkproofs");
                


# Now invoke it
$TEST->doit();

# print Dumper($TEST);

exit(0);


1;



