# A regression tester for Kettle
#
require 5.000;

# Packages to import.
use Getopt::Long;           # Command-line option processing
use File::Basename;         # File name parsing
use Cwd;                    # Directory navigation
use strict;
use Data::Dumper;
use FindBin;
use lib "$FindBin::Bin/../ocamlutil";

use RegTest;

print "Test infrastructure for Kettle\n";

# Create our customized test harness
my $TEST = KettleRegTest->new(AvailParams => { 'PROVE' => 1,
                                               'MEMORY' => 1,
                                               'SUCCESS' => 0 },
                              LogFile => "testkettle.log",
                              CommandName => "testkettle");

###
###
###
### Specialize RegTest
###
package KettleRegTest;

use strict;
use Data::Dumper;

BEGIN {
    use RegTest;
    @KettleRegTest::ISA = qw(RegTest);        # Inherit from RegTest
}

# The constructor
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@_);

    # print Dumper($self);

    return $self;
}



sub extraOptions {
    my($self) = @_;
    return "";
}

sub extraHelpMessage {
    my($self) = @_;
    
    my ($scriptname, $extra) = $self->SUPER::extraHelpMessage();
    return ("testkettle",
            $extra . << "EOF");

  Default log file is testkettle.log
EOF
}

sub errorHeading {
    my($self, $err) = @_;
    return "Not executed" if $err == -1;
    return "Success" if $err == 0;
    return "Initial" if $err == 1000;
    return "Parse error" if $err == 1001;
    return "Proving error" if $err == 1002;
    return $self->SUPER::errorHeading($err);
}

sub startParsingLog {
    my($self, $tst) = @_;
    $tst->{instage} = 1000;
    $tst->{ErrorCode} = 0;
}


sub availableParameters {
    my($self) = @_;
    return %::availpars;
}

# We watch the log and we remember in what stage we are (so that we can
# interpret the error)

# Stages:
#  1000 - Start
#  1001 - Parsing
#  1002 - Proving

my %commonerrors = 
    ("^Parsing file" => sub { $_[1]->{instage} = 1001; },

     "^Proving goal with ([\\d]+) points" 
     => sub { $_[1]->{instage} = 1002; 
              $_[1]->{POINTS}  = $_[2]; },
     "^Proving goal" 
     => sub { $_[1]->{instage} = 1002; },
# We are seeing an error from make. Try to classify it based on the stage
# in which we are
     "^make: \\*\\*\\*" => 
     sub { 
         if($_[1]->{ErrorCode} == 0) {
             $_[1]->{ErrorCode} = $_[1]->{instage};
         }},

    "^Segmentation fault" =>
     sub {
        if($_[1]->{ErrorCode} == 0) {
             $_[1]->{ErrorCode} = $_[1]->{instage};
         }},    

     "^Ran ([\\d.]+) iterations" => sub { $_[1]->{ITERS} = $_[2]; },
     
     "^TOTAL\\s+([\\d.]+) s" => sub { $_[1]->{PROVE} = $_[2]; },
     "^Memory.*total=([\\d.]+)M" => sub { $_[1]->{MEMORY} = $_[2]; },
     "^Error parsing:" => sub { $_[1]->{ErrorCode} = 1001; },
     "^Error .+:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; },
     "^Bug:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; 
                      if(! defined $_[1]->{ErrorMsg}) {
                          $_[1]->{ErrorMsg} = $_; 
                      }},
     "^Unimplemented:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; 
                      if(! defined $_[1]->{ErrorMsg}) {
                          $_[1]->{ErrorMsg} = $_; 
                      }},
    );

                                         

sub addTest {
    my($name, # The name of the file (withouth the extension)
       %args, # An optional list of key => data
              #   Group => ['grp1', 'grp2' ]
              #   Target => the target if different from name
              #   Enabled => 0
              #   ExtraArgs => "something to add to command line"
              #   ExtraKettleArgs => "something to add to Kettle command line"
              #   Comm => "a comment to be printed if it fails"
              #   ExpectPattern => "something to appear on output"
              #   MustFail => 1
       ) = @_;

    my $target = defined $args{Target} ? $args{Target} : $name;
    my $tst = 
        $TEST->newTest(
                  Name => $name,
                  Dir => ".",
                  Cmd => "make EXTRAARGS=\"-checkproofs -timing $args{ExtraKettleArgs} $ENV{EXTRAARGS}\" $target",
                  Group => [],
                  ExtraArgs => $args{ExtraArgs},
                  Patterns => \%commonerrors);
    delete $args{ExtraKettleArgs};
    foreach my $k (keys %args) {
        $tst->{$k} = $args{$k};
    }
    return $tst;
}

sub setTimeout {
    my($name, $value) = @_;
    my $cmd = $TEST->getField($name, 'Cmd');
    $cmd =~ s|-checkproofs|-checkproofs -timeout $value|;
    $TEST->setField($name, 'Cmd', $cmd);
}
sub noCheckProofs {
    my($name) = @_;
    my $cmd = $TEST->getField($name, 'Cmd');
    $cmd =~ s|-checkproofs||;
    $TEST->setField($name, 'Cmd', $cmd);
}

# test/equality 
foreach my $t ("0", "1", "2", "3", "4", "5", "7", "1_SATb", 
               "uninterpreted") {
    &addTest("prove/equality/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC");
}

       
foreach my $t ("uf0-true", "uf1-true", "uf-single", "uf-copies", 
               "uf-wrong", "uf2-false", "uf2-true", 
               "uf3-true", "uf3-false") {
    &addTest("prove/sumit/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC");
}         
$TEST->enable("prove/sumit/uf-wrong", 0);

foreach my $t ("uf1-false", "uf2-false", "uf3-false") {
    &addTest("prove/sumit/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC",
             MustFail => 1);
}         

foreach my $t ("0", "1", "2", "3", "4", "5", "6", "7", 
               "8", "9", "10", "11", "12", "13", "14", "15", 
               "16", "17", "18", "19", "20", "21", 
               "40", "greg", "hypax", "neq0", "unint",
               "w1vc") {
    &addTest("prove/cmu/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC");
}         
$TEST->addComment("prove/cmu/13", "integer programming");


foreach my $t ("simplex/Simplex") 
{
    &addTest("prove/tcj/" . $t, 
             Group => ['tcj'],
             ExtraArgs => "USE_TCJ=1",
             Comm => "works only if TCJ is enabled");
}


$TEST->addComment("prove/cmu/21", "parsing => as part of the goal");
$TEST->addComment("prove/cmu/hypax", 
                  "parsing complex pred as part of assumption");
$TEST->getTest("prove/cmu/7")->{MustFail} = 1;
$TEST->getTest("prove/cmu/9")->{MustFail} = 1;

foreach my $t ("1", "2", "3", "4", "5", "6", "7", "8", "9", "funcallbug",
               "simple0", "simple1", "simple2", "simple3", "simple4",
               "simple5", "simple6",
               "ex0", "ex1", 
               "dc0", "dc1", "dc2", "dc3", "dc4", "dc5", "dc6",
               "dc7", "dc8",
               "pfbug1") {
    # Regular arithmetic
    &addTest("prove/arith/" . $t,  
             Group => ['shostak', 'arith'],
             ExtraKettleArgs => "-disable SX -enable SH");

    # Test Simplex
    &addTest("prove/arith/simplex/" . $t,  
             Group => ['simplex', 'arith'],
             Target => "prove/arith/$t",
             ExtraKettleArgs => "-enable SX -disable SH");
}


foreach my $t ("arr0", "arr1", "arr2", "arr5", "arr6", "arr7", "arr8", 
               "arr9",
               "arrptr", "array1", "array2", "array3",
               "pldi05_3", "pldi05_1", "pldi05_2", "pldi05_4", 
               "pldi05_3_ax", "pldi05_1_ax", "pldi05_2_ax", "pldi05_4_ax", 
               "pldi05_ax", "pldi05_ns",
               "insert", "insert_ax") {
    &addTest("prove/arrays/$t", 
             Group => ['arrays'],
             ExtraKettleArgs => "-enable AR");
}
$TEST->addComment("prove/arrays/arr6", "We cannot prove equalities of memories");
$TEST->addComment("prove/arrays/arrptr", "Working on it");

$TEST->enable("prove/arrays/arrptr", 0);

foreach my $t ("case0", "case01", "bug1",
               "case1", "case2", "case3", "case4", "case5", 
               "case6", "case7", "case8") {
    &addTest("prove/cases/$t");
}

foreach my $t ("m0", "m1", "m2", "m3", "m4", "m5", "m6", "m7",
               "m8", "m10", "m11", "infinite") { 
    &addTest("prove/match/$t", Group => [ 'match' ]);
}
$TEST->enable("prove/match/infinite", 0);

foreach my $t ("doublylinked", "scheduler2", "gc1",
               "exist1", "exist2", "exist3",
               "consume1", "consume2", "consume2-false1",
               "dis1-fail1", "dis1", "dis2-fail1", "dis2", 
               "dis3-fail1", "dis3", "dis4-fail1", "dis4-fail2", 
               "dis4","f12", "fplus12", "fx","gf35","gfxy", "ggfxy", 
               "inf1", "mem1", "mem2", "notneq",
               "list3", "dis5",
               "qdis1", "qdis1-fail1", "trans", "trans2", "dis6",
               "dis7", "dis8") {
    &addTest("prove/scott/$t", Group => [ 'scott' ]);
}

# A test that must be run without checkproofs
&addTest("prove/scott/set_bad_case_opt", Group => [ 'scott' ]);
&noCheckProofs("prove/scott/set_bad_case_opt");

&addTest("prove/scott/gc_copy", Group => [ 'scott', 'slow' ],
         ExtraKettleArgs => ' -timeout 20');
&noCheckProofs("prove/scott/gc_copy");



$TEST->enable("prove/scott/inf1", 0);

foreach my $t ("driver", "expaper", "funcall-hard",
               "pointertest-bug", "pointertest", "sr", "swap", 
               "tfwdbwd", "trivial") {
    &addTest("prove/blast/$t", 
             Group => ['blast', 'simplex']);
}


&addTest("set_tests");
 
&addTest("prove/smtlib/QF_AUFLIA/bug1");
&addTest("prove/smtlib/QF_AUFLIA/pp-invariant");
&addTest("prove/smtlib/QF_AUFLIA/fb-var-6-12");
&addTest("prove/smtlib/QF_AUFLIA/fb-var-12-11", 
         Group => ['slow']);
&noCheckProofs("prove/smtlib/QF_AUFLIA/fb-var-12-11");

# this now runs a special target in test/Makefile.in
&addTest("prove/scott/cacheproblem");
&noCheckProofs("prove/scott/cacheproblem");# problem disappears with -checkproofs

# The bug only appears in the debug version (when we check a lot of internal
# proofs) and with -nostop
&addTest("prove/cmu/bug1", 
         ExtraKettleArgs => '-nostop',
         ExtraArgs => 'DEBUG=1',
         MustFail => 1,
         ExpectPattern => 'Kettle failure');

&addTest("prove/match/lambda1");
&addTest("prove/match/trig1");


#####################################################



# Now invoke it
$TEST->doit();

# print Dumper($TEST);

exit(0);


1;



