// c2v_eval.cc
// code for c2v_eval.ast

#include "c2v.ast.gen.h"        // includes c2v_eval.ast declarations for this module
#include "cc.ast.gen.h"         // C++ AST
#include "c2vml.h"              // C2VML
#include "exttype.h"            // ECompoundType
#include "strutil.h"            // parseQuotedString

#include <ctype.h>              // islower

// STUBEMIT: #include "c2v.ast.gen.h"


// ---------------------------- C2VML ---------------------------
C2VScope *C2VML::pushScope()
{
  C2VScope *s = new C2VScope;
  scopes.prepend(s);
  return s;
}

void C2VML::popScope(C2VScope *scope)
{
  C2VScope *s = scopes.removeFirst();
  xassert(s == scope);
  delete s;
}


void C2VML::promoteInnerScopeValues()
{                   
  C2VScope *inner = scopes.first();
  C2VScope *next = scopes.nth(1);

  C2VValueMap::Iter iter(inner->map);
  for (; !iter.isDone(); iter.adv()) {
    next->map.addReplace(iter.key(), new C2VValue(*(iter.value())));
  }
}


void C2VML::bind(StringRef name, C2VValue * owner value)
{
  C2VScope *scope = scopes.first();
  scope->map.addReplace(name, value);
}

C2VValue * /*nullable*/ C2VML::lookup(StringRef name)
{
  FOREACH_OBJLIST_NC(C2VScope, scopes, iter) {
    C2VValue *v = iter.data()->map.get(name);
    if (v) {
      return v;
    }
  }
  return NULL;     // not found in any scope
}

C2VValue * /*nullable*/ C2VML::lookupInner(StringRef name)
{
  return scopes.first()->map.get(name);
}

C2VValue *C2VML::lookupf(StringRef name)
{
  C2VValue *v = lookup(name);
  if (!v) {
    error(stringc << "unbound variable: " << name);
  }
  return v;
}


bool C2VML::variableAcceptsType(StringRef name, C2VValueKind kind)
{
  if (name[0]=='$') {
    if (islower(name[1])) {
      // lowercase initial letter: unconstrained, but cannot be used
      // in a pattern
      return true;
    }

    switch (name[1]) {
      case 'I':
      case 'N':
        return kind==VK_IDENTIFIER;

      case 'S': throw_XUnimp("$S");
      case 'E': throw_XUnimp("$E");
      case 'T': throw_XUnimp("$T");
      case 'F': throw_XUnimp("$F");

      case 'V':
        switch (name[2]) {
          case 'F':
          case 'N': return kind==VK_IDENTIFIER;
          case 'E': return kind==VK_VML_EXPRESSION;
          case 'P': return kind==VK_VML_PREDICATE;
          case 'S': return kind==VK_VML_STATEMENT;
        }
        break;

      default:
        error(stringc << "illegal variable name `" << name << "'");
    }
  }
  else {
    error(stringc << "illegal variable name `" << name << "'");
  }

  return false;    // silence warning
}

void C2VML::checkVariableType(StringRef name, C2VValueKind kind)
{
  if (!variableAcceptsType(name, kind)) {
    env.error(stringc << "variable " << name
                      << " assigned value of type " << toString(kind));
  }
}


// ------------------------ C2VStmt::eval -------------------------
// STUBCODE: return SER_FALSE;
// STUBPATTERN: ^StmtEvalResult CS_.*::ieval\b

StmtEvalResult C2VStmt::eval(C2VML &env) const          // stubify
{
  return ieval(env);
}


StmtEvalResult CS_varDecl::ieval(C2VML &env) const
{
  if (env.lookupInner(name)) {
    env.error(stringc << "duplicate declaration of " << name);
  }

  // evaluate
  C2VValue v = value->eval(env);

  // check that the expression's type is consistent with the variable name
  env.checkVariableType(name, v.kind);

  env.bind(name, new C2VValue(v));
  return SER_TRUE;
}

StmtEvalResult CS_expr::ieval(C2VML &env) const
{
  C2VValue val = expr->eval(env);
  return val.getInt()? SER_TRUE : SER_FALSE;
}

StmtEvalResult CS_skip::ieval(C2VML &env) const
{
  return SER_TRUE; 
}

StmtEvalResult CS_compound::ieval(C2VML &env) const
{
  FOREACH_ASTLIST(C2VStmt, stmts, iter) {
    StmtEvalResult res = iter.data()->eval(env);
    if (res != SER_TRUE) {
      // first false or return statement in a sequence aborts the whole sequence
      return res;
    }
  }

  return SER_TRUE;
}

StmtEvalResult CS_if::ieval(C2VML &env) const
{
  if (guard->eval(env).getInt()) {
    return thenBranch->eval(env);
  }
  else {
    return elseBranch->eval(env);
  }
}

StmtEvalResult CS_for_each_field::ieval(C2VML &env) const
{
  C2VValue v = type->eval(env);
  Type *t = v.getCType();
  ECompoundType *ct = ext(t->ifCompoundType());
  if (!ct) {
    env.error("for_each_field must be used with compound type");
  }

  FOREACH_OBJLIST_NC(Variable, ct->fieldSlots, iter) {
    Variable *field = iter.data();

    C2VScope *scope = env.pushScope();
    env.bind(var, new C2VValue(VK_C_FIELD, field));
                    
    // evaluate body for side effects
    body->eval(env);

    env.popScope(scope);
  }
  
  return SER_TRUE;
}

StmtEvalResult CS_emit::ieval(C2VML &env) const
{
  // replace pattern variable references with concrete AST
  VStatement * owner conc = env.substVStatement(stmt);

  // insert into VML output
  if (conc->isVS_compound()) {
    // the VS_compound arises from the braces used in the "emit"
    // syntax; strip them away so blocks can declare variables
    // that can be used later; a block can always put an extra
    // layer of braces is a VS_compound is truly desired
    VS_compound * cpd = conc->asVS_compound();
    while (cpd->body.isNotEmpty()) {
      VStatement * owner s = cpd->body.removeFirst();
      env.addStmt(s);
    }
    delete cpd;
  }
  else {
    // at the moment it is syntactically legal to leave off the
    // braces, and so we get here; I have not yet decided whether
    // I want to keep this possibility
    env.addStmt(conc);
  }

  return SER_TRUE;
}

StmtEvalResult CS_return::ieval(C2VML &env) const
{
  // get current binding for the return value
  C2VValue *retv = env.lookup(env.str("$__retval"));
  if (!retv) {
    env.error("'return' used in context where it is not allowed");
  }

  // evaluate the return value
  if (expr) {
    // the return value is a VML expression fragment that must have
    // its pattern variables substituted
    VExpr *conc = env.substVExpr(expr);
    C2VValue v(VK_VML_EXPRESSION, conc);
    if (retv->kind != v.kind) {
      // I don't think this can happen in the current design ..
      env.error("wrong return value type");
    }

    // modify the 'retv' object in-place because the code that
    // kicks off the evaluation doesn't redo the lookup
    *retv = v;
  }

  return SER_RETURN;
}


// --------------------- C2VExpr::eval ---------------------------
// STUBCODE: return C2VValue(0);
// STUBPATTERN: ^C2VValue CE_.*::eval\b

//C2VValue C2VExpr::eval(C2VML &env) const

C2VValue CE_matchStmt::eval(C2VML &env) const
{ xunimp("todo"); return C2VValue(0); }

C2VValue CE_matchExpr::eval(C2VML &env) const
{
  C2VValue concreteVal = concrete->eval(env);

  // isolate resulting bindings until we know the match has succeeded
  C2VScopePushPop scope(env);

  if (env.matchExpr(concreteVal.getCExpr(), pattern->cexpr)) {
    env.promoteInnerScopeValues();
    return C2VValue(1);
  }
  else {
    return C2VValue(0);
  }
}

C2VValue CE_matchType::eval(C2VML &env) const
{
  C2VValue concreteVal = concrete->eval(env);

  // isolate resulting bindings until we know the match has succeeded
  C2VScopePushPop scope(env);

  if (env.matchSemType(concreteVal.getCType(), pattern->ctype)) {
    env.promoteInnerScopeValues();
    return C2VValue(1);
  }
  else {
    return C2VValue(0);
  }
}

C2VValue CE_var::eval(C2VML &env) const
{
  C2VValue *val = env.lookup(name);
  if (!val) {
    env.error(stringc << "unbound variable: " << name);
    // not reached
  }               
  return *val;
}

C2VValue CE_funCall::eval(C2VML &env) const
{
  // evaluate arguments
  GrowArray<C2VValue> argVals(args.count());
  {
    int ct=0;
    FOREACH_ASTLIST(C2VExpr, args, iter) {
      argVals[ct++] = iter.data()->eval(env);
    }
  }

  if (0==strcmp(func, "breaker")) {
    breaker();
    return C2VValue(1);
  }

  if (0==strcmp(func, "makeTempVarName")) {
    if (argVals.size() != 1) {
      env.error("makeTempVarName accepts 1 argument");
    }
    StringRef fresh = env.makeTempVarName(argVals[0].getString());
    return C2VValue(VK_IDENTIFIER, fresh);
  }

  if (0==strcmp(func, "c2vmlExpr")) {
    if (argVals.size() != 1) {
      env.error("c2vmlExpr accepts 1 argument");
    }
    Expression const *ce = argVals[0].getCExpr();

    // it seems that my 'c2vml' methods are not marked 'const'...
    // I *think* they could be, so I should at some point go through
    // and fix this, but for now I hack it with a cast
    VExpr *ve = const_cast<Expression*>(ce)->c2vml(env);

    return C2VValue(VK_VML_EXPRESSION, ve);
  }

  if (0==strcmp(func, "modelVar")) {
    if (argVals.size() != 1) {
      env.error("modelVar accepts 1 argument");
    }
    Variable *var = argVals[0].getCVariable();

    return C2VValue(VK_IDENTIFIER, var->name);
  }

  if (0==strcmp(func, "metaVar")) {
    if (argVals.size() != 1) {
      env.error("metaVar accepts 1 argument");
    }
    Variable *var = argVals[0].getCVariable();

    return C2VValue(VK_IDENTIFIER, env.str(stringc << var->name << "_meta"));
  }

  if (0==strcmp(func, "typeof")) {
    if (argVals.size() != 1) {
      env.error("typeof accepts 1 argument");
    }
     
    switch (argVals[0].kind) {
      default: 
        env.error("typeof accepts expressions or variables or fields");

      case VK_C_EXPRESSION:
        return C2VValue(VK_C_TYPE, argVals[0].getCExpr()->type);

      case VK_C_VARIABLE:
        return C2VValue(VK_C_TYPE, argVals[0].getCVariable()->type);

      case VK_C_FIELD:
        return C2VValue(VK_C_TYPE, argVals[0].getCField()->type);
    }
  }

  if (0==strcmp(func, "isSplitModelStruct")) {
    if (argVals.size() != 1) {
      env.error("isSplitModelStruct accepts 1 argument");
    }
    Type *t = argVals[0].getCType();

    if (t->isCompoundType()) {
      ECompoundType *ct = ext(t->asCompoundType());
      return C2VValue(ct->flatMap? 0 : 1);
    }
    return C2VValue(0);
  }

  if (0==strcmp(func, "isFlatModelStruct")) {
    if (argVals.size() != 1) {
      env.error("isFlatModelStruct accepts 1 argument");
    }
    Type *t = argVals[0].getCType();

    if (t->isCompoundType()) {
      ECompoundType *ct = ext(t->asCompoundType());
      return C2VValue(ct->flatMap? 1 : 0);
    }
    return C2VValue(0);
  }

  if (0==strcmp(func, "tagName")) {
    if (argVals.size() != 1) {
      env.error("tagName accepts 1 argument");
    }
    Type *t = argVals[0].getCType();

    if (!t->isCompoundType()) {
      env.error(stringc << "tagName must be applied to a structure type, not `"
                        << t->toString() << "'");
    }

    ECompoundType *ct = ext(t->asCompoundType());
    return C2VValue(VK_IDENTIFIER, ct->tagVar->name);
  }

  if (0==strcmp(func, "fieldArrayNameOf")) {
    if (argVals.size() != 1) {
      env.error("fieldArrayNameOf accepts 1 argument");
    }
    Variable *field = argVals[0].getCField();

    return C2VValue(VK_IDENTIFIER, field->name);
  }

  // what is the difference between 'fieldArrayNameOf' and
  // 'fieldSlotName'?  the former works in for_each_field
  // loops... ugh.
  if (0==strcmp(func, "fieldSlotName")) {
    if (argVals.size() != 1) {
      env.error("fieldSlotName accepts 1 argument");
    }
    Variable *field = argVals[0].getCField();

    return C2VValue(VK_IDENTIFIER, env.fieldSlotName(field));
  }

  if (0==strcmp(func, "hasVarAttribute")) {
    if (argVals.size() != 2) {
      env.error("hasVarAttribute accepts 2 arguments");
    }
    EVariable *v = ext(argVals[0].getCVariable());
    StringRef quotedAttrName = argVals[1].getString();

    string attrName = parseQuotedString(quotedAttrName);
    VarAttr attr = stringToVarAttr(attrName.c_str());
    if (!attr) {
      env.error(stringc << "invalid variable attribute: " << quotedAttrName);
    }

    return C2VValue(v->hasVAttr(attr)? 1 : 0);
  }

  env.error(loc, stringc << "unknown function: " << func);
  return C2VValue(0);
}

C2VValue CE_string::eval(C2VML &env) const
{
  return C2VValue(VK_STRING, str);
}

C2VValue CE_intLit::eval(C2VML &env) const
{
  return C2VValue(VK_INT, n);
}

C2VValue CE_binary::eval(C2VML &env) const
{
  if (op == BIN_ASSIGN) {
    if (e1->isCE_var()) {
      StringRef target = e1->asCE_varC()->name;
      C2VValue v = e2->eval(env);
      env.bind(target, new C2VValue(v));
      return v;
    }
    else {
      env.error("assignment must have variable as LHS");
    }
  }

  // evaluate first argument
  C2VValue v1 = e1->eval(env);

  // short-circuit ops first
  if (op == BIN_OR) {
    if (v1.getInt()) {
      return C2VValue(1);     // no need to evaluate second argument
    }
    else {
      return C2VValue(e2->eval(env).getInt()? 1 : 0);
    }
  }

  if (op == BIN_AND) {
    if (!v1.getInt()) {
      return C2VValue(0);
    }
    else {
      return C2VValue(e2->eval(env).getInt()? 1 : 0);
    }
  }

  // evaluate second argument
  C2VValue v2 = e2->eval(env);

  switch (op) {
    default:
      xunimp(stringc << "binary operator " << toString(op));

    case BIN_EQUAL:
      return C2VValue(env.equalValues(v1, v2));

    case BIN_NOTEQUAL:
      return C2VValue(!env.equalValues(v1, v2));
      
    case BIN_PLUS:
      return C2VValue(v1.getInt() + v2.getInt());

    case BIN_LESS:
      return C2VValue(v1.getInt() < v2.getInt());

    case BIN_GREATER:
      return C2VValue(v1.getInt() > v2.getInt());
  }
}


// EOF
