// c2v_match.cc
// match C AST fragments against patterns

#include "c2vml.h"              // this module
#include "cc.ast.gen.h"         // C++ AST
#include "vml.ast.gen.h"        // VML AST
#include "strutil.h"            // parseQuotedString

#include <stdlib.h>             // strtoul


// string->int conversion for c2v patterns (the concrete syntax will
// not need this since it has already been through the type checker)
static unsigned long stringToInt(char const *str)
{
  return strtoul(str, NULL /*endptr*/, 0 /*radix*/);
}


// ----------------------------------------------------------
bool C2VML::matchExpr(Expression const *conc, Expression const *pat)
{
  // skip grouping parens on either side
  if (conc->isE_grouping()) {
    return matchExpr(conc->asE_groupingC()->expr, pat);
  }
  if (pat->isE_grouping()) {
    return matchExpr(conc, pat->asE_groupingC()->expr);
  }

  // match a pattern variable?
  if (pat->isE_patvar()) {
    StringRef patvname = pat->asE_patvarC()->name;
    if (patvname[1]=='N') {
      // 'conc' must be an E_variable
      if (conc->isE_variable()) {
        return matchBind(patvname,
          new C2VValue(VK_C_VARIABLE, conc->asE_variableC()->var));
      }
      else {
        return false;
      }
    }
    else {
      // 'conc' can be any kind of expression
      xassert(patvname[1]=='E');
      return matchBind(patvname,
        new C2VValue(VK_C_EXPRESSION, conc));
    }
  }

  // both are concrete, should be same type
  if (conc->kind() != pat->kind()) {
    return false;
  }

  ASTSWITCH2C(Expression, conc, pat) {
    ASTCASE2C(E_boolLit, c, p)
      return c->b == p->b;

    ASTNEXT2C(E_intLit, c, p)
      return c->i == stringToInt(p->text);

    ASTNEXT2CU(E_floatLit, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_stringLit, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_charLit, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_this, c, p)
      xunimp("todo");

    ASTNEXT2C(E_variable, c, p)
      // ignore qualifiers ....
      return c->name->getName() == p->name->getName();

    ASTNEXT2C(E_funCall, c, p)
      return matchExpr(c->func, p->func) &&
             matchArgExprs(c->args, p->args);

    ASTNEXT2CU(E_constructor, c, p)
      xunimp("todo");

    ASTNEXT2C(E_fieldAcc, c, p)
      return matchExpr(c->obj, p->obj) &&
             matchField(c->field, p->fieldName);

    ASTNEXT2CU(E_sizeof, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_unary, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_effect, c, p)
      xunimp("todo");

    ASTNEXT2C(E_binary, c, p)
      return matchExpr(c->e1, p->e1) &&
             c->op == p->op &&
             matchExpr(c->e2, p->e2);

    ASTNEXT2CU(E_addrOf, c, p)
      xunimp("todo");

    ASTNEXT2C(E_deref, c, p)
      return matchExpr(c->ptr, p->ptr);

    ASTNEXT2CU(E_cast, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_cond, c, p)
      xunimp("todo");

    ASTNEXT2C(E_sizeofType, c, p)
      return matchType(c->atype, p->atype);

    ASTNEXT2CU(E_assign, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_new, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_delete, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_throw, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_keywordCast, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_typeidExpr, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_typeidType, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_grouping, c, p)
      xunimp("todo");
    ASTNEXT2CU(E_arrow, c, p)
      xunimp("todo");

    ASTDEFAULT2C
      xunimp("todo");

    ASTENDCASE2C
  }
}


bool C2VML::matchExprWithInt(int conc, Expression const *pat)
{
  if (pat->isE_intLit()) {
    return conc == (int)stringToInt(pat->asE_intLitC()->text);
  }

  if (pat->isE_patvar()) {
    StringRef patName = pat->asE_patvarC()->name;
    
    // bind to an expression
    E_intLit *lit = new E_intLit(str(stringc << conc));
    lit->i = conc;    // make it look tcheck'd
    return matchBind(patName, new C2VValue(VK_C_EXPRESSION, lit));
  }

  // I could conceivably do crazy things like match 2 against 1+$E and
  // bind $E to 1.  But I think expression matching should be mostly
  // syntactic (whereas type matching is mostly semantic), hence I
  // will refrain.

  return false;
}


bool C2VML::matchArgExprs(FakeList<ArgExpression> const *conc,
                          FakeList<ArgExpression> const *pat)
{
  while (conc->isNotEmpty() && pat->isNotEmpty()) {
    if (!matchExpr(conc->firstC()->expr, pat->firstC()->expr)) {
      return false;
    }
    conc = conc->butFirstC();
    pat = pat->butFirstC();
  }

  return conc->isEmpty() && pat->isEmpty();
}


// if 't' is a named type, get the name, otherwise return NULL
StringRef getTypeName(Type *t)
{
  // both named types?
  if (!t->isCVAtomicType()) {
    return NULL;
  }
  CVAtomicType *cv = t->asCVAtomicType();
  if (!cv->isNamedAtomicType()) {
    return NULL;
  }
  return cv->asNamedAtomicType()->name;
}


bool C2VML::matchType(ASTTypeId const *conc, ASTTypeId const *pat)
{
  // This is slightly awkward because we will compare the semantic
  // type of 'conc' to the syntactic type of 'pat'.  We cannot use the
  // semantic type of 'pat' because it cannot be tchecked due to the
  // presence of pattern variables.  We *could* use the syntactic type
  // of 'conc', but then that would force writers of patterns to
  // include variations for each possible way a type can be expressed,
  // which would be annoying, so we instead use the semantic type as
  // a kind of normalized type.
  
  return matchSemType(conc->getType(), pat);
}

bool C2VML::matchSemType(Type *concType, ASTTypeId const *pat)
{
  // unravel the declarator structure of 'pat' so we can turn it
  // inside-out
  ArrayStack<IDeclarator*> patDeclarator;
  {
    IDeclarator *d = pat->decl->decl;
    while (d) {
      if (!d->isD_grouping()) {
        patDeclarator.push(d);
      }
      d = d->getBase();
    }
    
    // the final declarator will be something like D_name, which
    // we don't want to look at
    patDeclarator.pop();
  }

  // For 'pat' = "$T * const *", it is broken down like this:
  //
  //                 patDeclarator[1]
  //                          |
  //      patDeclarator[0]    |
  //               |          |
  // pat->spec     |          |
  //  |            |          |
  //  V            V          V
  // $T        * const        *
  //
  // So now we read 'patDeclarator' from right to left, as that
  // corresponds to the type constructor tree in 'concType()'.

  int patDeclIndex = patDeclarator.length()-1;

  for (; patDeclIndex >= 0 &&
         !concType->isCVAtomicType(); concType = concType->getAtType()) {
    IDeclarator *d = patDeclarator[patDeclIndex--];

    if (d->isD_reference() &&
        concType->isReferenceType()) {
      continue;
    }
    
    // any other condition, just skip past reference-ness
    if (concType->isReferenceType()) {
      concType = concType->getAtType();
    }

    // match 'd' with the topmost type constructor of 'concType'
    if (d->isD_pointer() &&
        concType->isPointerType()) {
      // just assume that they match, ignoring cv-flags ..
      continue;
    }

    if (d->isD_array() &&
        concType->isArrayType()) {
      D_array *da = d->asD_array();

      if (da->size) {
        ArrayType *at = concType->asArrayType();
        if (at->hasSize()) {
          if (!matchExprWithInt(at->size, da->size)) {
            return false;
          }
        }
        else {
          // this might happen if the module declares an extern array
          // with no size... not sure what the best way to handle it is
          xbase("concrete array type without size being compared "
                "against pattern with specified size");
        }
      }

      continue;
    }

    // TODO: other kinds of constructors?

    // no match
    return false;
  }

  if (patDeclIndex >= 0) {
    // pattern has extra unconsumed constructors, so no match
    return false;
  }

  // extract the PQName from the pattern type
  PQName *patName = NULL;
  if (pat->spec->isTS_name()) {
    patName = pat->spec->asTS_name()->name;
  }
  else if (pat->spec->isTS_elaborated()) {
    TS_elaborated *s = pat->spec->asTS_elaborated();
    patName = s->name;
    
    // the point of using an elaborated pattern type is to constrain
    // the kind of type 'concType' may be
    if (s->keyword == TI_ENUM) {
      if (!concType->isEnumType()) {
        return false;
      }
    }
    else {
      if (!concType->isCompoundTypeOf((CompoundType::Keyword)s->keyword)) {
        return false;
      }
    }
  }
  else {
    xunimp("unexpected kind of pattern type");
    return false;
  }

  // is the pattern specifier a pattern type variable?
  if (patName->isPQ_patvar()) {
    // yes, bind it to 'concType'
    return matchBind(patName->asPQ_patvar()->name,
      new C2VValue(VK_C_TYPE, concType));
  }
  else {
    // does the name match?
    return getTypeName(concType) == patName->getName();
  }
}


bool C2VML::matchField(Variable *conc, PQName const *pat)
{
  // ignore qualifiers..

  if (pat->isPQ_patvar()) {
    return matchBind(pat->asPQ_patvarC()->name,
      new C2VValue(VK_C_FIELD, conc));
  }
  else {
    return conc->name == pat->getName();
  }
}


bool C2VML::matchBind(StringRef patvname, C2VValue * owner value)
{
  C2VValue *existing = lookup(patvname);
  if (!existing) {
    bind(patvname, value);
    return true;
  }
  
  // since the name is already bound, require that 'value' be 
  // equal to the name's current binding
  bool ret = equalValues(*existing, *value);
  delete value;
  return ret;
}

bool C2VML::equalValues(C2VValue const &v1, C2VValue const &v2)
{
  // normalize: make v1.kind always the smaller if they are different
  if (v1.kind > v2.kind) {
    return equalValues(v2, v1);
  }

  // For now, I just have some ad-hoc pairs of different types that
  // can be equal.  A better scheme would be to have explicit
  // coercions between the types, and a type checker than could insert
  // implicit coercions to cover these ad-hoc pairings.
  if (v1.kind == VK_STRING && v2.kind == VK_C_VARIABLE) {
    return 0==strcmp(v2.getCVariable()->name, 
                     parseQuotedString(v1.getString()));
  }

  if (v1.kind != v2.kind) {
    return false;
  }
  
  switch (v1.kind) {
    case VK_INT:
      return v1.getInt() == v2.getInt();

    case VK_IDENTIFIER:
      return v1.getIdentifier() == v2.getIdentifier();

    case VK_STRING:
      return v1.getString() == v2.getString();

    case VK_C_VARIABLE:
      return v1.getCVariable() == v2.getCVariable();

    case VK_C_FIELD:
      return v1.getCField() == v2.getCField();

    case VK_C_EXPRESSION:
      // both are concrete
      return matchExpr(v1.getCExpr(), v2.getCExpr());

    case VK_C_TYPE:
      return v1.getCType()->equals(v2.getCType());

    case VK_C_STATEMENT:
    //case VK_VML_VARIABLE:
    case VK_VML_EXPRESSION:
    case VK_VML_PREDICATE:
    case VK_VML_STATEMENT:
    default:
      xunimp("todo");
      return false;
  }
}


// EOF
