// c2vml.h
// translation environment for c2vml.ast

#ifndef C2VML_H
#define C2VML_H

#include "c2vvalue.h"     // C2VValue
#include "array.h"        // ObjArrayStack
#include "vml.ast.gen.h"  // VML AST components
#include "ownerserf.h"    // owner
#include "objmap.h"       // ObjMap
#include "strobjdict.h"   // StringObjDict

class StringTable;        // strtable.h
class Variable;           // variable.h
class ECompoundType;      // exttype.h
class EVariable;          // exttype.h
class EFunctionType;      // exttype.h
class Function;           // cc.ast
class ArgExpression;      // cc.ast
class PQName;             // cc.ast
class Expression;         // cc.ast
class Condition;          // cc.ast
class C2VFile;            // c2v.ast
class CCLang;             // cc_lang.h

// fwd in this file
class C2VML;

// fwd for struct declared in c2vml.cc
struct SwitchInfo;        // c2vml.cc


// map from StringRefs to C2VValues, owning the values
typedef ObjMap<char const, C2VValue> C2VValueMap;


// set of C2V bindings
class C2VScope {
public:      // data
  C2VValueMap map;
  
public:      // funcs
  C2VScope();
  ~C2VScope();
};


// push a scope in ctor, pop it in dtor
class C2VScopePushPop {
public:      // data
  C2VML &env;                   // the environment
  C2VScope *scope;              // (serf) the scope

public:
  C2VScopePushPop(C2VML &env);
  ~C2VScopePushPop();
};


// main c2vml context
class C2VML {
public:      // data
  // reference to 'this' to allow code to move into and out of
  // this class w/o lots of simple rewriting
  C2VML &env;

  // string table for making identifiers
  StringTable &str;

  // stack of scopes; first element of list is innermost element
  ObjList<C2VScope> scopes;

  // pattern-based translation specification
  C2VFile *c2vfile;             // (owner)

  // when nonzero, we're in the middle of translating the body
  // of an annotation predicate; for example, in such situations,
  // we do not generate proof obligations for memory accesses
  int inPredicate;

  // counter for temporary names, one counter per base name
  StringObjDict<int> tempNameCounters;

  // counter for function addresses
  int functionAddressCounter;

  // counter for structure type tags
  int typeTagCt;

  // when true, we automatically add invariants for pointers
  bool automaticTypingAxioms;

  // when true, automatic typing axioms refer to 'objct'
  bool modelUsesObjct;

  // VML file we're building; this is initialized to an empty
  // file in the ctor, and deleted by the dtor unless NULL; the
  // client is expected to remove it and set this pointer to
  // NULL when it's done translating
  VMLFile *vmlfile;             // (owner)

  // function being translated, or NULL if none
  Function *currentFunc;        // (serf)

  // stack of compound statements being built
  ObjArrayStack<VS_compound> stmtStack;

  // stack of break/continue target labels; topmost is the
  // current target of the respective C statement
  ArrayStack<StringRef> breakLabelStack;
  ArrayStack<StringRef> continueLabelStack;

  // stack of open switch statements; topmost is active
  ArrayStack<SwitchInfo*> switchStack;

  // some predefined StringRefs
  StringRef str_malloc;
  StringRef str_memset;
  StringRef str_memmove;
  StringRef str_memcpy;
  StringRef str_free;
  StringRef str_kfree;
  StringRef str_metaVar;
  StringRef str_fresh_value;
  StringRef str_CLT_between;
  StringRef str_declare_partial_order;

public:      // funcs
  C2VML(StringTable &str, CCLang &lang);
  ~C2VML();

  // bracket function translations
  void beginFunc(Function *f);
  VS_compound *endFunc();

  // true if we're at toplevel, outside any function
  bool atTopLevel() const { return stmtStack.isEmpty(); }

  // -------- add components to in-progress VML fragments -------
  // add a toplevel form
  void addTopForm(VTopForm * owner vtf);

  // add a statement to the currently open (topmost) VS_compound
  void addStmt(VStatement * owner stmt);

  // like 'addStmt', except the contents of the compound are directly
  // added, and the argument compound itself deallocated
  void foldStmt(VS_compound * owner compound);

  // add a declaration to either the current statement list, or
  // else toplevel, depending on where we are
  void addDecl(VDeclaration *decl);

  // manipulate the stack of compounds
  void pushCompound(SourceLoc loc);
  VS_compound * owner popCompound();

  // add a proof obligation statement; uses loc()
  void prove(rostring why, Pred * owner pred);

  // add a declaration statement mapping a name to a value
  void define(rostring varName, VExpr * owner initVal);

  // add a declaration of a variable without an initial value
  void declare(rostring varName);

  // add an assignment statement
  void assign(rostring targetVar, VExpr * owner value);

  // similar to 'assign', but also set 'meta' to 'live'
  void assignLive(Variable *var, VExpr * owner value);

  // varName := varName{index := value};
  void assignUpd(rostring varName, VExpr * owner index, VExpr * owner val);

  // add a comment (does nothing yet)
  void comment(rostring text);

  // -------- compute parts of the VML translation -------
  // make a new name for a temporary; the temporary will include
  // the string 'basename', and subsequent temporaries with the
  // same basename will differ only incrementing a portion of the
  // name (so they are not affected by temporaries with other
  // basenames)
  StringRef makeTempVarName(rostring basename, rostring why);

  // get 'basename' from the first word in 'why'
  StringRef makeTempVarName(rostring why);

  // get 'basename' from 'e' if it is a variable, otherwise get
  // it from the first word in 'why'
  StringRef makeTempVarName(Expression const *e, rostring why);
  StringRef makeTempVarName(Condition const *e, rostring why);

  // make a new label name
  StringRef makeLabelName(rostring why);

  // id for the return label
  Identifier * owner returnLabelId();

  // enters 'id' into the string table
  Identifier * owner makeId(rostring id);
  VE_var * owner makeVar(rostring id);

  // name of a slot of a field
  StringRef fieldSlotName(Variable const *field) const;

  // get synthesized variable that stands for a given field of
  // a given variable (single object, not a slice)
  EVariable *getSynthField(EVariable *v, Variable *field);

  // surround a given object index with a predicate to say that that
  // object has the given class' tag; this will take into account
  // subtyping, i.e. allow the object to actually be a subclass
  Pred * owner hasClassTag(ECompoundType *ct, VExpr * owner objIndex);
                                                           
  // array[index]
  VExpr * owner sel(VExpr * owner array, VExpr * owner index);

  // array{index := value}  
  VExpr * owner upd(VExpr * owner array, VExpr * owner index, VExpr * owner value);

  // ptr(object, offset)
  VExpr * owner ptr(VExpr * owner object, VExpr * owner offset);

  // target := srcValue, but also returns srcValue
  VExpr * owner assignmentExpr
    (Expression *target, VExpr * owner srcValue);

  // not sure what this did in the old version...
  VExpr * owner rval(VExpr * owner p) { return p; }
  
  // given a function type, create a list of VML parameters
  ASTList<VD_var> *funcTypeToParams(EFunctionType *ft,
                                    ASTList<VD_var> * nullable returns);

  // build a VProcedure, getting most pieces from the type
  VProcedure *buildProcedure(Identifier *id, EFunctionType *ft,
                             VS_compound *vbody);

  // same as e_lit
  VExpr * owner intLit(int i);

  // see comments at implementations
  VStatement * owner buildConditional(SourceLoc loc, Condition *cond,
                                      VStatement * owner sTrue, 
                                      VStatement * owner sFalse);
  VStatement * owner buildConditional(SourceLoc loc, Expression *expr,
                                      VStatement * owner sTrue,
                                      VStatement * owner sFalse);
  VStatement * owner buildGoto(SourceLoc loc, VStatement *target);
  
  // emit a typing invariant fot <ct>::<slot>, which is a pointer
  // to <pointee>
  void emitTypingInvariant(SourceLoc loc, ECompoundType *ct, EVariable *slot,
                           ECompoundType *pointee);
                                        
  // assign a tag number to 'ct' and emit a VML definition of it
  void defineTypeTag(ECompoundType *ct);

  // -------------- C2V interpretation --------------
  // (implementations for these are in c2v_eval.cc)

  // scope manipulation
  C2VScope *pushScope();
  void popScope(C2VScope *scope);     // scope passed for a debugging check
  
  // every binding in the innermost scope gets moved into the
  // next-outermost, overwriting whatever is there
  void promoteInnerScopeValues();
    
  // c2v values
  void bind(StringRef name, C2VValue * owner value);
  C2VValue * /*nullable*/ lookup(StringRef name);
  C2VValue * /*nullable*/ lookupInner(StringRef name);    // innermost scope only
  C2VValue *lookupf(StringRef name);                      // throws error if unbound

  bool variableAcceptsType(StringRef name, C2VValueKind kind);
  void checkVariableType(StringRef name, C2VValueKind kind);

  // ----------------- matching --------------------
  // (implementations in c2v_match.cc)

  // match a concrete AST fragment with a pattern, binding pattern
  // variables in the process and returning 'true' if they ultimately
  // match; note that even in a failed match, pattern variables may be
  // bound (so "speculative" matching must be done by explicitly
  // adding an insulating scope)
  bool matchExpr(Expression const *conc, Expression const *pat);
  bool matchExprWithInt(int conc, Expression const *pat);
  bool matchArgExprs(FakeList<ArgExpression> const *conc,
                     FakeList<ArgExpression> const *pat);
  bool matchType(ASTTypeId const *conc, ASTTypeId const *pat);
  bool matchSemType(Type *concType, ASTTypeId const *pat);
  bool matchField(Variable *conc, PQName const *pat);

  // attempt to bind 'patvname' to 'value'; if it is already bound
  // to something different, return false
  bool matchBind(StringRef patvname, C2VValue * owner value);
  
  // true if values 'v1' and 'v2' are equal
  bool equalValues(C2VValue const &v1, C2VValue const &v2);

  // ---------------- substitution ------------------
  // (implementations in c2v_subst.cc)

  // given a VML AST fragment with pattern variable references, and
  // the current environment of C2V variables, yield a concrete VML
  // AST fragment with the variable references replaced with the
  // current values of the named variables
  VStatement * owner substVStatement(VStatement const *pat);
  VExpr * owner substVExpr(VExpr const *pat);

  // ------------- diagnostics ---------------
  // Report an error that prevents complete translation to VML.
  // This usually corresponds to a violation of the additional
  // static semantic requirements that some model imposes.  This
  // function throws an exception.
  void error(rostring msg);
  void error(SourceLoc loc, rostring msg);

  // returns NULL
  VExpr * owner todo(rostring msg);
};


// bracket predicates
class PredBracket {
private:
  C2VML &env;
  
public:
  PredBracket(C2VML &e)
    : env(e)
  {
    env.inPredicate++;
  }
  
  ~PredBracket()
  {
    env.inPredicate--;
  }
};


#endif // C2VML_H
