// doublylinked.c
// doubly-linked list, modeled after pale/Examples/doublylinked.pale

// --------- verifier preamble ----------
#include "selupd.model"     // sel/upd
#include "state.model"      // metaVar

int fresh_value();

verifier_option((no_objct));


// --------------- code to verify -----------------
struct Node;

typedef struct Head {
  struct Node *first;   // NULL or valid
  struct Node *last;    // NULL or last element in list
} Head;

typedef struct Node {
  int value;
  struct Node *next;    // invariant: NULL or valid
  struct Node *prev;    // thing whose 'next' points at me
  struct Head *head;    // ghost: pointer to my list's header object
} Node;

// invariant: NULL or valid
Node *garbage = NULL;


// this disables the effect of the pattern specs
#if 0
  #define PATS(...) /*nothing*/
  #define MPAT(...) /*nothing*/
#endif // 0


// global invariants
int global_inv() { return vc_expression(all_of {
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // garbage
  garbage!=NULL ==> all_of {
    tag[garbage]==Node_tag;
    garbage->prev==NULL;
    garbage->head==NULL;
  };


  // Head

  // 'first' is valid if not NULL

  forall(Head *h) PATS(MPAT(h->first,
                            TRIGGEREQ(tag[h], Head_tag))) .
        tag[h]==Head_tag && h->first!=NULL ==> tag[h->first] == Node_tag;

  forall(Head *h) PATS(MPAT(h->last, TRIGGEREQ(tag[h], Head_tag))) .
    tag[h]==Head_tag && h->last!=NULL ==> tag[h->last] == Node_tag;

  // declared type of 'next' is accurate
  forall(Node *n) PATS(MPAT(n->next, TRIGGEREQ(tag[n], Node_tag))) .
    tag[n]==Node_tag && n->next!=NULL ==> tag[n->next] == Node_tag;

  // declared type of 'prev' is accurate
  forall(Node *n) PATS(MPAT(n->prev, TRIGGEREQ(tag[n], Node_tag))) .
    tag[n] == Node_tag && n->prev!=NULL ==> tag[n->prev] == Node_tag;


  // list labels are accurate; 'first' is injective
  forall(Head *h) PATS(MPAT(h->first, TRIGGEREQ(tag[h], Head_tag))) .
    tag[h]==Head_tag && h->first!=NULL ==> h->first->head == h;

  // 'first' points to the first element (if not NULL)
  forall(Head *h) PATS(MPAT(h->first->prev, TRIGGEREQ(tag[h], Head_tag))) .
    tag[h]==Head_tag && h->first!=NULL ==> h->first->prev == NULL;

  // similar for 'last'
  
  forall(Head *h) PATS(MPAT(h->last, TRIGGEREQ(tag[h], Head_tag))) .
     tag[h]==Head_tag && h->last!=NULL ==> h->last->head == h;
  
  forall(Head *h) PATS(MPAT(h->last->next, TRIGGEREQ(tag[h], Head_tag))) .
    tag[h]==Head_tag && h->last!=NULL ==> h->last->next == NULL;

  // should have 'first' iff have 'next'; actually, I cannot ensure this
  // because my invariants about 'first' and 'last' are too weak
  //forall(Head *h).
  //  tag[h]==Head_tag ==> (h->first==NULL <==> h->last==NULL);

  // Node


  // transitivity of list label
  forall(Node *n) PATS(MPAT(TRIGGEREQ(tag[n], Node_tag), n->next, n->head),
                       MPAT(TRIGGEREQ(tag[n], Node_tag), n->next->head)) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->head == n->head;

  // injectivity of 'next'
  forall(Node *n) PATS(MPAT(TRIGGEREQ(tag[n], Node_tag), n->next)) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->prev == n;

  // injectivity of 'prev' (seems to be needed...)
  forall(Node *n) PATS(MPAT(TRIGGEREQ(tag[n], Node_tag), n->prev)) .
    tag[n]==Node_tag && n->prev!=NULL ==> n->prev->next == n;
});}


// ----------------------- delete -----------------------
void delete(Head *list, Node *p)
  pre(all_of {
        tag[list]==Head_tag;      // implied by 'data' keyowrd
        tag[p]==Node_tag;         // p!=null
        p->head == list;          // list.first<next*>p
        /* ? */                   // list.last<prev*>p
        /* ? */                   // p<next*>list.last
        /* ? */                   // p<prev*>list.first
        /* ? */                   // list.first<next*>S
      })
  post(all_of {
         p->head == NULL;         // !L.first<next*>P
         forall(Node *q)          // allpos q of Node: q in S <=> (L.first<next*>q | q=P)
           PATS(pre(q->head), q->head).
             q!=p ==> pre(q->head) == q->head;
       })
  changes(Head, Node, garbage)
{
  if (p == list->first) {
    list->first = p->next;
  }
#ifdef BUG
  // leave out the following conditional as a deliberately inserted bug
#else
  // it is correct to have this code here
  if (p == list->last) {
    list->last = p->prev;
  }
#endif
  if (p->prev != NULL) {
    p->prev->next = p->next;
  }
  if (p->next != NULL) {
    p->next->prev = p->prev;
  }

  // This is delete(p)
  if (garbage != NULL) {
    garbage->prev = p;
  }
  p->next = garbage;
  p->prev = NULL;
  p->head = NULL;
  garbage = p;
}


Node *search(Head *list, int value)
  pre(tag[list] == Head_tag)      // list!=null
  post(all_of {
    /* redundant! */              // L!=null
    result!=NULL ==> result->value == pre(value);     // approximates pale's postcondition...
  })
  changes()
{
  Node *t = list->first;
  while (t!=NULL && t->value != value) {

    invariant(all_of {
      tag[list] == Head_tag;      // list!=null
      pre(value) == value;        // V=value
      pre(list) == list;          // L=list
      t->head == list;            // list.first<next*>t
      // cannot state: allpos p of Node: p<next.next*>t => p.value!=value]
      
      // additional
      tag[t] == Node_tag;
    });

    t = t->next;
  }
  return t;
}


void insert(Head *list, Node *node)
  pre(all_of {
    tag[list] == Head_tag;          // list!=null 
    tag[node] == Node_tag;          // node!=null
    node->next == NULL;             // node.next=null
    node->prev == NULL;             // node.prev=null
    
    // additional
    node->head == NULL;
    node != garbage;
  })
  post(1==1)
  changes(Head, Node)
{
  if (list->first == NULL) {
    list->last = node;
  }
  else {
    list->first->prev = node;
  }
  node->head = list;
  node->next = list->first;
  list->first = node;
  node = NULL;
}


Head *concat(Head *p, Head *q)
  pre(all_of {
    tag[p] == Head_tag;       // p!=null
    tag[q] == Head_tag;       // q!=null
  })
  post(1==1)
  changes(Head, Node)
{
  if (p->last != NULL) {
    p->last->next = q->first;
  }
  else {
    p->first = q->first;
  }
  if (q->first != NULL) {
    q->first->prev = p->last;
    p->last = q->last;
  }
  q->first = NULL;
  q->last = NULL;
  
  // update all of 'head' in one shot (it is ghost)
  Node_head = fresh_value();
  assume(forall(Node *n) PATS(pre(n->head), n->head).
    (pre(n->head)!=q && n->head == pre(n->head)) ||
    (pre(n->head)==q && n->head == p)
  );
  
  return p;
}






// EOF
