// scheduler2.c
// revised example from PLDI submission

// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar


// --------------- code to verify -----------------
typedef struct Process Process;

enum State { SLP, RUN };

struct Process {
  int prio;
  State state;
  Process *next;
  Process *prev;
};

Process *runnable;
Process *sleeping;


// global invariants
int global_inv() { return vc_expression(all_of {
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // declared type of 'next' is accurate
  forall(Process *p) PATS(p->next).
    (tag[p]==Process_tag && p->next!=NULL) ==>
    (tag[p]==Process_tag && p->next!=NULL && tag[p->next] == Process_tag);

  // declared type of 'prev' is accurate
  forall(Process *p) PATS(p->prev).
    tag[p]==Process_tag ==>
    (tag[p]==Process_tag && p->prev==NULL) ||
    (tag[p]==Process_tag && p->prev!=NULL && tag[p->prev] == Process_tag);

  // A1: injectivity of 'next'
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag ==>
    (tag[p]==Process_tag && p->next==NULL) ||
    (tag[p]==Process_tag && p->next!=NULL && p->next->prev == p);

  // A2: reverse injectivity
  forall(Process *p) PATS(p->prev).
    tag[p]==Process_tag ==>
    (tag[p]==Process_tag && p->prev==NULL) ||
    (tag[p]==Process_tag && p->prev!=NULL && p->prev->next == p);

  // A3: state is transitive
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag ==>
    (tag[p]==Process_tag && p->next==NULL) ||
    (tag[p]==Process_tag && p->next!=NULL && p->state == p->next->state);

  // A4: runnable list is sorted
  forall(Process *p) PATS(p->next).
    tag[p]==Process_tag ==>
    (tag[p]==Process_tag && p->next==NULL) ||
    (tag[p]==Process_tag && p->next!=NULL && p->state!=RUN) ||
    (tag[p]==Process_tag && p->next!=NULL && p->state==RUN && p->prio >= p->next->prio);

  // A5: runnable dummy node priority is highest
  // (not needed if no nonrecursive calls)
  //forall(Process *p).
  //  tag[p]==Process_tag ==> runnable->prio >= p->prio;

  // list headers are never null
  tag[runnable]==Process_tag;
  tag[sleeping]==Process_tag;

  // list headers have proper state
  runnable->state == RUN;
  sleeping->state == SLP;

  // they point to first elements
  runnable->prev == NULL;
  sleeping->prev == NULL;
});}


// precondition: list.prio >= x.prio
// precondition: list is runnable (RUN)
// precondition: x is sleeping (SLP)
void insert(Process *list, Process *x)
  pre(all_of {
    tag[list]==Process_tag;
    tag[x]==Process_tag;
    list->state == RUN;
    x->state == SLP;         // 'x' is not already in the list
    list->prio >= x->prio;
    x->next == NULL;
    x->prev == NULL;
    x != sleeping;
  })
  changes(Process_next, Process_prev, Process_state)
{
  if (list->next != NULL && list->next->prio >= x->prio) {
    insert(list->next, x);
  }
  else {
    //            x
    //            |
    //            V
    //    list   +--+
    //     |     |  |
    //     V     +--+
    //    +--+         +--+
    //    |  |-------->|  |-->..
    //    +--+         +--+
    x->prev = list;
    x->next = list->next;
    if (list->next) {
      list->next->prev = x;
    }
    list->next = x;
    x->state = RUN;
  }
}


void remove(Process *x)
  pre(all_of {
    tag[x]==Process_tag;
    x->prev != NULL;
    x->state == RUN;
    x != runnable;
  })
  changes(Process_next, Process_prev, Process_state)
{
  // remove from runnable list
  x->prev->next = x->next;
  if (x->next) {
    x->next->prev = x->prev;
  }
  // set the state
  x->state = SLP;
  x->next = NULL;
  x->prev = NULL;
}


// EOF
