// exttype.cc
// code for exttype.h

#include "exttype.h"       // this module
#include "cc.ast.gen.h"    // C++ AST
#include "vc_env.h"        // VCEnv
#include "astbuild.h"      // e_and, etc.
#include "trace.h"         // trace
#include "smregexp.h"      // Regexp, regexpMatch
#include "cc_flags.h"      // bitmapString


// --------------------- ECompoundType ------------------
ECompoundType::ECompoundType(Keyword k, StringRef name)
  : CompoundType(k, name),
    sizeVar(NULL),
    tagVar(NULL),
    fieldSlots(),
    typeTag(-1),
    objectSize(0),
    sb(),
    flatMap(false),
    superclass(NULL),
    subclasses()
{}


ECompoundType::~ECompoundType()
{
  if (sizeVar) {
    delete sizeVar;
  }
  if (tagVar) {
    delete tagVar;
  }
}


void ECompoundType::setSuperclass(ECompoundType *super)
{                          
  xassert(superclass == NULL);     // otherwise I need to remove from other list first
  superclass = super;
  super->subclasses.append(this);
}


// --------------------- EFunctionType ------------------
EFunctionType::EFunctionType(Type *retType)
  : FunctionType(retType),
    precondition(NULL),
    postcondition(NULL),
    changes(NULL),
    changesConj(NULL),
    result(NULL)
{}

EFunctionType::~EFunctionType()
{
  if (precondition) {
    delete precondition;
  }
  if (postcondition) {
    delete postcondition;
  }
  if (result) {
    delete result;
  }
}


string stringizeE_variableList(ASTList<E_variable> &list)
{
  stringBuilder sb;
  int ct=0;
  FOREACH_ASTLIST(E_variable, list, iter) {
    if (ct++ > 0) {
      sb << ", ";
    }
    sb << iter.data()->name->getName();
  }
  return sb;
}

void EFunctionType::extraRightmostSyntax(stringBuilder &sb) const
{
  if (precondition) {
    sb << " pre(" << precondition->asString() << ")";
  }
  if (postcondition) {
    sb << " post(" << postcondition->asString() << ")";
  }
  if (changes) {
    sb << " changes(" << stringizeE_variableList(changes->vars) << ")";
  }
}


// moved EFunctionType::getPostcondition to vcgen.cc


// --------------------- EVariable ----------------------
EVariable::EVariable(SourceLoc L, StringRef n, Type *t, DeclFlags f)
  : Variable(L, n, t, f),
    attr(VA_NONE),
    changeGroup(NULL),
    synthFields(),
    originalStructField(NULL),
    modelVar(NULL),
    modelVarOffset(0)
{}

EVariable::~EVariable()
{}


char const * const varAttrFlagNames[] = {
  "parambyref",
  "argbyref",
  "addrtaken",
  "funcsym",
  "trigger",
  "array",
  "predsym",
  "pats_mpat",
  "noTypingAxiom"
};

string toString(VarAttr attrs)
{
  ASSERT_TABLESIZE(varAttrFlagNames, NUM_VATTRFLAGS);
  return bitmapString(attrs, varAttrFlagNames, NUM_VATTRFLAGS);
}


VarAttr stringToVarAttr(char const *str)
{
  for (int i=0; i<NUM_VATTRFLAGS; i++) {
    if (0==strcmp(str, varAttrFlagNames[i])) {
      return (VarAttr)(1 << i);
    }
  }
  return VA_NONE;         // not a known attribute
}


bool EVariable::attachAttribute(StringRef name)
{
  VarAttr a = stringToVarAttr(name);
  if (a) {
    attr |= a;
    return true;
  }
  else {
    return false;
  }
}


string EVariable::namePrintSuffix() const
{
  if (!attr) {
    return "";
  }

  stringBuilder sb;
  int ct=0;
  for (int i=0; i<NUM_VATTRFLAGS; i++) {
    if (attr & (1 << i)) {
      if (ct++ == 0) {
        sb << " ";
      }
      sb << "__varattribute__((" << varAttrFlagNames[i] << ")) ";
    }
  }
  
  return sb;
}


// --------------------- ETypeFactory -------------------
CompoundType *ETypeFactory::makeCompoundType
  (CompoundType::Keyword keyword, StringRef name)
{
  return new ECompoundType(keyword, name);
}

FunctionType *ETypeFactory::makeFunctionType
  (Type *retType)
{
  return new EFunctionType(retType);
}

Variable *ETypeFactory::makeVariable
  (SourceLoc L, StringRef n, Type *t, DeclFlags f, TranslationUnit *)
{
  return new EVariable(L, n, t, f);
}


// EOF
