// edag.cc
// code for edag.h

#include "edag.h"      // this module
#include "exc.h"       // xbase


// ------------------ SymbolTable ---------------------
SymbolTable::SymbolEntry::SymbolEntry(char const *n, int i, int a)
  : name(n),
    index((Symbol)i),
    arity(a)
{}

SymbolTable::SymbolEntry::~SymbolEntry()
{}


SymbolTable::SymbolTable()
  : symToIndex(&SymbolTable::getTableKey),
    indexToSym()
{
  // ensure the 0th symbol is always ""
  Symbol zero = get("", 0);
  xassert(zero == 0);
}

SymbolTable::~SymbolTable()
{}


STATICDEF char const* SymbolTable::getTableKey(void *data)
{
  SymbolEntry *entry = (SymbolEntry*)data;
  return entry->name;
}


Symbol SymbolTable::get(char const *sym, int arity)
{
  SymbolEntry *entry = (SymbolEntry*)symToIndex.get(sym);

  if (entry) {
    if (entry->arity != arity) {
      xbase(stringc << "mismatched arity for " << sym
                    << ": " << entry->arity
                    << " vs " << arity);
    }
  }
  else {
    entry = new SymbolEntry(sym, indexToSym.length(), arity);
    symToIndex.add(sym, entry);
    indexToSym.push(entry);
  }

  return entry->index;
}


char const *SymbolTable::name(Symbol s) const
{
  return indexToSym[s]->name;
}

int SymbolTable::arity(Symbol s) const
{
  return indexToSym[s]->arity;
}


// ---------------------- Term ------------------------
Term::Term(Symbol s)
  : symbol(s),
    args(),
    eqclass(NULL)
{}

Term::~Term()
{}


Term const *Term::reprC() const
{
  Term const *t = this;
  while (t->eqclass) {
    t = t->eqclass;
  }
  return t;
}


bool Term::equals(Term const *t) const
{
  return reprC() == t->reprC();
}


// --------------------- EDAG ------------------------
EDAG::EDAG(SymbolTable &table)
  : nodes(),
    symTable(table)
{}

EDAG::~EDAG()
{}


ObjList<Term> *EDAG::getNodeList(Symbol s)
{
  while (s >= nodes.length()) {
    nodes.push(new ObjList<Term>);     // make new lists
  }

  return nodes[s];
}


Term *EDAG::find(Symbol sym, SObjList<Term> const &args)
{
  xassert(symTable.arity(sym) == args.count());

  ObjList<Term> *list = getNodeList(sym);

  // look for a node that represents this
  FOREACH_OBJLIST_NC(Term, *list, nodeIter) {
    Term *n = nodeIter.data();

    // check for equality of arguments
    SObjListIter<Term> iter1(n->args);
    SObjListIter<Term> iter2(args);
    for (; !iter1.isDone(); iter1.adv(), iter2.adv()) {
      Term const *a1 = iter1.data();
      Term const *a2 = iter2.data();

      if (!a1->equals(a2)) {
        goto next_iteration;
      }
    }

    // found it!
    return n;

    next_iteration:;
  }
  
  return NULL;     // not found
}


Term *EDAG::get(Symbol sym, SObjList<Term> const &args)
{
  // already exists?
  Term *ret = find(sym, args);
  if (ret) {
    return ret;
  }

  ObjList<Term> *list = getNodeList(sym);

  // must make it anew
  ret = new Term(sym);
  SFOREACH_OBJLIST(Term, args, iter) {
    // unfortunate.. the constness is meant to apply to the list
    // structure, not the data elements..
    Term *a = const_cast<Term*>(iter.data());

    ret->args.append(a);
  }

  list->prepend(ret);

  return ret;
}


Term *EDAG::find0(Symbol sym)
{
  SObjList<Term> args;
  return find(sym, args);
}

Term *EDAG::find1(Symbol sym, Term *arg1)
{
  SObjList<Term> args;
  args.append(arg1);
  return find(sym, args);
}


Term *EDAG::get0(Symbol sym)
{
  SObjList<Term> args;
  return get(sym, args);
}

Term *EDAG::get1(Symbol sym, Term *arg1)
{
  SObjList<Term> args;
  args.append(arg1);
  return get(sym, args);
}


void EDAG::unify(Term *t1, Term *t2)
{
  Term *r1 = t1->repr();
  Term *r2 = t2->repr();

  // already the same?
  if (r1 == r2) {
    return;
  }

  // merge classes
  r1->eqclass = r2;

  // taking hint from Simplify, don't compress paths since I
  // may want this stuff to be undo-able
}


void EDAG::congruenceClosure()
{                         
  // iterate until fixpoint
  bool changes=1;
  while (changes) {
    changes=0;

    // consider all pairs of nodes with the same symbol
    for (int symIndex=0; symIndex < nodes.length(); symIndex++) {
      Symbol sym = (Symbol)symIndex;
      ObjList<Term> *list = getNodeList(sym);

      // nested loops
      for (ObjListIterNC<Term> iter1(*list); !iter1.isDone(); iter1.adv()) {
        Term *t1 = iter1.data();

        ObjListIterNC<Term> iter2(iter1);
        iter2.adv();       // don't compare an element to itself
        for (; !iter2.isDone(); iter2.adv()) {
          Term *t2 = iter2.data();

          if (t1->equals(t2)) {
            continue;      // already equal
          }

          // all arguments equal?
          SObjListIterNC<Term> argIter1(t1->args);
          SObjListIterNC<Term> argIter2(t2->args);
          for (; !argIter1.isDone(); argIter1.adv(), argIter2.adv()) {
            if (argIter1.data()->equals(argIter2.data())) {
              // argument is equal
            }
            else {
              // arguments not equal, congruence doesn't apply
              goto next_iteration;
            }
          }

          // equal arguments and symbol!
          unify(t1, t2);
          
          // bizarre gcc (2.95.3) bug: emits a warning if I say "changes++"!
          ++changes;

          next_iteration:;
        }
      }
    }
  }
}


void EDAG::getNodeList(ArrayStack<Term*> &dest)
{
  // iterate over function symbols
  int numSyms = symTable.numSymbols();
  for (int sym=0; sym<numSyms; sym++) {
    ObjList<Term> *list = getNodeList((Symbol)sym);
    
    // put all nodes on this list into 'dest'; that means we
    // will put all elements of each equivalence class, instead
    // of just one representative, but oh well
    FOREACH_OBJLIST_NC(Term, *list, iter) {
      dest.push(iter.data());
    }
  }
}


string EDAG::dotNodeName(Term const *t) const
{
  // name + address to make it unique
  return stringf("\"%s (0x%X)\"", symTable.name(t->symbol), (int)t);
}

string EDAG::toDOT(char const *graphName)
{
  stringBuilder sb;

  sb << "digraph \"" << graphName << "\" {\n";

  // loop over symbols
  for (int symIndex=0; symIndex < nodes.length(); symIndex++) {
    Symbol sym = (Symbol)symIndex;
    ObjList<Term> *list = getNodeList(sym);

    // loop over nodes
    FOREACH_OBJLIST(Term, *list, iter) {
      Term const *t = iter.data();

      // node itself
      sb << "  " << dotNodeName(t) << " [\n"
         << "    label = \"" << symTable.name(t->symbol) << "\"\n"
         << "  ]\n"
         ;

      // argument edges
      SFOREACH_OBJLIST(Term, t->args, argIter) {
        sb << "  " << dotNodeName(t) << " -> " << dotNodeName(argIter.data()) << " [\n"
           << "  ]\n"
           ;
      }

      // equivalence edge
      if (t->eqclass) {
        sb << "  " << dotNodeName(t) << " -> " << dotNodeName(t->eqclass) << " [\n"
           << "    style = dashed\n"
           << "  ]\n"
           ;
      }
    }
  }
  
  sb << "}\n";
  
  return sb;
}


// ------------------------- test code -------------------------
#ifdef TEST_EDAG

#include "match.h"     // MatchRule, etc.
#include "test.h"      // USUAL_MAIN

void test1()
{
  cout << "--------- test1 --------\n";

  SymbolTable table;
  EDAG edag(table);

  Symbol f = table.get("f", 1);
  Symbol x = table.get("x", 0);

  // f(f(f(x))) = x
  Term *tx = edag.get0(x);
  Term *f3x = edag.get1(f, edag.get1(f, edag.get1(f, tx)));
  edag.unify(tx, f3x);
  edag.congruenceClosure();

  // f(f(f(f(f(x))))) = x
  Term *f5x = edag.get1(f, edag.get1(f, edag.get1(f, edag.get1(f, edag.get1(f, tx)))));
  edag.unify(tx, f5x);
  edag.congruenceClosure();

  // f(x) = x ?
  Term *fx = edag.get1(f, tx);
  if (fx->equals(tx)) {
    cout << "f(x) = x  discovered\n";
  }
  else {
    cout << "f(x) = x  NOT discovered!\n";
  }

  //cout << edag.toDOT("f5x = f3x = x");
}


void test2()
{
  cout << "--------- test2 --------\n";

  SymbolTable table;
  EDAG edag(table);

  Symbol a = table.get("a", 1);
  Symbol b = table.get("b", 1);     // inverse of a
  Symbol x = table.get("x", 0);

  MatchRuleSet rules;

  //      p->a->b = p
  //      +--+
  MatchRule *rule = new MatchRule;
  rule->match[0].push(a);
  rule->extend[0].push(b);
  rules.add(rule);

  // x->a
  Term *tx = edag.get0(x);
  Term *txa = edag.get1(a, tx);
  rules.complete(edag);

  // x->a->b == x ?
  Term *txab = edag.get1(b, txa);
  if (txab->equals(tx)) {
    cout << "x->a->b = x  discovered\n";
  }
  else {
    cout << "x->a->b = x  NOT discovered!\n";
  }
}


void entry()
{
  test1();
  test2();
}

USUAL_MAIN

#endif // TEST_EDAG
