// prv_int.cc
// code for prv_int.h

#include "prv_int.h"   // this module
#include "str.h"       // stringc
#include "mypopen.h"   // popen_execvp
#include "unixutil.h"  // readString, writeAll
#include "trace.h"     // trace
#include "xassert.h"   // xassert
#include "exc.h"       // xbase
#include "strutil.h"   // replace
#include "syserr.h"    // xsyserror
#include "autofile.h"  // AutoFILE

#include <stdio.h>     // perror, FILE stuff
#include <string.h>    // strstr
#include <sys/types.h> // open
#include <sys/stat.h>  // open
#include <fcntl.h>     // open
#include <unistd.h>    // close



ProverInterface::ProverInterface(rostring progName)
  : toProver(-1), fromProver(-1),
    predicateLog(-1)
{
  char const *argv[2];
  argv[0] = toCStr(progName);
  argv[1]= NULL;

  // open prover as a child process, with pipes to communicate;
  // this calls exit(2) if it fails...
  popen_execvp(&toProver, &fromProver, &fromProver /*stderr==stdout*/,
               argv[0], argv);

  // derived class must call 'analyzeInit' next
}

// this is separated out from the superclass constructor because it
// invokes 'analyzeInitialPrompt', which is virtual, so if it were
// in the superclass ctor it would call the superclass version
void ProverInterface::analyzeInit()
{
  // read initial prompt
  stringBuilder response;
  ReadAction act = RA_CONTINUE;
  while (act == RA_CONTINUE) {
    if (!collectMoreInput(response)) {
      xfailure("prover closed pipe right at start");
    }
    act = analyzeInitialPrompt(response);      // virtual function call
  }
  
  if (act != RA_PROVED) {
    xbase(stringc << "unexpected initial response from prover:\n"
                  << response << "\n");
  }
}

ProverInterface::~ProverInterface()
{
  close(toProver);
  close(fromProver);

  if (predicateLog > -1) {
    close(predicateLog);
  }
}


void ProverInterface::openLog(rostring fname)
{
  xassert(predicateLog == -1);    // otherwise already open!

  predicateLog = creat(toCStr(fname), 0777);
  if (predicateLog < 0) {
    xsyserror("creat", fname);
  }

  string header = getLogfileHeader(fname);
  writeAll(predicateLog, toCStr(header), header.length());
}


void ProverInterface::feedFile(rostring fname)
{
  // open file
  AutoFILE bgpred(toCStr(fname), "r");

  // read the file and feed it to the prover
  enum { BUFSIZE = 4096 };
  char buf[BUFSIZE];
  int len;
  while ((len = fread(buf, 1, BUFSIZE, bgpred)) > 0) {
    if (!writeAll(toProver, buf, len)) {
      xsyserror("write", "to prover");
    }
  }
}


void ProverInterface::feed(rostring str)
{
  if (!writeAll(toProver, toCStr(str), strlen(str))) {
    xsyserror("write", "feed to prover");
  }

  if (predicateLog > -1) {
    // log the string
    writeAll(predicateLog, toCStr(str), strlen(str));
  }

  // make sure prover hasn't replied
  if (canRead(fromProver)) {
    fprintf(stderr, "unexpected reply from prover:\n");

    // what did it say?
    char *buf = new char[1024];
    if (!readString(fromProver, buf, 1024)) {
      xsyserror("read", "from prover");
    }
    fprintf(stderr, "%s\n", buf);
    delete[] buf;

    exit(2);     // obviates delete[], but whatever
  }
}


bool ProverInterface::prove(rostring str, string *counterexample)
{
  trace("prover") << str << endl;

  // hand the current predicate to prover
  if (!writeAll(toProver, toCStr(str), strlen(str))) {
    perror("write to prover");
    exit(2);
  }

  if (predicateLog > -1) {
    // log predicate; don't care about failure as much
    writeAll(predicateLog, toCStr(str), strlen(str));
  }

  // read prover's response
  stringBuilder response;
  ReadAction act = RA_CONTINUE;
  while (act == RA_CONTINUE) {
    // read next segment of response
    if (!collectMoreInput(response)) {
      xbase(stringc << "Unexpected response+EOF from prover:\n  "
                    << replace(response, "\n", "\n  ") << "\n");
    }
    
    // analyze it
    act = analyzeGoalResponse(response);
  }

  if (act == RA_ABORT) {
    fprintf(stderr, "\nabort; prover says: %s\n", response.c_str());
    exit(2);
  }

  afterProofCheck(response);

  if (act == RA_PROVED) {
    return true;           // formula is proven
  }

  else {
    if (counterexample) {
      // look for counterexample
      *counterexample = extractCounterexample(response);
    }

    return false;          // formula is not proven
  }
}


bool ProverInterface::collectMoreInput(stringBuilder &response)
{
  char buf[512];
  int count = read(fromProver, buf, sizeof(buf)-1);
  if (count < 0) {
    xsyserror("read", "from prover");
  }
  buf[count]=0;

  if (count==0) {
    if (!response.length()) {
      xfailure("prover closed the pipe");
    }
    return false;
  }
  response << buf;
  
  return true;
}


string ProverInterface::getLogfileHeader(rostring )
{
  return "";
}


void ProverInterface::afterProofCheck(rostring response)
{}


string ProverInterface::extractCounterexample(rostring response)
{ 
  return "";
}


// EOF
