; reach.thy
; reachability predicates and axioms
; requires: memory.thy

; for now, not parameterized by the access path; and I'm assuming
; memories are flat

; (canReach mem p q): from p, q is reachable, in memory state 'mem'
; equivalent to:
;   exists i >= 0, such that mem^i(p) = q
(DEFPRED (canReach mem p q))

; (canReachWO): from p, q is reachable w/o using mem(r)
; equivalent to:
;   exists i >= 0, such that mem^i(p) = q,
;                  and forall j < i, mem^j(p) != r
(DEFPRED (canReachWO mem p q r))

(BG_PUSH (AND

  ; ------------------ axioms for static memories --------------------

  ; canReach is a special case of canReachAvoiding
  (FORALL (mem p q)   (PATS (canReach mem p q))
    (IFF
      (canReach mem p q)
      (canReachWO mem p q q)
    ))

  ; introduction
  (FORALL (mem p q x)
    (IMPLIES
      (AND
        (EQ (sel mem p) q)               ; p->next == q
        (NEQ p x)                        ; p != x
      )
      (canReachWO mem p q x)             ; p ~> q /x
    ))

  ; reflexive
  (FORALL (mem p x)
    (canReachWO mem p p x)
  )

  ; transitive
  (FORALL (mem p q r x)
    (IMPLIES
      (AND
        (canReach mem p q)               ; p ~> q
        (canReach mem q r)               ; q ~> r
      )
      (canReach mem p r)                 ; p ~> r
    ))

  ; transitive, alternate form (don't see how to merge this with above)
  (FORALL (mem p q r x)
    (IMPLIES
      (AND
        (canReachWO mem p q x)           ; p ~> q /x
        (canReachWO mem q r x)           ; q ~> r /x
      )
      (canReachWO mem p r x)             ; p ~> r /x  
    ))

  ; a node with no path to the destination can always be avoided
  (FORALL (mem p q x h)
    (IMPLIES
      (AND
        (NOT (canReach mem q p))         ; q ~!> p
        (canReachWO mem h p x)           ; h ~> p /x
      )
      (canReachWO mem h p q)             ; h ~> p /q
    ))

  ; a node that leads nowhere cannot reach anything
  (FORALL (mem p q)
    (IMPLIES
      (AND
        (EQ (sel mem q) null)            ; q->next == null
        (NEQ p q)                        ; p != q
        (NEQ p null)                     ; p != null
      )
      (NOT (canReach mem q p))           ; q ~!> p
    ))

  ; if p reaches r, but p cannot reach h, then r cannot reach h
  (FORALL (mem p r h)
    (IMPLIES
      (AND
        (canReach mem p r)               ; p ~> r
        (NOT (canReach mem p h))         ; p ~!> h
      )
      (NOT (canReach mem r h))           ; r ~!> h
    ))

  ; ------------------ axioms for changing memories --------------------

  ; changing where the tail points doesn't affect the chain
  (FORALL (mem h p x q)
    (IMPLIES
      (canReachWO mem h p x)             ; mem |- h ~> p /x
      (canReachWO (upd mem p q) h p x)   ; mem{p:=q} |- h ~> p /x
    ))

  ; adding a link to a previously reachable node does not make
  ; new things reachable
  (FORALL (mem p h r q)
    (IMPLIES
      (AND
        (NOT (canReach mem p h))         ; mem |- p ~!> h
        (canReach mem p r)               ; mem |- p ~> r
      )
      (NOT (canReach (upd mem q r) p h)) ; mem{q:=r} |- p ~!> h
    ))
    
  ; setting a link to null does not make new things reachable
  (FORALL (mem p h q)
    (IMPLIES
      (NOT (canReach mem p h))              ; mem |- p ~!> h
      (NOT (canReach (upd mem q null) p h)) ; mem{q:=null} |- p ~!> h
    ))

  ; modifying an unused link doesn't affect reachability
  (FORALL (mem h p q x)
    (IMPLIES
      (canReachWO mem h p q)             ; mem |- h ~> p /q
      (canReachWO (upd mem q x) h p q)   ; mem{q:=x} |- h ~> p /q
    ))

  ; adding a new link to the head doesn't let the chain reach
  ; anything new
  (FORALL (mem r h q)
    (IMPLIES
      (AND
        (NOT (canReach mem r h))         ; mem |- r ~!> h
        (NEQ q h)                        ; q != h
      )
      (NOT (canReach (upd mem q r) q h)) ; mem{q:=r} |- q ~!> h
    ))

))








