// rewrite.h
// conditional rewrite system

#ifndef REWRITE_H
#define REWRITE_H

#include "term.h"      // Term, ObjList
#include "sobjlist.h"  // SObjList

class Unify;           // unify.h


// order on terms
enum Order {
  LT,      // a < b
  EQ,      // a = b
  NLE      // neither a<b nor a=b; we don't assume order is total
};
string toString(Order o);

// compare order of two terms; uses a knuth-bendix order for now
Order compare(Term const *t1, Term const *t2);

// the weight of a term (minimum weight, if has unif. vars)
int weight(Term const *t);


// an equality between two terms: 'lhs' = 'rhs'
class Equality {
public:  
  // equality id, useful for tracking what happens to
  // individual equations
  int id;

  // static counter for the above
  static int nextId;

  // the equation terms
  Term *lhs, *rhs;      // (owners)

public:
  Equality(Term *l, Term *r);
  ~Equality();
    
  // ensure that lhs >= rhs; return true if they are equal
  bool orient();
                                          
  string asString(bool disEq = false) const;
};


// a conditional axiom
class Axiom : public Equality {
public:      // data
  // term pairs that must be equal for this rule to work
  ObjList<Equality> positives;

  // term pairs that must be unequal for this rule to work
  ObjList<Equality> negatives;

private:     // funcs
  void checkOrder(Equality const *e) const;

public:      // funcs
  Axiom(Term *l, Term *r);
  ~Axiom();

  bool unconditional() const
    { return positives.isEmpty() && negatives.isEmpty(); }
                           
  // check that the premise terms are all smaller than 'lhs',
  // throwing an exception if they are not
  void checkPremiseOrder() const;

  string asString() const;
};


// set of facts and axioms to which rewriting is applied in
// search of a contradiction to the satisfiability hypothesis
class Rewriter {
public:      // data
  // axioms
  ObjList<Axiom> axioms;

  // set of ground equality facts
  ObjList<Equality> equalities;

  // set of ground *disequality* facts
  ObjList<Equality> disequalities;

  // true if a contradiction is discovered
  bool contradiction;

private:     // data
  // counter for possibly-destructive operations
  int changes;
                     
private:     // funcs
  // helpers for 'reduce'
  bool reduceEqList(ObjList<Equality> &list, bool positive);
  bool reduceEq(Equality *reducee);
  Term *reduceTerm(Term *t);
  Term *reduceTermVia(Term *t, Equality const *e, Axiom const *ax);
  bool satisfiedPremises(Unify &u, Axiom const *ax);
  bool haveDisequality(Term const *lhs, Term const *rhs) const;

  // helpers for 'mix'
  void getNontrivProperSubterms(SObjList<Term const> &dest,
                                Term const *t);
  void gnpsHelper(SObjList<Term const> &dest, Term const *t);
  Term *substituteWithReplacement
    (Unify const &u, Term const *t,
     Term const *slot, Term const *replacement) const;

public:      // funcs
  Rewriter();
  ~Rewriter();

  // apply every (unconditional) equality to reduce all the
  // terms in the system as much as possible; return true if
  // any changes were made
  bool reduce();
             
  // mix axioms with facts; return true if new facts generated
  bool mix();

  // print out current state
  void print(char const *title);
};


#endif // REWRITE_H
