// list.c
// linked list of non-transferrable owner pointers

// --------- verifier preamble ----------
// for now, use asm hack to put the includes in
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"div.vc\";");
asm("verifier: include \"reach.vc\";");      // canReach, noCycles
asm("verifier: include \"disjoint.vc\";");   // disjoint
asm("verifier: include \"objct.vc\";");      // reachAllocated
asm("verifier: include \"cycles.vc\";");     // noCycles

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);

// declarations of model functions
int sel(int mem, int p);
int ptr(int addr, int field);
int canReach(int mem, int field, int src, int dest);
int canReachWO(int mem, int field, int src, int dest, int avoid);
int noCycles(int mem, int field, int ptr);
int disjoint(int mem, int field, int p, int q);
int reachAllocated(int mem, int objct, int field, int head);

// need to somehow say this is a constant symbolic value, not a global..
// let's try 'const'
int const dead;
int const live;
//int const state;

// more or less true globals
int mem;
int meta;
int tag;
int objsize;
int objct;


// --------------- code to verify -----------------
typedef struct Node {              
  // TODO: adjust reachability predicates to use the split model
  int use_flat_struct_model;

  int data;
  struct Node * /*owner*/ next;    // invariant: NULL or valid
  struct Node * /*serf*/  prev;    // ghost: thing whose 'next' points at me
} Node;

define reach_field = Node_next;    // field for ~> syntax
int const reach_field;


// invariant: NULL or valid, and no serfs point within *head
Node * /*owner*/ head = NULL;


// properties of a pointer to (the address of) a list node; basically
// says it is (locally) a value of the correct type
#define isListPtr(addr) (                                           \
  addr==NULL ||        /* NULL or ... */                            \
  ( tag[addr] == Node_tag &&               /* tag is "Node" */      \
    0<addr<objct )                         /* and addr < objct */   \
)

// invariant properties of an address of a list node, whenever the
// role tag is "Node"; this breaks the recursive invariant over the
// list structure at the point where the role tag is examined
#define isListNode(addr) (                                                  \
  objsize[addr] == Node_size &&                      /* node size */        \
  isListPtr(addr->next)                              /* 'next' is ok */     \
)

// global invariants
int global_inv() { return vc_expression(all_of {
  NULL < objct;                         // next address isn't NULL
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  isListPtr(head);                      // head is ok
  head!=NULL ==> head->prev==NULL;      // and first element's 'prev' is NULL

  forall(Node *addr). tag[addr]==Node_tag ==> all_of {   // all nodes [36][35][34]
    //isListNode(addr);                                  //   are nodes
    objsize[addr] == Node_size;

    addr->next!=NULL ==> all_of {
      0 < addr->next < objct;
      tag[addr->next] == Node_tag;

      addr->next->prev == addr;
    };
  };
  noCycles(head);                                  // list has no cycles [37]
  reachAllocated(mem, objct, Node_next, head);     // can't reach unallocated nodes [38]
});}



// append node with data==val
void append(int val)
  changes(mem, meta, tag, objsize, objct, head)
{
  Node * /*owner*/ last = (Node * /*owner*/)malloc(sizeof(Node));
  memset(last, 0, sizeof(Node));
  last->data = val;

  assert(head !~> last);

  // last is valid, no serfs point at it

  if (!head) {
    head = last;                // ownership transfer

    assert(head->next==NULL && head->prev==NULL);
    return;
  }

  Node * /*serf*/ p = head;

  // p is a valid serf pointing at *head

  while (p->next) {
    // p is a valid serf pointing within *head
    invariant {
      global_inv;
      head != NULL;

      // p points to a valid list node
      p != NULL;
      p->next != NULL;
      isListPtr(p);
      isListNode(p);

      // so does last
      last != NULL;
      last->next == NULL;
      last->prev == NULL;
      isListPtr(last);
      isListNode(last);

      // head can reach p, but cannot reach last
      head ~> p;
      head !~> last;

      // there are no cycles at or below last, and it's disconnected
      // from 'head' and 'p'
      noCycles(last);
      last !~> p;

      // 'head' and 'last' are disjoint; this should be a key
      // property for allowing updates to each to proceed
      // independently
      disjoint(mem, Node_next, head, last);
    }

    p = p->next;
  }

  // p is a valid serf pointing within *head, and p->next == NULL

  p->next = last;               // ownership transfer
  last->prev = p;               // ghost

  // last is dead
}


// remove first node with data==val, if it exists
void remove(int val)
  changes(mem, meta, tag, head)
{
  if (!head) {
    return;
  }

  if (head->data == val) {
    Node * /*owner*/ temp = head;  // ownership transfer

    // head is dead

    head = temp->next;          // ownership transfer
    temp->prev = NULL;          // ghost
    if (head) {
      head->prev = NULL;        // ghost
    }

    // temp->next is dead

    // is this sufficient to make it happy?
    assert(!exists(Node *p). tag[p]==Node_tag && p->next==temp);

    free(temp);                 // ok, *temp is dead
    return;
  }

  Node * /*serf*/ p = head;

  // p is a valid serf pointing at *head

  while (p->next &&
         p->next->data != val) {
    // p is a valid serf pointing within *head
    invariant {
      global_inv;
      head ~> p;
      p != NULL;
      isListPtr(p);
      p->next != NULL;       // loop guard
      p->next->data != val;
    }

    p = p->next;
  }

  // p is a valid serf pointing within *head

  // no other serf points within *head; in particular,
  // no serf points within *p->next

  if (p->next) {
    Node * /*owner*/ temp2 = p->next;     // ownership transfer

    // p->next is dead

    // no serf points within *temp

    p->next = temp2->next;             // ownership transfer
    if (temp2->next) {
      temp2->next->prev = p;           // ghost
    }

    // temp->next is dead

    assert(temp2->prev == p);

    // *p is valid

    // is this sufficient to make it happy?
    assert(!exists(Node *q). tag[q]==Node_tag && q->next==temp2);

    free(temp2);                       // ok, *temp is dead
  }
}


/*
  Errors to catch:
    - failure to restore some node's validity
*/
