// threaded.c
// based on threaded.pale

// threaded binary tree: parent pointers, and links to post-order
// successor

// ------------------- preamble ----------------
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"minmax.vc\";");

// C functions that are modeled directly
void *malloc(int size);

// declarations of model functions
int sel(int mem, int p);
int min(int a, int b);
int max(int a, int b);
int fresh_value();

int tag;
int objct;
int objsize;


// ---------------- code to verify -------------
#define AND all_of
#define OR one_of


typedef int bool;
#define false 0
#define true 1

  
// node in tree
typedef struct Node {
  // primary neighbor pointers; mutually injective
  struct Node *left, *right;

  // inverse of left/right
  bool isLeft;                  // (ghost)
  struct Node *parent;          // this one isn't ghost

  // intermediate ghost field: pointer to the leftmost
  // leaf in this subtree; might point to itself
  struct Node *firstLeaf;       // (ghost)

  // difficult auxilliary pointer: points to the node that
  // would be visited after this one in a postorder traversal;
  // for the final such node, it points to the first node in
  // that traversal
  struct Node *postSucc;
} Node;

#define isNode(n) (0 < (n) < objct && tag[n]==Node_tag)


// node whose 'firstLeaf' might be wrong
Node *needsFixing;


// global invariant, codifying the field relationships
int global_inv() { return vc_expression(AND {
  0 < objct;
  tag[0] == 0;

  forall(Node *n). tag[n]==Node_tag ==> AND {
    // declared types are accurate
    #define ACC_TYPE(field) \
      n->field!=NULL ==> isNode(n->field) /*user ;*/
    ACC_TYPE(left);
    ACC_TYPE(right);
    ACC_TYPE(parent);
    ACC_TYPE(firstLeaf);
    ACC_TYPE(postSucc);
    #undef ACC_TYPE

    // isLeft/parent is inverse of left/right
    n->left!=NULL ==>  (n->left->isLeft == true &&
                        n->left->parent == n);
    n->right!=NULL ==> (n->right->isLeft == false &&
                        n->right->parent == n);

    // firstLeaf: same as first child's left leaf
    n!=needsFixing ==> AND {
      n->left!=NULL ==>
        n->firstLeaf == n->left->firstLeaf;
      n->left==NULL && n->right!=NULL ==>
        n->firstLeaf == n->right->firstLeaf;
      n->left==NULL && n->right==NULL ==>
        n->firstLeaf == n;
    };

    #if 0   // temporary
    // postSucc: successor in post-order traversal
    // postSucc ==
    //   if (!parent) {
    //     return firstLeaf;
    //   }
    //   else {
    //     if (isLeft) {
    //       if (parent->right) {
    //         return parent->right->firstLeaf;
    //       }
    //       else {
    //         return parent;
    //       }
    //     }
    //     else {
    //       return parent;
    //     }
    //   }

    // am I the root?  then it's my first leaf
    n->parent==NULL ==>
      n->postSucc == n->firstLeaf;

    // am I a left child with a right sibling?  then sibling's first leaf
    n->parent!=NULL && n->isLeft == true && n->parent->right!=NULL ==>
      n->postSucc == n->parent->right->firstLeaf;

    // am I a left child with no sibling?  then it's my parent
    n->parent!=NULL && n->isLeft == true && n->parent->right==NULL ==>
      n->postSucc == n->parent;

    // am I a right child?  then it's my parent
    n->parent!=NULL && n->isLeft == false ==>
      n->postSucc == n->parent;
    #endif // 0
  };


});}


void insert(Node *root)
  pre(isNode(root) && needsFixing==NULL)
  post(AND {
    // nothing deallocated
    forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];

    // the primary neighbor function is not modified, only extended
    forall(Node *n).  0 < n < pre(objct) ==> AND {
      pre(n->left)!=NULL ==> n->left == pre(n->left);
      pre(n->right)!=NULL ==> n->right == pre(n->right);
    };

    // objct monotonic
    pre(objct) <= objct;
    
    // the parent of this node might need to have its 'firstLeaf'
    // adjusted (there must be a wrapper around the recursion that
    // nullifies this at the very end)
    needsFixing == root->parent;
  })
  changes(objct, tag, objsize, Node_tag, needsFixing)
{
  if (fresh_value() /*nondeterministic choice*/) {
    // insert on left
    if (root->left) {
      insert(root->left);
    }
    else {
      root->left = (Node*)malloc(sizeof(Node));

      // fill in the new node
      root->left->left = NULL;
      root->left->right = NULL;
      root->left->isLeft = true;
      root->left->parent = root;
      root->left->firstLeaf = root->left;
      root->left->postSucc = root->firstLeaf;   // root's old first leaf

      // TODO: fix the node that has 'root' as its postSucc
    }
  }

  else {
    // insert on right
    if (root->right) {
      insert(root->right);
    }
    else {
      root->right = (Node*)malloc(sizeof(Node));

      // fill in the new node
      root->right->left = NULL;
      root->right->right = NULL;
      root->right->isLeft = false;
      root->right->parent = root;
      root->right->firstLeaf = root->right;
      root->right->postSucc = root;

      // TODO: fix the node that has 'root' as its postSucc
    }
  }

  // fix firstLeaf (just a copy of the logic in the invariant)
  if (root->left!=NULL) {
    root->firstLeaf = root->left->firstLeaf;
  }
  else if (root->right!=NULL) {
    root->firstLeaf = root->right->firstLeaf;
  }
  else {             
    // can't happen..
    root->firstLeaf = root;
    assert(0==1);     // verify it can't happen
  }

  needsFixing = root->parent;
}


// EOF
