// enum.cc
// code for enum.h, including the enumeration engine itself

#include "enum.h"       // this module
#include "rule.h"       // Rule
#include "proof.h"      // ProofTree, ProofNode
#include "sobjlist.h"   // SObjList
#include "ptrmap.h"     // PtrMap
#include "term.h"       // Term
#include "autofile.h"   // AutoFILE
#include "strutil.h"    // readLine
#include "exc.h"        // xformat

#include <iostream.h>   // cout
#include <string.h>     // strstr
#include <stdlib.h>     // atoi


// --------------------- EnumNode --------------------
EnumNode::EnumNode(EnumNode *p,
                   EnumNode *r,
                   int s,
                   Rule *ru)
  : parent(p),
    refinedRule(r),
    refinedSlot(s),
    rule(ru)
{}

EnumNode::~EnumNode()
{
  delete rule;
}


// --------------------- Enumerator --------------------
Enumerator::Enumerator()
  : axioms(),          // empty initially
    nextRuleIndex(0),
    queue(),
    nextQueueIndex(0),
    emitLimit(0)
{}

Enumerator::~Enumerator()
{
  // enumeration nodes in 'queue' automatically deallocated
}


// when this is 1, we:
//   - only consider proofs that begin with the first axiom
//   - thereafter, only try the first axiom on exactly the first
//     unrefined premise enountered
// this means we consider proofs that begin with two uses of
// the first axiom, and then drop the branching factor
#define FIRST_AX_HACK 0


// enumeration engine
void Enumerator::enumerateTheory(int lim, bool falseConcl)
{
  // reset
  emitLimit = lim;
  queue.deleteAll();
  nextQueueIndex = 0;

  // enqueue each kind of rule as a root
  {
    FOREACH_OBJLIST_NC(Rule, axioms, axIter) {
      Rule const *ax = axIter.data();
      Rule *r = ax->clone(nextRuleIndex++);
      EnumNode *en = new EnumNode(NULL, NULL, 0, r);
      queue.push(en);
      #if FIRST_AX_HACK
      break;
      #endif
    }
  }

  while (emitLimit>0 && nextQueueIndex<queue.length()) {
    // take the next node to process from the queue (which should
    // never be empty)
    EnumNode *node = queue[nextQueueIndex++];

    // build the proof tree represented by this node
    ProofTree *tree = buildProofTree(node);
    if (!tree) {
      // the tree could not be built because the rules to not
      // tile together properly (unification fails somewhere);
      // discard this one and go on
      //delete node;    // no! the queue will get it later
      continue;
    }

    // emit the formula corresponding to this tree
    cout << tree->getFormula(falseConcl) << endl;
    emitLimit--;

    // walk over the tree, and for each open hypothesis site and
    // each axiom, create an EnumNode representing the refinement
    // of the tree via that axiom at that site
    enqueueRefinements(node, tree->root);

    delete tree;
  }
}


ProofTree *Enumerator::buildProofTree(EnumNode *node)
{
  // walk back towards the root of the enumeration tree, collecting
  // the list of nodes on that path
  SObjList<EnumNode> path;
  {
    for (EnumNode *n = node; n; n = n->parent) {
      path.prepend(n);
    }
  }
  
  // pull out the enumeration root for convenience
  EnumNode *enumRoot = path.removeFirst();

  // maintain a map from the enumeration nodes to the corresponding
  // proof tree nodes, since the former is what we have in 'path'
  PtrMap<EnumNode, ProofNode> map;

  // build the proof tree root
  ProofNode *proofRoot = new ProofNode(enumRoot);
  map.add(enumRoot, proofRoot);

  // create the proof tree; this contains the substitutions arising
  // from the various unifications throughout the proof tree
  ProofTree *ptree = new ProofTree(proofRoot);

  // build the tree by walking forwards along the path
  SFOREACH_OBJLIST_NC(EnumNode, path, iter) {
    EnumNode *n = iter.data();

    // which proof node is being refined?
    ProofNode *refined = map.get(n->refinedRule);
    xassert(refined);

    // the slot in question should be empty
    int slot = n->refinedSlot;
    xassert(refined->child[slot] == NULL);

    // create a new proof node for that slot
    ProofNode *pnode = new ProofNode(n);
    map.add(n, pnode);
    refined->child[slot] = pnode;

    // unify the conclusion of 'pnode' with the designated premise
    // of 'refined'
    if (!ptree->subst.unify(pnode->rule()->conclusion,
                            refined->rule()->premises.nth(slot))) {
      // failed unification; abandon this construction
      delete ptree;      // this gets everything
      return NULL;
    }
  }
  
  return ptree;
}


#if FIRST_AX_HACK
bool canUseFirstAx = true;
#endif

void Enumerator::enqueueRefinements(EnumNode *enode, ProofNode *n)
{
  ObjListIter<Term> iter(n->rule()->premises);
  for (int i=0; i < n->child.size(); i++, iter.adv()) {
    if (n->child[i]) {
      // has refined premise; recurse
      enqueueRefinements(enode, n->child[i]);
    }
    else {
      #if FIRST_AX_HACK
      int ct=0;
      #endif

      // unrefined premise; enqueue refinements
      FOREACH_OBJLIST(Rule, axioms, axIter) {
        Rule const *ax = axIter.data();

        #if FIRST_AX_HACK
        if (ct++==0 && !canUseFirstAx) {
          continue;                 // skip first ax
        }
        canUseFirstAx = false;      // one shot
        #endif

        // does this axiom's conclusion unify with the premise?
        if (!unifies(ax->conclusion, iter.data())) {
          // no, forget it
          continue;
        }

        // make a copy of this rule with unique unification variables
        Rule *r = ax->clone(nextRuleIndex++);

        // make the refinement node
        EnumNode *ref = new EnumNode(enode,       // enumeration tree parent
                                     n->enumNode, // proof node to refine
                                     i,           // child to refine
                                     r);          // new proof rule

        // enqueue it
        queue.push(ref);
      }
    }
  }
  xassert(iter.isDone());
}


void Enumerator::parseAxioms(char const *fname)
{
  AutoFILE fp(fname, "r");
    
  // parsing state
  enum State {
    S_PREMISE,         // expecting premise terms
    S_CONCLUSION       // expecting conclusion term
  } state = S_PREMISE;

  // premises so far
  ObjList<Term> premises;

  string line;
  while (readLine(line, fp)) {
    // remove comments
    char *p = strstr(line.pchar(), "//");
    if (p) {
      *p = 0;
    }

    // skip blank lines
    line = trimWhitespace(line);
    if (line.length() == 0) {
      continue;
    }
    
    // conclusion separator?
    if (line[0] == '-') {
      state = S_CONCLUSION;
      continue;
    }
    
    // read the term
    Term *t = Term::parse(line);
    if (state == S_PREMISE) {
      premises.append(t);
    }
    else {
      // complete the rule
      Rule *r = new Rule(t);
      r->premises.concat(premises);    // move premises into the rule
      xassert(premises.isEmpty());
      axioms.append(r);
      state = S_PREMISE;
    }
  }
  
  if (premises.isNotEmpty()) {
    xformat("axioms file ends with an unfinished rule");
  }
}


void Enumerator::printAxioms()
{
  FOREACH_OBJLIST(Rule, axioms, axIter) {
    Rule const *ax = axIter.data();
    FOREACH_OBJLIST(Term, ax->premises, pIter) {
      cout << "; " << pIter.data()->asString() << "\n";
    }
    cout << "; ---------------------------\n";
    cout << "; " << ax->conclusion->asString() << "\n\n";
  }
}


// ----------------------- main -------------------------
int main(int argc, char *argv[])
{
  bool falseConcl = false;
  char const *progName = argv[0];

  #define SHIFT   \
    argc--;       \
    argv++

  SHIFT;

  while (argv[0] && argv[0][0] == '-') {
    if (0==strcmp(argv[0], "-false")) {
      falseConcl = true;
    }
    else {
      printf("unknown option: %s\n", argv[0]);
      return 2;
    }
    
    SHIFT;
  }

  if (argc != 2) {
    printf("usage: %s [options] <axiomFile> <limit>\n", progName);
    printf("  options:\n"
           "    -false       emit all conclusions as \"FALSE\"\n");
    return 0;
  }

  char const *axiomFile = argv[0];
  char const *limitStr = argv[1];

  // read axioms
  Enumerator e;
  e.parseAxioms(axiomFile);

  #if 0      // superceded by equality.ax
  // reflexivity
  e.axioms.append(
    makeRule0(new Term("EQ", new Term("X"), new Term("X")))
  );

  // symmetry
  e.axioms.append(
    makeRule1(new Term("EQ", new Term("X"), new Term("Y")),
              new Term("EQ", new Term("Y"), new Term("X")))
  );

  // transitivity
  e.axioms.append(
    makeRule2(new Term("EQ", new Term("X"), new Term("Z")),
              new Term("EQ", new Term("X"), new Term("Y")),
              new Term("EQ", new Term("Y"), new Term("Z")))
  );
  #endif // 0

  // print them out for comfirmation
  e.printAxioms();

  // start printing....
  e.enumerateTheory(atoi(limitStr), falseConcl);

  if (e.emitLimit > 0) {
    cout << "\n; enumeration stopped on its own" << endl;
  }
  else {
    cout << "\n; stopping due to limit" << endl;
  }

  return 0;
}

// EOF
