// vc_env.h
// verification condition generation environment

#ifndef VC_ENV_H
#define VC_ENV_H

#include "ownerserf.h"       // owner/serf
#include "cc.ast.gen.h"      // Function, Expression, etc.
#include "autofile.h"        // AutoFILE
#include "str.h"             // stringc

class TypeFactory;           // cc_type.h
class StringTable;           // strtable.h
class Variable;              // cc_type.h
class ECompoundType;         // exttype.h


class VCEnv {
public:      // data
  // ----------- global -----------
  // for occasional synthesis of new AST elements
  TypeFactory &tfac;
  StringTable &str;

  // output stream
  AutoFILE output;

  // variable for use when assigning to memory
  Variable * owner memVar;

  // counter for making up fresh names
  int tempCt;

  // counter for naming type tags
  int typeTagCt;

  // some predefined StringRefs
  StringRef str_malloc;
  StringRef str_memset;
  StringRef str_memmove;
  StringRef str_memcpy;
  StringRef str_free;
  StringRef str_kfree;
  StringRef str_metaVar;
  StringRef str_fresh_value;
  StringRef str_CLT_between;

  // list of global variables
  SObjList<Variable> globals;

  // ----------- per-function -----------
  // function we're analyzing
  Function * serf currentFunc;

  // variable that represents the value returned from the function
  // we're currently analyzing
  Variable * owner result;

  // list of parameters
  SObjList<Variable> params;

  // list of local variables
  SObjList<Variable> locals;

  // for convenience, give names to certain combinations of lists
  SObjList<Variable> * serf globalsAndParams[2];
  SObjList<Variable> * serf globalsAndParamsAndLocals[3];

  // ----------- per-path -----------
  // location of statement we're vcgen'ing inside, for error messages
  // when no other location is convenient
  SourceLoc lastStmtLoc;

  // ----------- per-expression -----------
  // true if we're inside a verifier predicate (may affect how
  // some things are rendered into VCs)
  bool inPredicate;

  // true if we're inside a pre() modifier
  bool inPreState;

private:     // funcs
  void proveOrLemma(rostring which, Expression * owner pred, rostring why);


public:      // funcs
  VCEnv(TypeFactory &tfac, StringTable &str, rostring outFname);
  ~VCEnv();

  void beginFunc(Function *f);
  void endFunc();

  void beginPath() {}
  void endPath() {}


  // yield an expression that means "this syntax isn't yet handled",
  // i.e. it's a to-do
  Expression * owner todo();

  // yield an E_intLit
  E_intLit * owner intLit(int i);

  // yield an E_variable with unknown location
  E_variable * owner makeVar(rostring name);

  // like 'makeVar', but add "_pre" if we're in the pre-state
  E_variable * owner makePVar(rostring name);

  // not sure what I want to do with this
  Expression * owner rval(Expression * owner e) { return e; }
    

  // some common functions; all pointers are owners
  Expression *ptr(Expression *addr, Expression *field);
  Expression *upd(Expression *mem, Expression *ptr, Expression *value);
  Expression *upd(Expression *mem, Expression *addr, Expression *field,
                  Expression *value)
    { return upd(mem, ptr(addr, field), value); }
  Expression *updrange(Expression *mem, Expression *addr, Expression *low,
                       Expression *high, Expression *value);


  // add to assumptions
  void assume(Expression * owner pred, rostring why);
  void assumeNot(Expression * owner pred, rostring why);
  void assumeSense(Expression * owner pred, bool sense,  // true means 'assume', false means 'assumeNot'
                   rostring why);

  // record a variable assignment
  void assign(rostring name, Expression * serf target);
  void assignDel(rostring name, Expression * owner target)
    { assign(name, target); delete target; }
  void assign(Variable *var, Expression * serf target);
  void assignLive(Variable *var, Expression * serf target);

  // proof obligation
  void prove(Expression * owner pred, rostring why);

  // prove, then assume
  void lemma(Expression * owner pred, rostring why);

  // deallocate a vcgen expression
  void discard(Expression * owner expr);

  // add a definition
  void define(rostring name, Expression * owner val, rostring why);


  // list of global variables as E_variables, for use as E_funCall args
  FakeList<ArgExpression> *globalVarList();

  // yield a comma-separated list of all global variable names
  string globalVarNameList(bool preState = false) const;

  // name of a slot of a field
  string fieldSlotName(Variable const *field) const;

  // add a trailing "_pre" if 'inPreState' is true
  string maybePre(rostring name) const;
  
  // surround a given stringified object index with a predicate
  // to say that that object has the given class' tag; this will
  // take into account subtyping, i.e. allow the object to actually
  // be a subclass
  string hasClassTag(ECompoundType *ct, rostring objIndex) const;


  // common part of assignment expression vcgen'ing
  Expression * owner assignmentExpr
    (Expression *target, int path, Expression * owner srcValue);


  // output
  void out(rostring str);
  VCEnv& operator<< (char const *str) { out(str); return *this; }
  VCEnv& operator<< (rostring str) { out(str); return *this; }
  VCEnv& operator<< (int i) { out(stringc << i); return *this; }
  VCEnv& operator<< (char c) { out(stringc << c); return *this; }

  // errors: usually, syntax that violates the chosen model
  void error(rostring msg);
  void error(SourceLoc loc, rostring msg);
  
  // warnings
  void warning(rostring msg);
  void warning(SourceLoc loc, rostring msg);

  // break conjunctions, printing them with 'indent' before and
  // a semicolon and newline after
  void printConjunctiveLines(rostring indent, Expression *pred);
};


#endif // VC_ENV_H
