// vcgen.cc
// vcgen methods on the C abstract syntax tree

#include "cc.ast.gen.h"         // C++ ast
//#include "absval.ast.gen.h"     // abstract domain values
#include "vc_env.h"             // VCEnv
#include "exttype.h"            // EFunctionType, etc.
#include "sobjlist.h"           // SObjList
#include "trace.h"              // tracingSys
#include "strutil.h"            // quoted, plural
#include "cfgpaths.h"           // countExprPaths
//#include "predicate.ast.gen.h"  // Predicate ast, incl. P_equal, etc.
#include "objstack.h"           // ObjStack
#include "owner.h"              // Owner
//#include "treeout.h"            // treeOut
#include "astbuild.h"           // apply*, etc.
#include "cc_print.h"           // PrintEnv
#include "smregexp.h"           // regexpMatch

#include <stdlib.h>             // getenv

// STUBEMIT: #include "cc.ast.gen.h"


#define IN_PREDICATE(env) Restorer<bool> restorer1(env.inPredicate, true)
#define IN_PRE_STATE(env) Restorer<bool> restorer2(env.inPreState, true)


// true if the Variable also has a 'meta' variable for its state
static bool hasMetaVar(Variable const *_v)
{
  EVariable const *v = extC(_v);
  return !v->hasVAttr(VA_PARAMBYREF) &&
         v->type->isPointerType();
}


// ------------------- TopForm -------------------
// STUBPATTERN: ^void.*TF_.*::vcgen\b

void TF_decl::vcgen(VCEnv &env)
{
  FAKELIST_FOREACH_NC(Declarator, decl->decllist, iter) {
    EVariable *v = ext(iter->var);
    Type *t = v->type;

    // exclude some global declarations
    if (t->isFunctionType()) {    // prototypes
      continue;
    }

    if (v->hasFlag(DF_TYPEDEF)) { // typedef
      continue;
    }

    if (t->isConst()) {           // (symbolic) constants
      if (v->value) {
        // hack: initialized constants will be regarded as new symbolic
        // constants, whereas uninitialized constants are taken to be
        // constants that are already declared in the .vc file... this needs
        // do be fixed
        env << "introduce " << v->name << ";\n";
      }
      continue;
    }

    if (t->isCompoundType() &&
        !v->hasVAttr(VA_ADDRTAKEN)) {
      // global instance of a struct, whose addr is not taken:
      // model as a bunch of separate globals, one for each field
      env << "// fields of " << v->name << "\n";
      FOREACH_OBJLIST_NC(EVariable, v->synthFields, fieldIter) {
        EVariable *synth = fieldIter.data();

        env << "var " << synth->name << ";\n";
        if (hasMetaVar(synth)) {
          env << "var " << synth->name << "_meta;\n";
        }
        env.globals.append(synth);
      }
      env << "\n";
      continue;
    }

    env << "var " << v->name << ";\n";
    if (hasMetaVar(v)) {
      env << "var " << v->name << "_meta;\n";
    }
    env.globals.append(v);
  }
}


void TF_func::vcgen(VCEnv &env)
{
  // functions are handled directly through the visitor, so that the
  // VCEnv doesn't have to worry about nesting the function
  // environments
}

void TF_template::vcgen(VCEnv &env)
{}

void TF_explicitInst::vcgen(VCEnv &env)
{}

void TF_linkage::vcgen(VCEnv &env)
{}

void TF_one_linkage::vcgen(VCEnv &env)
{}
               

void handleAsm(VCEnv &env, E_stringLit *text)
{
  string decoded = parseQuotedString(text->text);
  if (regexpMatch(decoded, "^verifier: ")) {
    // this 'asm' is a hack to let me pass vclang commands
    // straight through
    env << (decoded.c_str()+10) << "\n";
  }
}

void TF_asm::vcgen(VCEnv &env)
{
  handleAsm(env, text);
}


void TF_namespaceDefn::vcgen(VCEnv &env)
{
  // not implemented
}

void TF_namespaceDecl::vcgen(VCEnv &env)
{
  // not implemented
}


void TF_macro::vcgen(VCEnv &env)
{
  FOREACH_ASTLIST_NC(Macro, macros, iter) {
    iter.data()->vcgen(env);
  }
}


void TF_changeGroup::vcgen(VCEnv &env)
{
  env << "// change_group: " << name->getName() << "\n";
}


void TF_axiom::vcgen(VCEnv &env)
{
  // not implemented
}


void TF_option::vcgen(VCEnv &env)
{}


// -------------------- Function ----------------------
class FindLocals : public ASTVisitor {
public:
  VCEnv &env;

public:
  FindLocals(VCEnv &e) : env(e) {}
  bool visitFunction(Function *f);
  bool visitDeclaration(Declaration *d);
};

bool FindLocals::visitFunction(Function *f)
{
  // do not look inside any nested functions at this time
  return false;
}

bool FindLocals::visitDeclaration(Declaration *decl)
{
  FAKELIST_FOREACH_NC(Declarator, decl->decllist, iter) {
    Variable *v = iter->var;
    env << "  var " << v->name << ";\n";
    if (hasMetaVar(v)) {
      env << "  var " << v->name << "_meta;\n";
    }
    env.locals.append(v);
  }
  return true;
}


void Function::vcgen(VCEnv &env)          // stubify
{
  StringRef name = nameAndParams->var->name;

  // the special global_inv function body becomes the body of a macro
  // that will be used as part of the pre/post of every function
  if (0==strcmp(name, "global_inv")) {
    xassert(numPaths == 1);
    S_return *r = body->asS_compound()->stmts.first()->asS_return();
    Expression * owner pred = r->expr->vcgenPred(env, 0 /*path*/);
    env << "define global_inv(" << env.globalVarNameList() << ") {\n"
        << pred->tovcConjunctiveLines("  ")
        << "}\n\n";
    delete pred;
    return;
  }

  env << "// ------------------- " << name << " ----------------------\n";

  env << "introduce " << name << "_address;    // for taking address\n";

  EFunctionType *ft = ext(funcType);

  env.beginFunc(this);

  int numRoots = roots.count();
  env << "// roots: " << numRoots << "\n";
  env << "// paths: " << numPaths << "\n";
  env << "block \"" << name << "\" {\n";

  // give names to all the pre-state values of globals
  {
    env << "  // global pre-state values\n";
    SFOREACH_OBJLIST(Variable, env.globals, iter) {
      Variable const *v = iter.data();
      env << "  introduce " << v->name << "_pre;\n";
      if (hasMetaVar(v)) {
        env << "  introduce " << v->name << "_meta_pre;\n";
      }
    }
    env << "\n";
  }

  // declare all params
  {
    env << "  // params\n";
    
    // why did I do it this way?  it's clumsy and wrong
    //D_func *df = nameAndParams->decl->asD_func();
    //FAKELIST_FOREACH(ASTTypeId, df->params, iter) {
      //Variable *v = iter->decl->var;

    SFOREACH_OBJLIST_NC(Variable, ft->params, iter) {
      Variable *v = iter.data();
      env.params.append(v);

      env << "  var " << v->name << ";\n";

      // name the pre-state value
      env << "  introduce " << v->name << "_pre;\n";
      
      if (hasMetaVar(v)) {
        env << "  var " << v->name << "_meta;\n";
        env << "  introduce " << v->name << "_meta_pre;\n";
      }
    }
    env << "\n";
  }

  env << "  // precondition\n";
  env << "  assume global_inv(" << env.globalVarNameList(true /*pre*/) << ");\n";
  if (ft->precondition) {
    IN_PREDICATE(env);
    IN_PRE_STATE(env);
    env.assume(ft->precondition->vcgenPred(env, 0 /*path*/), "" /*why*/);
  }
  env << "\n";

  // find and declare all the locals
  {
    env << "  // locals\n";
    FindLocals f(env);
    body->traverse(f);
    env << "\n";
  }

  // return value
  if (ft->result) {
    env << "  // pseudovariable for return value\n";
    env << "  var " << ft->result->name << ";\n";
    env << "\n";
  }

  // for each root...
  int rootIndex=-1;
  SFOREACH_OBJLIST_NC(Statement, roots, rootIter) {
    Statement *root = rootIter.data();
    rootIndex++;

    env << "  // " << toLCString(root->loc)
        << ": start of root " << rootIndex
        << ", " << root->numPaths
        << plural(root->numPaths, " path") << "\n";

    // for each path from that root...
    for (int path=0; path < root->numPaths; path++) {
      // skip?
      {       
        char const *pathStr = getenv("SELPATH");
        if (pathStr && regexpMatch(pathStr, "^[0-9]+/[0-9]+$")) {
          int selRootIdx = atoi(pathStr);
          int selPathIdx = atoi(strchr(pathStr, '/')+1);
          if (rootIndex==selRootIdx && path==selPathIdx) {
            // go ahead
          }
          else {
            // bail
            continue;
          }
        }
      }

      env << "  // root " << rootIndex << ", path " << path << "\n";
      env << "  block \"" << name << "/" << rootIndex << "/" << path << "\" {\n";

      env.beginPath();

      // ------------- establish set of known facts ---------------
      // add the path's start predicate as an assumption
      if (root == body) {
        // all of the globals and parameters are equal to their pre-state values
        for (SObjListMultiIter<Variable> iter(env.globalsAndParams, 2);
             !iter.isDone(); iter.adv()) {
          Variable const *v = iter.data();
          env << "    assume " << v->name << " == " << v->name << "_pre;\n";
          if (hasMetaVar(v)) {
            env << "    assume " << v->name << "_meta == " << v->name << "_meta_pre;\n";
          }
        }
      }
      else {
        // the root should be an invariant statement, and that is
        // the start predicate
        S_invariant *inv = root->asS_invariant();

        //env.addFact(inv->expr->vcgenPred(env, 0 /*path*/), "invariant");
        IN_PREDICATE(env);
        env.assume(inv->vcgenPredicate(env), "invariant");
      }

      // information about variables, derived from their types
      for (SObjListMultiIter<Variable> iter(env.globalsAndParamsAndLocals, 3);
           !iter.isDone(); iter.adv()) {
        Variable const *v = iter.data();
        if (v->type->isSimple(ST_UNSIGNED_INT)) {
          env << "    assume " << v->name << " >= 0;\n";
        }
      }

      // -------------- interpret the code ------------
      // now interpret the function body
      SObjList<Statement> stmtList;
      root->vcgenPath(env, stmtList, path, false /*cont*/);

      // NOTE: the path's termination predicate will be proven
      // inside the vcgenPath call above

      // todo: enable this in the vclang tool
      env << "    //report_inconsistency;\n";

      env << "  }\n";

      env.endPath();
    }
  }

  env << "}\n\n";
  env.endFunc();
}


// --------------------- Declaration --------------------
#if 0
// NOTE: this is only used for globals; local declarations are
// taken apart by S_decl
void Declaration::vcgen(VCEnv &env)
{
  FOREACH_ASTLIST(Declarator, decllist, iter) {
    Declarator *d = iter.data();
    if (d->init) {
      d->vcgen(env, d->init->vcgen(env, d->var->type, 0));
    }
    else {
      // declarator's vcgen just acts like an assignment, so if
      // there's no initializing expression, we don't do anything here
    }
  }
}

AVvar *fieldRep(CompoundType::Field const *field)
{
  return new AVvar(field->name, "field value representative");
}

void Declarator::vcgen(VCEnv &env, AbsValue *value)
{
  // treat the declaration as an assignment; the variable was added
  // to the abstract environment in TF_func::vcgen
  env.updateVar(var, value);

  // add facts about the variable and its value which are a
  // consequence of being declared at this point
  env.addDeclarationFacts(var, value);
}
#endif // 0


// ----------------------- Statement ----------------------
// eval 'expr' in a predicate context
Expression * owner vcgenPredicate(VCEnv &env, Expression *expr, int path)
{
  if (path != 0) {
    // used to be just an assertion, but then I needed more info...
    //cout << "(path=" << path
    //     << ") predicate expr problem: " << expr->toString() << endl;
    xfailure("nonzero path count in vcgenPredicate");
  }

  IN_PREDICATE(env);
  return expr->vcgenPred(env, path);
}

#if 0
void proveNonOwning(VCEnv &env, AbsValue *ptrVal, char const *why)
{
  ptrVal = env.rval(ptrVal);
  env.prove(P_notEqual(env.avSel(ptrVal, env.avOwnerField_state()),
                       env.avOwnerState_owning()),
            why);
}

void variableLeavingScope(VCEnv &env, Variable const *var)
{
  if (var->type->isOwnerPtr()) {
    // OWNER: must be nonowning to leave scope
    proveNonOwning(env, env.get(var), "owner must be nonowning to leave scope");
  }
}
#endif // 0


// STUBEMIT: void Statement::vcgenPath(VCEnv &env, SObjList<Statement> &path,
// STUBEMIT:                           int index, bool isContinue) {}

// uber-vcgen: generic, as it uses the embedded CFG; but it
// calls the specific 'vcgen' for the particular statement kind
void Statement::vcgenPath(VCEnv &env, SObjList<Statement> &path,
                          int index, bool isContinue)
{
  env.lastStmtLoc = loc;

  // validate 'index'
  int exprPaths = countExprPaths(this, isContinue);
  xassert(exprPaths >= 1);
  xassert(0 <= index && index < (numPaths * exprPaths));

  // debugging check
  if (path.contains(this)) {
    env << "    // CIRCULAR path\n";
    return;
  }
  path.prepend(this);

  // print information about this statement
  if (!isS_compound()) {           // don't print single '{' chars everywhere
    env << "    // " << toLCString(loc) << ": ";
    stringBuilder sb;
    StringPrintEnv penv(sb);
    print(penv);
    penv.finish();

    // truncate the statement at the first newline; this ensures in
    // general that I don't print beyond the end of the line (b/c if
    // I did then it wouldn't be treated as part of the comment),
    // and also makes sure I abbreviate large compound statements and
    // loop constructs
    char const *newline = strchr(sb.c_str(), '\n');
    if (newline) {
      env << substring(sb.c_str(), newline-sb.c_str());
    }
    else {
      env << string(sb);    // no newline, print whole thing
    }

    env << "\n";
  }

  // pick which expression path we will follow
  int exprPath = index % exprPaths;
  index = index / exprPaths;

  // retrieve all successors of this node
  NextPtrList successors;
  getSuccessors(successors, isContinue);

  if (successors.isEmpty()) {
    // this is a return statement (or otherwise end of function)
    xassert(index == 0);

    // vcgen this statement, telling it no continuation path
    vcgen(env, isContinue, exprPath, NULL);

    // prove the function postcondition
    Expression *post = ext(env.currentFunc->funcType)->getPostcondition(env);
    if (post) {
      IN_PREDICATE(env);
      env.prove(post->vcgenPred(env, 0 /*path*/), "postcondition");
    }

    #if 0   // todo: hook this up again
    // implicit postconditions: do all scope-exit checks
    SFOREACH_OBJLIST(Variable, env.currentFunc->params, iter1) {
      variableLeavingScope(env, iter1.data());
    }
    SFOREACH_OBJLIST(Variable, env.currentFunc->locals, iter2) {
      variableLeavingScope(env, iter2.data());
    }
    #endif // 0
    
    // global_inv is effectively part of the postcondition
    env << "    prove \"postcondition: global_inv\"\n"
        << "      global_inv(" << env.globalVarNameList() << ");\n";
  }
  else {
    // consider each choice
    // largely DUPLICATED from cfgpaths.cc:printPathFrom
    for (int i=0; i < successors.length(); i++) {
      NextPtr np = successors[i];
      Statement *s = np.stmt();
      int pathsFromS = numPathsThrough(s);

      // are we going to follow 's'?
      if (index < pathsFromS) {
        // yes; vcgen 'node', telling it 's' as the continuation
        vcgen(env, isContinue, exprPath, s);

        // is 's' an invariant?
        if (s->isS_invariant()) {
          // terminate the path & prove the invariant
          xassert(exprPath == 0);              // bugfix: had "index" where meant "exprPath" (it's like vcgen call above)
          //env.prove(vcgenPredicate(env, s->asS_invariantC()->expr, exprPath),   // bugfix: had "index" where meant "exprPath" (it's like vcgen call above)
          env.prove(s->asS_invariant()->vcgenPredicate(env),
                    stringc << "invariant at " << toLCString(s->loc));
        }
        else {
          // continue the path
          s->vcgenPath(env, path, index, np.cont());
        }
        index = 0;       // remember that we found a path to follow
        break;
      }

      else {
        // no; factor out s's contribution to the path index
        index -= pathsFromS;
      }
    }

    // make sure we followed *some* path
    xassert(index == 0);
  }

  path.removeFirst();
}


// STUBPATTERN: ^void.*S_.*::vcgen\b

void S_skip::vcgen(STMT_VCGEN_PARAMS) { xassert(path == 0); }

// control-flow targets add nothing; implied predicates are
// added by the preceding statement
void S_label::vcgen(STMT_VCGEN_PARAMS) { xassert(path == 0); }
void S_case::vcgen(STMT_VCGEN_PARAMS) { xassert(path == 0); } 
//void S_caseRange::vcgen(STMT_VCGEN_PARAMS) { xassert(path == 0); }
void S_default::vcgen(STMT_VCGEN_PARAMS) { xassert(path == 0); }


void S_expr::vcgen(STMT_VCGEN_PARAMS)
{
  // evaluate and discard
  env.discard(expr->vcgen(env, path));
}


void S_compound::vcgen(STMT_VCGEN_PARAMS)
{
  // this statement does nothing, because successor info
  // is already part of the CFG
  xassert(path == 0);
}


// prototypical statement vcgen: hit the expression language in
// the guard, and then add a predicate based on outgoing edge
void S_if::vcgen(STMT_VCGEN_PARAMS)
{
  // evaluate the guard
  Expression * owner pred = cond->vcgenPred(env, path);

  if (next == thenBranch) {
    // remember that the guard was true
    env.assume(pred, stringc << "true guard of 'if' at " << toLCString(loc));
  }
  else {
    // guard is false
    xassert(next == elseBranch);
    env.assumeNot(pred, stringc << "false guard of 'if' at " << toLCString(loc));
  }
}


void S_switch::vcgen(STMT_VCGEN_PARAMS)
{
  // evaluate switch-expression
  env.discard(cond->vcgen(env, path));

  // for now, learn nothing from the outgoing edge
  env << "    // TODO: ignoring implication of followed switch edge\n";
}


void S_while::vcgen(STMT_VCGEN_PARAMS)
{
  Expression * owner pred = cond->vcgenPred(env, path);

  if (next == body) {
    env.assume(pred, stringc << "true guard of 'while' at " << toLCString(loc));
  }
  else {
    // 'next' is whatever follows the loop
    env.assumeNot(pred, stringc << "false guard of 'while' at " << toLCString(loc));
  }
}


void S_doWhile::vcgen(STMT_VCGEN_PARAMS)
{
  if (!isContinue) {
    // drop straight into body, learn nothing
    xassert(path == 0);
  }
  else {
    Expression * owner pred = expr->vcgenPred(env, path);
    if (next == body) {
      env.assume(pred, stringc << "true guard of 'do' at " << toLCString(loc));
    }
    else {
      env.assumeNot(pred, stringc << "false guard of 'do' at " << toLCString(loc));
    }
  }
}


void S_for::vcgen(STMT_VCGEN_PARAMS)
{
  int modulus = cond->numPaths1();

  if (!isContinue) {
    // init
    // UPDATE: don't do this, because it should get done when 'init'
    // is encountered in the CFG directly
    //init->vcgen(env, false, path / modulus, NULL);
    xassert(path < modulus);
  }
  else {
    // inc
    after->vcgen(env, path / modulus);
  }

  // guard
  Expression * owner pred = cond->vcgenPred(env, path % modulus);

  // body?
  if (next == body) {
    env.assume(pred, stringc << "true guard of 'for' at " << toLCString(loc));
  }
  else {
    xassert(next == this->next.stmt());
    env.assumeNot(pred, stringc << "false guard of 'for' at " << toLCString(loc));
  }
}


void S_break::vcgen(STMT_VCGEN_PARAMS) { xassert(path==0); }
void S_continue::vcgen(STMT_VCGEN_PARAMS) { xassert(path==0); }

void S_return::vcgen(STMT_VCGEN_PARAMS)
{
  xassert(next == NULL);
  if (expr) {
    // model as an assignment to 'result'
    Expression * owner val = expr->vcgen(env, path);
    env.assign(env.result, val);
    delete val;
  }
  else {
    xassert(path == 0);
  }
}


void S_goto::vcgen(STMT_VCGEN_PARAMS) { xassert(path==0); }


// NOTE: my design for handling variable declarations *sucks*.. it's
// spread out among half a dozen functions, with no particular rhyme
// or reason except I keep tweaking it in response to nonworking
// verifications, in the hopes I'll evolve a decent design... so far
// it is not working
// UPDATE: maybe it's working better now?  we'll see...
void S_decl::vcgen(STMT_VCGEN_PARAMS)
{
  int subexpPaths = 1;
  FAKELIST_FOREACH_NC(Declarator, decl->decllist, d) {
    if (d->init) {
      subexpPaths = countExprPaths(d->init);

      // evaluate 'init' to an abstract value; the initializer
      // will need to know which type it's constructing
      Expression * owner initVal = 
        d->init->vcgen(env, d->var->type, path % subexpPaths);

      path = path / subexpPaths;

      // add a binding for the variable
      env.assignLive(d->var, initVal);
      
      delete initVal;
    }
    else {
      // this shouldn't be needed in the current vcgen strategy, because
      // vclang will make up a name 

      // make up a new variable to stand for the initial value, etc.
      //env.initializeUninitVariable(d->var);
    }
  }

  // if we entered the loop at all, this ensures the last iteration
  // consumed all of 'path'; if we never entered the loop, this
  // checks that path was 0
  xassert(path < subexpPaths);
}


void S_try::vcgen(STMT_VCGEN_PARAMS) { xassert(path==0); }


void S_asm::vcgen(STMT_VCGEN_PARAMS)
{
  handleAsm(env, text);
}


void S_namespaceDecl::vcgen(STMT_VCGEN_PARAMS)
{
  // not implemented
}


void S_assert::vcgen(STMT_VCGEN_PARAMS)
{
  // map the expression
  xassert(path==0);
  Expression * owner p = vcgenPredicate(env, expr, path);

  // emit as lemma
  env.lemma(p, stringc << "assertion at " << toString(loc));
}


void S_assume::vcgen(STMT_VCGEN_PARAMS)
{
  // evaluate
  xassert(path==0);
  Expression * owner p = vcgenPredicate(env, expr, path);

  // remember it as a known fact
  env.assume(p, stringc << "assumption at " << toString(loc));
}


void S_invariant::vcgen(STMT_VCGEN_PARAMS)
{
  // we only call the 'vcgen' of an invariant at the *start* of a path
  //env.assume(vcgenPredicate(env, expr, path));

  // update: I'll let the path driver take care of both adding premises,
  // and proving obligations; so s_invariant itself just holds onto the
  // expression
}


Expression * owner S_invariant::vcgenPredicate(VCEnv &env)
{
  IN_PREDICATE(env);

  Expression * owner exprPred = expr->vcgenPred(env, 0 /*path*/);

  // make sure the 'changes' are available as a conjunction
  ext(env.currentFunc->funcType)->getPostcondition(env);

  // add in the conjunction of unchanged globals
  if (ext(env.currentFunc->funcType)->changesConj) {
    exprPred = e_and(exprPred,
      ext(env.currentFunc->funcType)->changesConj->vcgenPred(env, 0));
  }

  #if 0   // todo: add this back in
  if (inferFacts.isEmpty()) {
    return exprPred;
  }
  else {
    // construct a conjunction for expr and inferFacts
    P_and *conj = new P_and(NULL);
    conj->conjuncts.append(exprPred);
    SFOREACH_OBJLIST(Expression, inferFacts, iter) {
      conj->conjuncts.append(iter.data()->vcgenPred(env, 0 /*path*/));
    }
    return conj;
  }
  #endif // 0

  return exprPred;
}


#if 0
void S_thmprv::vcgen(STMT_VCGEN_PARAMS)
{
  xassert(path==0);
  IN_PREDICATE(env);
  s->vcgen(env, false, path, NULL);
}
#endif // 0


// ---------------------- Condition -----------------
// STUBPATTERN: ^Expression.*CN_.*::vcgen(Pred)?\b

Expression * owner CN_expr::vcgen(VCEnv &env, int path)
{
  return expr->vcgen(env, path);
}

Expression * owner CN_expr::vcgenPred(VCEnv &env, int path)
{
  return expr->vcgenPred(env, path);
}


Expression * owner CN_decl::vcgen(VCEnv &env, int path)
{
  return env.todo();
}

Expression * owner CN_decl::vcgenPred(VCEnv &env, int path)
{
  return env.todo();
}


// ---------------------- Expression -----------------
// STUBPATTERN: ^Expression.*E_.*::ivcgen\b

Expression * owner Expression::vcgen(VCEnv &env, int path)   // stubify
{            
  #if 0      // trying something else.. in vclang..
  if (path==0 && numPaths==0) {
    // side-effect free; is it evaluable to a literal?
    string msg;
    int value;
    if (constEval(msg, value)) {
      return env.intLit(value);
    }
  }
  #endif // 0
  
  return ivcgen(env, path);
}


Expression * owner E_boolLit::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);
  return clone();
}

Expression * owner E_intLit::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);
  return clone();
}

Expression * owner E_floatLit::ivcgen(VCEnv &env, int path) 
{ return env.todo(); }

Expression * owner E_stringLit::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);

  #if 0
  // make a new variable to stand for this string's address
  Expression * owner addr = env.freshVariable("str", stringc << "address of " << quoted(s));
  env.addDistinct(addr);

  // expression which stands for the object as a whole
  Expression * owner object = env.avSel(env.getMem(), addr);

  // assert the length of this object is the length of the string (w/ null)
  env.assume(P_equal(env.avLength(object),
                      env.avInt(strlen(s)+1)), "string literal length");

  // assert that the offset of the first zero is the length - 1
  env.assume(P_equal(env.avFirstZero(object),
                      env.avInt(strlen(s))), "string literal zero");

  // the result of this expression is a pointer to the string's start
  return env.avSub(addr, env.avSub(new AVint(0), env.avWhole()));
  #endif // 0
  
  //return env.todo();
  return env.intLit(6);    // whatever
}


Expression * owner E_charLit::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);
  return env.todo();
  //return env.grab(new AVint(c));
}


Expression * owner E_this::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);
  return env.todo();
}


//Expression * owner E_structLit::ivcgen(VCEnv &env, int path) { return avTodo(); }

Expression * owner E_variable::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);

  if (var && ext(var)->hasVAttr(VA_PARAMBYREF)) {
    env.error(name->loc, "read `parambyref' variable address");
  }

  if (var && ext(var)->hasVAttr(VA_ADDRTAKEN)) {
    if (!env.inPredicate) {
      env.error(name->loc, "raw access to `addrtaken' variable");
    }
    else {
      // HACK: since I don't have '&' in my VC syntax, just let
      // the unadorned name stand for its object id
    }
  }

  if (0==strcmp(name->getName(), "global_inv")) {
    // make the list of globals be the arguments to a function call node
    return new E_funCall(clone(), env.globalVarList());
  }

  #if 0
  if (var->type->isFunctionType()) {
    // this is the name of a function.. let's say for each global
    // function we have a logic variable of the same name which
    // represents its address
    return env.grab(new AVvar(var->name, "address of global function"));
  }

  if (!env.isMemVar(var)) {
    // ordinary variable: lval of that name
    return env.avLval(var, env.avWhole());
  }
  else {
    // memory variable: lval into 'mem' variable
    return env.avLval(env.mem, env.avSub(env.getMemVarAddr(var), env.avWhole()));
  }
  #endif // 0

  // get name, using type-checking results if available
  StringRef n = var? var->name : name->getName();
  DeclFlags df = var? var->flags : DF_NONE;

  // name of a function?
  if (type->asRval()->isFunctionType()) {
    return env.makeVar(stringc << n << "_address");
  }

  if (env.inPreState &&
      !(df & (DF_ENUMERATOR | DF_UNIVERSAL | DF_EXISTENTIAL)) &&
      (!var || !var->type->isConst())) {     // ugly
    return env.makeVar(stringc << n << "_pre");
  }
  
  if (df & DF_ENUMERATOR) {
    if (0==strcmp(n, "true")) {
      // the programmer's use of true/false need to be seen as 0/1
      return env.intLit(1);
    }
    if (0==strcmp(n, "false")) {
      return env.intLit(0);
    }
  }

  return env.makeVar(n);
}


#if 0
void deadenOwnerState(VCEnv &env, Expression * owner value, char const *why)
{
  if (!value->isAVlval()) {
    // not an lval; probably either NULL, or the return value
    // from an owner-returning function; it will have to be up
    // to the type system to prevent me from losing a reference
    // this way
    trace("owner") << "couldn't deaden rval: " << why << endl;
    return;
  }

  AVlval *lval = value->asAVlval();

  // modify source's 'state' field
  trace("owner") << "deadening state: " << why << endl;
  env.setLval(lval->progVar,
              env.avAppendIndex(lval->offset, env.avOwnerField_state()),
              env.avOwnerState_dead());
}
#endif // 0


// for debugger
char *exprToString(Expression *e)
{
  static char buf[200];
  strncpy(buf, e->asString().c_str(), 199);
  buf[199] = 0;
  return buf;
}

                                                               
// yield a list of owner pointers to expressions representing the
// result of vcgen'ing the argument list; the "Pure" means that
// we assume there is only one path (and no side effects)
FakeList<ArgExpression> *vcgenPureArgs(VCEnv &env, FakeList<ArgExpression> *args)
{
  // recursive to preserve list order
  if (!args) {
    return args;
  }
  else {
    return vcgenPureArgs(env, args->butFirst())
           ->prepend(new ArgExpression(args->first()->expr->vcgen(env, 0 /*path*/)));
  }
}


Expression * owner E_funCall::ivcgen(VCEnv &env, int path)
{
  trace("vcgen") << "E_funCall::vcgen: " << asString() << endl;

  // match_expr
  //   metaVar($n)     // the user-invoked metaVar function on a variable
  // translate {
  //   metaVar($n)
  // }
  if (path==0 &&
      func->isE_variable() &&
      func->asE_variable()->var->name == env.str_metaVar &&
      args->nth(0)->expr->isE_variable()) {
    Variable *var = args->nth(0)->expr->asE_variable()->var;

    return env.makeVar(stringc << var->name << "_meta");
  }

  // match_expr
  //   fresh_value()
  // translate {
  //   introduce $n;    // new name, fresh value
  //   return $n;
  // }
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_fresh_value) {
    string fresh = stringc << "temp" << env.tempCt++;
    env << "    introduce " << fresh << ";\n";
    return env.makeVar(fresh);
  }

  // match_expr
  //   CLT_between(x,y)
  // translate {
  //   return CLT_between(x,y);
  // }
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_CLT_between) {
    return new E_funCall(func->clone(), vcgenPureArgs(env, args));
  }

  // fn symbol application?
  if (env.inPredicate) {
    xassert(path == 0);

    // assume the function name is an E_variable
    return new E_funCall(func->clone(), vcgenPureArgs(env, args));
  }

  if (path==0 &&
      func->isE_variable() &&
      func->asE_variable()->var->name == env.str_memset) {
    // TODO: this really only makes sense when 'e_val' is 0 ...

    PointerType *pt = args->nth(0)->expr->type->asRval()->asPointerType();

    Expression * owner e_ptr = args->nth(0)->expr->vcgen(env, 0 /*path*/);
    Expression * owner e_val = args->nth(1)->expr->vcgen(env, 0 /*path*/);

    if (args->nth(2)->expr->isBinary(BIN_MULT)) {
      Expression * owner e_length = args->nth(2)->expr->asE_binary()->e1->vcgen(env, 0 /*path*/);

      if (pt->atType->isCompoundType()) {
        ECompoundType *ct = ext(pt->atType->asCompoundType());
        if (!ct->flatMap) {
          // match_expr
          //   memset($e_ptr, $e_val, $e_length * sizeof(struct $n))
          // where
          //   type_eq($e_ptr, struct $n *);
          //   struct $n uses the split model;
          // translate {
          //   prove forall(i). 0<=i<$e_length ==> tag[$e_ptr+i] == tagName(struct $n);
          //   for_each_field $f of struct $n {
          //     $f := updflatrange($f, $e_ptr, $e_ptr+$e_length, $e_val);
          //   }
          // }
          env << "    prove \"tag checks on memset[]\"\n"
              << "      forall(i). 0<=i<(" << e_length->tovc() << ") ==>\n"
              << "        tag[" << e_ptr->tovc() << "+i] == " << ct->name << "_tag;\n"
              ;
          FOREACH_OBJLIST(Variable, ct->fieldSlots, iter) {
            StringRef f = iter.data()->name;
            env << "    " << f << " := updflatrange("
                << f << ", "
                << e_ptr->tovc() << ", "
                << e_ptr->tovc() << " + " << e_length->tovc() << ", "
                << e_val->tovc() << ");\n"
                ;
          }
         
          delete e_length;
          delete e_val;
          delete e_ptr;
          return env.intLit(0);      // no return value
        }
      }

      // match_expr
      //   memset($e_ptr, $e_val, $e_length * sizeof($t_elt))
      // where
      //   type_eq($e_ptr, $t_elt *)
      // translate {
      //   mem := mem{$e_ptr, [0 ... $e_length] := $e_val};
      //   meta := meta{$e_ptr, [0 ... $e_length] := live};
      // }

      //Expression *width = args->nth(2)->expr->asE_binary()->e2;
      // TODO: check that type in 'width' matches what I expect

      env.assignDel("mem",
        apply5(env.makeVar("updrange"),
               env.makeVar("mem"),
               e_ptr->clone(),
               env.intLit(0),
               e_length->clone(),
               e_val));
      env.assignDel("meta",
        apply5(env.makeVar("updrange"),
               env.makeVar("meta"),
               e_ptr,
               env.intLit(0),
               e_length,
               env.makeVar("live")));

      // e_ptr, e_val, and e_length consumed above
      return env.intLit(0);      // no return value
    }

    if (args->nth(2)->expr->isE_sizeofType() &&
        args->nth(0)->expr->type->asRval()->isPointerType()) {
      if (pt->atType->isCompoundType()) {
        ECompoundType *ct = ext(pt->atType->asCompoundType());

        env << "    prove \"tag check on memset\"\n"
            << "      tag[" << e_ptr->tovc() << "] == " << ct->name << "_tag;\n";

        if (ct->flatMap) {
          // match_expr
          //   memset($e_ptr, $e_val, sizeof(struct $n))
          // where
          //   type_eq($e_ptr, struct $n *);
          //   struct $n uses the flat model;
          // translate {
          //   mem := mem{$e_ptr, [0 ... $e_length] := $e_val};
          //   meta := meta{$e_ptr, [0 ... $e_length] := live};
          // }

          string size = stringc << ct->name << "_size";

          env.assignDel("mem",
            apply5(env.makeVar("updrange"),
                   env.makeVar("mem"),
                   e_ptr->clone(),
                   env.intLit(0), env.makeVar(size),
                   e_val));
          env.assignDel("meta",
            apply5(env.makeVar("updrange"),
                   env.makeVar("meta"),
                   e_ptr,
                   env.intLit(0), env.makeVar(size),
                   env.makeVar("live")));
        }

        else {
          // match_expr
          //   memset($e_ptr, $e_val, sizeof(struct $n))
          // where
          //   type_eq($e_ptr, struct $n *);
          //   struct $n uses the split model;
          // translate {
          //   prove "memset" tag[$e1] == tagName(struct $n);
          //   for_each_field $f of struct $n {
          //     $f := $f{$e_ptr := $e_val};
          //   }
          // }
          FOREACH_OBJLIST(Variable, ct->fieldSlots, iter) {
            StringRef f = iter.data()->name;
            env << "    " << f << " := " << f << "{" << e_ptr->tovc()
                << " := " << e_val->tovc() << "};\n";
          }
          delete e_val;
          delete e_ptr;
        }

        return env.intLit(0);      // no return value
      }
    }
    delete e_val;
    delete e_ptr;
  }

  // match_expr
  //   free($e_arr[$e_idx])
  // translate {
  //   prove "precondition of free" {
  //     mem[$e_arr,$e_idx] != 0;
  //     meta[$e_arr,$e_idx] == live;
  //   }
  //   meta := meta{$e_arr,$e_idx := dead};
  // }
  if (path==0 &&
      func->isE_variable() &&
      func->asE_variable()->var->name == env.str_free &&
      args->nth(0)->expr->isE_deref()) {
    Expression *ptr = args->nth(0)->expr->asE_deref()->ptr;
    if (ptr->isBinary(BIN_PLUS)) {
      Expression *arr = ptr->asE_binary()->e1;
      Expression *idx = ptr->asE_binary()->e2;
      env.prove(
        e_and(
          e_notEqual(apply2(env.makeVar("sel"),
                            env.makeVar("mem"),
                            apply2(env.makeVar("ptr"),
                                   arr->vcgen(env, path),
                                   idx->vcgen(env, path))),
                     env.intLit(0)),
          e_equal(apply2(env.makeVar("sel"),
                         env.makeVar("meta"),
                         apply2(env.makeVar("ptr"),
                                arr->vcgen(env, path),
                                idx->vcgen(env, path))),
                  env.makeVar("live"))),
        "precondition of free");
      env.assignDel("meta",
        apply3(env.makeVar("upd"),
               env.makeVar("meta"),
               apply2(env.makeVar("ptr"),
                      arr->vcgen(env, path),
                      idx->vcgen(env, path)),
               env.makeVar("dead")));
      return env.intLit(0);     // no return value
    }
  }

  if (path==0 &&
      func->isE_variable() &&
      func->asE_variable()->var->name == env.str_free) {
    Expression *arg = args->nth(0)->expr;

    // the following patterns are *not* exclusive, so track
    // whether we've used at least one
    bool ok = false;

    // match_expr
    //   free($n)     // free called on a variable
    // translate {
    //   prove "precondition of free" {
    //     $n != 0;
    //     metaVar($n) == live;
    //   }
    //   metaVar($n) := dead;
    // }
    if (arg->isE_variable()) {
      Variable *var = arg->asE_variable()->var;

      env.prove(
        e_and(
          e_notEqual(env.makeVar(var->name),
                     env.intLit(0)),
          e_equal(env.makeVar(stringc << var->name << "_meta"),
                  env.makeVar("live"))),
        "precondition of free");
      env.assignDel(stringc << var->name << "_meta",
                    env.makeVar("dead"));

      ok = true;
    }

    // match_expr
    //   free($e)
    // where
    //   typeof($e) is struct $n *
    // translate {
    //   prove "precondition of free(struct ptr)" {
    //     $e != 0;
    //     tag[$e] == tagName($n);
    //   }
    //   tag := tag{$e := 0};
    // }
    if (arg->type->asRval()->isPointerType() &&
        arg->type->asRval()->asPointerType()->atType->isCompoundType()) {
      ECompoundType *ct = ext(arg->type->asRval()->asPointerType()->atType->asCompoundType());
      Expression * owner e = arg->vcgen(env, path);

      env << "    prove \"precondition of free(struct ptr)\" {\n"
          << "      " << e->tovc() << " != 0;\n"
          << "      " << env.hasClassTag(ct, e->tovc()) << ";\n"
          //<< "      tag[" << e->tovc() << "] == " << ct->name << "_tag;\n"
          << "    }\n"
          << "    tag := tag{" << e->tovc() << " := 0};\n"
          ;

      delete e;
      ok = true;
    }

    if (ok) {
      // matched at least one of the patterns
      return env.intLit(0);     // no return value
    }
  }

  // match_expr
  //   kfree($e)
  // translate {
  //   prove "precondition of kfree"
  //     tag[$e] != 0;
  //   tag := tag{$e := 0};
  // }
  if (path==0 &&
      func->isE_variable() &&
      func->asE_variable()->var->name == env.str_kfree) {
    Expression * owner e = args->nth(0)->expr->vcgen(env, path);

    env << "    prove \"precondition of kfree\"\n"
        << "      tag[" << e->tovc() << "] != 0;\n"
        << "    tag := tag{" << e->tovc() << " := 0};\n"
        ;

    delete e;
    return env.intLit(0);     // no return value
  }

  // match_expr
  //   memmove($eDestObj->$fDest + $eDest,       // (*$eDestObj).$fDest + $eDest
  //           $eSrcObj->$fSrc + $eSrc,          // (*$eSrcObj).$fSrc + $eSrc
  //           $eElts * sizeof($t))
  // or
  //   memcpy(...)
  // where
  //   type of $fDest is $t[], and
  //   type of $fSrc is $t[]
  // translate {
  //   prove "memmove: dest range check"
  //     0 <= $eDest <= $eDest+$eElts <= arraySize($fDest);
  //   prove "memmove: src range check"
  //     0 <= $eSrc  <= $eSrc+$eElts  <= arraySize($fSrc);
  //   if (func is memcpy && $fDest==$fSrc) {
  //     prove "memcpy: non-overlapping"
  //       $eDestObj != $eSrcObj       ||
  //       $eDest+$eElts <= $eSrc      ||
  //       $eSrc+$eElts <= $eDest      ;
  //   }
  //   $fDest := $fDest{$eDestObj :=
  //     copyrange($fDest[$eDestObj], $eDest, $fSrc[$eSrcObj], $eSrc, $eElts)};
  // }
  if (path==0 &&
      func->isE_variable() &&
      (func->asE_variable()->var->name == env.str_memmove ||
       func->asE_variable()->var->name == env.str_memcpy) &&
      args->count() == 3 &&
      args->nth(0)->expr->isBinary(BIN_PLUS) &&
      args->nth(1)->expr->isBinary(BIN_PLUS) &&
      args->nth(2)->expr->isBinary(BIN_MULT)) {
    E_binary *arg0 = args->nth(0)->expr->asE_binary();
    E_binary *arg1 = args->nth(1)->expr->asE_binary();
    E_binary *arg2 = args->nth(2)->expr->asE_binary();

    if (arg0->e1->isE_fieldAcc() &&
        arg1->e1->isE_fieldAcc() &&
        arg2->e2->isE_sizeofType()) {
      E_fieldAcc *fieldAcc0 = arg0->e1->asE_fieldAcc();
      E_fieldAcc *fieldAcc1 = arg1->e1->asE_fieldAcc();
      E_sizeofType *sizeof2 = arg2->e2->asE_sizeofType();

      if (fieldAcc0->obj->isE_deref() &&
          fieldAcc1->obj->isE_deref() &&
          fieldAcc0->field->type->isArrayType() &&
          fieldAcc0->field->type->isArrayType()) {
        E_deref *deref0 = fieldAcc0->obj->asE_deref();
        E_deref *deref1 = fieldAcc1->obj->asE_deref();
        ArrayType *at0 = fieldAcc0->field->type->asArrayType();
        ArrayType *at1 = fieldAcc1->field->type->asArrayType();

        if (at0->eltType->equals(at1->eltType) &&
            at0->eltType->equals(sizeof2->atype->getType())) {
          StringRef funcName = func->asE_variable()->var->name;

          Expression * owner eDestObj = deref0->ptr->vcgen(env, 0);
          string fDest = env.fieldSlotName(fieldAcc0->field);
          Expression * owner eDest = arg0->e2->vcgen(env, 0);

          Expression * owner eSrcObj = deref1->ptr->vcgen(env, 0);
          string fSrc = env.fieldSlotName(fieldAcc1->field);
          Expression * owner eSrc = arg1->e2->vcgen(env, 0);

          Expression * owner eElts = arg2->e1->vcgen(env, 0);

          // debugging aid
          env << "    // " << funcName
              << ": eDestObj=" << eDestObj->exprToString()
              << " fDest=" << fDest
              << " eDest=" << eDest->exprToString()
              << " eSrcObj=" << eSrcObj->exprToString()
              << " fSrc=" << fSrc
              << " eSrc=" << eSrc->exprToString()
              << " eElts=" << eElts->exprToString()
              << "\n";

          env << "    prove \"" << funcName << ": dest range check\"\n"
              << "      0 <= " << eDest->tovc() << " <= "
              << eDest->tovc() << " + " << eElts->tovc() << " <= "
              << at0->size << ";\n";

          env << "    prove \"" << funcName << ": src range check\"\n"
              << "      0 <= " << eSrc->tovc() << " <= "
              << eSrc->tovc() << " + " << eElts->tovc() << " <= "
              << at1->size << ";\n";

          if (funcName == env.str_memcpy &&
              fieldAcc0->field == fieldAcc1->field) {
            env << "    prove \"memcpy: non-overlapping\"\n"
                << "      " << eDestObj->tovc() << " != " << eSrcObj->tovc() << " ||\n"
                << "      " << eDest->tovc() << "+" << eElts->tovc()
                << " <= " << eSrc->tovc() << " ||\n"
                << "      " << eSrc->tovc() << "+" << eElts->tovc() 
                << " <= " << eDest->tovc() << ";\n";
          }

          env << "    " << fDest << " := " << fDest << "{"
              << eDestObj->tovc() << " := copyrange(" << fDest << "["
              << eDestObj->tovc() << "], " << eDest->tovc() << ", "
              << fSrc << "[" << eSrcObj->tovc() << "], "
              << eSrc->tovc() << ", " << eElts->tovc() << ")};\n";
              
          delete eDestObj;
          delete eDest;
          delete eSrcObj;
          delete eSrc;
          delete eElts;

          return env.intLit(0);
        }
      }
    }
  }

  // check for specific functions that we expect to model directly
  if (func->isE_variable()) {
    Variable *v = func->asE_variable()->var;
    if (v->name == env.str_malloc ||
        v->name == env.str_free ||
        v->name == env.str_kfree) {
      cout << "warning: unmodeled call to " << v->name << "\n";
    }
  }

  // evaluate the function, but then discard it; we'll only use its type
  int subexpPaths = func->numPaths1();
  env.discard(func->vcgen(env, path % subexpPaths));
  path = path / subexpPaths;

  // get type of called function
  EFunctionType *ft;
  {
    Type *t = func->type->asRval();
    if (t->isPointerType()) {
      t = t->asPointerTypeC()->atType;
    }
    ft = ext(t->asFunctionType());
  }

  if (ft->params.count() != args->count()) {
    env << "    // TODO: function call with num args != num params\n";
    return env.todo();
  }

  // if the function has a return value, introduce a new name now
  // for that value
  string fresh = stringc << "temp" << env.tempCt++;
  if (ft->result) {
    env << "    introduce " << fresh << ";      // return value of call\n";
  }

  // introduce parambyref parameter variables
  {
    SFOREACH_OBJLIST(Variable, ft->params, iter) {
      EVariable const *v = extC(iter.data());
      if (v->hasVAttr(VA_PARAMBYREF)) {
        env << "    var " << v->name << ";      // parambyref\n";
      }
    }
  }

  // open a block in which to bind the names that the
  // function pre- and postcondition refer to
  // NOTE: the prefix string "call:" is special!
  env << "    block \"call: " << func->exprToString() << "\" {\n";

  // evaluate the arguments, binding each argument to its
  // parameter name; these bindings should be processed
  // in parallel (each binding not seeing the others) by
  // vclang, though in fact that's not done currently (bug)
  env << "      // bind params to args\n";

  // track argbyrefs that have been passed
  SObjList<Variable> passedArgbyref;
  SObjList<Variable> receivedParambyref;

  // simultaneously walk params and args
  SObjListIterNC<Variable> paramIter(ft->params);
  FAKELIST_FOREACH_NC(ArgExpression, args, argIter) {
    EVariable *param = ext(paramIter.data());    
    Expression *arg = argIter->expr;

    if (param->hasVAttr(VA_PARAMBYREF)) {
      // argument must be the address of a VA_ARGBYREF variable
      if (arg->isE_addrOf() &&
          arg->asE_addrOf()->expr->isE_variable() &&
          ext(arg->asE_addrOf()->expr->asE_variable()->var)->hasVAttr(VA_ARGBYREF)) {
        Variable *v = arg->asE_addrOf()->expr->asE_variable()->var;
        if (passedArgbyref.contains(v)) {
          env.error("passed same `argbyref' more than once");
        }

        // hack: assignment is what's going on, but my hack for
        // call scopes makes that not work (improper balancing
        // of BG_PUSH and BG_POP)..
        env << "      assume " << param->name << " == " << v->name << ";\n";

        passedArgbyref.prepend(v);
        receivedParambyref.prepend(param);
      }
      // this case is largely DUPLICATED from the one just above; we allow
      // a by-ref parameter to be passed again to another by-ref parameter
      else if (arg->isE_variable() &&
               ext(arg->asE_variable()->var)->hasVAttr(VA_PARAMBYREF)) {
        Variable *v = arg->asE_variable()->var;
        if (passedArgbyref.contains(v)) {
          env.error("passed same `parambyref' more than once");
        }

        // hack: assignment is what's going on, but my hack for
        // call scopes makes that not work (improper balancing
        // of BG_PUSH and BG_POP)..
        env << "      assume " << param->name << " == " << v->name << ";\n";

        passedArgbyref.prepend(v);
        receivedParambyref.prepend(param);
      }
      else {
        env.error("argument for `parambyref' parameter must be &`argbyref'"
                  " or another `parambyref'");
      }
    }

    else {
      // evaluate
      subexpPaths = arg->numPaths1();
      Expression * owner val = arg->vcgen(env, path % subexpPaths);
      path = path / subexpPaths;

      // bind
      if (param->name) {     // if no name, pre/post can't refer to it
        env << "      define " << param->name << " = " << val->tovc() << ";\n";
      }
      delete val;
    }

    paramIter.adv();
  }
  xassert(path == 0);
  xassert(paramIter.isDone());    // otherwise mismatched number of params/args

  // assert the precondition, of which global_inv is a part
  env << "      prove \"call precondition: global_inv\"\n"
      << "        global_inv(" << env.globalVarNameList() << ");\n";
  if (ft->precondition) {
    IN_PREDICATE(env);
    Expression * owner vc = ft->precondition->vcgenPred(env, 0 /*path*/);
    env << "      prove \"call precondition\"\n"
        << "        " << vc->tovc() << ";\n";
    delete vc;
  }

  // now, bind the current values of the globals and params such
  // that they'll be seen as the pre-state values when we evaluate
  // the postcondition
  env << "      define {           // pre-state w.r.t. upcoming postcondition\n";
  SObjList<Variable> *arr[2] = { &env.globals, &(ft->params) };
  for (SObjListMultiIter<Variable> iter(arr, 2);
       !iter.isDone(); iter.adv()) {
    Variable const *v = iter.data();
    if (v->name) {
      env << "        " << v->name << "_pre = " << v->name << ";\n";
    }
  }
  env << "      }\n";

  // freshen the global variables, as they may be changed by the call;
  // the postcondition is free to claim that some have not changed
  env << "      freshen " << env.globalVarNameList() << ";\n";

  // freshen parambyref parameter variables
  {
    SFOREACH_OBJLIST(Variable, ft->params, iter) {
      EVariable const *v = extC(iter.data());
      if (v->hasVAttr(VA_PARAMBYREF)) {
        env << "      freshen " << v->name << ";      // parambyref\n";
      }
    }
  }

  // bind the name 'result', since that's what the postcondition refers to
  if (ft->result) {
    env << "      introduce result;\n";
  }

  // finally, assume the postcondition
  env << "      // assume postcondition\n";
  env << "      assume global_inv(" << env.globalVarNameList() << ");\n";
  if (ft->getPostcondition(env)) {
    IN_PREDICATE(env);
    Expression * owner vc = ft->getPostcondition(env)->vcgenPred(env, 0 /*path*/);
    env << "      assume " << vc->tovc() << ";\n";
    delete vc;
  }

  // communicate the return value to the caller
  if (ft->result) {
    env << "      assume " << fresh << " == result;\n";
  }

  env << "    }\n";

  // similarly, copy passbyref values
  {
    SObjListIterNC<Variable> argIter(passedArgbyref);
    SObjListIterNC<Variable> paramIter(receivedParambyref);
    while (!argIter.isDone()) {
      env << "    " << argIter.data()->name << " := "
          << paramIter.data()->name << ";\n";
          
      argIter.adv();
      paramIter.adv();
    }
  }

  return env.makeVar(fresh);

  #if 0
  env << "    // TODO: function call\n";
  return env.todo();
  #endif // 0

  #if 0
  // get type of called function
  FunctionType const &ft = func->type->asRval()->asFunctionTypeC();

  // evaluate the argument expressions, taking into consideration
  // which path to follow
  SObjList<AbsValue> argExps;
  {
    // 'func'
    int subexpPaths = func->numPaths1();
    func->vcgen(env, path % subexpPaths);   // ignore value; my analysis only uses type info
    path = path / subexpPaths;

    // args
    ObjListIter<FunctionType::Param> paramIter(ft.params);
    FOREACH_ASTLIST(Expression, args, iter) {
      Expression *arg = iter.data();

      // evaluate
      subexpPaths = arg->numPaths1();
      Expression * owner val = arg->vcgen(env, path % subexpPaths);

      // if 'val' is an lval, grab the current value, so that the
      // owner modification below will not affect what the function
      // precondition sees
      Expression * owner valToPass = env.rval(val);

      if (arg->type->asRval()->isOwnerPtr() &&
          paramIter.data()->type->isOwnerPtr()) {
        // OWNER: when an owner pointer is passed to an owner param,
        // the argument becomes dead
        deadenOwnerState(env, val,
                         stringc << "passed to owner param: " << arg->asString());
      }
      // TODO: add typechecking rules to prevent passing a serf pointer
      // into an owner param

      argExps.prepend(valToPass);
      path = path / subexpPaths;

    }
    xassert(path < subexpPaths);

    argExps.reverse();      // when it's easy to stay linear..
  }

  // -------------- modify environment for pre/post ----------------
  // if we are calling the same function we're in (recursively), then
  // the parameter bindings will clobber the caller params; so save
  // them before setting, and restore at the end (use of AbsVariable
  // is just because it happens to be a <Variable,AbsValue> pair)
  ObjStack<AbsVariable> prevValues;

  // the pre- and postconditions get evaluated to an abstract value in
  // an environment augmented to include parameters (and in the case
  // of the postcondition, 'result')
  SObjListMutator<AbsValue> argExpIter(argExps);
  FOREACH_OBJLIST(FunctionType::Param, ft.params, paramIter) {
    Variable const *param = paramIter.data()->decl;
    Expression * owner value = argExpIter.data();

    if (env.hasBinding(param)) {
      // conflict between callee and caller param; remember the old value
      prevValues.push(new AbsVariable(param, env.get(param),
                                      false /*memvar; not used*/));
    }

    // bind the names to the expressions that are passed
    env.set(param, value);

    argExpIter.adv();
  }

  // let pre_mem = mem
  //newEnv.set(env.str("pre_mem"), newEnv.getMem());

  // ----------------- prove precondition ---------------
  IN_PREDICATE(env);

  #if 0  // until pre/post implemented
  if (ft.precondition) {
    // include precondition bindings in the environment
    FOREACH_ASTLIST(Declaration, ft.precondition->decls, iter) {
      iter.data()->vcgen(env);
    }

    // finally, interpret the precondition to arrive at a predicate to
    // prove
    Predicate *predicate = ft.precondition->expr->vcgenPred(env, 0);
    xassert(predicate);      // tcheck should have verified we can represent it

    // prove this predicate
    env.prove(predicate, stringc << "precondition of call: " << toString());
  }
  #endif // 0

  // -------------- conservatively destroy state ------------
  // NOTE: by keeping the environment from before, we are interpreting
  // all references to parameters as their values in the *pre* state
  env.forgetAcrossCall(this);

  // ----------------- assume postcondition ---------------
  // make a new variable to stand for the value returned
  Expression * owner resultVal = NULL;
  if (!ft.retType->isVoid()) {
    resultVal = env.freshVariable("result",
                  stringc << "function call result: " << toString());

    // add this to the environment so if the programmer talks about
    // the result, it will state something about the result variable
    env.set(ft.result, resultVal);
  }

  #if 0  // until pre/post implemented
  if (ft.postcondition) {
    // evaluate it
    Predicate *predicate = ft.postcondition->expr->vcgenPred(env, 0);
    xassert(predicate);

    // assume it
    env.assume(predicate, stringc << "postcondition from call: " << toString());
  }
  #endif // 0

  // restore local parameter values
  while (prevValues.isNotEmpty()) {
    Owner<AbsVariable> av(prevValues.pop());
    env.set(av->decl, av->value);
  }

  // result of calling this function is the result variable
  return resultVal;
  #endif // 0
}


Expression * owner E_constructor::ivcgen(VCEnv &env, int path)
  { return env.todo(); }


EVariable *getSynthField(EVariable *v, Variable *field)
{
  FOREACH_OBJLIST_NC(EVariable, v->synthFields, sfiter) {
    if (sfiter.data()->originalStructField == field) {
      // found the right synth field
      return sfiter.data();
    }
  }                            

  return NULL;     // not found
}

Expression * owner E_fieldAcc::ivcgen(VCEnv &env, int path)
{
  if (obj->type->asRval()->isCompoundType()) {
    ECompoundType *ct = ext(obj->type->asRval()->asCompoundType());
    if (obj->isE_deref() ) {
      Expression * owner e = obj->asE_deref()->ptr->vcgen(env, path);

      if (!env.inPredicate) {
        env << "    prove \"$L: ptr->field read: " << exprToString() << "\"\n"
            << "      " << env.hasClassTag(ct, e->tovc()) << ";\n";
            //<< "      tag[" << e->tovc() << "] == " << ct->name << "_tag;\n";
      }

      if (ct->flatMap) {
        // match_expr
        //   $e->$f                       // (*$e1).$f
        // where
        //   match_type $e with struct $n *
        //   struct $n uses the flat model
        // translate {
        //   prove tag[$e] == tagName(struct $n);
        //   return mem[$e, fieldName($f)];
        // }
        return apply2(env.makeVar("sel"),
                      env.makeVar("mem"),
                      apply2(env.makeVar("ptr"),
                             e,
                             env.makeVar(env.fieldSlotName(field))
                            ));
      }
      else {
        // match_expr
        //   $e->$f                       // (*$e1).$f
        // where
        //   match_type $e with struct $n *
        //   struct $n uses the split model
        // translate {
        //   prove tag[$e] == tagName(struct $n);
        //   return fieldName($f)[$e];
        // }
        return apply2(env.makeVar("sel"),
                      env.makeVar(env.maybePre(env.fieldSlotName(field))),
                      e);
      }
    }

    if (obj->isE_variable()) {
      EVariable *v = ext(obj->asE_variable()->var);
      if (v->hasVAttr(VA_ADDRTAKEN)) {
        // match_expr
        //   $n.$f
        // where
        //   $n has attribute `addrtaken'
        // translate {
        //   return $f[$n];
        // }
        return apply2(env.makeVar("sel"),
                      env.makeVar(env.maybePre(env.fieldSlotName(field))),
                      env.makeVar(v->name));
      }
      else {
        // match_expr
        //   $n.$f
        // where
        //   $n does not have attribute `addrtaken'
        // translate {
        //   return instanceFieldName($n,$f)
        // }
        EVariable *synth = getSynthField(v, field);
        if (synth) {
          return env.makeVar(synth->name);
        }
      }
    }
  }

  return env.todo();
  
  #if 0
  // get val for the object
  Expression * owner objVal = obj->vcgen(env, path);

  // and field index for the field
  Expression * owner fieldIndex = env.avInt(field->index);

  // is it an lvalue?  (it almost always is)
  if (objVal->isAVlval()) {
    AVlval *objLval = objVal->asAVlval();

    // append this field's index as an offset
    objLval->offset = env.avAppendIndex(objLval->offset, fieldIndex);
    return objLval;
  }

  else {
    // unusual, but not impossible; just select within the rvalue
    return env.avSel(objVal, fieldIndex);
  }
  #endif // 0

  #if 0   // old
    if (obj->isE_variable()) {
      Variable *var = obj->asE_variable()->var;
      if (!var->hasAddrTaken()) {
        // modeled as an unaliased tuple
        // retrieve named piece of the (whole) structure value
        return env.avGetElt(env.avInt(field->index), env.get(var));
      }
    }

    if (obj->isE_deref()) {
      E_deref *deref = obj->asE_derefC();

      // partially DUPLICATED from E_assign::vcgen; reflective of general
      // problem with lvalues

      // get an abstract value for the address of the object being modified
      Expression * owner addr = deref->ptr->vcgen(env, path);

      if (0==strcmp(field->compound->name, "OwnerPtrMeta")) {
        // OWNER: model the 'OwnerPtrMeta' struct as a tuple (obviously
        // I need to generalize the decision of what to model as a tuple)

        // get the tuple value
        Expression * owner tuple =
           env.avSelect(
             env.getMem(),
             env.avObject(addr),
             env.avOffset(addr));

        // select the field I want
        return env.avGetElt(env.avInt(field->index), tuple);
      }

      #if 0    // old: from before I used + to form offset
      if (!env.inPredicate) {
        // the address should have offset 0
        env.prove(P_equal(env.avOffset(addr), env.avInt(0)),
                  "object field access at offset 0");

        // TODO: assign type tags to allocated regions, and prove here
        // that the type tag is correct
      }

      // since we've proved the offset is 0, assume it (this really for
      // the predicate case, where we *don't* prove it, so it acts like
      // an unstated assumption about a pointer to which a field
      // accessor is applied)
      env.assume(P_equal(env.avOffset(addr), env.avInt(0)),
                  "implicit offset=0 assumption for field access pointer");
      #endif // 0

      // read from memory
      return env.avSelect(env.getMem(),
        env.avObject(addr),                // object being accessed
        env.avSum(env.avOffset(addr),      // offset is pointer's offset (usually 0)
                  env.get(field->decl)));  //  + symbolic offset of the field
    }

    cout << "TODO: unhandled structure field access: " << toString() << endl;
    return env.freshVariable("field",
             stringc << "structure field access: " << toString());
  #endif // 0
}


Expression * owner E_sizeof::ivcgen(VCEnv &env, int path)
{
  // whether 'expr' has side effects or not is irrelevant; they
  // won't be executed at run time; we are only interested in the type
  return env.intLit(size);
}

Expression * owner E_unary::ivcgen(VCEnv &env, int path)
{
  return new E_unary(op, expr->vcgen(env, path));
}


Expression * owner E_effect::ivcgen(VCEnv &env, int path)
{
  BinaryOp binop = (op==EFF_POSTINC || op==EFF_PREINC)? BIN_PLUS : BIN_MINUS;

  if (expr->numPaths1() != 1) {
    // the current handling would evaluate the expr twice...
    cout << "TODO: unhandled effect operator\n";
    return env.todo();
  }


  Expression * owner current = expr->vcgen(env, path);

  if (op == EFF_POSTINC || op == EFF_POSTDEC) {
    // yield value of expression before the inc/dec
    string fresh = stringc << "temp" << env.tempCt++;
    env << "    define " << fresh << " = " << current->tovc() << ";\n";

    delete env.assignmentExpr(expr, path,
      new E_binary(current, binop, env.intLit(1)));
    return env.makeVar(fresh);
  }
  else {
    // yield value of expression after the inc/dec
    xassert(op == EFF_PREINC || op == EFF_PREDEC);
    return env.assignmentExpr(expr, path,
      new E_binary(current, binop, env.intLit(1)));
  }
}


#if 0
Expression * owner coerceArrayToPointer(VCEnv &env, Expression * owner v)
{
  xassert(v->isAVlval());     // must be a reference to the array
  AVlval *lval = v->asAVlval();
  return env.avAppendIndex(lval->offset, env.avInt(0));
}    
#endif // 0


// usually this is op(val1,val2), but it might be something else
// depending on the type of 'e1' (the expression that yielded 'val1');
// both expression values may be rvals initially
Expression * owner arithmeticCombinator(
  VCEnv &env, Expression * owner val1, Expression *e1,
  BinaryOp op, Expression * owner val2)
{
  if (op == BIN_COMMA) {
    // discard LHS argument
    delete val1;
    return val2;
  }

  #if 0
  if (e1->type->asRval()->isArrayType()) {
    // coerce to pointer
    val1 = coerceArrayToPointer(env, val1);
  }
  // make them rvals now, after the array coercion opportunity
  #endif // 0

  if (e1->type/*hack*/ && 
      e1->type->asRval()->isPointerType() &&
      op==BIN_PLUS) {
    // don't do this if it's a pointer to "Scull_dev"
    Type *at = e1->type->asRval()->asPointerType()->atType;
    if (at->isCompoundType() &&
        0==strcmp(at->asCompoundType()->name, "Scull_Dev")) {
      // don't use the 'ptr' function, just make an ordinary sum
    }
    else {
      // pointer + offset: ptr
      return apply2(env.makeVar("ptr"), val1, val2);
    }
  }

  #if 0
  val1 = env.rval(val1);
  val2 = env.rval(val2);

  if (e1->type->asRval()->isPointerType()) {
    if (op==BIN_PLUS) {
      // use the addOffset function
      return env.avAddOffset(val1, val2);
    }
    if (op==BIN_MINUS) {
      // similar but with negation
      return env.avAddOffset(val1, env.grab(new AVunary(UNY_MINUS, val2)));
    }
    xfailure("bad pointer operation");
  }

  return env.grab(new AVbinary(val1, op, val2));
  #endif // 0

  // '==' or '!=', but *not* in a boolean context
  if (op==BIN_EQUAL || op==BIN_NOTEQUAL) {
    // modeled after '?:' handling

    // this is perhaps a bit too clever.... whatever
    static char const onezero_src[3] = { '0', '1', '0' };
    char const *onezero = (op==BIN_EQUAL? onezero_src : onezero_src+1);

    string r = stringc << "temp" << env.tempCt++;
    env << "    introduce " << r << ";      // result of " << toString(op) << "\n";
    env << "    assume (" << val1->tovc() << "==" << val2->tovc() << ") ==> "
        <<               r << " == " << onezero[1] << ";\n";
    env << "    assume (" << val1->tovc() << "!=" << val2->tovc() << ") ==> "
        <<               r << " == " << onezero[0] << ";\n";
    delete val1;
    delete val2;
    return env.makeVar(r);
  }

  // for now
  return new E_binary(val1, op, val2);
}


Expression * owner E_binary::ivcgen(VCEnv &env, int path)
{
  if (e2->numPaths == 0) {
    // simple side-effect-free binary op
    return arithmeticCombinator(env, e1->vcgen(env, path), e1,
                                op, e2->vcgen(env, 0));
  }

  // consider operator
  if (op==BIN_AND || op==BIN_OR) {
    // evaluate LHS
    int modulus = e1->numPaths1();
    Expression * owner lhs = e1->vcgenPred(env, path % modulus);
    path = path / modulus;

    #if 0   // investigating alternative solution...
    if (e2->numPaths==0) {
      xassert(path==0);

      // e2 will be evaluated in a temporary context, since we
      // aren't using a separate path
      env << "    block \"RHS temp context\" {\n";

      // this context assumes we eval rhs
      env.assumeSense(lhs, op==BIN_AND,
        stringc << "LHS guard given RHS is eval'd " << asString());

      Expression * owner ret = e2->vcgen(env, 0);

      env << "    }\n";
      return ret;
    }
    #endif // 0

    // eval RHS *if* it's followed
    if (path > 0) {
      // we follow rhs; this implies something about the value of lhs:
      // if it's &&, lhs had to be true; false for ||
      env.assumeSense(lhs, op==BIN_AND,
        stringc << "LHS guard given RHS is eval'd " << asString());

      // the true/false value is now entirely determined by rhs
      return e2->vcgen(env, path-1);
    }
    else {
      // we do *not* follow rhs; this also implies something about lhs
      env.assumeSense(lhs, op==BIN_OR,
        stringc << "LHS guard given RHS is *not* eval'd " << asString());

      // it's a C boolean, yielding 1 or 0 (and we know which)
      return env.intLit(op==BIN_OR? 1 : 0);
    }
  }

  else {
    // evaluate LHS
    int modulus = e1->numPaths1();
    Expression * owner lhs = e1->vcgen(env, path % modulus);
    path = path / modulus;

    // non-short-circuit: always evaluate RHS
    return arithmeticCombinator(env, lhs, e1, op, e2->vcgen(env, path));
  }
}


Expression * owner E_addrOf::ivcgen(VCEnv &env, int path)
{
  // match_expr
  //   &*$e
  // translate {
  //   $e
  // }
  if (expr->isE_deref()) {
    return expr->asE_deref()->ptr->vcgen(env, path);
  }

  if (expr->isE_variable()) {
    EVariable *v = ext(expr->asE_variable()->var);
    if (v->hasVAttr(VA_ADDRTAKEN)) {
      // match_expr
      //   &$n
      // where
      //   $n has attribute `addrtaken'
      // translate {
      //   $n            // name is treated like an object id in memory
      // }
      return env.makeVar(v->name);
    }

    //env.warning(stringc << "address of " << v->name);
  }

  return env.todo();

  #if 0
  xassert(path==0);

  Expression * owner exprVal = expr->vcgen(env, path);
  if (exprVal->isAVlval()) {
    AVlval *lval = exprVal->asAVlval();
    xassert(lval->progVar == env.mem);
    return lval->offset;       // pointer value relative to 'mem'
  }

//    if (expr->isE_variable()) {
//      return env.avPointer(env.getMemVarAddr(expr->asE_variable()->var), // obj
//                           new AVint(0));                                // offset
//    }
  else {
    cout << "TODO: unhandled addrof: " << asString() << endl;
    return avTodo();
  }    
  #endif // 0
}


void verifyPointerAccess(VCEnv &env, Expression *expr, Expression * serf addr)
{
  #if 0
  // for a while I was doing ok checking these even inside predicates,
  // but it induces an ordering constraint on the conjuncts in invariants,
  // and that is hard to respect during automatic strengthening
  if (!env.inPredicate) {
    env.prove(P_named2(env.str("okSelOffset"),
                       env.getMem(), addr),
              stringc << "pointer access check: " << expr->asString());

    #if 0     // old
    env.prove(new P_relation(env.avOffset(addr), RE_GREATEREQ,
                             env.avInt(0)),
                             stringc << "pointer lower bound: " << expr->asString());
    env.prove(new P_relation(env.avOffset(addr), RE_LESS,
                             env.avLength(env.avObject(addr))),
                             stringc << "pointer upper bound: " << expr->asString());
    #endif // 0
  }
  #endif // 0
}

Expression * owner E_deref::ivcgen(VCEnv &env, int path)
{
  if (ptr->isE_variable()) {
    EVariable *v = ext(ptr->asE_variable()->var);
    if (v->hasVAttr(VA_PARAMBYREF)) {
      // match_expr
      //   *$n
      // where
      //   $n has attribute VA_PARAMBYREF
      // translate {
      //   return $n;
      // }
      return env.makePVar(v->name);
    }
  }

  if (ptr->isBinary(BIN_PLUS)) {
    E_binary *bin = ptr->asE_binary();

    int modulus = bin->e1->numPaths1();
    int e1Path = path % modulus;
    int e2Path = path / modulus;

    // match_expr
    //   ($e1->$f)[$e2]               // *( ( (*$e1).$f + $e2 )
    // where {
    //   isArrayType($e1->$f);
    //   match_type $e1 with struct $n *;
    // }
    // translate {
    //   prove tag[$e1] == tagName(struct $n);
    //   prove 0 <= $e2 < arraySize($f);
    //   return mem[$e1,($f + $e2)]
    // }
    if (bin->e1->isE_fieldAcc()) {
      E_fieldAcc *acc = bin->e1->asE_fieldAcc();
      if (acc->obj->isE_deref()) {
        if (acc->field->type->isArrayType() &&
            acc->obj->type->asRval()->isCompoundType()) {
          EVariable *field = ext(acc->field);
          ArrayType *at = field->type->asArrayType();
          ECompoundType *ct = ext(acc->obj->type->asRval()->asCompoundType());

          Expression * owner e1 = acc->obj->asE_deref()->ptr->vcgen(env, e1Path);
          Expression * owner e2 = bin->e2->vcgen(env, e2Path);

          env << "    // structure array field read\n";
          env << "    prove " << env.hasClassTag(ct, e1->tovc()) << ";\n";
          //env << "    prove tag[" << e1->tovc() << "] == " << ct->name << "_tag;\n";
          env << "    prove 0 <= " << e2->tovc() << " < " << at->size << ";\n";

          if (ct->flatMap) {
            return apply2(env.makeVar("sel"),
                          env.makeVar("mem"),
                          apply2(env.makeVar("ptr"),
                                 e1,
                                 e_plus(env.makeVar(field->modelVar->name), e2)
                                )
                         );
          }
          else {          
            // return fieldName($f)[$e1][$e2];
            Expression * owner objectArray =
              apply2(env.makeVar("sel"),
                     env.makeVar(env.fieldSlotName(field)),
                     e1);
            return apply2(env.makeVar("sel"),
                          objectArray,
                          e2);
          }
        }
      }
    }

    // match_expr
    //   $n[$e2]             // *($n + $e2)
    // where
    //   typeof($n) == $t [$eSz]      // directly-modeled array
    // translate {
    //   prove "array var read"
    //     0 <= $e2 < $eSz;
    //   return $n[$e2];
    // }
    if (bin->e1->isE_variable() &&
        bin->e1->type->asRval()->isArrayType()) {
      Variable *n = bin->e1->asE_variable()->var;
      int eSz = n->type->asArrayType()->size;
      if (eSz != ArrayType::NO_SIZE) {
        Expression * owner e2 = bin->e2->vcgen(env, e2Path);
        env << "    prove \"array var read: " << exprToString() << "\"\n"
            << "      0 <= " << e2->tovc() << " < " << eSz << ";\n";
        return apply2(env.makeVar("sel"),
                      env.makeVar(n->name),
                      e2);
      }
    }

    // match_expr
    //   $e1[$e2]             // *($e1 + $e2)
    // translate {
    //   prove "array ptr read"
    //     0 <= $e2 < objsize[$e1];
    //   return mem[$e1,$e2];
    // }
    Expression * owner e1 = bin->e1->vcgen(env, e1Path);
    Expression * owner e2 = bin->e2->vcgen(env, e2Path);
    env << "    prove \"array ptr read: " << exprToString() << "\"\n"
        << "      0 <= " << e2->tovc() << " < objsize[" << e1->tovc() << "];\n";
    return apply2(env.makeVar("sel"),
                  env.makeVar("mem"),
                  apply2(env.makeVar("ptr"),
                         e1,
                         e2)
                 );
  }

  Expression * owner addr = env.rval(ptr->vcgen(env, path));

  #if 0
  if (ptr->type->asRval()->isOwnerPtr()) {
    // OWNER: pointer must be in owning state to use
    trace("owner") << "owner pointer being dereferenced for read\n";
    env.prove(P_equal(env.avSel(addr, env.avOwnerField_state()),
                      env.avOwnerState_owning()),
              "owner must be in owner state to dereference");

    // dereferencing an owner needs to actually dereference the 'ptr' field
    addr = env.avSel(addr, env.avOwnerField_ptr());
  }

  // emit a proof obligation for safe access through 'addr'
  verifyPointerAccess(env, this, addr);
  #endif // 0

  return apply2(env.makeVar("sel"), env.makeVar("mem"), addr);

  //return env.avLval(env.mem, addr);

  //return env.avSelect(env.getMem(), env.avObject(addr), env.avOffset(addr));
}


Expression * owner E_cast::ivcgen(VCEnv &env, int path)
{
  // is this a call to malloc with a typecast in front?
  if (path==0 &&
      ctype->getType()->isPointerType() &&
      expr->isE_funCall()) {
    Type *eltType = ctype->getType()->asPointerType()->atType;
    E_funCall *f = expr->asE_funCall();
    if (f->func->isE_variable() &&
        f->func->asE_variable()->var->name == env.str_malloc) {
      Expression *a = f->args->first()->expr;
      if (a->isE_binary() &&
          a->asE_binary()->op == BIN_MULT) {
        Expression *lhs = a->asE_binary()->e1;
        Expression *rhs = a->asE_binary()->e2;
        if (rhs->isE_sizeofType() &&
            rhs->asE_sizeofType()->atype->getType()->equals(eltType)) {
          if (eltType->isCompoundType() &&
              0==strcmp(eltType->asCompoundType()->name, "Scull_Dev")) {
            // match: (Scull_Dev *)malloc($e * sizeof(Scull_Dev))
            // replacement:
            //   define $fresh1 = objct;
            //   objct := objct + $e;
            //   tag := updflatrange(tag, $fresh1, $fresh1+$e, tagName(struct Scull_Dev));
            //   return $fresh1;      // becomes the yielded expression
            string fresh = stringc << "temp" << env.tempCt++;
            Expression * owner e = lhs->vcgen(env, 0);

            env << "    define " << fresh << " = objct;\n";
            env << "    objct := objct + " << e->tovc() << ";\n";
            env << "    tag := updflatrange(tag, " << fresh << ", "
                << fresh << " + " << e->tovc() << ", "
                << "Scull_Dev_tag);\n";

            delete e;
            return env.makeVar(fresh);
          }

          // match: ($t *)malloc($e * sizeof($t))
          // replacement:
          //   define $fresh1 = objct;
          //   objct := objct + 1;
          //   meta := meta{$fresh1,[0..$e] := dead};
          //   objsize := objsize{$fresh1 := $e};
          //   return $fresh1;      // becomes the yielded expression
          string fresh = stringc << "temp" << env.tempCt++;
          env.define(fresh, env.makeVar("objct"), "malloc result");
          env.assign("objct", e_plus(env.makeVar("objct"), env.intLit(1)));
          env.assign("meta",
            env.updrange(env.makeVar("meta"),
                         env.makeVar(fresh),
                         env.intLit(0), lhs->vcgen(env,0),
                         env.makeVar("dead")));
          env.assign("objsize",
            env.upd(env.makeVar("objsize"),
                    env.makeVar(fresh),
                    lhs->vcgen(env,0)));
          return env.makeVar(fresh);
        }
      }

      if (a->isE_sizeofType() &&
          eltType->isCompoundType()) {
        Type *typeInSizeof = a->asE_sizeofType()->atype->getType();
        if (typeInSizeof->equals(eltType)) {
          ECompoundType *ct = ext(eltType->asCompoundType());
          string numFields = stringc << ct->name << "_size";
          string tagName = stringc << ct->name << "_tag";

          // match_expr
          //   (struct $n *)malloc(sizeof(struct $n))
          // translate {
          //   define $fresh1 = objct;
          //   objct := objct + 1;
          //   tag := tag{$fresh1 := tagName(struct $n)};
          //   static_if (struct $n uses flat model) {
          //     meta := meta{$fresh1,[0 ... numFields(struct $n)] := dead};
          //   }
          //   objsize := objsize{$fresh1 := numFields(struct $n)};
          //   return $fresh1;
          // }
          string fresh = stringc << "temp" << env.tempCt++;
          env.define(fresh, env.makeVar("objct"), "malloc result");
          env.assign("objct", e_plus(env.makeVar("objct"), env.intLit(1)));
          env.assign("tag",
            env.upd(env.makeVar("tag"),
                    env.makeVar(fresh),
                    env.makeVar(tagName)));
          if (ct->flatMap) {
            env.assign("meta",
              env.updrange(env.makeVar("meta"),
                           env.makeVar(fresh),
                           env.intLit(0), env.makeVar(numFields),
                           env.makeVar("dead")));
          }
          env.assign("objsize",
            env.upd(env.makeVar("objsize"),
                    env.makeVar(fresh),
                    env.makeVar(numFields)));
          return env.makeVar(fresh);
        }
      }
    }
  }

  Expression * owner v = expr->vcgen(env, path);
                              
  // cast to pointer?
  if (type->isPointerType() &&
      type->asPointerType()->atType->isCompoundType()) {
    // to demonstrate that the cast is valid, prove that either the
    // pointer is NULL or else that the tag matches
    ECompoundType *ct = ext(type->asPointerType()->atType->asCompoundType());
    
    env << "    prove \"type cast to structure pointer\" {\n"
        << "      (" << v->tovc() << ")==0 ||\n"
        << "      " << env.hasClassTag(ct, v->tovc()) << ";\n"
        //<< "      tag[" << v->tovc() << "]==" << ct->name << "_tag;\n"
        << "    }\n";
  }


  #if 0
  // if the target of the assignment has pointer type and
  // the source has integer type, encode it as a pointer with
  // object 0; this is copied from E_assign, and should replace
  // that code eventually
  if (ctype->type->asRval()->isPointerType() &&
      expr->type->isSimple(ST_INT)) {
    // encode as pointer: should be null..
    env.proveIsZero(env.rval(v), "only zeroes can be cast to pointers");
    v = env.nullPointer();

    // OWNER: if the target is an owner pointer, construct
    // a tuple for it
    if (ctype->type->asRval()->isOwnerPtr()) {
      // first of all, we can only assign the integer zero to owners
      // (redundant given above check, but won't hurt)
      env.proveIsNullPointer(v, "can only assign zero to owners");

      // construct the tuple
      trace("owner") << "yielding a null-owner tuple\n";
      v = env.avUpd(
            env.avInt(0),                // start with no-information tuple
            env.avOwnerField_ptr(),      // field: ptr
            v);                          // value: pointer(0,0)
      v = env.avUpd(
            v,                           // value to update
            env.avOwnerField_state(),    // field: state
            env.avOwnerState_null());    // value: nullowner
    }
  }

  if (!ctype->type->asRval()->isOwnerPtr() &&
      expr->type->asRval()->isOwnerPtr()) {
    // OWNER: if casting from owner to nonowner, select 'ptr' field
    v = env.avSel(env.rval(v), env.avOwnerField_ptr());
  }

  // if casting from array to pointer, get pointer to element 0
  if (ctype->type->asRval()->isPointerType() &&
      expr->type->asRval()->isArrayType()) {
    v = coerceArrayToPointer(env, v);
  }
  #endif // 0

  return v;
}


Expression * owner E_cond::ivcgen(VCEnv &env, int path)
{
  // evaluate guard
  int modulus = cond->numPaths1();
  Expression * owner guard = cond->vcgenPred(env, path % modulus);
  path = path / modulus;

  if (th->numPaths == 0 && el->numPaths == 0) {
    // match_expr
    //   $e1 ? $e2 : $e3
    // translate {
    //   introduce $r;       // result of this conditional
    //   assume $e1 ==> $r == $e2;
    //   assume !$e1 ==> $r == $e3;
    //   return $r;
    // }

    Expression * owner e2 = th->vcgen(env,0);
    Expression * owner e3 = el->vcgen(env,0);
    string r = stringc << "temp" << env.tempCt++;

    env << "    introduce " << r << ";      // result of conditional: " << exprToString() << "\n";
    env << "    assume (" << guard->tovc() << ") ==> "
        <<               r << " == (" << e2->tovc() << ");\n";
    env << "    assume !(" << guard->tovc() << ") ==> "
        <<               r << " == (" << e3->tovc() << ");\n";
    return env.makeVar(r);
  }

  else {
    int thenPaths = th->numPaths1();
    //int elsePaths = el->numPaths1();     // not needed, but informative anyway

    if (path < thenPaths) {
      // taking 'then' branch, so we can assume guard is true
      env << "    assume " << guard->tovc() << ";    // true guard of conditional\n";
      return th->vcgen(env, path);
    }
    else {
      // taking 'else' branch, so guard is false
      env << "    assume !(" << guard->tovc() << ");    // false guard of conditional\n";
      return el->vcgen(env, path - thenPaths);
    }
  }
}


#if 0
Expression * owner E_gnuCond::ivcgen(VCEnv &env, int path)
{
  // again on the assumption there are no side effects in the 'else' branch
  // TODO: make a deep-copier for astgen
  return avTodo();
}
#endif // 0


Expression * owner E_sizeofType::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);
  return env.intLit(size);
}


// return true if this variable's value is modeled as a tuple
// that cannot have interior pointers
bool modelAsTuple(Variable *var)
{
  // old condition
  if (var->hasAddrTaken()) { return false; }
  if (var->type->isStructType()) { return true; }

  // OWNER: model owner pointers this way too
  if (var->type->isOwnerPtr()) { return true; }

  return false;
}


Expression * owner E_assign::ivcgen(VCEnv &env, int path)
{
  // get source value
  int modulus = src->numPaths1();
  Expression * owner srcValue = src->vcgen(env, path % modulus);
  path = path / modulus;

  if (op != BIN_ASSIGN) {
    // combine new with old
    srcValue = new E_binary(target->vcgen(env, path), op, srcValue);
  }

  return env.assignmentExpr(target, path, srcValue);
}


Expression *skipGrouping(Expression *e)
{
  while (e->isE_grouping()) {
    e = e->asE_grouping()->expr;
  }
  return e;
}


Expression * owner VCEnv::assignmentExpr
  (Expression *target, int path, Expression * owner srcValue)
{
  VCEnv &env = *this;

  target = skipGrouping(target);

  #if 0
  // get target lvalue
  Expression * owner targetValue = target->vcgen(env, path);
  xassert(targetValue->isAVlval());     // otherwise can't modify!
  AVlval *targetLval = targetValue->asAVlval();

  // do compound assignment
  if (op != BIN_ASSIGN) {
    srcValue = env.grab(new AVbinary(env.rval(targetLval), op, env.rval(srcValue)));
  }

  if (target->type->asRval()->isOwnerPtr()) {
    // OWNER: verify it is nonowning before reassignment
    proveNonOwning(env, env.rval(targetLval),
                   "verify owner pointer is nonowning before reassignment");
  }

  // do assignment
  env.setLval(targetLval, env.rval(srcValue));

  if (src->type->asRval()->isOwnerPtr()) {
    // OWNER: deaden state of source
    deadenOwnerState(env, srcValue,
                     stringc << "source in assignment: " << asString());
  }

  return env.dup(srcValue);
  #endif // 0

  #if 0      // I insert explicit casts now
  // HACK: if the target of the assignment has pointer type and
  // the source has integer type, encode it as a pointer with
  // object 0 (the right solution is to, during typechecking,
  // insert explicit coercions; then this code would be a part
  // of the cast-to-pointer coercion)
  if (target->type->asRval()->isPointerType() &&
      src->type->isSimple(ST_INT)) {
    // encode as pointer: some integer offset from the null object
    v = env.avPointer(env.avInt(0), v);
  }
  #endif // 0

  #if 0     // old
  // HACK: if both the target and source are owners, and the source
  // is a dereference of another pointer, then I need to deaden the
  // state of the owner stored in memory (proper solution: do lvals
  // right)
  if (target->type->asRval()->isOwnerPtr() &&
      src->type->asRval()->isOwnerPtr() &&
      src->isE_deref()) {
    E_deref *srcDeref = src->asE_derefC();
    if (src->numPaths != 0) {    // no side effects
      cout << "WARNING: unhandled ownership transfer: " << asString() << "\n";
    }
    else {
      trace("owner") << "deadening state of owner read through pointer\n";

      // get address of the owner pointer losing ownership
      Expression * owner addr = srcDeref->ptr->vcgen(env, 0 /*path*/);

      // update its 'state' field
      env.setMem(env.avUpdate(env.getMem(),        // starting memory
        env.avObject(addr), env.avOffset(addr),    // object to modify
        env.avSetElt(                              // new value
          env.avOwnerField_state(),                  // modify 'state' field
          v,                                         // prior tuple value
          env.avOwnerState_dead()                    // new state is 'dead'
        )));
    }
  }
  #endif // 0

  if (target->isE_variable()) {
    xassert(path==0);
    Variable *var = target->asE_variable()->var;

    if (ext(var)->hasVAttr(VA_PARAMBYREF)) {
      env.error("write to `parambyref' variable address");
    }

    if (ext(var)->hasVAttr(VA_ADDRTAKEN)) {
      env.error("write to `addrtaken' raw variable");
    }

    #if 0
    // extract old value
    Expression * owner old = env.get(var);

    if (var->type->isOwnerPtr()) {
      // OWNER: verify it is nonowning before reassignment
      proveNonOwning(env, old, "verify owner pointer is nonowning before reassignment");
    }
    #endif // 0

    env.assignLive(var, srcValue);

    return srcValue;
  }

  if (target->isE_deref()) {
    E_deref *deref = target->asE_deref();
    //string syntax = target->asString();

    if (deref->ptr->isE_variable()) {
      EVariable *v = ext(deref->ptr->asE_variable()->var);
      if (v->hasVAttr(VA_PARAMBYREF)) {
        // match_expr
        //   *$n = $e
        // where
        //   $n has attribute VA_PARAMBYREF
        // translate {
        //   $n := $e;
        //   return $e;
        // }
        env << "    " << v->name << " := " << srcValue->tovc() << ";\n";
        return srcValue;
      }
    }
  
    if (deref->ptr->isBinary(BIN_PLUS)) {
      E_binary *bin = deref->ptr->asE_binary();

      int modulus = bin->e1->numPaths1();
      int e1Path = path % modulus;
      int e2Path = path / modulus;

      // match_expr
      //   ($e1->$f)[$e2] = $e3        // *( ( (*$e1).$f + $e2 ) = $e3
      // where {
      //   isArrayType($e1->$f);
      //   match_type $e1 with struct $n *;
      // }
      // translate {
      //   prove tag[$e1] == tagName(struct $n);
      //   prove 0 <= $e2 < arraySize($f);
      //   mem := mem{$e1,($f + $e2) := $e3};
      //   meta := meta{$1,($f + $e2) := live};
      //   return $e3;                 // implicitly, each expr eval'd only once
      // }
      if (bin->e1->isE_fieldAcc()) {
        E_fieldAcc *acc = bin->e1->asE_fieldAcc();
        if (acc->obj->isE_deref()) {
          if (acc->field->type->isArrayType() &&
              acc->obj->type->asRval()->isCompoundType()) {
            EVariable *field = ext(acc->field);
            ArrayType *at = field->type->asArrayType();
            ECompoundType *ct = ext(acc->obj->type->asRval()->asCompoundType());

            // I ensure that each expression is evaluated only once,
            // and the result of vcgen of course has no remaining side
            // effects, so could be freely used multiple times in the
            // translated code.. I'd just have to make sure my
            // translator was smart enough to only evaluate each
            // expression once (and perhaps emit a warning if an
            // expression isn't evaluated at all)
            Expression * owner e1 = acc->obj->asE_deref()->ptr->vcgen(env, e1Path);
            Expression * owner e2 = bin->e2->vcgen(env, e2Path);
            // 'srcValue' is e3

            env << "    // structure array field write\n";
            env << "    prove " << env.hasClassTag(ct, e1->tovc()) << ";\n";
            //env << "    prove tag[" << e1->tovc() << "] == " << ct->name << "_tag;\n";
            env << "    prove 0 <= " << e2->tovc() << " < " << at->size << ";\n";
            
            if (ct->flatMap) {
              env << "    mem := mem{" << e1->tovc() << ",("
                  << field->modelVar->name << " + "
                  << e2->tovc() << ") := " << srcValue->tovc() << "};\n";
              env << "    meta := meta{" << e1->tovc() << ",("
                  << field->modelVar->name << " + "
                  << e2->tovc() << ") := live};\n";
            }
            else {
              // fieldName($f) := fieldName($f){$e1 := 
              //   fieldName($f)[$e1]{$e2 := $e3}};
              string fieldName = env.fieldSlotName(field);
              env << "    " << fieldName << " := " << fieldName << "{" 
                  << e1->tovc() << " := " << fieldName << "[" << e1->tovc()
                  << "]{" << e2->tovc() << " := " << srcValue->tovc() << "}};\n";
            }

            delete e1;
            delete e2;
            return srcValue;
          }
        }
      }

      // match_expr
      //   $n[$e2] = $e3                // *($n + $e2) = $e3
      // where
      //   typeof($n) == $t [$eSz]      // directly-modeled array
      // translate {
      //   prove "array var write"
      //     0 <= $e2 < $eSz;
      //   $n := $n{$e2 := $e3};
      //   return $e3;
      // }
      if (bin->e1->isE_variable() &&
          bin->e1->type->asRval()->isArrayType()) {
        Variable *n = bin->e1->asE_variable()->var;
        int eSz = n->type->asArrayType()->size;
        if (eSz != ArrayType::NO_SIZE) {
          Expression * owner e2 = bin->e2->vcgen(env, e2Path);
          env << "    prove \"array var write: " << exprToString(target) << "\"\n"
              << "      0 <= " << e2->tovc() << " < " << eSz << ";\n";
          env << "    " << n->name << " := " << n->name 
              << "{" << e2->tovc() << " := " << srcValue->tovc() << "};";
          delete e2;
          return srcValue;
        }
      }

      // match_expr
      //   $e1[$e2] = $e3            // *($e1 + $e2) = $e3
      // translate {
      //   prove "array ptr write"
      //     0 <= $e2 < objsize[$e1];
      //   mem := mem{$e1,$e2 := $e3};
      //   meta := meta{$e1,$e2 := live};
      //   return $e3;
      // }
      Expression * owner e1 = bin->e1->vcgen(env, e1Path);
      Expression * owner e2 = bin->e2->vcgen(env, e2Path);
      env << "    prove \"array ptr write: " << exprToString(target) << "\"\n"
          << "      0 <= " << e2->tovc() << " < objsize[" << e1->tovc() << "];\n";
      env << "    mem := mem{" << e1->tovc() << "," << e2->tovc()
          <<                 " := " << srcValue->tovc() << "};\n";
      env << "    meta := meta{" << e1->tovc() << "," << e2->tovc()
          <<                   " := live};\n";
      delete e1;
      delete e2;
      return srcValue;
    }

    // get a symbolic value for the address being modified
    Expression * owner addr = deref->ptr->vcgen(env, path);

    // want: "mem := mem{addr := srcValue};"
    env.assignDel("meta",
      env.upd(env.makeVar("meta"), addr->clone(), env.makeVar("live")));
    env.assignDel("mem",
      env.upd(env.makeVar("mem"), addr, srcValue->clone()));

    return srcValue;

    #if 0
    // COPIED from E_deref
    if (deref->ptr->type->asRval()->isOwnerPtr()) {
      // OWNER: pointer must be in owning state to use
      trace("owner") << "owner pointer being dereferenced for write\n";
      env.prove(P_equal(env.avGetElt(env.avOwnerField_state(), addr),
                        env.avOwnerState_owning()),
                "owner must be in owner state to dereference");

      // dereferencing an owner needs to actually dereference the 'ptr' field
      addr = env.avGetElt(env.avOwnerField_ptr(), addr);
    }

    // emit a proof obligation for safe access
    verifyPointerAccess(env, this, addr);

    // memory changes
    env.setMem(env.avUpdate(env.getMem(), env.avObject(addr),
                            env.avOffset(addr), v));
    #endif // 0
  }

  if (target->isE_fieldAcc()) {
    E_fieldAcc *fldAcc = target->asE_fieldAcc();
    string f = env.fieldSlotName(fldAcc->field);
    if (fldAcc->obj->type->asRval()->isCompoundType()) {
      ECompoundType *ct = ext(fldAcc->obj->type->asRval()->asCompoundType());

      // interpret field accesses to objects in memory
      string objid;
      if (fldAcc->obj->isE_deref()) {
        Expression * owner e1 = fldAcc->obj->asE_deref()->ptr->vcgen(env, path);
        objid = e1->tovc();
        delete e1;
      }
      else if (fldAcc->obj->isE_variable()) {
        EVariable *v = ext(fldAcc->obj->asE_variable()->var);
        if (v->hasVAttr(VA_ADDRTAKEN)) {
          objid = v->name;
        }
        else {
          // field of non-addr-taken structure instance
          // match_expr
          //   $n.$f = $e
          // where
          //   $n does not have attr `addrtaken'
          // translate {
          //   instanceFieldName($n,$f) := $e;
          //   return $e;
          // }
          EVariable *synth = getSynthField(v, fldAcc->field);
          if (synth) {
            // found the right synth field
            env << "    " << synth->name << " := " 
                << srcValue->tovc() << ";\n";
            return srcValue;
          }
        }
      }

      if (objid.length() > 0) {
        env << "    prove \"ptr->field write: " << exprToString(target) << "\"\n"
            << "      " << env.hasClassTag(ct, objid) << ";\n";
            //<< "      tag[" << objid << "] == " << ct->name << "_tag;\n";

        if (ct->flatMap) {
          // match_expr
          //   $e1->$f = $e2            // (*$e1).f = $e2
          // where
          //   match_type $e1 with struct $n *;
          //   struct $n uses the flat model;
          // translate {
          //   prove tag[$e1] == tagName(struct $n);
          //   mem := mem{$e1,$f := $e2};
          //   return $e2;
          // }
          env << "    mem := mem{" << objid << "," << f << " := "
              << srcValue->tovc() << "};\n";
        }

        else {
          // match_expr
          //   $e1->$f = $e2            // (*$e1).f = $e2
          // where
          //   match_type $e1 with struct $n *;
          //   struct $n uses the split model;
          // translate {
          //   prove tag[$e1] == tagName(struct $n);
          //   $f := $f{$e1 := $e2};
          //   return $e2;
          // }
          env << "    " << f << " := " << f << "{" << objid << " := "
              << srcValue->tovc() << "};\n";
        }
          
        // also covers:
          // match_expr
          //   $n.$f = $e
          // where
          //   $n has attribute `addrtaken'

        return srcValue;
      }
    }
  }

  #if 0
  // structure field access
  if (target->isE_fieldAcc()) {
    E_fieldAcc *fldAcc = target->asE_fieldAcc();

    // accessing a field of a local or global variable
    if (fldAcc->obj->isE_variable()) {
      Variable *var = fldAcc->obj->asE_variable()->var;

      if (modelAsTuple(var)) {
        // modeled as an unaliased tuple

        // get current (whole) structure value
        Expression * owner current = env.get(var);

        // compute updated structure value
        Expression * owner updated =
          env.avSetElt(
            env.avInt(fldAcc->field->index),      // field index
            current,                              // value being updated
            v);                                   // new field value

        if (var->type->isOwnerPtr() &&
            fldAcc->field->index == 0 /*ptr*/) {
          // OWNER: verify it is nonowning before reassignment
          proveNonOwning(env, current,
                         "verify owner pointer is nonowning before reassignment");
        }

        // replace old with new in abstract environment
        env.set(var, updated);
      }
      else {
        goto unhandled;
      }
    }

    // accessing a field through an object pointer
    else if (fldAcc->obj->isE_deref()) {
      // get an abstract value for the address of the object being modified
      Expression * owner addr = fldAcc->obj->asE_deref()->ptr->vcgen(env, path);

      #if 0    // old: before using + for field offsets
      if (!env.inPredicate) {
        // the address should have offset 0
        env.prove(P_equal(env.avOffset(addr), env.avInt(0)),
                  "object field access at offset 0");

        // TODO: assign type tags to allocated regions, and prove here
        // that the type tag is correct
      }                                                
      #endif // 0

      // update memory
      env.setMem(env.avUpdate(env.getMem(),
        env.avObject(addr),                          // object being modified
        env.avSum(env.avOffset(addr),                // offset: pointer offset
                  env.get(fldAcc->field->decl)),     //  + symbolic field offset
        v));                                         // new value
    }

    else {
    unhandled:
      // other forms aren't handled yet
      cout << "TODO: unhandled structure field assignment: " << asString() << endl;
    }
  }
  #endif // 0

  env.warning("unhandled assignment");
  return env.todo();
}


Expression * owner E_new::ivcgen(VCEnv &env, int path)
{
  return env.todo();

  #if 0
  xassert(path==0);

  // to model allocation:
  //   - make up a new logic variable to stand for the address
  //   - state that this address is distinct from any others we know
  //     about, including NULL
  //   - state that no pointer fields in the program currently point
  //     to the new object

  // new variable
  Expression * owner v = env.freshVariable("newAlloc",
    stringc << "result of " << asString());

  // distinct from others
  env.addDistinct(v);

  // we yield a memory offset
  Expression * owner ret = env.avSub(v, env.avWhole());

  // nothing points to this new thing
  env.assumeNoFieldPointsTo(ret);

  return ret;
  #endif // 0
}


Expression * owner E_delete::ivcgen(VCEnv &env, int path)
  { return env.todo(); }
Expression * owner E_throw::ivcgen(VCEnv &env, int path)
  { return env.todo(); }
Expression * owner E_keywordCast::ivcgen(VCEnv &env, int path)
  { return env.todo(); }
Expression * owner E_typeidExpr::ivcgen(VCEnv &env, int path)
  { return env.todo(); }
Expression * owner E_typeidType::ivcgen(VCEnv &env, int path)
  { return env.todo(); }


Expression * owner E_grouping::ivcgen(VCEnv &env, int path)
{
  return expr->vcgen(env, path);
}


Expression * owner E_arrow::ivcgen(VCEnv &env, int path)
{ 
  xfailure("should not get here");
  return NULL;
}


Expression * owner E_quantifier::ivcgen(VCEnv &env, int path)
{
  // this is quite nonideal, but at the moment I don't know exactly
  // what syntactic rule to use, so I just wait until the problem
  // shows up, and then bomb
  //SourceLoc loc = vars->first()->decl->getDeclaratorId()->name.loc;
  xfailure(stringc //<< toString(loc)
                   << "can't use quantifier(forall/exists) in a non-obviously-boolean context");
  return NULL;   // silence warning
}


Expression * owner E_let::ivcgen(VCEnv &env, int path)
{                                            
  xassert(path==0);
  return new E_let(name.clone(), binding->vcgen(env, 0), body->vcgen(env, 0));
}


Expression * owner E_pre::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);                                       
  
  IN_PRE_STATE(env);
  return expr->vcgen(env, path);

  // assume that this is in a context where I've already declared
  // something to represent the pre-state value
  //return env.makeVar(stringc << var->name->getName() << "_pre");
}


Expression * owner E_vc_expression::ivcgen(VCEnv &env, int path)
{
  xassert(path==0);                                       
  
  IN_PREDICATE(env);
  return expr->vcgen(env, path);
}


// --------------- Expression::vcgenPred -------------------
// STUBPATTERN: ^Expression.*E_.*::vcgenPred\b

Expression * owner Expression::vcgenPredDefault(VCEnv &env, int path)     // stubify
{
  Expression * owner val = env.rval(vcgen(env, path));
  
  return e_notEqual(env.intLit(0), val);

  #if 0
  // use exprToPred to give this another chance to map into the
  // predicate space; the vcgenPred functions below do a good job
  // when it's obvious something is being used in a predicate context,
  // but if I save a value into a variable and *then* try to treat
  // it as a predicate, I won't learn until later that I need to
  // map it into predicate space
  return exprToPred(env.rval(vcgen(env, path)));
  #endif // 0
}


Expression * owner E_boolLit::vcgenPred(VCEnv &env, int path)
{
  return vcgen(env, path);
}

Expression * owner E_intLit::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_floatLit::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_stringLit::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_charLit::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_this::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_variable::vcgenPred(VCEnv &env, int path)
{
  // sort of awkward...
  if (0==strcmp(name->getName(), "global_inv")) {
    return vcgen(env, path);
  }

  return vcgenPredDefault(env, path);
}


Expression * owner E_funCall::vcgenPred(VCEnv &env, int path)
{
  if (env.inPredicate) {
    return vcgen(env, path);
  }

  #if 0
  E_variable *funcVar = func->ifE_variableC();
  if (funcVar) {
    // is this function specially declared to be a predicate that
    // Simplify knows about?
    if (funcVar->var->hasFlag(DF_PREDICATE)) {
      // yes, so emit directly
      xassert(path==0);

      // collect the list of argument values
      ASTList<AbsValue> *avArgs = new ASTList<AbsValue>;
      FOREACH_ASTLIST(Expression, args, iter) {
        avArgs->append(env.rval(iter.data()->vcgen(env, path)));
      }
      return new P_named(funcVar->var->name, avArgs);
    }
  }
  #endif // 0

  // not a simple function name, or not the right kind
  return vcgenPredDefault(env, path);
}


Expression * owner E_constructor::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_fieldAcc::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_sizeof::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_unary::vcgenPred(VCEnv &env, int path)
{
  if (op == UNY_NOT) {
    // push the predicate context one level down
    return e_not(expr->vcgenPred(env, path));
  }
  else {
    return vcgenPredDefault(env, path);
  }
}

Expression * owner E_effect::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }


Expression * owner E_binary::vcgenPred(VCEnv &env, int path)
{
  int modulus = e1->numPaths1();
  int lhsPath = path % modulus;
  int rhsPath = path / modulus;

  if (op == BIN_COMMA) {
    // evaluate and discard
    env.discard(e1->vcgen(env, lhsPath));

    // evaluate and keep
    return e2->vcgenPred(env, rhsPath);
  }

  if (isPredicateCombinator(op)) {
    if (e2->numPaths == 0) {
      xassert(rhsPath == 0);

      // update: I now split paths as if derefs had side effects,
      // so side-effect-free means I can do straightforward combination
      return new E_binary(e1->vcgenPred(env, lhsPath), op,
                          e2->vcgenPred(env, rhsPath));

      #if 0
      // simple side-effect-free combinator, so there's no path stuff;
      // *but* we do have to assume something about the LHS while
      // analyzing the RHS, e.g. for "if (p && p->foo) ..."
      Expression * owner lhs = e1->vcgenPred(env, lhsPath);

      Expression * owner rhs;
      if (op != BIN_OR) {
        env.pushExprFact(lhs);
        rhs = e2->vcgenPred(env, 0);
        env.popExprFact();
      }
      else {
        // for "or", assume the negation
        P_not notLHS(lhs);
        try {
          env.pushExprFact(&notLHS);
          rhs = e2->vcgenPred(env, 0);
          env.popExprFact();
        }
        catch (...) {
          notLHS.p = NULL;
          throw;
        }
        notLHS.p = NULL;
      }

      return P_combinator(op, lhs, rhs);
      #endif // 0
    }

    else {
      // RHS has a side effect, so we need to analyze paths
      xassert(op == BIN_AND || op == BIN_OR);     // we don't allow ==>, <==> in side-effecting exprs

      // evaluate LHS
      Expression * owner lhs = e1->vcgenPred(env, lhsPath);

      // eval RHS *if* it's followed
      if (rhsPath > 0) {
        // we follow rhs; this implies something about the value of lhs:
        // if it's &&, lhs had to be true; false for ||
        env.assumeSense(lhs, op==BIN_AND,
          stringc << "LHS guard given RHS is eval'd " << asString());

        // the true/false value is now entirely determined by rhs
        return e2->vcgenPred(env, rhsPath-1);
      }
      else {
        // we do *not* follow rhs; this also implies something about lhs
        env.assumeSense(lhs, op==BIN_OR,
          stringc << "LHS guard given RHS is *not* eval'd " << asString());

        // it's a C boolean (and we know which)
        return new E_boolLit(op==BIN_OR? true : false);
      }
    }
  }

  else if (isRelational(op)) {
    // non-short-circuit: always evaluate RHS
    Expression * owner lhs = env.rval(e1->vcgen(env, lhsPath));
    Expression * owner rhs = env.rval(e2->vcgen(env, rhsPath));
    return new E_binary(lhs, op, rhs);
  }

  else {
    return vcgenPredDefault(env, path);
  }
}


Expression * owner E_addrOf::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_deref::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }

Expression * owner E_cast::vcgenPred(VCEnv &env, int path)
{
  // assume casts preserve truth or falsehood (I think that's sound..)
  return expr->vcgenPred(env, path);
}

Expression * owner E_cond::vcgenPred(VCEnv &env, int path)
{
  return vcgenPredDefault(env, path);
  
  #if 0
  // condition
  int modulus = cond->numPaths1();
  Expression * owner guard = cond->vcgenPred(env, path % modulus);
  path = path / modulus;

  if (th->numPaths == 0 && el->numPaths == 0) {
    // no side effects in either branch;
    // map ?: as a pair of implications
    return P_and2(new P_impl(cond->vcgenPred(env, 0),
                             th->vcgenPred(env, 0)),
                  new P_impl(new P_not(cond->vcgenPred(env, 0)),
                             el->vcgenPred(env, 0)));
  }

  else {
    int thenPaths = th->numPaths1();
    int elsePaths = el->numPaths1();

    if (path < thenPaths) {
      // taking 'then' branch, so we can assume guard is true
      env.assume(guard, stringc << "true guard of '?:': " << asString());
      return th->vcgenPred(env, path);
    }
    else {
      // taking 'else' branch, so guard is false
      env.assumeNot(guard, stringc << "false guard of '?:': " << asString());
      return el->vcgenPred(env, path - elsePaths);
    }
  }    
  #endif // 0
}


Expression * owner E_sizeofType::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }


Expression * owner E_assign::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }


Expression * owner E_new::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_delete::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_throw::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_keywordCast::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_typeidExpr::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }
Expression * owner E_typeidType::vcgenPred(VCEnv &env, int path)
  { return vcgenPredDefault(env, path); }


Expression * owner E_grouping::vcgenPred(VCEnv &env, int path)
{
  return expr->vcgenPred(env, path);
}


Expression * owner E_arrow::vcgenPred(VCEnv &env, int path)
{
  xfailure("should not get here");
  return NULL;
}


Expression * owner E_quantifier::vcgenPred(VCEnv &env, int path)
{
  xassert(path==0);

  return new E_quantifier(loc, isForall, label, cloneFakeList(vars),
                          patterns? patterns->vcgen(env, 0 /*path*/) : NULL,
                          body->vcgenPred(env, path));
  
  #if 0
  // gather up the quantifiers
  P_quantifier *ret = new P_quantifier(NULL, NULL, forall);

  FOREACH_ASTLIST(Declaration, decls, outer) {
    FOREACH_ASTLIST(Declarator, outer.data()->decllist, inner) {
      Variable *var = inner.data()->var;

      // this will instantiate the variables
      ret->variables.append(env.get(var)->asAVvar());
    }
  }
  
  // make a new list of facts, into which path-facts will be
  // accumulated when they are generated inside the body
  env.pushPathFactsFrame();

  // analyze the body
  Expression * owner body = pred->vcgenPred(env, path);

  // pop off that new frame so we can look at what was adde
  ObjList<Predicate> *localFacts = env.popPathFactsFrame();    // (owner)

  // search the local fact list for anything which refers to the
  // quantified variables; if we find a fact which does *not*
  // refer to them, put that into the outer path-facts
  ObjListMutator<Predicate> factIter(*localFacts);
  while (!factIter.isDone()) {
    Expression * owner fact = factIter.data();

    // does this fact refer to any of the variables in 'ret->variables'?
    bool refers = false;
    FOREACH_ASTLIST(AVvar, ret->variables, varIter) {
      AVvar const *var = varIter.data();

      // does it refer to 'var'?
      if (predicateRefersToAV(fact, var)) {
        // yes; fine, will leave it here
        refers = true;
        break;
      }                    
    }
    
    if (!refers) {
      // the fact doesn't refer to any quantified variables; currently
      // this happens when type-related facts happen to get introduced
      // while inside a quantifier; that will be fixed soon, but I'll
      // leave the logic operational; move this fact to the outer
      // path-facts
      env.assume(factIter.remove(),   // advances factIter
                  "reason lost; was inside a quantifier");
    }
    else {
      // simple advance
      factIter.adv();
    }
  }

  // now, associate any facts remaining in 'localFacts' with the
  // quantified body
  if (localFacts->isNotEmpty()) {
    P_and *newFacts = new P_and(new ASTList<Predicate>);
    while (localFacts->isNotEmpty()) {
      newFacts->conjuncts.append(localFacts->removeFirst());
    }

    if (forall) {
      // if the quantifier is a forall, then we construct an implication
      // so the body can assume the facts we've pushed
      body = new P_impl(newFacts, body);
    }
    else {
      // if the quantifier is an exists, we simply conjoin the facts,
      // effectively (again) making them facts which can be assumed while
      // considering the body
      newFacts->conjuncts.append(body);
      body = newFacts;
    }
  }
  delete localFacts;

  // attach the body to the quantifier node itself
  ret->body = body;

  return ret;
  #endif // 0
}


Expression * owner E_let::vcgenPred(VCEnv &env, int path)
{                                                  
  xassert(path == 0);
  return new E_let(name.clone(), binding->vcgen(env, 0), body->vcgenPred(env, 0));
}


Expression * owner E_pre::vcgenPred(VCEnv &env, int path)
{ 
  IN_PRE_STATE(env);
  return expr->vcgenPred(env, path); 
}

Expression * owner E_vc_expression::vcgenPred(VCEnv &env, int path)
{
  IN_PREDICATE(env);
  return expr->vcgenPred(env, path);
}


// --------------------- Initializer --------------------
// STUBPATTERN: ^Expression.*IN_.*::vcgen\b

Expression * owner IN_expr::vcgen(VCEnv &env, Type *, int path)
{
  return e->vcgen(env, path);
}


Expression * owner IN_compound::vcgen(VCEnv &env, Type */*type*/, int /*path*/)
{
  return env.todo();
}

Expression * owner IN_ctor::vcgen(VCEnv &env, Type */*type*/, int /*path*/)
{
  return env.todo();
}


// ------------------- Macro -------------------
// tovc.cc
string astTypeIdList_tovc(FakeList<ASTTypeId> *vars);

void Macro::vcgen(VCEnv &env)
{
  // copied and modified a little from Macro::tovc

  stringBuilder sb;
  sb << "define " << name.str;
  if (!params) {
    sb << " = " << body->tovc() << ";\n";
  }
  else {
    if (0==strcmp(name.str, "global_inv")) {
      // use computed list instead of what was supplied explicitly,
      // so I don't have to maintain that explicit list
      sb << "(" << env.globalVarNameList() << ")";
    }
    else {
      sb << astTypeIdList_tovc(params);
    }
    sb << " {\n";
    IN_PREDICATE(env);
    Expression * owner pred = body->vcgenPred(env, 0 /*path*/);
    sb << pred->tovcConjunctiveLines("  ");
    delete pred;
    sb << "}\n\n";
  }
  env << sb;
}


// -------------------- EFunctionType -----------------------
void foldConj(Expression * owner nullable &conjunction, 
              Expression * owner conjunct)
{
  if (!conjunction) {
    conjunction = conjunct;
  }
  else {
    conjunction = e_and(conjunction, conjunct);
  }
}


// is 'v' among those that can change ('list')?
bool amongChangesList(VCEnv &env, Variable const *v,
                      ASTList<E_variable> const &list)
{
  Regexp istag("_tag$");

  FOREACH_ASTLIST(E_variable, list, varIter) {
    EVariable const *chVar = extC(varIter.data()->var);
    SourceLoc loc = varIter.data()->name->loc;
    StringRef iterName = varIter.data()->name->getName();

    if (istag.match(iterName)) {
      // the 'changes' variable is a tag, meaning it refers to an
      // entire type; see if 'v' is a field of that type
      string theType = substring(iterName, strlen(iterName)-4);  // strip "_tag"
      if (regexpMatch(v->name, stringc << "^" << theType << "_")) {
        // 'v' is a field of 'theType', so it can change
        trace("changes") << ::toString(loc) << ": "
                         << v->name << " can change b/c of " << iterName << endl;
        return true;
      }
    }

    else if (chVar->changeGroup) {
      // refers to a change group, look in the group for the name
      if (amongChangesList(env, v, chVar->changeGroup->group)) {
        return true;
      }
    }

    else if (v->name == iterName) {
      trace("changes") << ::toString(loc) << ": "
                       << v->name << " can change by explicit mention" << endl;
      return true;
    }
  }

  return false;     // not in the list
}


Expression * nullable serf EFunctionType::getPostcondition(VCEnv &env)
{
  if (changes) {
    // create a conjunction to represent the 'changes' spec, by
    // saying that all the unchanged variables are equal to their
    // pre-state values
    Expression * owner nullable conjunction = NULL;
    SFOREACH_OBJLIST(Variable, env.globals, iter) {
      Variable const *v = iter.data();

      // is 'v' among those that can change?
      if (amongChangesList(env, v, changes->vars)) {
        continue;    // yes it is, so don't emit a condition on it
      }

      foldConj(conjunction,
        e_equal(env.makeVar(v->name), e_pre(env.makeVar(v->name))));

      if (v->type->isPointerType()) {
        string meta = stringc << v->name << "_meta";
        foldConj(conjunction,
          e_equal(env.makeVar(meta), e_pre(env.makeVar(meta))));
      }
    }

    // fold this conjunction into the postcondition
    if (conjunction) {
      changesConj = conjunction->clone();
      foldConj(postcondition, conjunction);
    }

    changes = NULL;
  }

  return postcondition;
}


// EOF
