// verify.cc
// toplevel driver for verifier

#include <iostream.h>     // cout
#include <stdlib.h>       // exit

#include "trace.h"        // traceAddSys
#include "ckheap.h"       // malloc_stats
#include "srcloc.h"       // SourceLoc
#include "cc.ast.gen.h"   // C ast
#include "cc_env.h"       // Env
//#include "aenv.h"         // AEnv
#include "strutil.h"      // plural
//#include "factflow.h"     // factFlow
#include "cc_lang.h"      // CCLang
//#include "treeout.h"      // treeOut
#include "parsetables.h"  // ParseTables
#include "cc.gr.gen.h"    // CCParse
#include "lexer.h"        // Lexer
#include "glr.h"          // GLR
#include "cfg.h"          // computeUnitCFG
#include "cfgpaths.h"     // countPaths
#include "vc_env.h"       // VCEnv
#include "exttype.h"      // ETypeFactory
#include "mysig.h"        // printSegfaultAddrs


void if_malloc_stats()
{
  if (tracingSys("malloc_stats")) {
    malloc_stats();
  }
}


// ------------------ Verifier ------------------------
class Verifier : public ASTVisitor {
public:
  VCEnv &env;

public:
  Verifier(VCEnv &e) : env(e) {}
  bool visitFunction(Function *f);
  bool visitTopForm(TopForm *t);
  bool visitTypeSpecifier(TypeSpecifier *ts);
};

bool shouldSkipFunc(char const *name, char const *stage)
{
  static char const *selectedFunc = getenv("SELFUNC");
  if (selectedFunc && 
      0!=strcmp(name, selectedFunc) &&
      0!=strcmp(name, "global_inv")) {     // global_inv always processed
    //traceProgress() << "skipping " << stage << " of "
    //                << name << " due to SELFUNC" << endl;
    return true;
  }
  else {
    return false;
  }
}

bool Verifier::visitFunction(Function *f)
{
  PQName const *name = f->nameAndParams->getDeclaratorId();
  if (!name) {
    return true;   // skip it
  }

  if (shouldSkipFunc(f->nameAndParams->var->name, "vcgen")) {
    // might need this, even though the function body isn't analyzed
    env << "introduce " << f->nameAndParams->var->name 
        << "_address;    // for taking address\n";
    return true;
  }

  int n = countPaths(f);
  cout << "function " << name->toString() << " has " << pluraln(n, "path") << "\n";

  if (tracingSys("paths")) {
    printPaths(f);
  }

  // vcgen along all paths in the function
  f->vcgen(env);

  return true;     // keep going, *do* look at children
}

bool Verifier::visitTopForm(TopForm *f)
{
  f->vcgen(env);
  return true;
}

bool Verifier::visitTypeSpecifier(TypeSpecifier *ts)
{
  if (ts->isTS_classSpec()) {
    ECompoundType *ct = ext(ts->asTS_classSpec()->ctype);
    if (ct->name) {
      env << "// " << ct->keywordAndName() << " at " << toString(ts->loc) << "\n";

      // declare the type tag for this struct
      env << "define " << ct->name << "_tag = " << env.typeTagCt++ << ";\n";

      // emit field offsets/memories and total size, using the string
      // that was computed during tcheck
      env << ct->sb;

      env << "\n";
      
      if (!ct->flatMap) {
        // each of the field memories is a global
        FOREACH_OBJLIST_NC(Variable, ct->fieldSlots, iter) {
          env.globals.append(iter.data());
        }
      }
    }
  }

  if (ts->isTS_enumSpec()) {
    TS_enumSpec *es = ts->asTS_enumSpec();

    env << "// enum " << (es->name? es->name : "(anonymous)") << " at "
        << toString(ts->loc) << "\n"
        << "define {\n";
    FAKELIST_FOREACH(Enumerator, es->elts, iter) {
      if (0!=strcmp(iter->name, "false") &&
          0!=strcmp(iter->name, "true")) {    // hack: don't emit defn's for true &false
        env << "  " << iter->name << " = " << iter->enumValue << ";\n";
      }
    }
    env << "}\n\n";
  }

  return true;
}


// ------------------ ExtTCEnv -------------------
// extended type-check environment for type checking constructs that
// the base Elsa parser doesn't know about
class ExtTCEnv : public Env {
  StringRef flat_model;
  StringRef str_superclass;

public:
  ExtTCEnv(StringTable &str, CCLang &lang, TypeFactory &tfac,
           TranslationUnit *tunit)
    : Env(str, lang, tfac, tunit),
      flat_model(str("use_flat_struct_model")),
      str_superclass(str("superclass"))
    {}

  // declare a metavariable, including adding it to the tc environment
  EVariable *declMetaVar(char const *name, CVFlags cv);
  void addMetaVar(Variable *v);

  // Env functions
  virtual void checkFuncAnnotations(FunctionType *ft, D_func *syntax);
  virtual void addedNewCompound(CompoundType *ct);
  virtual void addedNewVariable(Scope *s, Variable *v);
};


void tcheckE_variableList(Env &env, ASTList<E_variable> &list);
#if 0    // moved to annot_aux.cc
{
  FOREACH_ASTLIST_NC(E_variable, list, iter) {
    E_variable *e = iter.data();
    xassert(!e->ambiguity);        // should never be ambiguous

    e->name->tcheck(env);
    e->var = env.lookupPQVariable(e->name);
    if (!e->var) {
      env.error(e->name->loc, stringc
        << "there is no variable called `" << *(e->name) << "'");
      continue;
    }

    if (e->var->hasFlag(DF_TYPEDEF) &&      // would mean a field changegroup
        !e->var->type->isCompoundType()) {
      env.error(e->name->loc,
        "if a type is mentioned in a 'changes' clause, it must be "
        "a struct/class/union");
    }
  }
}
#endif // 0


void ExtTCEnv::checkFuncAnnotations(FunctionType *_ft, D_func *syntax)
{
  EFunctionType *ft = ext(_ft);

  // if the function can return a value, make a Variable that
  // will represent the location where that return value is stored
  if (!ft->retType->isVoid()) {
    ft->result = tfac.makeVariable(syntax->loc, str("result"),
                                   ft->retType, DF_NONE, NULL /*tunit*/);

    // add it to the environment
    if (!addVariable(ft->result)) {
      error("parameter called 'result' conflicts with the pseudovariable "
            "by the same name that represents the return value; please "
            "change the name of the parameter (I apologize...)");
    }
  }

  FOREACH_ASTLIST_NC(FuncAnnotation, syntax->ann, iter) {
    FuncAnnotation *fa = iter.data();

    // type check the annotation expressions, and them attach them
    // to the computed function types
    ASTSWITCH(FuncAnnotation, fa) {
      ASTCASE(FA_pre, f)
        f->expr->tcheck(*this, f->expr);
        Expression * owner e = f->expr->clone();
        e->tcheck(*this, e);                  // convenient if already tcheck'd
        if (ft->precondition) {
          e = e_and(ft->precondition, e);     // conjoin
        }
        ft->precondition = e;

      ASTNEXT(FA_post, f)
        f->expr->tcheck(*this, f->expr);
        Expression * owner e = f->expr->clone();
        e->tcheck(*this, e);                   // convenient if already tcheck'd
        if (ft->postcondition) {
          e = e_and(ft->postcondition, e);     // conjoin
        }
        ft->postcondition = e;

      ASTNEXT(FA_changes, c)
        // check the 'changes' clause; importantly, this binds the
        // 'var' fields of the E_variables to the Variables that
        // they refer to
        tcheckE_variableList(*this, c->vars);

        if (ft->changes) {
          error("multiple 'changes' clauses");
        }
        ft->changes = c;

      ASTENDCASED
    }
  }

  // experiment: let's say that if there is no 'changes' clause
  // then it's as if there was an empty one
  if (!ft->changes) {
    ft->changes = new FA_changes(NULL);   // will be leaked
  }
}


void ExtTCEnv::addedNewCompound(CompoundType *_ct)
{
  ECompoundType *ct = ext(_ct);

  // flat model?
  if (ct->lookupVariable(flat_model, *this)) {
    ct->flatMap = true;
  }

  // superclass?        
  {
    Variable *superfield = ct->lookupVariable(str_superclass, *this);
    if (superfield) {
      if (ct->flatMap) {
        error("`superclass' directive is incompatible with flat struct model");
      }
      if (superfield->type->isPointer() &&
          superfield->type->asPointerType()->atType->isCompoundType()) {
        ct->setSuperclass(ext(superfield->type->asPointerType()->atType->asCompoundType()));
      }
      else {
        error(superfield->loc, stringc
          << "`superclass' field must have type `pointer to struct', not `"
          << superfield->type->toString() << "'");
      }
    }
  }

  // declare offsets or memories for all of the fields, depending on model
  int offset=0;
  int fieldNum=0;
  SFOREACH_OBJLIST_NC(Variable, ct->dataMembers, iter) {
    EVariable *field = ext(iter.data());
    if (field->name == flat_model ||
        field->name == str_superclass) continue;     // just a directive

    if (ct->superclass &&
        fieldNum < ct->superclass->dataMembers.count()) {
      // superclass field
      EVariable const *sfield = extC(ct->superclass->dataMembers.nthC(fieldNum));
      if (field->name == sfield->name &&
          field->type->equals(sfield->type)) {
        // ok; use the superclass' slot name so that all accesses to
        // the field are using the same model variable
        field->modelVar = sfield->modelVar;
      }
      else {
        error(field->loc, stringc
          << "superclass field " << fieldNum << " is `" << sfield->toString()
          << "', but subclass field is `" << field->toString()
          << "'; these must be identical");
      }
    }
    else {
      // not a superclass field
      CVFlags cv = CV_NONE;
      string slotName = stringc << ct->name << "_" << field->name;
      if (ct->flatMap) {
        ct->sb << "define " << slotName << " = " << offset << ";\n";
        cv = CV_CONST;
      }
      else {
        ct->sb << "var " << slotName << ";\n";
      }
      field->modelVar = declMetaVar(slotName.c_str(), cv);
      ct->fieldSlots.append(field->modelVar);
    }

    // flatten arrays
    if (field->type->isArrayType()) {
      offset += field->type->asArrayType()->size;
    }
    else {
      offset++;
    }

    fieldNum++;
  }

  // the size of all fields
  ct->objectSize = offset;
  ct->sb << "define " << ct->name << "_size = " << offset << ";\n";

  xassert(!ct->sizeVar && !ct->tagVar);
  ct->sizeVar = declMetaVar(stringc << ct->name << "_size", CV_CONST);
  ct->tagVar = declMetaVar(stringc << ct->name << "_tag", CV_CONST);
}


EVariable *ExtTCEnv::declMetaVar(char const *name, CVFlags cv)
{
  Variable *v = makeVariable(loc(), str(name),
                             getSimpleType(ST_INT, cv),
                             DF_DEFINITION);
  addMetaVar(v);
  return ext(v);
}

void ExtTCEnv::addMetaVar(Variable *v)
{
  if (!addVariable(v)) {
    error(stringc << "collision with metavariable: " << v->name);
  }
}


void ExtTCEnv::addedNewVariable(Scope *s, Variable *_v)
{
  EVariable *v = ext(_v);
  if (!v->hasFlag(DF_TYPEDEF) &&
      v->type->isCompoundType() &&
      !v->hasVAttr(VA_ADDRTAKEN)) {
    // global *instance* of a struct, whose addr is not taken:
    // model as a bunch of separate globals, one for each field
    SFOREACH_OBJLIST_NC(Variable, v->type->asCompoundType()->dataMembers, fieldIter) {
      Variable *field = fieldIter.data();

      // synthesize a variable
      StringRef synthName = str(stringc << v->name << "_" << field->name);
      EVariable *synth = new EVariable(v->loc, synthName, field->type, DF_NONE);
      synth->originalStructField = ext(field);
      v->synthFields.append(synth);
      addMetaVar(synth);
    }
  }
}


// ------------------ main, etc. ------------------
TranslationUnit *parseFile(StringTable &strTable, CCLang &lang,
                           char const *fname)
{
  // lexer
  Lexer lexer(strTable, lang, fname);

  // parser context
  CCParse ctx(strTable, lang);

  // parse tables (automatically deallocated)
  Owner<ParseTables> tables(ctx.makeTables());

  // parser
  GLR glr(&ctx, tables);

  // parse the input
  SemanticValue treeTop;
  if (!glr.glrParse(lexer, treeTop)) {
    xbase(stringc << "parse error in file " << fname);
  }

  return (TranslationUnit*)treeTop;
}


int doit(int argc, char **argv)
{
  traceAddSys("progress");

  if_malloc_stats();

  printSegfaultAddrs();

  SourceLocManager mgr;

  // string table for storing parse tree identifiers
  StringTable strTable;
  //globalStringTable = &strTable;

  // parsing language options
  CCLang lang;
  lang.ANSI_C89();

  // process arguments
  char const *progname = argv[0];
  TRACE_ARGS();

  if (argc < 2) {
    cout << "usage: " << progname << " [-tr flags] input.c\n"
      "  useful tracing flags:\n"
      "    parse              print shift/reduce steps of parsing algorithm\n"
      "    stopAfterParse     stop after parsing\n"
      "    printAST           print AST after parsing\n"
      "    printTypedAST      print AST with type info\n"
      "    stopAfterTCheck    stop after typechecking\n"
      "    disableFactFlow    don't do invariant strengthening\n"
      "    factflow           print details of factflow computation\n"
      "    stopAfterVCGen     stop after vcgen\n"
      "    printAnalysisPath  print each path that is analyzed\n"
      "    predicates         print all predicates (proved or not)\n"
      "    paths              print all paths in each function\n"
      "    absInterp          print results of abstract interpretation\n"
      "    tcheck             print typechecking info\n"
      ;
    return 0;
  }

  // parse
  TranslationUnit *unit = parseFile(strTable, lang, argv[1]);

  checkHeap();

  // print abstract syntax tree
  if (tracingSys("printAST")) {
    unit->debugPrint(cout, 0);
  }

  if (tracingSys("stopAfterParse")) {
    return 0;
  }


  // --------- declarations provided automatically -------
  #if 0
  SourceLoc dummyLoc = HERE_SOURCELOC;
  Variable mem(dummyLoc, strTable.add("mem"),
               new PointerType(PO_POINTER, CV_NONE,
                 &CVAtomicType::fixed[ST_INT]), DF_NONE);
  globalMemVariable = &mem;
  #endif // 0


  // ---------------- typecheck -----------------
  ETypeFactory factory;
  {
    traceProgress() << "type checking...\n";
    ExtTCEnv env(strTable, lang, factory, unit);
    //env.addVariable(mem.name, &mem);
    unit->tcheck(env);
    traceProgress(2) << "done type checking\n";

    int n = env.numErrors();
    n += computeUnitCFG(unit);

    // print abstract syntax tree annotated with types
    if (tracingSys("printTypedAST")) {
      unit->debugPrint(cout, 0);
    }

    // print errors and warnings
    env.errors.print(cout);
    
    if (n) {
      cout << "there " << plural(n, "was") << " " << n
           << " typechecking " << plural(n, "error") << "\n";
      return 4;
    }

    if (tracingSys("stopAfterTCheck")) {
      return 0;
    }
  }


  #if 0
  // ------------- automatic invariant strengthening ---------
  traceProgress() << "automatic invariant strengthening...\n";
  if (!tracingSys("disableFactFlow")) {
    FOREACH_ASTLIST_NC(TopForm, unit->topForms, tf) {
      if (tf.data()->isTF_func()) {
        factFlow(*( tf.data()->asTF_func() ));
      }
    }
  }
  #endif // 0


  // path analysis and verification condition generation
  {                                          
    string outFname = stringc << argv[1] << ".vc";
    cout << "writing vclang file: " << outFname << "\n";

    VCEnv vcenv(factory, strTable, outFname);
    vcenv << "// " << outFname << "\n"
          << "// verification condition for " << argv[1] << "\n"
          << "// automatically generated by " << progname << "\n"
          << "\n"
          ;

    Verifier v(vcenv);
    unit->traverse(v);
  }


  #if 0
  // --------------- abstract interp ------------
  {
    traceProgress() << "verification condition generation...\n";
    AEnv env(strTable, &mem);

    unit->vcgen(env);

    traceProgress(2) << "done with vcgen\n";

    if (env.failedProofs != 0) {
      treeOut(1) << "there were " << env.failedProofs << " failed proofs\n";
      return 5;
    }

    if (tracingSys("stopAfterVCGen")) {
      return 0;
    }
  }
  #endif // 0


  //malloc_stats();

  // delete the tree
  delete unit;
  strTable.clear();

  checkHeap();
  //malloc_stats();

  traceRemoveAll();

  return 0;
}


int main(int argc, char **argv)
{
  int ret;
  try {
    ret = doit(argc, argv);
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
  return ret;
}
