// vml_env.cc
// code for vml_env.h

#include "vml_env.h"         // this module
#include "vml.ast.gen.h"     // VML AST
#include "trace.h"           // tracingSys


VMLEnv::VMLEnv(StringTable &table)
  : scopes(),                // empty stack
    procs(),                 // empty map
    labels(),                // empty map
    nexts(),
    errors(0),
    str(table),
    preState(0),
    quantifiers(0),
    cfgGraph(tracingSys("cfgGraph"))
{                            
  // make the global scope
  enterScope();

  // make the initial pending-next set
  pushPendingNexts();
}

VMLEnv::~VMLEnv()
{
  // make this explicit so I can set a breakpoint, etc.
  scopes.deleteAll();
}


bool VMLEnv::addVar(StringRef name, VDeclaration *decl)
{
  if (scope()->isMapped(name)) {
    return false;            // collision
  }
  else {
    scope()->add(name, decl);
    return true;
  }
}

VDeclaration *VMLEnv::lookup(StringRef name)
{
  // work from innermost scope out
  for (int i = scopes.length()-1; i >= 0; i--) {
    VDeclaration *d = scopes[i]->queryif(name);
    if (d) {
      return d;
    }
  }
  return NULL;               // not found
}


void VMLEnv::enterScope()
{
  scopes.push(new StringSObjDict<VDeclaration>);
}

void VMLEnv::exitScope()
{
  delete scopes.pop();
}


void VMLEnv::addProcedure(VProcedure *p)
{
  if (procs.isMapped(p->id->name)) {
    error(p->id->loc, stringc << "duplicate procedure: " << p->id->name);
  }
  else {
    procs.add(p->id->name, p);
  }
}

VProcedure *VMLEnv::lookupProcedure(StringRef name)
{
  return procs.queryif(name);
}


void VMLEnv::addLabel(VS_label *sl)          
{
  if (labels.isMapped(sl->id->name)) {
    error(sl->id->loc, stringc << "duplicate label: " << sl->id->name);
  }
  else {
    labels.add(sl->id->name, sl);
  }
}

void VMLEnv::clearLabelInfo()
{
  labels.empty();
}

VS_label *VMLEnv::lookupLabel(StringRef name)
{
  return labels.queryif(name);
}


string cfgLabel(VStatement *s)
{
  if (s) {
    return stringc << "\"" << s->kindName() << "\\n"
                   << toString(s->loc) << "\"";
  }
  else {
    return "end";
  }
}

void cfgEdge(VStatement *src, VStatement *target)
{
  cout << "  " << cfgLabel(src) << " -> " << cfgLabel(target) << "\n";
}

void VMLEnv::resolvePendingNexts(VStatement *target)
{
  SFOREACH_OBJLIST_NC(VStatement, *(nexts.top()), iter) {
    VStatement *src = iter.data();
    src->next = target;

    if (cfgGraph) {
      // This code emits edges in Dot format as they are built.
      // Dot is at http://www.research.att.com/sw/tools/graphviz/.

      if (src->isVS_if()) {
        #if 0    // this doesn't work.. not sure why
        // diamond shape
        cout << "  node  [\n"
             << "    label = " << cfgLabel(src) << "\n"
             << "    shape = diamond\n"
             << "  ]\n"
             ;
        #endif // 0

        // two outgoing edges
        VS_if *srcIf = src->asVS_if();
        cfgEdge(srcIf, srcIf->sThen);
        if (srcIf->sElse) {
          cfgEdge(srcIf, srcIf->sElse);
        }
        else {
          cfgEdge(srcIf, target);
        }
      }
      else {
        // one edge
        cfgEdge(src, target);
      }
    }
  }

  nexts.top()->removeAll();
}

void VMLEnv::addPendingNext(VStatement *source)
{
  nexts.top()->prepend(source);
}

void VMLEnv::pushPendingNexts()
{
  nexts.push(new SObjList<VStatement>);
}

void VMLEnv::popAndMergeNexts()
{
  SObjList<VStatement> *top = nexts.pop();
  nexts.top()->concat(*top);
  delete top;
}

void VMLEnv::clearPendingNexts()
{
  resolvePendingNexts(NULL /*target*/);
}


void VMLEnv::error(SourceLoc loc, char const *msg)
{
  errors++;
  cout << toString(loc) << ": error: " << msg << endl;
}

void VMLEnv::error(char const *msg)
{
  errors++;
  cout << "error: " << msg << endl;
}


// EOF
