// vml_paths.cc
// code for vml_paths.h

#include "vml_paths.h"       // this module
#include "vml_env.h"         // VMLEnv
#include "vml.ast.gen.h"     // VML AST
#include "sobjset.h"         // SObjSet


class PathCounter {
public:      // data
  VMLEnv &env;
  VProcedure *proc;
  SObjSet<VStatement*> visited;     // for detecting cycles

public:      // funcs
  PathCounter(VMLEnv &e, VProcedure *p) : env(e), proc(p) {}
                                    
  // depth-first search to count paths; returns the count
  int dfs(VStatement *s) throw(int);
  int dfsNext(VStatement *s) throw(int);
};


int PathCounter::dfs(VStatement *s) throw(int)
{
  // cycle?
  if (visited.contains(s)) {
    env.error(s->loc, stringc << proc->id->name << " contains a cyclic path");
    throw 0;
  }
  visited.add(s);

  if (!s->isVS_if()) {
    // one successor, same # of paths
    s->paths = dfsNext(s->next);
  }
  else {
    VS_if *i = s->asVS_if();

    // two successors, count is the sum
    s->paths = dfsNext(i->sThen) +
               dfsNext(i->sElse? i->sElse : i->next);
  }
  
  // it would ok to reach 's' along a different path
  visited.remove(s);

  return s->paths;
}

int PathCounter::dfsNext(VStatement *s) throw(int)
{
  // base case: procedure end
  if (!s) {
    return 1;
  }

  // base case: path root
  if (isRoot(s)) {
    // do not set 's->paths'; that will be set when 's' itself
    // is used as the root of a search
    return 1;
  }
  
  return dfs(s);
}


void countPaths(VMLEnv &env, VProcedure *proc)
{
  try {
    SFOREACH_OBJLIST_NC(VStatement, proc->roots, iter) {
      PathCounter pc(env, proc);
      pc.dfs(iter.data());
    }
  }
  catch (int) {
    // error already reported, keep going
  }
}


bool isRoot(VStatement *s)
{
  if (s->isVS_pred()) {
    VS_pred *p = s->asVS_pred();
    if (p->isInvariant()) {
      return true;
    }
  }

  return false;
}


// EOF
