// vml_print.cc
// code for vml_print.ast

#include "vml.ast.gen.h"      // VML AST
#include "trace.h"            // tracingSys


// fwds in this file
void printPred(BoxPrint &bp, Pred const *p, int parentPrec, bool allowEq);
void printExpr(BoxPrint &bp, VExpr const *e, int parentPrec, bool allowEq);
void printSemiConj(BoxPrint &bp, Pred const *p);


// --------------------- global -----------------------
BoxPrint& operator<< (BoxPrint &bp, Identifier const &i)
{
  bp << i.name;
  return bp;
}

BoxPrint& operator<< (BoxPrint &bp, VStatement const &s)
{
  s.iprint(bp);
  return bp;
}

BoxPrint& operator<< (BoxPrint &bp, Pred const &p)
{
  p.iprint(bp);
  return bp;
}

BoxPrint& operator<< (BoxPrint &bp, VExpr const &e)
{
  if (*((unsigned int*)(&e)) == 0xBBBBBBBB) {
    return bp << "0xBBBBBBBB";    // dangling reference!
  }

  e.iprint(bp);
  return bp;
}


template <class T>
void printParenList(BoxPrint &bp, ASTList<T> const &list)
{
  bp << "(" << bp.seq;
  int ct=0;
  FOREACH_ASTLIST(T, list, iter) {
    if (ct++ > 0) {
      bp << "," << bp.br;
    }
    bp << *(iter.data());
  }
  bp << bp.end << ")";
}


// lists of VD_vars are slightly different, because in that case we
// just want the names, not their full stand-alone declarations
void printParenList(BoxPrint &bp, ASTList<VD_var> const &list)
{
  bp << "(" << bp.seq;
  int ct=0;
  FOREACH_ASTLIST(VD_var, list, iter) {
    if (ct++ > 0) {
      bp << "," << bp.br;
    }
    bp << *(iter.data()->id);
  }
  bp << bp.end << ")";
}


// --------------------- VMLFile ------------------------
string VMLFile::print() const
{
  BoxPrint bp;
  iprint(bp);

  if (tracingSys("vmlfileDebugPrint")) {
    bp.debugPrintCout();
  }

  return BPRender().takeAndRender(bp);
}

void VMLFile::iprint(BoxPrint &bp) const
{
  bp << bp.vert;
  FOREACH_ASTLIST(VTopForm, forms, iter) {
    if (iter.data()->isVTF_comment()) {
      bp << bp.br;    // blank line before comments
    }

    iter.data()->iprint(bp);
    bp << bp.br;
  }
  bp << bp.end;
}


// --------------------- VTopForm ------------------------
void VTF_decl::iprint(BoxPrint &bp) const
{
  d->iprint(bp);
}


void VTF_inv::iprint(BoxPrint &bp) const
{
  bp << bp.vert << "invariant {" << bp.ind;
  printSemiConj(bp, p);
  bp << bp.und << "}" << bp.end;
}


void VTF_axiom::iprint(BoxPrint &bp) const
{
  bp << bp.vert << "axiom ";
  if (label) {
    bp << label << " ";
  }
  bp << "{" << bp.ind;

  printSemiConj(bp, p);
  bp << bp.und << "}" << bp.end;
}


void VTF_proc::iprint(BoxPrint &bp) const
{
  p->iprint(bp);
}


void VTF_comment::iprint(BoxPrint &bp) const
{
  bp << "// " << c;
}


// -------------------- VDeclaration ---------------------
void VD_var::iprint(BoxPrint &bp) const
{
  bp << "var" << bp.sp << *id;
  if (init) {
    bp << bp.sp << ":=" << bp.sp << bp.seq << *init << bp.end;
  }
  bp << ";";
}


void VD_const::iprint(BoxPrint &bp) const
{
  bp << "const" << bp.sp << *id << bp.sp
     << "=" << bp.sp << bp.seq << *value << bp.end << ";";
}


void VD_funcsym::iprint(BoxPrint &bp) const
{
  bp << "funcsym" << bp.sp << *id;
  printParenList(bp, params);
  bp << ";";
}


// -------------------- VFunction ------------------------
// print a predicate that might be big
void printBigPred(BoxPrint &bp, Pred const *p)
{
  if (p->isBinary(BP_AND)) {
    // big enough
    bp << bp.hv;
    bp << "AND {" << bp.ind;
    printSemiConj(bp, p);
    bp << bp.und << "}";
    bp << bp.end;
  }
  else {
    // assume small
    bp << *p;
  }
}


void VProcedure::iprint(BoxPrint &bp) const
{
  bp << bp.vert << "procedure" << bp.sp;

  // name/params/returns; might put a break after the '->'
  bp << bp.seq << *id;
  printParenList(bp, params);
  bp << bp.op("->");
  printParenList(bp, returns);
  bp << bp.end;

  // pre/post/reads/writes
  bp << bp.ind;

  bp << "pre(";
  printBigPred(bp, pre);
  bp << ")" << bp.br;

  bp << "post(";
  printBigPred(bp, post);
  bp << ")" << bp.br;

  bp << "reads";
  printParenList(bp, reads);
  bp << bp.br;

  bp << "writes";
  printParenList(bp, writes);

  // body, which is an S_compound  
  bp << bp.und;
  if (body) {
    bp << *body;
  }
  else {
    bp << ";";
  }
  bp << bp.end;
}


// -------------------- VStatement -----------------------
string VStatement::print() const
{
  BoxPrint bp;
  iprint(bp);

  return BPRender().takeAndRender(bp);
}


// As a general principle, a print function only emits a 'br' if it
// will be enclosed by a box created by that function.


void VS_comment::iprint(BoxPrint &bp) const
{
  bp << "// " << c;
}


void VS_decl::iprint(BoxPrint &bp) const
{
  d->iprint(bp);
}


void VS_pred::iprint(BoxPrint &bp) const
{
  static char const * const map[] = {
    "assume",
    "assert",
    "pure_assert",
    "invariant",
    "standalone_invariant"
  };
  ASSERT_TABLESIZE(map, PS_STANDALONE_INVARIANT+1);
  xassert((unsigned)psk < TABLESIZE(map));

  bp << map[psk] << bp.sp << *p << ";";
}


void VS_assign::iprint(BoxPrint &bp) const
{
  bp << *target << bp.sp << ":=" << bp.sp << bp.seq << *source << bp.end << ";";
}


void VS_procCall::iprint(BoxPrint &bp) const
{
  // targets
  printParenList(bp, targets);

  bp << bp.sp << ":=" << bp.sp << *procName;

  // arguments
  printParenList(bp, args);

  bp << ";";
}

           
// print compound w/o braces
void printCStmt(BoxPrint &bp, VStatement const *s)
{
  if (s->isVS_compound()) {
    s->asVS_compoundC()->iiprint(bp);
  }
  else {
    // not expected, but tolerable I suppose
    s->iprint(bp);
  }
}


void VS_if::iprint(BoxPrint &bp) const
{
  bp << bp.vert;
  
  bp << "if" << bp.sp << "(" << *p << ")" << bp.sp << "{" << bp.ind;
  printCStmt(bp, sThen);
  bp << bp.und << "}";

  if (sElse) {
    bp << bp.br << "else" << bp.sp << "{" << bp.ind;
    printCStmt(bp, sElse);
    bp << bp.und << "}";
  }

  bp << bp.end;
}


void VS_loop::iprint(BoxPrint &bp) const
{
  bp << bp.vert << "while (true) {" << bp.ind;
  printCStmt(bp, body);
  bp << bp.und << "}" << bp.end;
}


void VS_label::iprint(BoxPrint &bp) const
{
  bp << *id << ":";
}


void VS_goto::iprint(BoxPrint &bp) const
{
  bp << "goto " << *id << ";";
}


void VS_compound::iprint(BoxPrint &bp) const
{
  bp << bp.vert << "{" << bp.ind;
  iiprint(bp);
  bp << bp.und << "}" << bp.end;
}

void VS_compound::iiprint(BoxPrint &bp) const
{
  int ct=0;
  FOREACH_ASTLIST(VStatement, body, iter) {
    if (ct++ > 0) {
      bp << bp.br;
      
      if (iter.data()->isVS_comment()) {
        bp << bp.br;     // blank line before comments
      }
    }
    bp << *(iter.data());
  }
}

#ifdef VML_PAT_INCLUDED
void VS_patvar::iprint(BoxPrint &bp) const
{
  bp << id->name;
}
#endif // VML_PAT_INCLUDED


// ----------------------- Pred -------------------------
string Pred::print() const
{
  BoxPrint bp;
  iprint(bp);

  return BPRender().takeAndRender(bp);
}


// numbers from vml.gr
int Pred::prec() const
{
  ASTSWITCHC(Pred, this) {
    ASTCASEC1(P_not) {
      return 60;
    }
    ASTNEXTC(P_binary, b) {
      switch (b->con) {
        default: xfailure("bad code");
        case BP_AND:      return 50;
        case BP_OR:       return 40;
        case BP_IMPLIES:  return 30;
        case BP_IFF:      return 20;
      }
    }
    ASTNEXTC1(P_compare) {
      return 65;
    }
    ASTDEFAULTC {
      return 200;    // high precedence
    }
    ASTENDCASEC
  }
}

Associativity Pred::assoc() const
{
  ASTSWITCHC(Pred, this) {
    ASTCASEC(P_binary, b) {
      return b->con == BP_IMPLIES? AS_RIGHT :     // right associative
             b->con == BP_IFF? AS_NONE :          // parenthesize these fully
             AS_ASSOC;                            // semantically associative
    }
    ASTDEFAULTC {
      return AS_NONE;
    }
    ASTENDCASEC
  }
}

bool includes(Associativity assocSet, Associativity assocElt)
{
  if (assocSet == AS_ASSOC) {
    return true;            // AS_LEFT and AS_RIGHT are included in AS_ASSOC
  }
  else {
    return assocSet == assocElt;
  }
}

void printPred(BoxPrint &bp, Pred const *p, int parentPrec, bool allowEq)
{
  if (allowEq? p->prec() < parentPrec :
               p->prec() <= parentPrec) {
    bp << "(" << bp.seq << *p << bp.end << ")";
  }                        
  else {
    bp << *p;
  }
}



void P_lit::iprint(BoxPrint &bp) const
{
  bp << (b? "true" : "false");
}


void P_not::iprint(BoxPrint &bp) const
{
  // stack "!" without parens, otherwise parenthesize all
  if (p->isP_not()) {
    bp << "!" << *p;
  }
  else {
    bp << "!(" << *p << ")";
  }
}


// this one puts breaks into the box made by the caller, but
// does not put a final break
void printSemiConj(BoxPrint &bp, Pred const *p)
{
  if (p->isBinary(BP_AND)) {
    P_binary const *b = p->asP_binaryC();
    printSemiConj(bp, b->p1);
    bp << bp.br;
    printSemiConj(bp, b->p2);
  }
  else {
    bp << *p << ";";
  }
}


void P_quant::iprint(BoxPrint &bp) const
{
  bp << bp.seq;

  // print line info for Kettle
  {
    char const *fname;
    int line, col;
    sourceLocManager->decodeLineCol(loc, fname, line, col);

    bp << bp.lineStart
       << "// file=" << fname << " line=" << line
       << bp.fbr;
  }

  bp << (universal? "forall " : "exists ");
  if (label) {
    if (label[0]=='\"') {
      bp << label << " ";               // label is already quoted
    }
    else {
      // when I am synthesizing quantifier it's a little annoying
      // to put in the quotes, and this tolerance seems harmless
      bp << "\"" << label << "\" ";     // not already quoted
    }
  }

  bp << "(" << bp.seq;
  int ct=0;
  FOREACH_ASTLIST(VD_var, vars, iter) {
    if (ct++) {
      bp << "," << bp.br;
    }
    bp << iter.data()->id->name;
  }
  bp << bp.end << ")";

  if (patterns) {
    bp << bp.ind;
    patterns->iprint(bp);
    bp << "." << bp.br;
  }
  else {
    bp << "." << bp.ind;
  }

  bp << bp.hv;
  if (body->isBinary(BP_IMPLIES)) {
    P_binary *pbin = body->asP_binary();
    bp << "if {" << bp.ind;
    printSemiConj(bp, pbin->p1);
    bp << bp.und << "} /*==>*/ {" << bp.ind;
    printSemiConj(bp, pbin->p2);
    bp << bp.und << "}";
  }

  else {
    bp << "{" << bp.ind;
    printSemiConj(bp, body);
    bp << bp.und << "}";
  }
  bp << bp.end << bp.und;

  bp << bp.end;
}


char const *opString(BPropCon op)
{
  static char const * const map[] = {
    "&&", "||", "==>", "<==>"
  };
  xassert((unsigned)op < TABLESIZE(map));
  return map[op];
}

void P_binary::iprint(BoxPrint &bp) const
{
  bp << bp.seq;
  printPred(bp, p1, prec(), includes(assoc(), AS_LEFT));

  bp << bp.sp << opString(con);

  if (assoc() == AS_ASSOC) {
    // no need to indent when there's no semantic nesting
    bp << bp.br;
  }
  else {
    // by adding an indentation after the connective, we ensure that
    // the second argument will be indented if the line is broken
    bp << bp.ind;
  }

  printPred(bp, p2, prec(), includes(assoc(), AS_RIGHT));
  bp << bp.end;
}


char const *opString(ComparisonOp op)
{
  static char const * const map[] = {
    "==", "!=", "<", "<=", ">", ">="
  };
  xassert((unsigned)op < TABLESIZE(map));
  return map[op];
}

void P_compare::iprint(BoxPrint &bp) const
{
  bp << bp.seq;
  printExpr(bp, e1, 65, true /*allowEq*/);
  bp << bp.op(opString(op));
  printExpr(bp, e2, 65, true /*allowEq*/);
  bp << bp.end;
}

void P_declare_partial_order::iprint(BoxPrint &bp) const
{
  bp << "declare_partial_order(" << bp.seq;
  bp << lt << "," << bp.br << le;
  bp << bp.end << ")";
}

#ifdef VML_PAT_INCLUDED
void P_patvar::iprint(BoxPrint &bp) const
{
  bp << id->name;
}
#endif // VML_PAT_INCLUDED


// ----------------------- VExpr -------------------------
string VExpr::print() const
{
  BoxPrint bp;
  iprint(bp);

  return BPRender().takeAndRender(bp);
}


// numbers from vml.gr
int VExpr::prec() const
{
  ASTSWITCHC(VExpr, this) {
    ASTCASEC1(VE_neg) {
      return 70;
    }
    ASTNEXTC(VE_arith, a) {
      return a->op==AO_TIMES? 80 : 70;
    }
    ASTDEFAULTC {
      return 200;    // high precedence
    }
    ASTENDCASEC
  }
}

Associativity VExpr::assoc() const
{
  return isVE_arith()? AS_LEFT : AS_NONE;
}

void printExpr(BoxPrint &bp, VExpr const *e, int parentPrec, bool allowEq)
{
  if (allowEq? e->prec() < parentPrec :
               e->prec() <= parentPrec) {
    bp << "(" << bp.seq << *e << bp.end << ")";
  }
  else {
    bp << *e;
  }
}


void VE_lit::iprint(BoxPrint &bp) const
{
  bp << n;
}

void VE_var::iprint(BoxPrint &bp) const
{
  bp << id->name;
}

void VE_neg::iprint(BoxPrint &bp) const
{
  bp << "-";
  printExpr(bp, e, prec(), false /*allowEq*/);
}

void VE_apply::iprint(BoxPrint &bp) const
{
  if (0==strcmp(func->name, "sel")) {
    printExpr(bp, args.nthC(0), 200, true);
    bp << "[" << *(args.nthC(1)) << "]";
    return;
  }

  if (0==strcmp(func->name, "upd")) {
    printExpr(bp, args.nthC(0), 200, true);
    bp << "{" << *(args.nthC(1)) << " := " << *(args.nthC(2)) << "}";
    return;
  }

  bp << func->name;
  
  printParenList(bp, args);
}

char const *opString(ArithOp op)
{
  switch (op) {
    default: xfailure("bad code");
    case AO_PLUS:  return "+";
    case AO_MINUS: return "-";
    case AO_TIMES: return "*";
  }
}

void VE_arith::iprint(BoxPrint &bp) const
{
  printExpr(bp, e1, prec(), assoc() == AS_LEFT);

  bp << bp.op(opString(op));

  printExpr(bp, e2, prec(), assoc() == AS_RIGHT);
}

void VE_pre::iprint(BoxPrint &bp) const
{
  bp << "pre(" << bp.seq << *e << bp.end << ")";
}

void VE_cond::iprint(BoxPrint &bp) const
{
  // intended layout:
  //
  //   ~~~~~~~ ?
  //     ~~~~~~~~~~~
  //     ~~~~~~~~ :
  //     ~~~~~~~~~~~
  //     ~~~~~~~~

  bp << bp.hv << *p << bp.sp << "?" << bp.ibr(2)
       << bp.seq << *a << bp.end << bp.op(":")
       << bp.seq << *b << bp.end 
     << bp.end;
}

void VE_pat::iprint(BoxPrint &bp) const
{
  bp << "__pat(" << bp.seq << *e << bp.end << ")";
}

#ifdef VML_PAT_INCLUDED
void VE_patvar::iprint(BoxPrint &bp) const
{
  bp << id->name;
}
#endif // VML_PAT_INCLUDED


// EOF
