// vml_tocnf.h
// convert a VML Pred (predicate) to CNF

// Conversion to CNF is captured by the following five rewrite rules:
//   
//   A => B             -->      !A || B
//   A <=> B            -->      (B || !A) && (A || !B)
//   !(A && B)          -->      !A || !B
//   !(A || B)          -->      !A && !B
//   A || (B && C)      -->      (A || B) && (A || C)
//
// So, in principle, an algorithm could simply apply these rules
// until no more rules apply.  However, this would be slow.
//
// This module takes a more sophisticated approach.  First, every
// literal (atomic formula that will not be broken down) and its
// negation is assigned an integer code.  The code is dense; if there
// are 'n' literals, the integers will go from 0 to 2n-1.  The use of
// this coding scheme allows literals to be easily manipulated without
// concerns like allocating and copying, and makes literal negation
// (which is a common operation in the algorithm) very fast.
//
// Next, we represent a CNF formula as a set of sets of codes, i.e., a
// conjunction of disjunctions of possibly negated literals.  The
// disjunction sets are represented with bit vectors, making union and
// testing for tautology fast.  The conjunction sets are represented
// with linked lists.
//
// Given these representation choices, we then have a recursive
// algorithm that takes as input a Pred and a boolean specifying
// whether the CNF of the predicate or its negation is desired.  The
// algorithm recursively CNFs its subformulas, passing the negation
// flags appropriate, and then joins the results in a manner dependent
// on the formula and the negation flag.
//
// Finally, the CNF representation is converted back into a Pred,
// making deep copies of the original formula's literals where they
// appear in the output formula.
//
// (Aside: The CNF representation used here has potentially wider
// application, so it is worth considering the possibility of using it
// more generally in the vcgen and even the prover.  But for now I
// only use it within this module.)

#ifndef VML_TOCNF_H
#define VML_TOCNF_H

#include "ownerserf.h"   // owner

class Pred;              // vml.ast


// external interface
Pred * owner toCNF(Pred const *input, bool negate = false);


// ---------- internal mechanics -----------
#ifdef VML_TOCNF_IMPLEMENTATION

#include "objlist.h"     // ObjList
#include "bitarray.h"    // BitArray
#include "array.h"       // ArrayStack
#include "vptrmap.h"     // VoidPtrMap


// codes for a possibly-negated literal
typedef int LiteralCode;


// codes for a disjunction of possibly-negated literals
// ("disjunction" is a synonym for "clause" in this context)
class Disj {
private:     // data
  BitArray bits;
  
public:      // types
  class Iter : public BitArray::Iter {
  public:
    Iter(Disj const &disj) : BitArray::Iter(disj.bits) {}
  };
  friend class Iter;

public:      // funcs
  Disj(int numLiteralCodes);
  Disj(Disj const &obj) : bits(obj.bits) {}
  ~Disj();
  
  void addDisjunct(LiteralCode lit)
    { bits.set(lit); }
    
  bool equals(Disj const &obj) const
    { return bits == obj.bits; }
    
  void unionWith(Disj const &obj)
    { bits |= obj.bits; }
                           
  // true if some literal and its negation both appear
  bool tautological() const
    { return bits.anyEvenOddBitPair(); }
};


// codes for a formula in CNF
class Conj {
public:      // data
  // conjunction of disjunctions
  ObjList<Disj> set;            
  
  // when this is false, the entire object codes for "false"; in
  // that case 'set' is empty
  bool consistent;

  // if 'consistent' is true and the set is empty, then the
  // conjunction as a whole is equivalent to "true"

public:
  Conj();
  ~Conj();
  
  void makeInconsistent()
    { consistent = false; set.deleteAll(); }
    
  void addConjunct(Disj *d);
                                 
  // set 'this' equal to the conjunction of it and 'c', and
  // destroy 'c' in the process
  void conjoinAndDelete(Conj *c);   
  
  // same but for a disjunction
  void disjoinAndDelete(Conj *c);
};


// class to manage algorithm state, esp. the literal table
class VML2CNF {
private:     // data
  // NOTE: The literalTable maps integers that are half the size
  // of the corresponding LiteralCode.  So, if a literal has
  // code 5 in the literalTable, then the LiteralCode for that
  // literal is 10, and the LiteralCode for its negation is 11.

  // int -> literal
  ArrayStack<Pred const *> literalTable;

  // literal -> int
  // (I am slightly abusing VoidPtrMap, in that I will use integers
  // as the values, and cast to/from the expected void*)
  VoidPtrMap literalCodeMap;

  // when true, we cannot accept new literals in the table
  bool tableLocked;

private:     // funcs
  LiteralCode encode(Pred const *p, bool negate);
  void makeLT_helper(Pred const *input);

  Pred * owner materializeLiteral(LiteralCode code);
  Pred * owner materializeDisj(Disj const *disj);

  Disj *makeDisj();
  Conj *literalToConj(LiteralCode code);

public:      // funcs
  VML2CNF();
  ~VML2CNF();

  int numLiterals() const 
    { return literalTable.length(); }

  // pre-pass
  void makeLiteralTable(Pred const *input);

  // primary conversion
  Conj * owner toCNF(Pred const *input, bool negate);

  // extraction
  Pred * owner materialize(Conj const *cnf);
};


#endif // VML_TOCNF_IMPLEMENTATION

#endif // VML_TOCNF_H
