// vml_tocvc.cc
// code for vml_tocvc.ast

// working from manual at
//   http://verify.stanford.edu/CVC/
// and also George's cvcparse.mly, cvclex.mll

// NOTE: CVC technically wants me to declare variables, but I do not.
//
// The real CVC spec does not have RETRACT, instead it has PUSH/POP,
// and the manual says that declarations are global across POPs.  One
// problem is that means I have to ensure I don't re-declare names from
// different functions, but otherwise I could do that.
//
// George's implementation has RETRACT, which does not affect decls,
// so again it's global.  But he also doesn't really use the decls, so
// for now I'm not going to bother.


#include "vml.ast.gen.h"      // VML AST
#include "exc.h"              // throw_XFatal


// ------------------------ CVCEnv ------------------------
CVCEnv::CVCEnv()
{}

CVCEnv::~CVCEnv()
{}


CVCEnv& CVCEnv::operator<< (Pred const &p)
{
  p.cvc(*this);
  return *this;
}

CVCEnv& CVCEnv::operator<< (VExpr const &e)
{
  e.cvc(*this);
  return *this;
}


// ------------------------- Pred --------------------------
string Pred::toCVC() const
{
  CVCEnv se;
  cvc(se);
  return se.sb;
}

void P_lit::cvc(CVCEnv &se) const
{
  se << (b? "TRUE" : "FALSE");
}

void P_not::cvc(CVCEnv &se) const
{
  se << "(NOT " << *p << ")";
}

bool isFunctionApp(VExpr const *e, char const *name)
{
  if (!e->isVE_apply()) {
    return false;
  }

  VE_apply const *a = e->asVE_applyC();
  return 0==strcmp(a->func->name, name);
}

          
void emitPatternTerm(CVCEnv &se, VExpr const *term)
{
  if (term->isVE_apply()) {
    VE_apply const *ap = term->asVE_applyC();

    if (0==strcmp(ap->func->name, "TRIGGEREQ")) {
      se << *(ap->args.nthC(0)) << "=" << *(ap->args.nthC(1));
      return;
    }

    if (0==strcmp(ap->func->name, "TRIGGERNEQ")) {
      se << *(ap->args.nthC(0)) << "/=" << *(ap->args.nthC(1));
      return;
    }
  }

  se << "have(" << *term << ")";
}

void P_quant::cvc(CVCEnv &se) const
{
  // number of parens to close at end
  int ct=0;

  // 2005-02-15: Kettle has trouble with existentials, so translate
  // them into universals
  //
  // but this does not work because of problems with the trigger...
  //if (!universal) {
  //  se << "(NOT ";
  //  ct++;
  //}

  FOREACH_ASTLIST(VD_var, vars, iter) {
    se << "("
       //<< "FORALL"         // this is if we use the NOT above and below
       << (universal? "FORALL" : "EXISTS")     // this is for when no NOTs are added
       << " " << iter.data()->id->name << ":exp. ";
    ct++;
  }

  //if (!universal) {
  //  se << "NOT (";
  //  ct++;
  //}

  if (patterns) {
    // translate Simplify pattern exp into Kettle pattern exp
    xassert(isFunctionApp(patterns, "PATS"));
    VE_apply const *pats = patterns->asVE_applyC();

    xassert(pats->args.count() > 0);
#define TRIGGER_PATTERNS
#ifdef TRIGGER_PATTERNS    
    se << "TRIGGER((";
    ct += 2;
#else
    se << "(";
#endif    
    // if there is more than one pattern in the PATS, that means that
    // any of the given terms is sufficient to trigger instantiation;
    // so we build an OR to go before the implication
    int ct1=0;
    FOREACH_ASTLIST(VExpr, pats->args, iter1) {
      VExpr const *pat = iter1.data();

      if (ct1++ > 0) {
        se << " OR ";
      }

      if (isFunctionApp(pat, "MPAT")) {
        // a multipattern is a set of terms all of which must be
        // present to trigger instantiation
        VE_apply const *mpat = pat->asVE_applyC();

        // connect them with AND, unless it's degenerate
        xassert(mpat->args.count() > 0);
        se << "(";

        int ct2=0;
        FOREACH_ASTLIST(VExpr, mpat->args, iter2) {
          VExpr const *term = iter2.data();

          if (ct2++ > 0) {
            se << " AND ";
          }

          emitPatternTerm(se, term);
        }

        se << ")";
      }
      else {
        emitPatternTerm(se, pat);
      }
    }
#ifdef TRIGGER_PATTERNS
    se << "), (";
#else
    se << ") => (";
    ct += 1;    // cause final parens to be emitted at the end
#endif    
  }
  else {
    cout << "patternless axiom:\n"
         << this->print() << "\n";
    throw_XFatal("axiom requires pattern for Kettle");
  }

  se << *body;

  while (ct--) {
    se << ")";
  }
}

static char const *opmap(BPropCon con)
{
  static char const * const map[] = {
    "AND",
    "OR",
    "=>",
    "<=>"
  };
  ASSERT_TABLESIZE(map, NUM_BPROPCONS);

  xassert((unsigned)con < NUM_BPROPCONS);
  return map[con];
}

void P_binary::cvc(CVCEnv &se) const
{
  se << "(" << *p1 << " " << opmap(con) << " " << *p2 << ")";
}

static char const *opmap(ComparisonOp op)
{
  static char const * const map[] = {
    "=",
    "/=",
    "<",
    "<=",
    ">",
    ">="
  };
  ASSERT_TABLESIZE(map, NUM_COMPARISONOPS);

  xassert((unsigned)op < NUM_COMPARISONOPS);
  return map[op];
}

void P_compare::cvc(CVCEnv &se) const
{
  // hack: translate DLT/DLE into <,<= for Kettle since, at the moment,
  // it does not support arbitrary partial orders, and it happens that
  // my use of DLT/DLE does not (should not) run into problems with
  // integer heuristics
  //
  // "DLT(x,y)==1" is emitted as "(x < y)"  
  // "DLE(x,y)==1" is emitted as "(x <= y)"
  if (op == CO_EQUAL &&
      e1->isVE_apply() &&
      e2->isVE_lit() &&
      e2->asVE_litC()->n == 1) {
    VE_apply const *app = e1->asVE_applyC();
    if (0==strcmp(app->func->name, "DLT")) {
      se << "(" << *(app->args.nthC(0)) << " < " << *(app->args.nthC(1)) << ")";
      return;
    }
    if (0==strcmp(app->func->name, "DLE")) {
      se << "(" << *(app->args.nthC(0)) << " <= " << *(app->args.nthC(1)) << ")";
      return;
    }
  }

  se << "(" << *e1 << " " << opmap(op) << " " << *e2 << ")";
}

void P_declare_partial_order::cvc(CVCEnv &se) const
{
  se << "TRUE \n% omitting due to lack of syntax: (ORDER " << lt << " " << le << ")\n";
}

#ifdef VML_PAT_INCLUDED
void P_patvar::cvc(CVCEnv &se) const
{
  xfailure("E_patvar::cvc should not be called");
}
#endif // VML_PAT_INCLUDED


// -------------------------- VExpr ----------------------------
string VExpr::toCVC() const
{
  CVCEnv se;
  cvc(se);
  return se.sb;
}

void VE_lit::cvc(CVCEnv &se) const
{
  if (n >= 0) {
    se << n;
  }
  else {
    // George's CVC parser doesn't like negative literals
    se << "(0 - " << -n << ")";
  }
}

void VE_var::cvc(CVCEnv &se) const
{
  se << id->name;
}

void VE_neg::cvc(CVCEnv &se) const
{
  // negation is sometimes flaky
  // (true of CVC? don't know)
  se << "(0 - " << *e << ")";
}

void VE_apply::cvc(CVCEnv &se) const
{
  se << func->name << "(";
  
  if (0==strcmp(func->name, "sel")) {
    // special case: change the argument order since Kettle does it
    // differently than Simplify
    se << args.nthC(1)[0] << ", " << args.nthC(0)[0];
  }
  else if (0==strcmp(func->name, "upd")) {
    // another special case
    se << args.nthC(1)[0] << ", " << args.nthC(2)[0] << ", " << args.nthC(0)[0];
  }
  else {
    // normal case: emit in order
    int ct=0;
    FOREACH_ASTLIST(VExpr, args, iter) {
      if (ct++ > 0) {
        se << ", ";
      }
      se << *(iter.data());
    }
  }

  se << ")";
}

void VE_arith::cvc(CVCEnv &se) const
{
  se << "(" << *e1 << " " << opString(op) << " " << *e2 << ")";
}

// vcgen should never yield these
void VE_pre::cvc(CVCEnv &se) const
{ xfailure("should not happen: VE_pre::cvc"); }
void VE_cond::cvc(CVCEnv &se) const
{ xfailure("should not happen: VE_cond::cvc"); }


void VE_pat::cvc(CVCEnv &se) const
{
  se << "pat(" << *e << ")";
}

#ifdef VML_PAT_INCLUDED
void VE_patvar::cvc(CVCEnv &se) const
{
  xfailure("VE_patvar::cvc should not be called");
}
#endif // VML_PAT_INCLUDED


// EOF
